/*
 * LambdaAbstractionRule.java
 *
 * Created on June 5, 2007, 8:36 PM
 *
 */

package lambdacalc.lf;

import lambdacalc.logic.Expr;
import lambdacalc.logic.Lambda;
import lambdacalc.logic.TypeEvaluationException;
import lambdacalc.logic.Var;


/**
 *
 * @author champoll
 */
public class LambdaAbstractionRule extends CompositionRule {
    public static final LambdaAbstractionRule INSTANCE
            = new LambdaAbstractionRule();
    
    private LambdaAbstractionRule() {
        super("Lambda abstraction");
    }
    
    /**
     * As a side effect, determines index and body.
     */
    @Override
	public boolean isApplicableTo(Nonterminal node) {
        
        if (node.size() != 2) return false;

        if (node.getLeftChild() instanceof BareIndex) {
            return !(node.getRightChild() instanceof BareIndex);
        } else {
            return (node.getRightChild() instanceof BareIndex);
        }
    }
    
    //We ignore the parameter onlyIfApplicable because there is just no way to "apply" this
    //rule in non-applicable cases.
    @Override
	public Expr applyTo(Nonterminal node, AssignmentFunction g, boolean onlyIfApplicable) throws MeaningEvaluationException {
     
        if (!this.isApplicableTo(node)) {
            throw new MeaningEvaluationException
                    ("The lambda abstraction rule is only " +
                    "applicable on a nonterminal that has exactly " +
                    "two children of which one is a bare index or a " +
                    "word like \"which\" or \"such\".");
        }

        BareIndex index = null;
        LFNode body = null;

        LFNode left = node.getLeftChild();
        LFNode right = node.getRightChild();
        
        if (left instanceof BareIndex) {
            index = (BareIndex) left;
            body = right;
        } else if (right instanceof BareIndex) {
            index = (BareIndex) right;
            body = left;
        } else { //should never be reached
            throw new RuntimeException();
        }
        
        Var var;
        
        if (g == null) {
            // We are evaluating bottom-up.
            // Get a fresh variable based on the meaning that we know we will eventually get.
            // Choose a variable that is not in use in the simplified inner expression.
            // See MeaningBracketExpr.evaluate().
            try {
                Expr bodyMeaning = body.getMeaning();
                try { bodyMeaning = 
                        MeaningBracketExpr.
                        replaceAllMeaningBrackets(bodyMeaning).simplifyFully(); 
                } catch (TypeEvaluationException e) {
                } // shouldn't throw since getMeaning worked
                var = bodyMeaning.createFreshVar();

            // But if we can't get a meaning, choose a default variable
                //TODO I think this is dangerous and we should provide a warning -Lucas
            } catch (MeaningEvaluationException mee) {
                mee.printStackTrace();
                var = new Var("x", lambdacalc.logic.Type.E, false);
            }
            
        } else {
            // We are evaluating top-down.
            // Choose a variable that is not in the range of the assignment function
            // being passed to us. Since we will add to the assignment function,
            // expressions within us will be sure to not create independent variables
            // that conflict with the one we choose.
            
            var = Expr.createFreshVar(Var.X, new java.util.HashSet(g.values()));
        }
        
        // Copy the assignment function being given to us and add the
        // new mapping from the bare index to a fresh variable.
        AssignmentFunction g2 = (g == null ? new AssignmentFunction() : new AssignmentFunction(g));
        g2.put(index, var);
        
        // When we evaluate the meaning bracket expression, we need to know whether
        // we've chosen a fresh variable based on what's above (top-down) or
        // below (bottom-up).
        boolean topDown = (g != null);
        
        return new Lambda(var, new MeaningBracketExpr(body, g2, topDown), true);
    }
}


