/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.data;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;

import lambdacalc.logic.Expr;
import lambdacalc.logic.TypeEvaluationException;

/**
 * @author Wei LU
 * @version 1.0
 */

public class Instance implements Serializable{

	private static final long serialVersionUID = 7793563431578051148L;
	
	private int _id;
	private double _weight;
	private Expr _expr;
	private ArrayList<Word> _words;
	private LambdaProd _prod;
	private String _mrl;//the raw mrl string as stored in the file..
	private boolean _isPrior;//whether this is a prior instance..i.e. one instance that provides some prior knowledge, but is not a real instance.
	
	private ArrayList<LambdaProd> _allProds = null;
	
	/**
	 * Constructor
	 */
	public Instance(int id, double weight, Expr expr, ArrayList<Word> words, String mrl, boolean isPrior){
		this._id = id;
		this._weight = weight;
		this._expr = expr;
		this._words = words;
		this._mrl = mrl;
		this._isPrior = isPrior;
	}
	
	/**
	 * This method returns in reverse order the list of all LambdaMRProds.
	 * @param fac
	 * @return
	 * @throws UnsupportedEncodingException
	 * @throws TypeEvaluationException
	 */
	public ArrayList<LambdaProd> getAllProds(LambdaProdFactory fac){
		
		if(this._allProds == null){
			this._allProds = new ArrayList<LambdaProd>();
			LambdaProd rootProd = this.getMRProd(fac);
			this._allProds = rootProd.getAllProds();
		}
		return this._allProds;
		
	}
	
	/**
	 * Get the MRProd based on a factory
	 */
	public LambdaProd getMRProd_single(LambdaProdFactory fac){
		if(this._prod == null){
			if(this._isPrior){
				this._prod = fac.decomposeAsRoot_prior(this._expr);
			} else {
				this._prod = fac.decomposeAsRoot_single(this._expr);
			}
		}
		return this._prod;
	}
	
	/**
	 * Get the MRProd based on a factory
	 */
	public LambdaProd getMRProd(LambdaProdFactory fac){
		if(this._prod == null){
			if(this._isPrior){
				this._prod = fac.decomposeAsRoot_prior(this._expr);
			} else {
				this._prod = fac.decomposeAsRoot(this._expr);
			}
		}
		return this._prod;
	}
	
	public boolean isPrior(){
		return this._isPrior;
	}
	
	public int getId(){
		return this._id;
	}
	
	public double getWeight(){
		return this._weight;
	}
	
	public Expr getExpr(){
		return this._expr;
	}
	
	public ArrayList<Word> getWords(){
		return this._words;
	}
	
	public String getMRL(){
		return this._mrl;
	}
	
	public String getSent(){
		StringBuffer sb = new StringBuffer();
		for(int i = 0; i<this._words.size(); i++){
			if(i!=0){
				sb.append(' ');
			}
			sb.append(this._words.get(i).getWord());
		}
		return sb.toString();
	}
	
	@Override
	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append(this._id);
		sb.append('\n');
		sb.append(this._weight);
		sb.append('\n');
		sb.append(this._expr);
		sb.append('\n');
		for(int i = 0; i<this._words.size(); i++){
			if(i>0)
				sb.append(' ');
			sb.append(this._words.get(i));
		}
		return sb.toString();
	}
	
	/**
	 * Private methods
	 */
	/**
	 * Serialization
	 */
	private void writeObject(ObjectOutputStream out) throws IOException{
		out.writeInt(this._id);
		out.writeDouble(this._weight);
		out.writeObject(this._expr);
		out.writeInt(this._words.size());
		for(int i=0; i<this._words.size(); i++){
			out.writeObject(this._words.get(i));
		}
		out.writeUTF(this._mrl);
		out.writeBoolean(this._isPrior);
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
		this._id = in.readInt();
		this._weight = in.readDouble();
		this._expr = (Expr)in.readObject();
		int size = in.readInt();
		this._words = new ArrayList<Word>(size);
		for(int i = 0; i<size; i++){
			this._words.add((Word)in.readObject());
		}
		this._mrl = in.readUTF();
		this._isPrior = in.readBoolean();
	}
	
}
