/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.data;

import lambdacalc.logic.*;
import java.util.*;
import java.io.*;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaCalculusProcessor {
	
    public static int RULE_MINMAX_1 = 1001;
    public static int RULE_MINMAX_2 = 1002;
    
    public static Expr reGroup(Expr exp) {
        return reGroupHelper(exp, new Vector<Identifier>());
    }
    
    private static Expr reGroupHelper(Expr exp, Vector<Identifier> boundedVars) {
        if (exp instanceof Binder) {
            Binder binder = (Binder) exp;
            Identifier var = binder.getVariable();
            boundedVars.add(0, var);
            Expr inner = reGroupHelper(binder.getInnerExpr(), boundedVars);
            binder.setInnerExpr(inner);
            boundedVars.remove(0);
            return binder;
        } else if (exp instanceof Unary) {
            Unary unary = (Unary) exp;
            Expr inner = reGroupHelper(unary.getInnerExpr(), boundedVars);
            unary.setInnerExpr(inner);
            return unary;
        } else if (exp instanceof And) {
            And and = (And) exp;
            And newAnd = reGroupAnd(and, boundedVars);
            return newAnd;
        } else if (exp instanceof Binary) {
            Binary binary = (Binary) exp;
            Expr left = binary.getLeft();
            Expr right = binary.getRight();
            binary.setLeft(reGroupHelper(left, boundedVars));
            binary.setRight(reGroupHelper(right, boundedVars));
            return binary;
        } else if (exp instanceof ArgList) {
            ArgList arglist = (ArgList) exp;
            Expr[] args = arglist.getArgs();
            Expr[] newArgs = new Expr[args.length];
            for (int i = 0; i < args.length; i++) {
                Expr arg = args[i];
                newArgs[i] = reGroupHelper(arg, boundedVars);
            }
            arglist.setArgs(newArgs);
            return arglist;
        } else {
            return exp;
        }
    }

    private static And reGroupAnd(And oldAnd, Vector<Identifier> boundedVars) {
        Vector<Expr> elements = new Vector<Expr>();
        
        Vector<Expr> varUnary = new Vector<Expr>();
        Vector<Expr> unaryPredicates = new Vector<Expr>();
        /*
                 Vector<Expr> singleBinaryPredicates = new Vector<Expr>();
         */
        Vector<Expr> doubleBinaryPredicates = new Vector<Expr>();
        Vector<Expr> others = new Vector<Expr>();

        andSplitter(oldAnd, boundedVars, elements);

        for (int i = 0; i < elements.size(); i++) {
            Expr element = elements.get(i);
            if (isVarUnary(element, boundedVars)) {
                varUnary.add(element);
            } else if (isUnaryPredicate(element, boundedVars)) {
                unaryPredicates.add(0, element);
            } else if (isDoubleBinaryPredicate(element, boundedVars)) {
                doubleBinaryPredicates.add(element);
            } else {
                others.add(element);
            }
        }

        Vector<Vector<Expr>> exps = new Vector<Vector<Expr>>();
        exps.add(varUnary);
        exps.add(unaryPredicates);
        exps.add(others);
        exps.add(doubleBinaryPredicates);

        Vector<Expr> allExps = merge(exps);

        Expr result = constructAnd(allExps);

        return (And) result;
    }

    private static Vector<Expr> merge(Vector<Vector<Expr>> vs) {
        Vector<Expr> allExps = new Vector<Expr>();
        for (int i = 0; i < vs.size(); i++) {
            Vector<Expr> exps = vs.elementAt(i);
            for (int j = 0; j < exps.size(); j++) {
                allExps.add(exps.elementAt(j));
            }
        }
        return allExps;
    }

    // P Q R => P ^ ( Q ^ R)
    private static Expr constructAnd(Vector<Expr> elements) {
        Expr exp = elements.remove(elements.size() - 1);
        if (elements.size() == 0) {
            return exp;
        }

        return new And(constructAnd(elements), exp);
    }

    private static void andSplitter(And oldAnd, Vector<Identifier> boundedVars, Vector<Expr> elements) {

        Expr left = oldAnd.getLeft().stripParens();
        Expr right = oldAnd.getRight().stripParens();

        if (left.getExprCat() == Expr.AND) {
            andSplitter((And) left, boundedVars, elements);
        } else {
            elements.add(reGroupHelper(left, boundedVars));
        }

        if (right.getExprCat() == Expr.AND) {
            andSplitter((And) right, boundedVars, elements);
        } else {
            elements.add(reGroupHelper(right, boundedVars));
        }

    }

    /**
     * check if it is f(x)
     */
    private static boolean isVarUnary(Expr exp, Vector<Identifier> vars) {
        Identifier var = vars.get(0);
        exp = exp.stripParens();
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr func = funapp.getFunc();
            Expr arg = funapp.getArg();
            if (!vars.contains(func)) {
                return false;
            }
            if (arg.equals(var)) {
                return true;
            }
        }
        return false;
    }

    /**
     * check if it is p(x,y)
     */
    private static boolean isDoubleBinaryPredicate(Expr exp, Vector<Identifier> vars) {
        Identifier var = vars.get(0);
        exp = exp.stripParens();
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr func = funapp.getFunc();
            if (vars.contains(func)) {
                return false;
            }
            Expr arg = funapp.getArg();
            if (arg.getExprCat() == Expr.ARGLIST) {
                Expr[] args = ((ArgList) arg).getArgs();
                if (args.length == 2) {
                    boolean flag1 = vars.contains(args[0]) && vars.contains(args[1]) && args[0] != args[1];
                    boolean flag2 = var.equals(args[0]) || var.equals(args[1]);
                    return flag1 && flag2;
                }
            }
        }
        return false;
    }
    
    /**
     * check if it is p(x)
     */
    private static boolean isUnaryPredicate(Expr exp, Vector<Identifier> vars) {
        Identifier var = vars.get(0);
        exp = exp.stripParens();
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr func = funapp.getFunc();
            Expr arg = funapp.getArg();
            if (vars.contains(func)) {
                return false;
            }
            if (arg.equals(var)) {
                return true;
            }
        }
        return false;
    }
    
    public static void main(String args[])throws IOException{
    	
    	PrintStream ps = new PrintStream(System.out, true, "UTF-8");
    	
    	String input = "(lambda $0 e (and (state:t $0) (next_to:t $0 texas:s)))";
    	
    	String output = preprocess(input);
    	
    	ps.println(input);
    	ps.println(output);
    	
    }
    
    private static Vector vars = new Vector();

    private static String genChar(String s) {
        String var = "x-";
        if (!vars.contains(s)) {
            var += vars.size();
            vars.add(s);
        } else {
            var += vars.indexOf(s);
        }
        return var;
    }

    
    public static String preprocess(String input) {
        StringBuffer sb;

        Node shallowTree = toTree(input);

        sb = preprocessHelper(shallowTree);

        return sb.toString();
    }
    
    private static StringBuffer preprocessHelper(Node node) {
        StringBuffer sb = new StringBuffer();
        if (node.getNodeName().startsWith("$")) {
            String var = genChar(node.getNodeName());
            sb.append(var);
            return sb;
        }

        String nodeName = node.getNodeName();
        if (nodeName.equals("and")) {
            NodeVector children = node.getChildren();
            for (int i = 0; i < children.size(); i++) {
                Node child = children.getNode(i);
                if (i != 0) {
                    sb.append('&');
                }
                sb.append(preprocessHelper(child));
            }
            return sb;
        }

        if (nodeName.equals("or")) {
            NodeVector children = node.getChildren();
            for (int i = 0; i < children.size(); i++) {
                Node child = children.getNode(i);
                if (i != 0) {
                    sb.append('|');
                }
                sb.append(preprocessHelper(child));
            }
            return sb;
        }

        if (nodeName.equals("not")) {
            NodeVector children = node.getChildren();

            Node child1 = children.getNode(0);

            sb.append('~');
            sb.append(preprocessHelper(child1));

            return sb;
        }

        if (nodeName.equals("=")) {
            NodeVector children = node.getChildren();

            Node child1 = children.getNode(0);
            sb.append(preprocessHelper(child1));

            sb.append('=');

            Node child2 = children.getNode(1);
            sb.append(preprocessHelper(child2));

            return sb;
        }

        if (nodeName.equals(">")) {
            NodeVector children = node.getChildren();

            Node child1 = children.getNode(0);
            sb.append(preprocessHelper(child1));

            sb.append('>');
            sb.append('>');

            Node child2 = children.getNode(1);
            sb.append(preprocessHelper(child2));

            return sb;
        }

        if (nodeName.equals("<")) {
            NodeVector children = node.getChildren();

            Node child1 = children.getNode(0);
            sb.append(preprocessHelper(child1));

            sb.append('<');
            sb.append('<');

            Node child2 = children.getNode(1);
            sb.append(preprocessHelper(child2));

            return sb;
        }

        if (nodeName.equals("exists")) {
            NodeVector children = node.getChildren();
            sb.append('E');
            for (int i = 0; i < children.size(); i++) {
                Node child = children.getNode(i);
                sb.append(preprocessHelper(child));
                sb.append(' ');
            }
            return sb;
        }

        if (nodeName.equals("all")) {
            NodeVector children = node.getChildren();
            sb.append('A');
            for (int i = 0; i < children.size(); i++) {
                Node child = children.getNode(i);
                sb.append(preprocessHelper(child));
                sb.append(' ');
            }
            return sb;
        }

        if (nodeName.equals("count")) {
            NodeVector children = node.getChildren();
            sb.append('C');
            for (int i = 0; i < children.size(); i++) {
                Node child = children.getNode(i);
                sb.append(preprocessHelper(child));
                sb.append(' ');
            }
            return sb;
        }

        if (nodeName.equals("(")) {
            NodeVector children = node.getChildren();
            sb.append('(');
            for (int i = 0; i < children.size(); i++) {
                Node child = children.getNode(i);
                sb.append(preprocessHelper(child));
            }
            sb.append(')');
            return sb;
        }

        if (nodeName.equals("lambda")) {
            NodeVector children = node.getChildren();
            sb.append('L');
            Node var = children.getNode(0);
            sb.append(preprocessHelper(var));
            for (int i = 1; i < children.size(); i++) {
                Node child = children.getNode(i);
                if (child.getNodeName().equals("e")) {
                    continue;
                }
                sb.append(preprocessHelper(child));
            }
            return sb;
        }

        if (nodeName.equals("the")) {
            NodeVector children = node.getChildren();
            sb.append('I');
            Node var = children.getNode(0);
            sb.append(preprocessHelper(var));
            for (int i = 1; i < children.size(); i++) {
                Node child = children.getNode(i);
                sb.append(preprocessHelper(child));
            }
            return sb;
        }

        if (nodeName.equals("sum")) {
            NodeVector children = node.getChildren();
            sb.append('S');
            if (children.size() == 3) {
                Node var = children.getNode(0);
                sb.append(preprocessHelper(var));
                sb.append('.');
                sb.append("sum_e");
                sb.append('(');
                sb.append('L');
                sb.append(preprocessHelper(var));
                Node child1 = children.getNode(1);
                sb.append(preprocessHelper(child1));
                sb.append(',');
                sb.append('L');
                sb.append(preprocessHelper(var));
                Node child2 = children.getNode(2);
                sb.append(preprocessHelper(child2));
                sb.append(')');
            } else {
                System.err.println("max number of args is 2 for: ");
                System.exit(1);
            }
            return sb;
        }

        if (nodeName.equals("max")) {
            NodeVector children = node.getChildren();
            sb.append('M');
            if (children.size() == 2) {
                Node var = children.getNode(0);
                sb.append(preprocessHelper(var));
                sb.append('.');
                sb.append("max_e");
                sb.append('(');
                Node child1 = children.getNode(1);
                sb.append(preprocessHelper(child1));
                sb.append(')');
            } else {
                System.err.println("max number of args is 1 for: ");
                System.exit(1);
            }
            return sb;
        }

        if (nodeName.equals("min")) {
            NodeVector children = node.getChildren();
            sb.append('V');
            if (children.size() == 2) {
                Node var = children.getNode(0);
                sb.append(preprocessHelper(var));
                sb.append('.');
                sb.append("min_e");
                sb.append('(');
                Node child1 = children.getNode(1);
                sb.append(preprocessHelper(child1));
                sb.append(')');
            } else {
                System.err.println("max number of args is 1 for: ");
                System.exit(1);
            }
            return sb;
        }

        if (nodeName.equals("argmax")) {
            NodeVector children = node.getChildren();
            sb.append('N');
            if (children.size() == 3) {
                Node var = children.getNode(0);
                sb.append(preprocessHelper(var));
                sb.append('.');
                sb.append("argmax_e");
                sb.append('(');
                sb.append('L');
                sb.append(preprocessHelper(var));
                Node child1 = children.getNode(1);
                sb.append(preprocessHelper(child1));
                sb.append(',');
                sb.append('L');
                sb.append(preprocessHelper(var));
                Node child2 = children.getNode(2);
                sb.append(preprocessHelper(child2));
                sb.append(')');
            } else {
                System.err.println("max number of args is 2 for: ");
                System.exit(1);
            }
            return sb;
        }

        if (nodeName.equals("argmin")) {
            NodeVector children = node.getChildren();
            sb.append('U');
            if (children.size() == 3) {
                Node var = children.getNode(0);
                sb.append(preprocessHelper(var));
                sb.append('.');
                sb.append("argmin_e");
                sb.append('(');
                sb.append('L');
                sb.append(preprocessHelper(var));
                Node child1 = children.getNode(1);
                sb.append(preprocessHelper(child1));
                sb.append(',');
                sb.append('L');
                sb.append(preprocessHelper(var));
                Node child2 = children.getNode(2);
                sb.append(preprocessHelper(child2));
                sb.append(')');
            } else {
                System.err.println("min number of args is 2 for: ");
                System.exit(1);
            }
            return sb;
        }

        //general case...
        {
            NodeVector children = node.getChildren();
            String varName = toValidExpVar(nodeName, children);
            sb.append(varName);

            if (children.size() == 0) {
            } else if (children.size() == 1 || children.size() == 2) {
                sb.append('(');
                Node child1 = children.getNode(0);
                sb.append(preprocessHelper(child1));
                if (children.size() == 2) {
                    sb.append(',');
                    Node child2 = children.getNode(1);
                    sb.append(preprocessHelper(child2));
                }
                sb.append(')');
            } else {
                System.err.println("max number of args is 2 for: " + varName);
                System.exit(1);
            }
            return sb;
        }
    }

    private static String toValidExpVar(String s, NodeVector children) {
        StringBuffer sb = new StringBuffer();
        char[] ch = s.toCharArray();
        String type = "t";
        for (int i = 0; i < ch.length; i++) {
            char c = ch[i];
            if (c == '_') {
                sb.append('-');
            } else if (c == ':') {
                type = s.substring(i + 1).trim();
                break;
            } else {
                sb.append(c);
            }
        }

        if (children.size() == 1) {
            if (type.equals("e") || type.equals("t")) {
                sb.append("_<e," + type + ">");
            } else {
                sb.append("-" + type + "_<e,e>");
            }
        } else if (children.size() == 2) {
            sb.append("_<e*e," + type + ">");
        } else {
            if (type != "e") {
                sb.append('-');
                sb.append(type);
                sb.append('_');
                sb.append('e');
            } else {
                sb.append('_');
                sb.append('e');
            }
        }

        String result = sb.toString();
        char first = result.charAt(0);
        if (Character.isDigit(first)) {
            result = "number-" + result;
        }

        return result;
    }

    public static Node toTree(String mrl) {
        if (mrl.startsWith("(") && mrl.endsWith(")")) {
            mrl = mrl.substring(1, mrl.length() - 1);
            Vector tokens = tokenize(mrl);
            String rootToken = (String) tokens.elementAt(0);
            Node rootNode = new Node(rootToken);
            for (int i = 1; i < tokens.size(); i++) {
                String token = (String) tokens.elementAt(i);
                Node childNode = toTree(token);
                rootNode.addChildren(childNode);
                childNode.setParent(rootNode);
            }
            Node branNode = new Node("(");
            NodeVector branChildren = new NodeVector();
            branChildren.add(rootNode);
            branNode.setChildren(branChildren);
            rootNode.setParent(branNode);
            return branNode;
        } else {
            Node rootNode = new Node(mrl);
            return rootNode;
        }
    }

    private static Vector tokenize(String s) {
        Vector results = new Vector();
        while (true) {
            String result[] = split(s);
            results.add(result[0]);

            if (result[1].equals("")) {
                break;
            } else {
                s = result[1];
            }
        }
        return results;
    }

    private static String[] split(String s) {
        if (s.startsWith("(")) {
            int lb = 1;
            int sp = -1;
            for (int i = 1; i < s.length(); i++) {
                char ch = s.charAt(i);
                if (ch == '(') {
                    lb++;
                } else if (ch == ')') {
                    lb--;
                }
                //System.err.println(ch+"\t"+lb);
                if (lb == 0) {
                    sp = i + 1;
                    break;
                }
            }
            //System.err.println(s+","+sp);
            String[] results = {s.substring(0, sp).trim(), s.substring(sp).trim()};
            return results;
        } else {
            int sp = s.length();
            for (int i = 0; i < s.length(); i++) {
                char ch = s.charAt(i);
                if (ch == '(' || ch == ' ') {
                    sp = i;
                    break;
                }
            }
            String[] results = {s.substring(0, sp).trim(), s.substring(sp).trim()};
            return results;
        }
    }

    private static class Node {

        private String node;
        private Node parent;
        private NodeVector children;

        public Node(String node) {
            this.node = node;
            this.children = new NodeVector();
        }

        public void setParent(Node parent) {
            this.parent = parent;
        }

        public void setChildren(NodeVector children) {
            this.children = children;
        }

        public void addChildren(Node child) {
            this.children.add(child);
        }

        public NodeVector getChildren() {
            return this.children;
        }

        public int countChildrenSize() {
            return this.children.size();
        }

        public Node getChildAt(int i) {
            return this.children.getNode(i);
        }

        public String getNodeName() {
            return this.node;
        }

        @Override
		public String toString() {
            return toString(0).toString();
        }

        public StringBuffer toString(int indent) {
            StringBuffer sb = new StringBuffer();

            for (int i = 0; i < indent; i++) {
                sb.append('\t');
            }
            sb.append(this.node);
            sb.append('\n');

            for (int i = 0; i < this.countChildrenSize(); i++) {
                Node child = this.getChildAt(i);
                sb.append(child.toString(indent + 1).toString());
            }

            if (this.node.equals("(")) {
                for (int i = 0; i < indent; i++) {
                    sb.append('\t');
                }
                sb.append(')');
                sb.append('\n');
            }

            return sb;
        }

        private void collectConstantsHelper(Vector<String>consts, Node n){
            String name = n.getNodeName();
            int index = name.indexOf(":");
            if(index!=-1){
                String type = name.substring(index+1).trim();
                if(!type.equals("i") && !type.equals("e") && !type.equals("t")){
                    consts.add(n.getNodeName());
                }
            }

            Vector<Node> cNodes = n.getChildren();
            for(int i=0; i<cNodes.size(); i++){
                Node cnode = cNodes.get(i);
                this.collectConstantsHelper(consts, cnode);
            }

        }

    }


    private static class NodeVector extends Vector {
        public NodeVector() {
            super();
        }

        public Node getNode(int i) {
            return (Node)super.elementAt(i);
        }
    }

}
