/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.data;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;

import lambdacalc.logic.*;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaGProd implements Serializable, Comparable<LambdaGProd>{
	
	private static final long serialVersionUID = -5642535085653859218L;
	
	/**
	 * Static fields
	 */
	public static final int _TERM = 0; //terminal, i.e. no RHS : Expr
	public static final int _APP0 = 1; //no application, i.e. 1 RHS : (Y)
	public static final int _APP1 = 2; //one application, i.e. 1 RHS : (Expr Y)
	public static final int _APP2 = 3; //two applications, normal order, i.e. 2 RHS : ((Expr Y) Z)
	public static final int _APP3 = 4; //one application, i.e. 2 RHS : (Y Z)
	
	/**
	 * Private object fields
	 */
	private int _id = -1;
	private int _constructType;
	private TypeToken _lhsT;
	private ExprToken _exprT;
	private TypeToken[] _rhsT;
	
	//a pre-defined small value
	private static double _mu = 0.1;
	private static double _phi = 1E-100;
//	private static double _alpha_trans = 0.00001;//the Dirichlet prior for transition 
//	private static double _alpha_pattern = 0.0001;//the Dirichlet prior for pattern
//	private static double _alpha_emission = 0.001;//the Dirichlet prior for emission
	
	/**
	 * Check if it is Lx.x
	 * @return
	 * @throws TypeEvaluationException 
	 */
	public boolean isLxx(){
		if(this.isRoot()){
			return false;
		}
		
		if(this._exprT == null){
			return false;
		}
		if(this.getArity()!=1){
			return false;
		}
		
		Expr exp = this._exprT.getExpr();
        Var x = new Var("u", Type.E, true);
        Expr Lxx = new Lambda(x, x, true);
        if(Lxx.alphaEquivalent(exp)){
        	return true;
        }
        return false;
	}
	
	/**
	 * Constructor
	 * @param id id of this GMRToken
	 * @param lambdaCalc the lambda calculus expression
	 * @param type type of the LHS
	 * @param childrenTypes the types of the RHS nonterminals
	 * @param constructType
	 */
	public LambdaGProd(int constructType, TypeToken lhsT, ExprToken exprT, TypeToken[] rhsT){
		this._constructType = constructType;
		this._lhsT = lhsT;
		this._exprT = exprT;
		this._rhsT = rhsT;
	}
	
	public boolean isRoot(){
		return this._lhsT.equals(TypeToken._rootToken);
	}
	
	/**
	 * Get the arity of the gm. 
	 * i.e. the number of RHS type tokens.
	 * @return the arity of this gm.
	 */
	public int getArity(){
		return this._rhsT.length;
	}
	
	public void setId(int id){
		this._id = id;
	}
	
	public int getId(){
		return this._id;
	}
	
	public int getConstructType(){
		return this._constructType;
	}
	
	public TypeToken getLHSTypeToken(){
		return this._lhsT;
	}
	
	public TypeToken[] getRHSTypeTokens(){
		return this._rhsT;
	}
	
	public ExprToken getExprToken(){
		return this._exprT;
	}
	
	@Override
	public boolean equals(Object o){
		if(o instanceof LambdaGProd){
			LambdaGProd gm = (LambdaGProd)o;
			if(this.compareTo(gm)==0){
				return true;
			}
			return false;
		}
		return false;
	}
	
	/**
	 * This method tries to order the LambdaGMRProds..
	 * @param lgm
	 * @return
	 */
	@Override
	public int compareTo(LambdaGProd gm){
		
		//if both ids are already assigned.
		if(this._id!=-1 && gm._id!=-1){
			if(this._id > gm._id){
				return +1;
			} else if(this._id < gm._id){
				return -1;
			} else {
				return 0;
			}
		}
		
		int val;

		//check the lhs
		val = this._lhsT.compareTo(gm._lhsT);
		if(val!=0) return val;
		
		//check the expr
		if(this._exprT!= null && gm._exprT!=null){
			val = this._exprT.compareTo(gm._exprT);
			if(val!=0) return val;
		} else if(this._exprT!=null){
			return +1;
		} else if(gm._exprT!=null){
			return -1;
		}
		
		//check the rhs size
		if(this._rhsT.length > gm._rhsT.length){
			return +1;
		} else if(this._rhsT.length < gm._rhsT.length){
			return -1;
		}
		
		//check the rhs types
		for(int i =0; i<this._rhsT.length; i++){
			val = this._rhsT[i].compareTo(gm._rhsT[i]);
			if(val!=0){
				return val;
			}
		}
		
		//check the construct type
		if(this._constructType > gm._constructType){
			return +1;
		} else if(this._constructType < gm._constructType){
			return -1;
		}
		
		return 0;
	}
	

	/**
	 * The put methods...
	 */
	private void putValue(HashMap<Integer, Double> map, int key, double val){
		if(map.containsKey(key)){
			double oldVal = map.get(key);
			map.put(key, oldVal + val);
		} else {
			map.put(key, val);
		}
	}
	
	public void putPatternCount(LambdaPattern p, double d) {
		this.putValue(this._patternTable_new, p.getId(), d);
	}

	public void putUnigramCount(int word, double d) {
		this.putValue(this._unigramTable_new, word, d);
	}

	public void putBigramCount(int word1, int word2, double d) {
		if(this._bigramTable_new.containsKey(word1)){
			HashMap<Integer, Double> subMap = this._bigramTable_new.get(word1);
			this.putValue(subMap, word2, d);
		} else {
			HashMap<Integer, Double> subMap = new HashMap<Integer, Double>();
			this._bigramTable_new.put(word1, subMap);
			this.putValue(subMap, word2, d);
		}
	}

	public void putMRTransCount(int pos, int id, double d) {
		if(pos == 0){
			this.putValue(this._mrTable_a_new, id, d);
		} else {
			this.putValue(this._mrTable_b_new, id, d);
		}
	}
	
	/**
	 * The get methods...
	 */
	public double getPatternProb_train(LambdaPattern p) {
		
		if(this.isLxx()){
			if(p.equals(LambdaPattern._X)){
				return 1.0;
			}
			return 0.0;
		}
		
		//in case it's the first time to run...
		if(this._patternTable == null){
			int arity = this.getArity();
			if(arity == 0){
				if(p.getArity() == 0){
					return 1.0;
				}
				return 0.0;
			} else if(arity == 1){
				if(p.getArity() == 1){
//					return 1.0;
					return 1.0 / LambdaPattern._patterns_arity_1.length;
				}
				return 0.0;
			} else if(arity == 2){
				if(p.getArity() == 2){
//					return 1.0;
					return 1.0 / LambdaPattern._patterns_arity_2.length;
				}
				return 0.0;
			}
		}
		return this.getVal(this._patternTable, p.getId());
	}
	
	/**
	 * The get methods...
	 */
	public double getPatternProb_test(LambdaPattern p) {
		
		if(this.isLxx()){
			if(p.equals(LambdaPattern._X)){
				return 1.0;
			}
			return 0.0;
		}
		
		//if the table is not actually learned, assume uniform..
		if(this._patternTable == null){
			int arity = this.getArity();
			if(arity == 0){
				if(p.getArity() == 0){
					return 1.0;
				}
				return 0.0;
			} else if(arity == 1){
				if(p.getArity() == 1){
					return 1.0 / LambdaPattern._patterns_arity_1.length;
				}
				return 0.0;
			} else if(arity == 2){
				if(p.getArity() == 2){
					return 1.0 / LambdaPattern._patterns_arity_2.length;
				}
				return 0.0;
			}
		}
		//if it's already learned, use the learned value, but also return a small value if assigned zero.
		double prob = this.getVal(this._patternTable, p.getId());
		return prob > _phi ? prob : _phi;
	}

	public double getUnigramProb_train(int word) {
		
		//in case it's the first time to run...
		if(this._unigramTable == null){
			//a pre-defined small value...
			return _mu;
		}
		
		return this.getVal(this._unigramTable, word);
	}

	public double getUnigramProb_test(int word) {
		if(this._unigramTable == null){
			return 0;
		}
		
		double prob = this.getVal(this._unigramTable, word);
		return prob > _phi ? prob : _phi;
	}

	public double getBigramProb_train(int word1, int word2) {
		
		/**
		 * in case it's the first time to run...
		 * 
		 * The story is as follows:
		 * if bigram table is null, it means it's the first time to run
		 * then check the unigram table.
		 * 
		 * if unigram table is again null, it means ibm1 is not run.
		 * then return the default value...
		 * 
		 * if unigram table is not null, return the value from there as a back-off model..
		 * 
		 */
		if(this._bigramTable == null){
			if(this._unigramTable == null){
				//a pre-defined small value...
				return _mu;
			} else {
				return this.getUnigramProb_train(word2);
			}
		}

		if(this._bigramTable.containsKey(word1)){
			HashMap<Integer, Double> subMap = this._bigramTable.get(word1);
			if(subMap.containsKey(word2)){
				return subMap.get(word2);
			} else {
				return 0;
			}
		} else {
			return 0;
		}
	}
	
	public double getBigramProb_test(int word1, int word2) {

		//should i do back-off??
		if(this._bigramTable == null){
			//a pre-defined small value...
			return _phi;
		}
		
		if(this._bigramTable.containsKey(word1)){
			HashMap<Integer, Double> subMap = this._bigramTable.get(word1);
			if(subMap.containsKey(word2)){
				double prob = subMap.get(word2);
				return prob > _phi ? prob : _phi;
			} else {
				return _phi;
			}
		} else {
			return _phi;
		}
	}
	
	public double getMRTransProb_train(int pos, LambdaGProd m2) {
		
		//in case it's the first time to run...
		if(this._mrTable_a == null && this._mrTable_b == null){
			if(this.getRHSTypeTokens()[pos].equals(m2.getLHSTypeToken())){
				//a pre-defined small value...
				return _mu;
			} else {
				return 0;
			}
		}
		
		if(pos == 0){
			return this.getVal(this._mrTable_a, m2.getId());
		} else {
			return this.getVal(this._mrTable_b, m2.getId());
		}
	}
	
	
	public double getMRTransProb_test(int pos, LambdaGProd m2) {
		
		if(this._mrTable_a == null && this._mrTable_b == null){
			if(this.getRHSTypeTokens()[pos].equals(m2.getLHSTypeToken())){
				//a pre-defined small value...
				return _phi;
			} else {
				return 0;
			}
		}
		
		if(pos == 0){
			double prob = this.getVal(this._mrTable_a, m2.getId());
			if(prob > _phi){
				return prob;
			}
			if(this.getRHSTypeTokens()[pos].equals(m2.getLHSTypeToken())){
				//a pre-defined small value...
				return _phi;
			} else {
				return 0;
			}
		} else {
			double prob = this.getVal(this._mrTable_b, m2.getId());
			if(prob > _phi){
				return prob;
			}
			if(this.getRHSTypeTokens()[pos].equals(m2.getLHSTypeToken())){
				//a pre-defined small value...
				return _phi;
			} else {
				return 0;
			}
		}
	}
	
	private double getVal(HashMap<Integer, Double> map, int key){
		Double val = map.get(key);
		if(val == null){
			return 0;
		}
		return val;
	}
	
	/**
	 * normalization for the HashMap...
	 * @param map
	 */
	private boolean norm(HashMap<Integer, Double> map){
		if(map.size() == 0){
			return false;
		}
		Iterator<Integer> keys;
		keys = map.keySet().iterator();
		double sum = 0;
		while(keys.hasNext()){
			int key = keys.next();
			double val = map.get(key);
			sum += val;
		}
		keys = map.keySet().iterator();
		while(keys.hasNext()){
			int key = keys.next();
			double val = map.get(key);
			map.put(key, val / sum);
		}
		return true;
	}
	
	private boolean norm2(HashMap<Integer, HashMap<Integer, Double>> map){
		if(map.size() == 0){
			return false;
		}
		Iterator<Integer> keys = map.keySet().iterator();
		while(keys.hasNext()){
			int key = keys.next();
			HashMap<Integer, Double> subMap = map.get(key);
			this.norm(subMap);
		}
		return true;
	}
	
	/**
	 * Update function.
	 */
	public void updates(){
		
		if(this.norm(this._patternTable_new)){
			this._patternTable = this._patternTable_new;
			this._patternTable_new = new HashMap<Integer, Double>();
		}

		if(this.norm(this._unigramTable_new)){
			this._unigramTable = this._unigramTable_new;
			this._unigramTable_new = new HashMap<Integer, Double>();
		}
		
		if(this.norm2(this._bigramTable_new)){
			this._bigramTable = this._bigramTable_new;
			this._bigramTable_new = new HashMap<Integer, HashMap<Integer, Double>>();
		}
		
		if(this.norm(this._mrTable_a_new)){
			this._mrTable_a = this._mrTable_a_new;
			this._mrTable_a_new = new HashMap<Integer, Double>();
		}

		if(this.norm(this._mrTable_b_new)){
			this._mrTable_b = this._mrTable_b_new;
			this._mrTable_b_new = new HashMap<Integer, Double>();
		}
		
	}
	
	public void initProbs(){
		//for the patterns..
		int arity = this.getArity();
		if(arity == 0){
			this._patternTable.put(LambdaPattern._w.getId(), 1.0);
		} else if(arity == 1){
			LambdaPattern[] patterns = LambdaPattern._patterns_arity_1;
			for(int i = 0; i<patterns.length; i++){
				this._patternTable.put(patterns[i].getId(), 1.0/patterns.length);
			}
		} else if(arity == 2){
			LambdaPattern[] patterns = LambdaPattern._patterns_arity_2;
			for(int i = 0; i<patterns.length; i++){
				this._patternTable.put(patterns[i].getId(), 1.0/patterns.length);
			}
		}
	}
	
	@Override
	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append(this.getId()+" + \t");
		sb.append(this._lhsT.toString());
		sb.append(" => ");
		if(this._exprT != null){
			sb.append(this._exprT.toString());
		}
		for(int i =0; i<this._rhsT.length; i++){
			sb.append(' ');
			sb.append(this._rhsT[i].toString());
		}
		return sb.toString();
	}

	/**
	 * Fields for tables storing probabilities.
	 */
	private HashMap<Integer, Double> _patternTable = null;
	private HashMap<Integer, Double> _unigramTable = null;
	private HashMap<Integer, HashMap<Integer, Double>> _bigramTable = null;
	private HashMap<Integer, Double> _mrTable_a = null;
	private HashMap<Integer, Double> _mrTable_b = null;
	
	private HashMap<Integer, Double> _patternTable_new = new HashMap<Integer, Double>();
	private HashMap<Integer, Double> _unigramTable_new = new HashMap<Integer, Double>();
	private HashMap<Integer, HashMap<Integer, Double>> _bigramTable_new = new HashMap<Integer, HashMap<Integer, Double>>();
	private HashMap<Integer, Double> _mrTable_a_new = new HashMap<Integer, Double>();
	private HashMap<Integer, Double> _mrTable_b_new = new HashMap<Integer, Double>();
	
	/**
	 * Private methods...
     * For serialization
     */
	private void writeObject(ObjectOutputStream out) throws IOException{
		out.writeInt(this._id);
		out.writeInt(this._constructType);
		out.writeObject(this._exprT);
		out.writeObject(this._lhsT);
		out.writeInt(this._rhsT.length);
		for(int i =0 ; i< this._rhsT.length; i++){
			out.writeObject(this._rhsT[i]);
		}
		//patternTable
		this.writeHashMap(this._patternTable, out);
		//unigramTable
		this.writeHashMap(this._unigramTable, out);
		//bigramTable
		this.writeHashMap2(this._bigramTable, out);
		//mrTable_a
		this.writeHashMap(this._mrTable_a, out);
		//mrTable_b
		this.writeHashMap(this._mrTable_b, out);
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
		this._id = in.readInt();
		this._constructType = in.readInt();
		this._exprT = (ExprToken)in.readObject();
		this._lhsT = (TypeToken)in.readObject();
		this._rhsT = new TypeToken[in.readInt()];
		for(int i =0; i<this._rhsT.length; i++){
			this._rhsT[i] = (TypeToken)in.readObject();
		}
		//patternTable
		this._patternTable = this.readHashMap(in);
		//unigramTable
		this._unigramTable = this.readHashMap(in);
		//bigramTable
		this._bigramTable = this.readHashMap2(in);
		//mrTable_a
		this._mrTable_a = this.readHashMap(in);
		//mrTable_b
		this._mrTable_b = this.readHashMap(in);
		
		//important: initialize the new tables..
		this._patternTable_new = new HashMap<Integer, Double>();
		this._unigramTable_new = new HashMap<Integer, Double>();
		this._bigramTable_new = new HashMap<Integer, HashMap<Integer, Double>>();
		this._mrTable_a_new = new HashMap<Integer, Double>();
		this._mrTable_b_new = new HashMap<Integer, Double>();
	}

	/**
	 * aux read/write methods for HashMaps...
	 */
	private void writeHashMap(HashMap<Integer, Double> map, ObjectOutputStream out) throws IOException{
		if (map == null) {
			out.writeInt(-1);
			return;
		}
		out.writeInt(map.size());
		Iterator<Integer> ids = map.keySet().iterator();
		while(ids.hasNext()){
			int id = ids.next();
			double val = map.get(id);
			out.writeInt(id);
			out.writeDouble(val);
		}
	}

	private void writeHashMap2(HashMap<Integer, HashMap<Integer, Double>> map, ObjectOutputStream out) throws IOException{
		if(map == null){
			out.writeInt(-1);
			return;
		}
		out.writeInt(map.size());
		Iterator<Integer> ids = map.keySet().iterator();
		while(ids.hasNext()){
			int id = ids.next();
			HashMap<Integer, Double> subMap = map.get(id);
			out.writeInt(id);
			this.writeHashMap(subMap, out);
		}
	}
	
	private HashMap<Integer, Double> readHashMap(ObjectInputStream in) throws IOException{
		int size = in.readInt();
		if(size == -1){
			return null;
		}
		HashMap<Integer, Double> map = new HashMap<Integer, Double>();
		for(int i = 0; i<size; i++){
			int key = in.readInt();
			double val = in.readDouble();
			map.put(key, val);
		}
		return map;
	}
	
	private HashMap<Integer, HashMap<Integer, Double>> readHashMap2(ObjectInputStream in) throws IOException{
		int size = in.readInt();
		if(size == -1){
			return null;
		}
		HashMap<Integer, HashMap<Integer, Double>> map = new HashMap<Integer, HashMap<Integer, Double>>();
		for(int i = 0; i<size; i++){
			int id = in.readInt();
			HashMap<Integer, Double> subMap = this.readHashMap(in);
			map.put(id, subMap);
		}
		return map;
	}


	
	/**
	 * Some other methods.
	 * @throws UnsupportedEncodingException 
	 */
	public void viewBigramTable(Vocabulary voc) throws UnsupportedEncodingException{
		
		PrintStream ps = new PrintStream(System.out, true, "UTF-8");
		StringBuffer sb = new StringBuffer();
		
		sb.append(this.toString());
		sb.append('\n');
		
		if(this._bigramTable==null){
			sb.append("-EMPTY-");
			ps.println(sb.toString());
			return;
		}
		
		ArrayList<keyPair_value> kvs = new ArrayList<keyPair_value>();
		Iterator<Integer> words1 = this._bigramTable.keySet().iterator();
		while(words1.hasNext()){
			int word1 = words1.next();
			Word w1 = voc.getWordById(word1);
			HashMap<Integer, Double> subMap = this._bigramTable.get(word1);
			
			Iterator<Integer> words2 = subMap.keySet().iterator();
			while(words2.hasNext()){
				int word2 = words2.next();
				double val = subMap.get(word2);
				Word w2 = voc.getWordById(word2);
				keyPair_value kv = new keyPair_value(w1, w2, val);
				
				int index = Collections.binarySearch(kvs, kv);
				if(index < 0){
					kvs.add(-1-index, kv);
				} else {
					throw new RuntimeException("Identical!"+kvs.get(index)+"\t"+kv);
				}
			}
		}
		
		for(int i =kvs.size() - 1; i >= Math.max(0, kvs.size()-kvs.size()); i--){
			sb.append(kvs.get(i));
			sb.append('\n');
		}
		
		ps.println(sb.toString());
	}
	
	/**
	 * Some other methods.
	 * @throws UnsupportedEncodingException 
	 */
	public void viewUnigramTable(Vocabulary voc) throws UnsupportedEncodingException{
		
		PrintStream ps = new PrintStream(System.out, true, "UTF-8");
		StringBuffer sb = new StringBuffer();
		
		sb.append(this.toString());
		sb.append('\n');
		
		if(this._unigramTable==null){
			sb.append("-EMPTY-");
			ps.println(sb.toString());
			return;
		}
		
		ArrayList<key_value> kvs = new ArrayList<key_value>();
		Iterator<Integer> keys = this._unigramTable.keySet().iterator();
		while(keys.hasNext()){
			int key = keys.next();
			Word word = voc.getWordById(key);
			double val = this._unigramTable.get(key);
			key_value kv = new key_value(word, val);
			int index = Collections.binarySearch(kvs, kv);
			if(index < 0){
				kvs.add(-1-index, kv);
			} else {
				throw new RuntimeException("Identical!"+kvs.get(index)+"\t"+kv);
			}
		}
		
		sb.append("size:"+this._unigramTable.size());
		sb.append('\n');
		for(int i =kvs.size() - 1; i >= Math.max(0, kvs.size() - 11); i--){
			sb.append(kvs.get(i));
			sb.append('\n');
		}
		
		ps.println(sb.toString());
	}

	class keyPair_value implements Comparable<keyPair_value>{
		public Word _word1;
		public Word _word2;
		public double _val;
		public keyPair_value(Word word1, Word word2, double val){
			this._word1 = word1;
			this._word2 = word2;
			this._val = val;
		}
		@Override
		public int compareTo(keyPair_value kv) {
			if(this._val > kv._val){
				return +1;
			} else if(this._val < kv._val){
				return -1;
			}
			if(this._word1.getId() > kv._word1.getId()){
				return +1;
			} else if(this._word1.getId() < kv._word1.getId()){
				return -1;
			}
			if(this._word2.getId() > kv._word2.getId()){
				return +1;
			} else if(this._word2.getId() < kv._word2.getId()){
				return -1;
			}
			return 0;
		}
		
		@Override
		public String toString(){
			StringBuffer sb = new StringBuffer();
			sb.append(this._word1.getWord());
			sb.append('\t');
			sb.append(this._word2.getWord());
			sb.append('\t');
			sb.append(this._val);
			return sb.toString();
		}
	}
	
	class key_value implements Comparable<key_value>{
		public Word _key;
		public double _val;
		public key_value(Word key, double val){
			this._key = key;
			this._val = val;
		}
		@Override
		public int compareTo(key_value kv) {
			return this._val > kv._val ? +1 : this._val < kv._val ? -1 : this._key.getId() > kv._key.getId() ? +1 : 
				this._key.getId() < kv._key.getId() ? -1 : 0;
		}
		
		@Override
		public String toString(){
			StringBuffer sb = new StringBuffer();
			sb.append(this._key.getWord());
			sb.append('\t');
			sb.append(this._val);
			return sb.toString();
		}
	}
	
}




