/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.data;

import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;

import lambdacalc.logic.Expr;
import lambdacalc.logic.FunApp;
import lambdacalc.logic.TypeEvaluationException;
import lambdacalc.logic.Var;
import lambdagen.learn.LambdaLearnCell;
import lambdagen.viterbi.LambdaAlignCell;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaProd {
	
	private Expr _exp = null;
	
	private int _seq = -1;
	
	/**
	 * Each LambdaMRProd is associated with one LambdaGMRProd
	 */
	private LambdaGProd _gp;
	
	/**
	 * The two list of children LambdaProds, where the second list could be null
	 */
	private ArrayList<LambdaProd> _children1 = null;
	private ArrayList<LambdaProd> _children2 = null;
	
	private LambdaLearnCell _learnCell;
	private LambdaAlignCell _alignCell;
	private ArrayList<LambdaProd> _allProds = null;
	
	public LambdaProd(LambdaGProd gp){
		this._gp = gp;
	}
	
	public void labelSeqs(){
		this._allProds = null;
		ArrayList<LambdaProd> prods = this.getAllProds();
		for(int i = 0; i<prods.size(); i++){
			LambdaProd prod1 = prods.get(i);
			prod1.setSeq(i);
			for(int j = i+1; j<prods.size(); j++){
				LambdaProd prod2 = prods.get(j);
				if(prod1.equals(prod2)){
					prods.remove(j);
					j--;
				}
			}
		}
		
	}
	
	public int getSeq(){
		return this._seq;
	}
	
	public void setSeq(int seq){
		this._seq = seq;
	}
	
	public int countTrees(){
		int arity = this.getArity();
		if(arity == 0){
			return 1;
		} else if(arity == 1){
			int size = 0;
			for(int i = 0; i<this._children1.size(); i++){
				size += this._children1.get(i).countTrees();
			}
			return size;
		} else if(arity == 2){
			int size1 = 0;
			for(int i = 0; i<this._children1.size(); i++){
				size1 += this._children1.get(i).countTrees();
			}
			int size2 = 0;
			for(int i = 0; i<this._children2.size(); i++){
				size2 += this._children2.get(i).countTrees();
			}
			return size1*size2;
		}
		throw new RuntimeException("Invalid !! 44.");
	}
	
	public void setLearnCell(LambdaLearnCell learnCell){
		this._learnCell = learnCell;
	}
	
	public LambdaLearnCell getLearnCell(){
		return this._learnCell;
	}
	
	public void setAlignCell(LambdaAlignCell alignCell){
		this._alignCell = alignCell;
	}
	
	public LambdaAlignCell getAlignCell(){
		return this._alignCell;
	}
	
	public void setChildren(ArrayList<LambdaProd> children){
		this._children1 = children;
	}
	
	public void setChildren(ArrayList<LambdaProd> children1, ArrayList<LambdaProd> children2){
		this._children1 = children1;
		this._children2 = children2;
	}
	
	public LambdaGProd getGP(){
		return this._gp;
	}
	
	/**
	 * Return the number of RHS
	 * @return
	 */
	public int getArity(){
		if(this._children2 != null){
			return 2;
		} 
		if(this._children1 != null){
			return 1;
		}
		return 0;
	}
	
	/** The two lists of children **/
	
	public ArrayList<LambdaProd> getFirstChildren(){
		return this._children1;
	}
	
	public ArrayList<LambdaProd> getSecondChildren(){
		return this._children2;
	}
	
	/** Auxiliary methods for printing the Production 
	 * @throws UnsupportedEncodingException **/
	
	public void viewTree(){
		try{
			PrintStream ps = new PrintStream(System.out, true, "UTF-8");
			ps.println(this.toString());
		}catch(Exception e){
			System.err.println(e);
		}
	}
	
	@Override
	public String toString(){
		StringBuffer sb = new StringBuffer();
		this.toStringHelper(0, sb);
		return sb.toString();
	}
	
	private void toStringHelper(int indent, StringBuffer sb){
		for(int i =0; i<indent; i++){
			sb.append('\t');
		}
		sb.append(this._gp.toString()+"\t("+this.countTrees()+")["+this._seq+"]");
		sb.append('\n');
		if(this._children1!=null){
			for(int i = 0 ; i<this._children1.size(); i++){
				sb.append("C1:"+i+":");
				this._children1.get(i).toStringHelper(indent+1, sb);
			}
		}
		if(this._children2!=null){
			for(int i = 0 ; i<this._children2.size(); i++){
				sb.append("C2:"+i+":");
				this._children2.get(i).toStringHelper(indent+1, sb);
			}
		}
	}
	
	public void toInOrder(ArrayList<LambdaProd> list){
		
		int arity = this.getArity();
		if(arity==0){
			list.add(this);
		} else if(arity==1){
			LambdaProd prod_child1 = this.getFirstChildren().get(0);
			prod_child1.toInOrder(list);
			list.add(this);
		} else {
			LambdaProd prod_child1 = this.getFirstChildren().get(0);
			prod_child1.toInOrder(list);
			list.add(this);
			LambdaProd prod_child2 = this.getSecondChildren().get(0);
			prod_child2.toInOrder(list);
		}
		
	}

	
	public void toPostOrder(ArrayList<LambdaProd> list){

		int arity = this.getArity();
		if(arity==0){
			list.add(this);
		} else if(arity==1){
			LambdaProd prod_child1 = this.getFirstChildren().get(0);
			prod_child1.toPostOrder(list);
			list.add(this);
		} else {
			LambdaProd prod_child1 = this.getFirstChildren().get(0);
			prod_child1.toPreOrder(list);
			LambdaProd prod_child2 = this.getSecondChildren().get(0);
			prod_child2.toPostOrder(list);
			list.add(this);
		}
		
	}
	
	public void toPreOrder(ArrayList<LambdaProd> list){
		
		int arity = this.getArity();
		if(arity==0){
			list.add(this);
		} else if(arity==1){
			list.add(this);
			LambdaProd prod_child1 = this.getFirstChildren().get(0);
			prod_child1.toPreOrder(list);
		} else {
			list.add(this);
			LambdaProd prod_child1 = this.getFirstChildren().get(0);
			prod_child1.toPreOrder(list);
			LambdaProd prod_child2 = this.getSecondChildren().get(0);
			prod_child2.toPreOrder(list);
		}
		
	}
	
	/**
	 * @param prods the list of Productions
	 */
	public ArrayList<LambdaProd> getAllProds(){
//		if(this._allProds == null){
			
			this._allProds = new ArrayList<LambdaProd>();

			int arity = this.getArity();
			if(arity >=2){
				for(int i = 0; i<this._children2.size(); i++){
					ArrayList<LambdaProd> cProds = this._children2.get(i).getAllProds();
					for(LambdaProd cProd : cProds){
						this._allProds.add(cProd);
					}
				}
			} 
			if(arity >=1){
				for(int i = 0; i<this._children1.size(); i++){
					ArrayList<LambdaProd> cProds = this._children1.get(i).getAllProds();
					for(LambdaProd cProd : cProds){
						this._allProds.add(cProd);
					}
				}
			}
			this._allProds.add(this);
//		}
		
		return this._allProds;
	}

	/**
	 * @throws  
	 * Get the Expr associated with the subtree rooted by this LambdaMRProd
	 * @return
	 * @throws  
	 */
	public Expr toExpr(){
		if(this._exp != null){
			return this._exp;
		}

		try{
//			PrintStream ps = new PrintStream(System.err, true, "UTF-8");

			int constructType = this._gp.getConstructType();
			
			if(constructType == LambdaGProd._TERM){
				this._exp = this._gp.getExprToken().getExpr();
//				ps.println(">>1");
			} else if(constructType == LambdaGProd._APP0){
				LambdaProd child = this._children1.get(0);
				this._exp = child.toExpr();
//				ps.println(">>2");
			} else if(constructType == LambdaGProd._APP1){
				LambdaProd child = this._children1.get(0);
				Expr exp = this._gp.getExprToken().getExpr().simplifyFully().stripParens();
				Expr exp_child = child.toExpr();
				this._exp = new FunApp(exp, exp_child);
				//				ps.println("-------");
				//ps.println(exp);
				//ps.println(exp_child);
				//ps.println(">>3");
			} else if(constructType == LambdaGProd._APP2){
				LambdaProd child1 = this._children1.get(0);
				LambdaProd child2 = this._children2.get(0);
				Expr exp = this._gp.getExprToken().getExpr();
				Expr exp_child1 = child1.toExpr();
				Expr exp_child2 = child2.toExpr();
				this._exp = new FunApp(new FunApp(exp, exp_child1), exp_child2);
				//ps.println(">>4");
			} else if(constructType == LambdaGProd._APP3){
				LambdaProd child1 = this._children1.get(0);
				LambdaProd child2 = this._children2.get(0);
				Expr exp_child1 = child1.toExpr();
				Expr exp_child2 = child2.toExpr();
				this._exp = new FunApp(exp_child1, exp_child2);
				//ps.println(">>5");
			} else {
				throw new RuntimeException("Construct type not valid : "+ constructType);
			}
//			ps.println(this._exp);

			this._exp = this._exp.simplifyFully().stripParens();
			
			return this._exp;
//			Expr result = this._exp;
//			ps.println("++:"+result);
//			return result;
		}catch(TypeEvaluationException e){
			throw new RuntimeException(e);
		}
		
	}
	
	/**
	 * To expression, but don't realize the expression covered by Prod with seq : excludeSeq
	 * @param excludeSeq
	 * @return
	 */
	public Expr toExpr(int excludeSeq, ArrayList<Var> vars, ArrayList<Integer> seqIds){

//		System.err.println("OK:"+seqIds+"\t"+excludeSeq);
		
		try{
			
			if(this._seq == excludeSeq){
				Var var = new Var(LambdaProdFactory.genVar(), this.toExpr().getType(), true);
				vars.add(var);
				return var;
			}
			
			Expr result = null;
			
			int constructType = this._gp.getConstructType();
			
			if(constructType == LambdaGProd._TERM){
				result = this._gp.getExprToken().getExpr();
			} else if(constructType == LambdaGProd._APP0){
				int expected_id = seqIds.get(0);
				boolean found = false;
				for(int k=0; k<this._children1.size(); k++){
					LambdaProd child = this._children1.get(k);
//					System.err.println(child.getSeq()+"\t"+expected_id);
					if(child.getSeq() == expected_id){
						found = true;
						seqIds.remove(0);
						result = child.toExpr(excludeSeq, vars, seqIds);
					}
				}
				if(!found){
					throw new RuntimeException("path not found! 292 : "+seqIds);
				}
			} else if(constructType == LambdaGProd._APP1){
				if(seqIds == null){
					LambdaProd child = this._children1.get(0);
					Expr exp_child = child.toExpr(excludeSeq, vars, null);
					Expr exp = this._gp.getExprToken().getExpr();
					result = new FunApp(exp, exp_child);
				} else {
					int expected_id = seqIds.get(0);
					boolean found = false;
					for(int k=0; k<this._children1.size(); k++){
						LambdaProd child = this._children1.get(k);
						Expr exp = this._gp.getExprToken().getExpr().simplifyFully().stripParens();
						if(child.getSeq() == expected_id){
							found = true;
							seqIds.remove(0);
							Expr exp_child = child.toExpr(excludeSeq, vars, seqIds);
							result = new FunApp(exp, exp_child);
						}
					}
					if(!found){
						throw new RuntimeException("path not found! 314");
					}
				}
			} else if(constructType == LambdaGProd._APP2){
				//if this path is not relevant...
				if(seqIds == null){
					LambdaProd child1 = this._children1.get(0);
					LambdaProd child2 = this._children2.get(0);
					Expr exp_child1 = child1.toExpr(excludeSeq, vars, null);
					Expr exp_child2 = child2.toExpr(excludeSeq, vars, null);
					Expr exp = this._gp.getExprToken().getExpr();
					result = new FunApp(new FunApp(exp, exp_child1), exp_child2);
				} else {
					int expected_id = seqIds.get(0);
					Expr exp_child1 = null, exp_child2 = null;
					boolean found = false;
					for(int k=0; k<this._children1.size(); k++){
						LambdaProd child1 = this._children1.get(k);
						if(child1.getSeq() == expected_id){
							found = true;
							seqIds.remove(0);
							exp_child1 = child1.toExpr(excludeSeq, vars, seqIds);
						}
					}
					if(!found){
						LambdaProd child1 = this._children1.get(0);
						exp_child1 = child1.toExpr(excludeSeq, vars, null);
						
						for(int k=0; k<this._children2.size(); k++){
							LambdaProd child2 = this._children2.get(k);
							if(child2.getSeq() == expected_id){
								seqIds.remove(0);
								exp_child2 = child2.toExpr(excludeSeq, vars, seqIds);
							}
						}
					} else {
						LambdaProd child2 = this._children2.get(0);
						exp_child2 = child2.toExpr(excludeSeq, vars, null);
					}
					Expr exp = this._gp.getExprToken().getExpr();
					result = new FunApp(new FunApp(exp, exp_child1), exp_child2);
				}
			} else {
				/*
				LambdaMRProd child1 = this._children1.get(0);
				LambdaMRProd child2 = this._children2.get(0);
				Expr exp_child1 = child1.toExpr(excludeSeq, vars);
				Expr exp_child2 = child2.toExpr(excludeSeq, vars);
				result = new FunApp(exp_child1, exp_child2);
			} else {
			*/
				throw new RuntimeException("Construct type not valid : "+ constructType);
			}

			result = result.simplifyFully().stripParens();
			
			return result;
		}catch(TypeEvaluationException e){
			throw new RuntimeException(e);
		}
		
	}
	
	
	
	
}



