/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.data;

import java.io.*;
import java.util.*;

import lambdacalc.logic.*;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaProdFactory implements Serializable{

	private static final long serialVersionUID = 6914820857544659781L;
	public static int varCount = 0;
	
	private ArrayList<Type> _allTypes = new ArrayList<Type>(); // this list stores all unique types
	private ArrayList<Expr> _allExprs = new ArrayList<Expr>(); // this list stores all unique expressions
	private ArrayList<LambdaGProd> _allGMRs; // this list stores all unique LambdaGMRProds
	
	private HashMap<Type, TypeToken> _typeTable = new HashMap<Type, TypeToken>(); // Type - TypeToken table
	private HashMap<Expr, ExprToken> _exprTable = new HashMap<Expr, ExprToken>(); // Expr - ExprToken table
	
	private boolean debug = true;
	
	public ArrayList<LambdaGProd> getAllGProds(){
		return this._allGMRs;
	}
	
    public static String genVar() {
        return "v-" + (varCount++);
    }
    
	public LambdaProdFactory(){
		this._allTypes = new ArrayList<Type>();
		this._allExprs = new ArrayList<Expr>();
		this._allGMRs = new ArrayList<LambdaGProd>();
	}

	/**
	 * Sequentially update the list of gms.
	 */
	public void updateGMRs(){
		for(LambdaGProd gm : this._allGMRs){
			gm.updates();
		}
	}
	
	/**
	 * Display the gms
	 * @throws UnsupportedEncodingException 
	 */
	public void displayGMRs_unigram(Vocabulary voc) throws UnsupportedEncodingException{
		for(LambdaGProd gm : this._allGMRs){
			gm.viewUnigramTable(voc);
		}
	}

	/**
	 * Display the gms
	 * @throws UnsupportedEncodingException 
	 */
	public void displayGMRs_bigram(Vocabulary voc) throws UnsupportedEncodingException{
		for(LambdaGProd gm : this._allGMRs){
			gm.viewBigramTable(voc);
		}
	}

	/**
	 * Return a TypeToken for a given Type
	 * @param t given Type
	 * @return a TypeToken for the given Type
	 */
	public TypeToken toTypeToken(Type t){
		if(this._typeTable.containsKey(t)){
			return this._typeTable.get(t);
		}
		TypeToken token = new TypeToken(t);
		this._allTypes.add(t);
		token.setId(this._allTypes.size()+1);//0 and 1 are reserved..
		this._typeTable.put(t, token);
		return token;
	}
	
	/**
	 * Return a ExprToken for a given Expr. Note that this is expensive.
	 * However, I don't know of a better way to deal with this..
	 * @param e given Expr
	 * @return an ExprToken for the given Type
	 */
	public ExprToken toExprToken(Expr e){
		int index = -1;
		//this step is inefficient... needs to find a smarter way to deal with it..
		for(int i = 0 ; i<this._allExprs.size(); i++){
			Expr exp = this._allExprs.get(i);
			if(e.alphaEquivalent(exp)){
				index = i;
				break;
			}
		}
		if(index == -1){
			ExprToken token = new ExprToken(e);
			token.setId(this._allExprs.size());
			this._allExprs.add(e);
			this._exprTable.put(e, token);
			return token;
		} else {
			Expr exp = this._allExprs.get(index);
			return this._exprTable.get(exp);
		}
	}
	
	/**
	 * Note that this method does not assign ids to the gms
	 * @param gm
	 * @return
	 */
	private LambdaGProd getGMR(LambdaGProd gm){
		int index = Collections.binarySearch(this._allGMRs, gm);
		if(index < 0){
			gm.setId(this._allGMRs.size());
			this._allGMRs.add(-1-index, gm);
//			System.err.println("new:"+gm+"\t"+this._allGMRs.size()+"\t"+gm.getId());
//			System.exit(1);
		} else {
			gm = this._allGMRs.get(index);
		}
		
		this.checkGP(gm);
		return gm;
	}
	
	private LambdaGProd getGMR_test(LambdaGProd gm){
		int index = Collections.binarySearch(this._allGMRs, gm);
		if(index < 0){
			return null;
		} else {
			gm = this._allGMRs.get(index);
		}
		
		this.checkGP(gm);
		return gm;
	}
	
	private void checkGP(LambdaGProd gm){
		try{
			if(gm.getExprToken()==null){
				return;
			} else if(gm.isRoot()){
				return;
			}
			Type lhs = gm.getLHSTypeToken().getType();
			TypeToken[] rhstokens = gm.getRHSTypeTokens();
			Expr exp = gm.getExprToken().getExpr();
			if(rhstokens.length==0){
				if(!exp.getType().equals(lhs)){
					throw new RuntimeException("GP "+gm.toString()+" is invalid!");
				}
			} else if(rhstokens.length == 1){
				Var v = new Var("x", rhstokens[0].getType(),true);
				Expr e = new FunApp(exp,v).simplifyFully();
				if(!e.getType().equals(lhs)){
					throw new RuntimeException("GP "+gm.toString()+" is invalid!");
				}
			} else if(rhstokens.length == 2){
				Var v1 = new Var("x", rhstokens[0].getType(),true);
				Var v2 = new Var("y", rhstokens[1].getType(),true);
				Expr e = new FunApp(new FunApp(exp,v1),v2).simplifyFully();
				if(!e.getType().equals(lhs)){
					throw new RuntimeException("GP "+gm.toString()+" is invalid!");
				}
			}
		}catch(TypeEvaluationException e){
			throw new RuntimeException(e+"\t"+gm.toString());
		}catch(NullPointerException e){
			throw new RuntimeException(e+"\t"+gm.toString());
		}
	}

	/*
	//the position of the GM is just its id
	public void assignIds2GMRs(){
		for(int i = 0 ; i< this._allGMRs.size(); i++){
			this._allGMRs.get(i).setId(i);
		}
	}
	*/
	
	public LambdaGProd toLambdaGProd(TypeToken type, ExprToken expr){
		LambdaGProd gm = new LambdaGProd(LambdaGProd._TERM, type, expr, new TypeToken[]{});
		return this.getGMR(gm);
	}

	public LambdaGProd toLambdaGProd(TypeToken type1, ExprToken token1, TypeToken type2){
		LambdaGProd gm = new LambdaGProd(LambdaGProd._APP1, type1, token1, new TypeToken[]{type2});
		return this.getGMR(gm);
	}
	public LambdaGProd toLambdaGProd_test(TypeToken type1, ExprToken token1, TypeToken type2){
		LambdaGProd gm = new LambdaGProd(LambdaGProd._APP1, type1, token1, new TypeToken[]{type2});
		return this.getGMR_test(gm);
	}
	public LambdaGProd toLambdaGProd_test(TypeToken type1, ExprToken expr){
		LambdaGProd gm = new LambdaGProd(LambdaGProd._APP1, type1, expr, new TypeToken[]{});
		return this.getGMR_test(gm);
	}

	/*
	private LambdaGMRProd toGMRToken(TypeToken type1, TypeToken type2, TypeToken type3){
		LambdaGMRProd gm = new LambdaGMRProd(LambdaGMRProd._APP2, type1, null, new TypeToken[]{type2, type3});
		return this.getGMR(gm);
	}
	*/
	
	public LambdaGProd toLambdaGProd(TypeToken type1, ExprToken token1, TypeToken type2, TypeToken type3){
		LambdaGProd gm = new LambdaGProd(LambdaGProd._APP2, type1, token1, new TypeToken[]{type2, type3});
		return this.getGMR(gm);
	}
	
	public LambdaProd toLambdaProd(TypeToken type1, ExprToken expr1){
		return new LambdaProd(this.toLambdaGProd(type1, expr1));
	}

	/*
	private LambdaMRProd toTreeLambdaRule(TypeToken type1, TypeToken type2, TypeToken type3){
		return new LambdaMRProd(this.toGMRToken(type1, type2, type3));
	}
	*/

	public LambdaProd toLambdaProd_test(TypeToken type1, ExprToken expr1){
		LambdaGProd gp = this.toLambdaGProd_test(type1, expr1);
		if(gp == null){
			return null;
		} else {
			return new LambdaProd(gp);
		}
	}
	
	public LambdaProd toLambdaProd_test(TypeToken type1, ExprToken expr1, TypeToken type2){
		LambdaGProd gp = this.toLambdaGProd_test(type1, expr1, type2);
		if(gp == null){
			return null;
		} else {
			return new LambdaProd(gp);
		}
	}
	
	public LambdaProd toLambdaProd(TypeToken type1, ExprToken expr1, TypeToken type2){
		return new LambdaProd(this.toLambdaGProd(type1, expr1, type2));
	}
	
	public LambdaProd toLambdaProd(TypeToken type1, ExprToken expr1, TypeToken type2, TypeToken type3){
		return new LambdaProd(this.toLambdaGProd(type1, expr1, type2, type3));
	}
	
	public LambdaProd decomposeAsRoot_prior(Expr rootExpr){
		try{
			ArrayList<LambdaProd> children = this.decompose(rootExpr);
			
			Type X = rootExpr.getType();
			TypeToken token_X = this.toTypeToken(X);
			
			Var x = new Var(genVar(), X, true);
			
			//Lx.x
			Lambda Lxx = new Lambda(x, x, true);
			
			ExprToken token_Lxx = this.toExprToken(Lxx);
			
			LambdaGProd rootGMR = new LambdaGProd(LambdaGProd._APP1, TypeToken._rootToken, token_Lxx, new TypeToken[]{token_X});
			rootGMR = this.getGMR(rootGMR);
			LambdaProd rootMR = new LambdaProd(rootGMR);
			rootMR.setChildren(children);
			
			return rootMR;
		} catch(TypeEvaluationException e){
			throw new RuntimeException(e);
		} catch(UnsupportedEncodingException e){
			throw new RuntimeException(e);
		}
	}
	
	public LambdaProd decomposeAsRoot(Expr rootExpr){
		try{
			ArrayList<LambdaProd> children = this.decompose(rootExpr);
			
			Type X = rootExpr.getType();
			TypeToken token_X = this.toTypeToken(X);
			
			LambdaGProd rootGMR = new LambdaGProd(LambdaGProd._APP0, TypeToken._rootToken, null, new TypeToken[]{token_X});
			rootGMR = this.getGMR(rootGMR);
			LambdaProd rootMR = new LambdaProd(rootGMR);
			rootMR.setChildren(children);
			
			return rootMR;
		} catch(TypeEvaluationException e){
			throw new RuntimeException("type eval error: "+e);
		} catch(UnsupportedEncodingException e){
			throw new RuntimeException("encoding error: "+e);
		} catch(NotImplementedException e){
			return null;
		}
	}
	

	
	public LambdaProd decomposeAsRoot_single(Expr rootExpr){
		try{
			ArrayList<LambdaProd> children = this.decompose_single(rootExpr);
			
			Type X = rootExpr.getType();
			TypeToken token_X = this.toTypeToken(X);
			
			LambdaGProd rootGMR = new LambdaGProd(LambdaGProd._APP0, TypeToken._rootToken, null, new TypeToken[]{token_X});
			rootGMR = this.getGMR(rootGMR);
			LambdaProd rootMR = new LambdaProd(rootGMR);
			rootMR.setChildren(children);
			
			return rootMR;
		} catch(TypeEvaluationException e){
			throw new RuntimeException(e);
		} catch(UnsupportedEncodingException e){
			throw new RuntimeException(e);
		}
	}
	
	private ArrayList<LambdaProd> decompose(Expr exp) throws UnsupportedEncodingException, TypeEvaluationException{
		
		ArrayList<LambdaProd> treeRules = new ArrayList<LambdaProd>();
		
		exp = exp.stripAnyDoubleParens().stripOutermostParens();
		
        TypeToken token_X = this.toTypeToken(exp.getType());
    	ExprToken token_self = this.toExprToken(exp);
        
        /**
         * If it is a constant, add one transit prod.
         */
    	/**/
        if(exp instanceof Const){
            LambdaProd expProd = this.toLambdaProd(token_X, token_self);
            Var x = new Var(genVar(), exp.getType(), true);
            Expr Lxx = new Lambda(x, x, true);
        	ExprToken token_Lxx = this.toExprToken(Lxx);
        	LambdaProd transProd = this.toLambdaProd(token_X, token_Lxx, token_X);
        	ArrayList<LambdaProd> children = new ArrayList<LambdaProd>();
        	children.add(expProd);
        	transProd.setChildren(children);
        	treeRules.add(transProd);
        	if(debug) checkRestore(Lxx, exp, exp);
        	return treeRules;
        }
        /**/
        
        int size_before = treeRules.size();
		
		//ps.println(exp);

        /** (1000)
         * If the outer most is constant
         * X : toTree ( c )
         * <=>
         * X : c
         */
        if (exp.getExprCat() == Expr.CONST) {
        	//do nothing, already added..
        }
        /** (2000)
         * If the outer most is Count
         * X : toTree ( Cx.P(x) )
         * <=>
         * X : ( Lf.Cx.f(x) Y )
         * Y : toTree ( Lx.P(x) )
         */
        else if (exp.getExprCat() == Expr.COUNT) {
        	Count count = (Count) exp;
            Expr inner = count.getInnerExpr().stripParens();
            
            Identifier x = count.getVariable();
            
            Var f = new Var(genVar(), Type.ET, true);
            Lambda LfCxfx = new Lambda(f, new Count(x, new FunApp(f, x), true), true);
            Lambda LxPx = new Lambda(x, inner, true);
            
            Type Y = LxPx.getType();
            
            ExprToken token_exp = this.toExprToken(LfCxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_exp, token_Y);
            ArrayList<LambdaProd> treeRule_children = this.decompose(LxPx);
            
            treeRule.setChildren(treeRule_children);
            
            if(debug) checkRestore(LfCxfx, LxPx, exp);
            
            treeRules.add(treeRule);
        }
        /** (3000)
         * If the outer most is FunApp
         */
        else if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr p = funapp.getFunc().stripParens();
            Expr EXP = funapp.getArg().stripParens();
            
            /** (3001)
             * X : toTree ( p(EXP) )
             * <=>
             * X : ( ( Lf.Lx.f(x) Y ) Z )
             * Y : toTree ( Lx.p(x) )
             * Z : toTree ( EXP )
             * 
             * or
             * 
             * X : ( Lx.p(x) Z )
             * Z : toTree( EXP )
             */

            if (EXP.getExprCat() != Expr.ARGLIST && p.getExprCat() == Expr.CONST) {
                Var x = new Var(genVar(), EXP.getType(), true);
                Lambda Lxpx = new Lambda(x, new FunApp(p, x), true);
                Type Y = Lxpx.getType();
                Type Z = EXP.getType();
                
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                Var f = new Var(genVar(), Y, true);
                Expr LfLxfx = new Lambda(f, new Lambda(x, new FunApp(f,x), true), true);
                ExprToken token_LfLxfx = this.toExprToken(LfLxfx);
                
                ExprToken token_Lxpx = this.toExprToken(Lxpx);
                
                LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLxfx, token_Y, token_Z);
                ArrayList<LambdaProd> dec_Lxpx = this.decompose(Lxpx);
                ArrayList<LambdaProd> dec_EXP = this.decompose(EXP);
                treeRule1.setChildren(dec_Lxpx, dec_EXP);
                
                LambdaProd treeRule2 = this.toLambdaProd(token_X, token_Lxpx, token_Z);
                treeRule2.setChildren(dec_EXP);
                
                if(debug) checkRestore(LfLxfx, Lxpx, EXP, exp);
                if(debug) checkRestore(Lxpx, EXP, exp);
                
                treeRules.add(treeRule1);
                treeRules.add(treeRule2);
            }
            /** (3002)
             * X : toTree ( p(EXP1, EXP2) )
             * <=>
             * X : ( (Lx.Ly.p(x,y) Y) Z)
             * Y : toTree (EXP1)
             * Z : toTree (EXP2)
             */
            else if (EXP.getExprCat() == Expr.ARGLIST) {
                Expr[] args = ((ArgList) EXP).getArgs();
                Expr EXP1 = args[0];
                Expr EXP2 = args[1];

                Var x = new Var(genVar(), EXP1.getType(), true);
                Var y = new Var(genVar(), EXP2.getType(), true);

                Type Y = EXP1.getType();
                Type Z = EXP2.getType();

                Lambda LxLypxy = new Lambda(x, new Lambda(y, new FunApp(p, new ArgList(new Expr[] {x, y})), true), true);
                
                ExprToken token_LxLypxy = this.toExprToken(LxLypxy);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                LambdaProd treeRule = this.toLambdaProd(token_X, token_LxLypxy, token_Y, token_Z);
                
                ArrayList<LambdaProd> dec_EXP1 = this.decompose(EXP1);
                ArrayList<LambdaProd> dec_EXP2 = this.decompose(EXP2);
                
                treeRule.setChildren(dec_EXP1, dec_EXP2);

                if(debug) checkRestore(LxLypxy, EXP1, EXP2, exp);
                
                treeRules.add(treeRule);
            }
            else{
            	throw new NotImplementedException("3003 not implemented!");
            }
        }
        /** (4000)
         * If the outer most is Iota
         * X : toTree ( Ix.EXP )
         * <=>
         * X : ( Lf.Ix.f(x) Y )
         * Y : toTree ( Lx.EXP )
         */
        else if (exp.getExprCat() == Expr.IOTA) {

            Iota iota = (Iota) exp;
            Identifier x = iota.getVariable();
            Expr EXP = iota.getInnerExpr().stripParens();
            Lambda lambda = new Lambda(x, EXP, true);
            Type Y = lambda.getType();
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXP.getType()), true);
            Lambda LfIxfx = new Lambda(f, new Iota(x, new FunApp(f, x), true), true);
            Lambda LxEXP = new Lambda(x, EXP, true);

            ExprToken token_LfIxfx = this.toExprToken(LfIxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfIxfx, token_Y);
            treeRule.setChildren(this.decompose(LxEXP));
            
            if(debug) checkRestore(LfIxfx, LxEXP, exp);
            
            treeRules.add(treeRule);
        }
        /** (5000)
         * If the outer most is Not
         * X : toTree ( ~EXP )
         * <=>
         * X : ( Lv.~v Y )
         * Y : toTree ( EXP )
         */
        else if (exp.getExprCat() == Expr.NOT) {
            Not not = (Not) exp;
            Expr EXP = not.getInnerExpr().stripParens();
            Type X = exp.getType();
            Type Y = EXP.getType();
            
            Var v = new Var(genVar(), X, true);
            Lambda LvNOTv = new Lambda(v, new Not(v), true);

            ExprToken token_LvNOTv = this.toExprToken(LvNOTv);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LvNOTv, token_Y);
            treeRule.setChildren(this.decompose(EXP));
            
            if(debug) checkRestore(LvNOTv, EXP, exp);

            treeRules.add(treeRule);
        }
        /** (6000)
         * If the outer most is Exists
         * X : toTree ( Ex.EXP )
         * <=>
         * X : ( Lf.Ex.f(x) Y )
         * Y : toTree ( Lx.EXP )
         */
        else if (exp.getExprCat() == Expr.EXISTS) {
            Exists exists = (Exists) exp;
            Expr EXP = exists.getInnerExpr().stripParens();
            Identifier x = exists.getVariable();
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXP.getType()), true);
            Lambda LfExfx = new Lambda(f, new Exists(x, new FunApp(f, x), true), true);
            Lambda LxEXP = new Lambda(x, EXP, true);
            Type Y = LxEXP.getType();

            ExprToken token_LfExfx = this.toExprToken(LfExfx);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfExfx, token_Y);
            treeRule.setChildren(this.decompose(LxEXP));
            
            if(debug) checkRestore(LfExfx, LxEXP, exp);

            treeRules.add(treeRule);
        }
        /** (7000)
         * If the outer most is Sum
         * X : toTree ( S(EXP1, EXP2) )
         * <=>
         * X : ( ((Lf.Lg.S(f,g) Y) Z) )
         * Y : toTree ( EXP1 )
         * Z : toTree ( EXP2 )
         * 
         * or 
         * 
         * X : ( (Lf.S(f,EXP2) Y) )
         * Y : toTree ( EXP1 )
         */
        else if (exp.getExprCat() == Expr.SUM) {
            Sum sum = (Sum) exp;
            Identifier x = sum.getVariable();
            FunApp fa = (FunApp) sum.getInnerExpr().stripParens();
            Expr sum_e = fa.getFunc(); //sum
            Expr[] args = ((ArgList) fa.getArg()).getArgs();
            Expr EXP1 = args[0].stripParens();
            Expr EXP2 = args[1].stripParens();
            Type Y = EXP1.getType();
            Type Z = EXP2.getType();

            Var f = new Var(genVar(), Y, true);
            Var g = new Var(genVar(), Z, true);

            Lambda LfLgSumfg = new Lambda(f, new Lambda(g, new Sum(x, new FunApp(sum_e, new ArgList(new Expr[] {f, g})), true), true), true);
            
            ExprToken token_LfLgSumfg = this.toExprToken(LfLgSumfg);
            TypeToken token_Y = this.toTypeToken(Y);
            TypeToken token_Z = this.toTypeToken(Z);
            
            LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgSumfg, token_Y, token_Z);
            ArrayList<LambdaProd> children1 = this.decompose(EXP1);
            ArrayList<LambdaProd> children2 = this.decompose(EXP2);
            treeRule1.setChildren(children1, children2);
            
            Lambda LfSumfEXP2 = new Lambda(f, new Sum(x, new FunApp(sum_e, new ArgList(new Expr[] {f, EXP2})),true),true);
            
            ExprToken token_LgSumfEXP2 = this.toExprToken(LfSumfEXP2);
            LambdaProd treeRule2 = this.toLambdaProd(token_X, token_LgSumfEXP2, token_Y);
            treeRule2.setChildren(children1);
            
            if(debug) checkRestore(LfLgSumfg, EXP1, EXP2, exp);
            if(debug) checkRestore(LfSumfEXP2, EXP1, exp);
            
            treeRules.add(treeRule1);
            treeRules.add(treeRule2);
        }
        /** (8000)
         * If the outer most is ArgMax
         * X : toTree ( ArgMax(EXP1, EXP2) )
         * <=>
         * X : ( ((Lf.Lg.ArgMax(f,g) Y) Z) )
         * Y : toTree ( EXP1 )
         * Z : toTree ( EXP2 )
         * 
         * or
         * X : ( (Lf.ArgMax(f,EXP2) Y) )
         * Y : toTree ( EXP1 )
         */
        else if (exp.getExprCat() == Expr.ARGMAX) {
        	
            ArgMax argmax = (ArgMax) exp;
            Identifier x = argmax.getVariable();
            Expr inner = argmax.getInnerExpr().stripParens();
            FunApp funapp = (FunApp) inner;
            Expr argmax_e = funapp.getFunc(); //argmax
            Expr[] args = ((ArgList) funapp.getArg()).getArgs();
            Expr EXP1 = args[0].stripParens();
            Expr EXP2 = args[1].stripParens();
            
            
            Type Y = EXP1.getType();
            Type Z = EXP2.getType();
            
            Var f = new Var(genVar(), EXP1.getType(), true);
            Var g = new Var(genVar(), EXP2.getType(), true);
            Lambda LfLgArgMaxfg = new Lambda(f, new Lambda(g, new ArgMax(x, new FunApp(argmax_e, new ArgList(new Expr[] {f, g})), true), true), true);
            
            ExprToken token_LfLgArgMaxfg = this.toExprToken(LfLgArgMaxfg);
            TypeToken token_Y = this.toTypeToken(Y);
            TypeToken token_Z = this.toTypeToken(Z);
            
            LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgArgMaxfg, token_Y, token_Z);
            ArrayList<LambdaProd> children1 = this.decompose(EXP1);
            ArrayList<LambdaProd> children2 = this.decompose(EXP2);
            treeRule1.setChildren(children1, children2);

            treeRules.add(treeRule1);
            if(debug) checkRestore(LfLgArgMaxfg, EXP1, EXP2, exp);
            
            /** (8001)
             * X : toTree ( ArgMax(EXP1, Lx.Cy.EXP2(x.y)) )
             * <=>
             * X : ( ((Lf.Lg.ArgMax(f, Lx.Cy.g(x)(y)) Y) Z) )
             * Y : toTree ( EXP1 )
             * Z : toTree ( Lx.Ly.EXP2(x.y) )
             */
            
            boolean alternative = true;
            
            if(EXP2.simplifyFully().stripParens().getExprCat() == Expr.LAMBDA){
            	Lambda lambda = (Lambda)EXP2;
            	Expr inner_lambda = lambda.getInnerExpr().simplifyFully().stripParens();
            	if(inner_lambda.getExprCat() == Expr.COUNT){
            		alternative = false;
            	}
            }
            
            if(alternative){
                Lambda LfArgMaxfEXP2 = new Lambda(f, new ArgMax(x, new FunApp(argmax_e, new ArgList(new Expr[]{f, EXP2})),true),true);
                ExprToken token_LfArgMaxfEXP2 = this.toExprToken(LfArgMaxfEXP2);
                
                LambdaProd treeRule2 = this.toLambdaProd(token_X, token_LfArgMaxfEXP2, token_Y);
                treeRule2.setChildren(children1);
                
                if(debug) checkRestore(LfArgMaxfEXP2, EXP1, exp);

                treeRules.add(treeRule2);
            }
            
            /** (8010)
             * If the outer most is ArgMax
             * X : toTree ( ArgMax(EXP1, Lx.Cy.()) )
             * <=>
             * X : ( ((Lf.Lg.ArgMax(f,g) Y) Z) )
             * Y : toTree ( EXP1 )
             * Z : toTree ( EXP2 )
             * 
             * or
             * X : ( (Lf.ArgMax(f,EXP2) Y) )
             * Y : toTree ( EXP1 )
             */
            /*
            if(EXP2.simplifyFully().stripParens().getExprCat() == Expr.LAMBDA){
            	Lambda lmd = (Lambda)EXP2;
            	if(lmd.getInnerExpr().simplifyFully().stripParens().getExprCat() == Expr.COUNT){
            	}
            }
            */
        }

        /** (9000)
         * If the outer most is ArgMin
         * X : toTree ( ArgMin(EXP1, EXP2) )
         * <=>
         * X : ( ((Lf.Lg.ArgMin(f,g) Y) Z) )
         * Y : toTree ( EXP1 )
         * Z : toTree ( EXP2 )
         * 
         * or
         * X : ( (Lf.ArgMin(f,EXP2) Y) )
         * Y : toTree ( EXP1 )
         */
        else if (exp.getExprCat() == Expr.ARGMIN) {

        	ArgMin argmin = (ArgMin) exp;
            Identifier x = argmin.getVariable();
            Expr inner = argmin.getInnerExpr().stripParens();
            FunApp funapp = (FunApp) inner;
            Expr argmin_e = funapp.getFunc(); //argmin
            Expr[] args = ((ArgList) funapp.getArg()).getArgs();
            Expr EXP1 = args[0].stripParens();
            Expr EXP2 = args[1].stripParens();

            Type Y = EXP1.getType();
            Type Z = EXP2.getType();

            Var f = new Var(genVar(), EXP1.getType(), true);
            Var g = new Var(genVar(), EXP2.getType(), true);
            Lambda LfLgArgMinfg = new Lambda(f, new Lambda(g, new ArgMin(x, new FunApp(argmin_e, new ArgList(new Expr[] {f, g})), true), true), true);
            
            ExprToken token_LfLgArgMinfg = this.toExprToken(LfLgArgMinfg);
            TypeToken token_Y = this.toTypeToken(Y);
            TypeToken token_Z = this.toTypeToken(Z);
            
            LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgArgMinfg, token_Y, token_Z);
            ArrayList<LambdaProd> children1 = this.decompose(EXP1);
            ArrayList<LambdaProd> children2 = this.decompose(EXP2);
            treeRule1.setChildren(children1, children2);
            
            //Lambda LfArgMinfEXP2 = new Lambda(f, new ArgMin(x, new FunApp(argmin_e, new ArgList(new Expr[]{f, EXP2})),true),true);
            //ExprToken token_LfArgMinfEXP2 = this.toExprToken(LfArgMinfEXP2);
            
            //LambdaProd treeRule2 = this.toLambdaProd(token_X, token_LfArgMinfEXP2, token_Y);
            //treeRule2.setChildren(children1);
            
            if(debug) checkRestore(LfLgArgMinfg, EXP1, EXP2, exp);
            //if(debug) checkRestore(LfArgMinfEXP2, EXP1, exp);

            treeRules.add(treeRule1);
            //treeRules.add(treeRule2);
            

            /** (8001)
             * X : toTree ( ArgMax(EXP1, Lx.Cy.EXP2(x.y)) )
             * <=>
             * X : ( ((Lf.Lg.ArgMax(f, Lx.Cy.g(x)(y)) Y) Z) )
             * Y : toTree ( EXP1 )
             * Z : toTree ( Lx.Ly.EXP2(x.y) )
             */
            
            boolean alternative = true;
            
            if(EXP2.simplifyFully().stripParens().getExprCat() == Expr.LAMBDA){
            	Lambda lambda = (Lambda)EXP2;
            	Expr inner_lambda = lambda.getInnerExpr().simplifyFully().stripParens();
            	if(inner_lambda.getExprCat() == Expr.COUNT){
            		alternative = false;
            	}
            }
            
            if(alternative){
                Lambda LfArgMaxfEXP2 = new Lambda(f, new ArgMin(x, new FunApp(argmin_e, new ArgList(new Expr[]{f, EXP2})),true),true);
                ExprToken token_LfArgMaxfEXP2 = this.toExprToken(LfArgMaxfEXP2);
                
                LambdaProd treeRule2 = this.toLambdaProd(token_X, token_LfArgMaxfEXP2, token_Y);
                treeRule2.setChildren(children1);
                
                if(debug) checkRestore(LfArgMaxfEXP2, EXP1, exp);

                treeRules.add(treeRule2);
            }
        }
        /** (80000)
         * X : toTree( Max x. EXP(x))
         * <=>
         * X : ( Lf.Max x.f(x) Y )
         * Y : toTree(Lx.EXP(x))
         */
        else if (exp.getExprCat() == Expr.MAX) {
            Max max = (Max) exp;
            FunApp fa = (FunApp) max.getInnerExpr();
            Expr max_e = fa.getFunc();
            Expr EXPx = fa.getArg();
            Identifier x = max.getVariable();
            
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
            Lambda LfMaxxfx = new Lambda(f, new Max(x, new FunApp(max_e, new ArgList(new Expr[] {f, x})), true), true);
            Lambda LxEXPx = new Lambda(x, EXPx, true);
            
            Type Y = LxEXPx.getType();
            
            ExprToken token_LfMaxxfx = this.toExprToken(LfMaxxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfMaxxfx, token_Y);
            treeRule.setChildren(this.decompose(LxEXPx));
            
            if(debug) checkRestore(LfMaxxfx, LxEXPx, exp);

            treeRules.add(treeRule);
        }

        /** (90000)
         * X : toTree( Min x. EXP(x))
         * <=>
         * X : ( Lf.Min x.f(x) Y )
         * Y : toTree(Lx.EXP(x))
         */
        else if (exp.getExprCat() == Expr.MIN) {
        	Min min = (Min) exp;
            FunApp fa = (FunApp) min.getInnerExpr();
            Expr min_e = fa.getFunc();
            Expr EXPx = fa.getArg();
            Identifier x = min.getVariable();
            
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
            Lambda LfMinxfx = new Lambda(f, new Min(x, new FunApp(min_e, new ArgList(new Expr[] {f, x})), true), true);
            Lambda LxEXPx = new Lambda(x, EXPx, true);
            
            Type Y = LxEXPx.getType();
            
            ExprToken token_LfMinxfx = this.toExprToken(LfMinxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfMinxfx, token_Y);
            treeRule.setChildren(this.decompose(LxEXPx));
            
            if(debug) checkRestore(LfMinxfx, LxEXPx, exp);

            treeRules.add(treeRule);
        }


        /** (10000)
         * If the outer most is Lambda
         */
        else if (exp.getExprCat() == Expr.LAMBDA) {
            Lambda lambda = (Lambda) exp;
            Expr inner = lambda.getInnerExpr().stripParens();
            Identifier x = lambda.getVariable();

            /** (10100)
             * X : toTree ( Lx.EXP1 ^ EXP2 )
             * <=>
             * X : ( ((Lf.Lg.Lx.f(x) ^ g(x) Y) Z) )
             * Y : toTree ( Lx.EXP1 )
             * Z : toTree ( Lx.EXP2 )
             */
            if (inner.getExprCat() == Expr.AND) {
                And and = (And) inner;
                Expr EXP1 = and.getLeft().stripParens();
                Expr EXP2 = and.getRight().stripParens();

                /*
                if(EXP1.simplifyFully().stripParens().getExprCat() == Expr.AND){
                	PrintStream ps = new PrintStream(System.err, true, "UTF-8");
                	ps.println(EXP1);
                	ps.println(EXP2);
                	ps.println();
//                	System.exit(1);
                }
                */
                
                Lambda LxEXP1 = new Lambda(x, EXP1, true);
                Lambda LxEXP2 = new Lambda(x, EXP2, true);
                Type Y = LxEXP1.getType();
                Type Z = LxEXP2.getType();
                
                Var f = new Var(genVar(), Y, true);
                Var g = new Var(genVar(), Z, true);
                
                FunApp fx = new FunApp(f, x);
                FunApp gx = new FunApp(g, x);
                
                Lambda LfLgLxfxANDgx = new Lambda(f, new Lambda(g, new Lambda(x, new And(fx, gx), true), true), true);
                
                ExprToken token_LfLgLxfxANDgx = this.toExprToken(LfLgLxfxANDgx);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                ArrayList<LambdaProd> children1 = this.decompose(LxEXP1);
                ArrayList<LambdaProd> children2 = this.decompose(LxEXP2);
                
                LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxfxANDgx, token_Y, token_Z);
                treeRule1.setChildren(children1, children2);
                
                LambdaProd treeRule2 = this.toLambdaProd(token_X, token_LfLgLxfxANDgx, token_Z, token_Y);
                treeRule2.setChildren(children2, children1);
                
                if(debug) checkRestore(LfLgLxfxANDgx, LxEXP1, LxEXP2, exp);
                
//                System.err.println(">>1:"+treeRule1.getGM());
//                System.err.println(">>2:"+treeRule2.getGM());
                
                treeRules.add(treeRule1);
                treeRules.add(treeRule2);
            }

            /** (10200)
             * X : toTree ( Lx.EXP1 v EXP2 )
             * <=>
             * X : ( ((Lf.Lg.Lx.f(x) v g(x) Y) Z) )
             * Y : toTree ( Lx.EXP1 )
             * Z : toTree ( Lx.EXP2 )
             */
            else if (inner.getExprCat() == Expr.OR) {
                Or or = (Or) inner;
                Expr EXP1 = or.getLeft();
                Expr EXP2 = or.getRight();
                
                Lambda LxEXP1 = new Lambda(x, EXP1, true);
                Lambda LxEXP2 = new Lambda(x, EXP2, true);
                Type Y = LxEXP1.getType();
                Type Z = LxEXP2.getType();

                Var f = new Var(genVar(), Y, true);
                Var g = new Var(genVar(), Z, true);

                FunApp fx = new FunApp(f, x);
                FunApp gx = new FunApp(g, x);

                Lambda LfLgLxfxORgx = new Lambda(f, new Lambda(g, new Lambda(x, new Or(fx, gx), true), true), true);

                ExprToken token_LfLgLxfxORgx = this.toExprToken(LfLgLxfxORgx);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                ArrayList<LambdaProd> children1 = this.decompose(LxEXP1);
                ArrayList<LambdaProd> children2 = this.decompose(LxEXP2);
                
                LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxfxORgx, token_Y, token_Z);
                treeRule1.setChildren(children1, children2);
                
                LambdaProd treeRule2 = this.toLambdaProd(token_X, token_LfLgLxfxORgx, token_Z, token_Y);
                treeRule2.setChildren(children2, children1);
                
                if(debug) checkRestore(LfLgLxfxORgx, LxEXP1, LxEXP2, exp);
                
                treeRules.add(treeRule1);
                treeRules.add(treeRule2);
            }

            /** (10300)
             * X : toTree ( Lx.Ey.P(y)^R(x,y))
             * <=>
             * X : ( ( Lg.Lf.Lx.Ey.f(y)^g(x)(y) Y ) Z )
             * Y : toTree(Lx.Ly.R(x,y))
             * Z : toTree(Ly.P(y))
             */
            else if (inner.getExprCat() == Expr.EXISTS) {
                Exists exists = (Exists) inner;
                Identifier y = exists.getVariable();

                Expr innerExists = exists.getInnerExpr().stripParens();

                if (innerExists.getExprCat() == Expr.AND) {
                	
                    And and = (And) innerExists;
                    Expr Py = and.getLeft();
                    Expr Rxy = and.getRight();

                    /** (10301)
                     * X : toTree ( Lx.Ey.P(y)^R(x,y) )
                     * <=>
                     * X : ( ( Lg.Lf.Lx.Ey.f(y)^g(x)(y) Y ) Z )
                     * Y : toTree( Lx.Ly.R(x,y) )
                     * Z : toTree( Ly.P(y) )
                     */
                    if (involvesTwoVars(Rxy, x, y)) {
                        Var g = new Var(genVar(), new CompositeType(x.getType(), new CompositeType(y.getType(), Rxy.getType())), true);
                        Var f = new Var(genVar(), new CompositeType(y.getType(), Py.getType()), true);

                        Expr fy = new FunApp(f, y);
                        Expr gxy = new FunApp(new FunApp(g, x), y);
                        Lambda LgLfLxEyfygxy = new Lambda(g, new Lambda(f, new Lambda(x, new Exists(y, new And(fy, gxy), true), true), true), true);

                        Lambda LxLyRxy = new Lambda(x, new Lambda(y, Rxy, true), true);
                        Lambda LyPy = new Lambda(y, Py, true);
                        Type Y = LxLyRxy.getType();
                        Type Z = LyPy.getType();
                        
                        ExprToken token_LgLfLxCyfygxy = this.toExprToken(LgLfLxEyfygxy);
                        TypeToken token_Y = this.toTypeToken(Y);
                        TypeToken token_Z = this.toTypeToken(Z);
                        
                        ArrayList<LambdaProd> children1 = this.decompose(LxLyRxy);
                        ArrayList<LambdaProd> children2 = this.decompose(LyPy);
                        
                        LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LgLfLxCyfygxy, token_Y, token_Z);
                        treeRule1.setChildren(children1, children2);

                        if(debug) checkRestore(LgLfLxEyfygxy, LxLyRxy, LyPy, exp);
                        
                        treeRules.add(treeRule1);
                    }

                    /** (10302)
                     * X : toTree ( Lx.Ey.P(y)^Q(x)>=<R(y) )
                     * <=>
                     * ??
                     */
                    else if (Rxy.getExprCat() == Expr.EQUALITY) {
                        System.err.println(Rxy.getFreeVars());
                        throw new NotImplementedException("10302\t" + exp);
                    }

                    else {
                    	throw new NotImplementedException(">>10300\t" + exp);
                    }
                }
            }

            /** (10350)
             * X : toTree ( Lx.P(x) >=< EXP )
             * <=>
             * X : ( ( Lf.Lg.Lx.f(x) >=< g Y) Z)
             * Y : toTree(Lx.P(x))
             * Z : toTree(EXP)
             */
            else if (inner.getExprCat() == Expr.EQUALITY) {
                Equality eq = (Equality) inner;
                int equality = eq.getEquality();
                Expr left = eq.getLeft();
                Expr right = eq.getRight();

                /** (10351)
                 * X : toTree ( Lx.P(x) >=< EXP )
                 * <=>
                 * X : ( ( Lf.Lg.Lx.f(x) >=< g Y) Z)
                 * Y : toTree(Lx.P(x))
                 * Z : toTree(EXP)
                 */
                if (involvesOneVar(left, x) && !involvesOneVar(right, x)) {

                    Expr Px = left;
                    Expr EXP = right;
                    Var f = new Var(genVar(), new CompositeType(x.getType(), Px.getType()), true);
                    Var g = new Var(genVar(), EXP.getType(), true);

                    FunApp fx = new FunApp(f, x);

                    Lambda LfLgLxfxEQg = new Lambda(f, new Lambda(g, new Lambda(x, new Equality(fx, g, equality), true), true), true);
                    Lambda LxPx = new Lambda(x, Px, true);

                    Type Y = LxPx.getType();
                    Type Z = EXP.getType();

                    ArrayList<LambdaProd> children1 = this.decompose(LxPx);
                    ArrayList<LambdaProd> children2 = this.decompose(EXP);
                    
                    ExprToken token_LfLgLxfxEQg = this.toExprToken(LfLgLxfxEQg);
                    TypeToken token_Y = this.toTypeToken(Y);
                    TypeToken token_Z = this.toTypeToken(Z);
                    
                    LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxfxEQg, token_Y, token_Z);
                    treeRule1.setChildren(children1, children2);

                    if(debug) checkRestore(LfLgLxfxEQg, LxPx, EXP, exp);

                    treeRules.add(treeRule1);
                }

                /** (10352)
                 * X : toTree ( Lx.EXP >=< P(x) )
                 * <=>
                 * X : ( ( Lg.Lf.Lx.g >=< f(x) Y) Z)
                 * Y : toTree(EXP)
                 * Z : toTree(Lx.P(x))
                 */
                else if (!involvesOneVar(left, x) && involvesOneVar(right, x)) {

                    Expr EXP = left;
                    Expr Px = right;
                    Var f = new Var(genVar(), new CompositeType(x.getType(), Px.getType()), true);
                    Var g = new Var(genVar(), EXP.getType(), true);

                    FunApp fx = new FunApp(f, x);

                    Lambda LfLgLxgEQfx = new Lambda(g, new Lambda(f, new Lambda(x, new Equality(g, fx, equality), true), true), true);
                    Lambda LxPx = new Lambda(x, Px, true);

                    Type Y = EXP.getType();
                    Type Z = LxPx.getType();

                    ArrayList<LambdaProd> children1 = this.decompose(EXP);
                    ArrayList<LambdaProd> children2 = this.decompose(LxPx);
                    
                    ExprToken token_LfLgLxgEQfx = this.toExprToken(LfLgLxgEQfx);
                    TypeToken token_Y = this.toTypeToken(Y);
                    TypeToken token_Z = this.toTypeToken(Z);
                    
                    LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxgEQfx, token_Y, token_Z);
                    treeRule1.setChildren(children1, children2);

                    if(debug) checkRestore(LfLgLxgEQfx, EXP, LxPx, exp);

                    treeRules.add(treeRule1);
                }
                else {
                	throw new NotImplementedException("10350\t" + exp);
                }
            }

            /** (10400)
             * X : toTree ( Lx.p(x.) )
             */
            else if (inner.getExprCat() == Expr.FUNAPP) {
                FunApp funapp = (FunApp) inner;
                Expr p = funapp.getFunc().stripParens();
                Expr arg = funapp.getArg().stripParens();

                /** (10401)
                 * X : toTree ( Lx.p(x) )
                 * <=>
                 * X : Lx.p(x)
                 */
                if (arg.equals(x)) {
                	//do nothing.. it's already added
                }

                /** (10402)
                 * X : toTree ( Lx.p(x,EXP) )
                 * <=>
                 * X : ( Ly.Lx.p(x,y) Y )
                 * Y : toTree(EXP)
                 */
                else if (isLeftBinaryFunc(inner, x)) {
                	
                    Expr[] args = ((ArgList) arg).getArgs();
                    Expr EXP = args[1].stripParens();
                    Var y = new Var(genVar(), EXP.getType(), true);
                    Lambda LyLxpxy = new Lambda(y, new Lambda(x, new FunApp(p, new ArgList(new Expr[] {x, y})), true), true);
                    Type Y = EXP.getType();
                    
                    ExprToken token_LyLxpxy = this.toExprToken(LyLxpxy);
                    TypeToken token_Y = this.toTypeToken(Y);
                    
                    ArrayList<LambdaProd> children = this.decompose(EXP);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LyLxpxy, token_Y);
                    treeRule.setChildren(children);

                    if(debug) checkRestore(LyLxpxy, EXP, exp);

                    treeRules.add(treeRule);
                }

                /** (10403)
                 * X : toTree ( Lx.p(EXP,x) )
                 * <=>
                 * X : ( Ly.Lx.p(y,x) Y )
                 * Y : toTree(EXP)
                 */
                else if (isRightBinaryFunc(inner, x)) {

                    Expr[] args = ((ArgList) arg).getArgs();
                    Expr EXP = args[0].stripParens();
                    Var y = new Var(genVar(), EXP.getType(), true);
                    Lambda LyLxpyx = new Lambda(y, new Lambda(x, new FunApp(p, new ArgList(new Expr[] {y, x})), true), true);
                    Type Y = EXP.getType();
                    
                    ExprToken token_LyLxpyx = this.toExprToken(LyLxpyx);
                    TypeToken token_Y = this.toTypeToken(Y);
                    
                    ArrayList<LambdaProd> children = this.decompose(EXP);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LyLxpyx, token_Y);
                    treeRule.setChildren(children);

                    if(debug) checkRestore(LyLxpyx, EXP, exp);

                    treeRules.add(treeRule);
                }

                /** (10409)
                 * X : toTree ( Lx.p(EXP(x)) )
                 * <=>
                 * X : ((Lf.Lg.Lx.f(g(x)) Y) Z)
                 * Y : toTree(Lx.p(x))
                 * Z : toTree(Lx.EXP(x))
                 */
                else if (!arg.equals(x) && arg.getExprCat() != Expr.ARGLIST && involvesOneVar(arg, x)) {

                    Expr EXPx = arg;

                    Var g = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
                    FunApp gx = new FunApp(g, x);
                    Var f = new Var(genVar(), new CompositeType(gx.getType(), funapp.getType()), true);

                    FunApp fgx = new FunApp(f, gx);
                    FunApp px = new FunApp(p, x);

                    Lambda LfLgLxfgx = new Lambda(f, new Lambda(g, new Lambda(x, fgx, true), true), true);
                    Lambda Lxpx = new Lambda(x, px, true);
                    Lambda LxEXPx = new Lambda(x, EXPx, true);

                    Type Y = Lxpx.getType();
                    Type Z = LxEXPx.getType();

                    ArrayList<LambdaProd> children1 = this.decompose(Lxpx);
                    ArrayList<LambdaProd> children2 = this.decompose(LxEXPx);
                    
                    ExprToken token_LfLgLxfgx = this.toExprToken(LfLgLxfgx);
                    TypeToken token_Y = this.toTypeToken(Y);
                    TypeToken token_Z = this.toTypeToken(Z);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLgLxfgx, token_Y, token_Z);
                    treeRule.setChildren(children1, children2);

                    checkRestore(LfLgLxfgx, Lxpx, LxEXPx, exp);

                    treeRules.add(treeRule);
                }
                else{
                	throw new NotImplementedException("10400\t" + exp);
                }
            }

            /** (10500)
             * X : toTree ( Lx.~EXP(x) )
             * <=>
             * 
             * Lf.~f Lx.EXP(x)
             * 
             * X : ( Lf.Lx.~f(x) Y )
             * Y : toTree(Lx.EXP(x))
             */
            else if (inner.getExprCat() == Expr.NOT) {
            	
                Not not = (Not) inner;
                Expr EXPx = not.getInnerExpr();
                Var f = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
                Lambda LfLxNOTfx = new Lambda(f, new Lambda(x, new Not(new FunApp(f, x)), true), true);
                Expr LxEXPx = new Lambda(x, EXPx, true);
                Type Y = LxEXPx.getType();
                
                ArrayList<LambdaProd> children = this.decompose(LxEXPx);
                
                ExprToken token_LfLxNOTfx = this.toExprToken(LfLxNOTfx);
                TypeToken token_Y = this.toTypeToken(Y);
                LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLxNOTfx, token_Y);
                treeRule.setChildren(children);
                
                if(debug) checkRestore(LfLxNOTfx, LxEXPx, exp);

                treeRules.add(treeRule);
            }

            /** (10600)
             * X : toTree ( Lx.Cy.P(y)^R(x,y))
             * <=>
             * 
             * X : ( ( Lg.Lf.Lx.Cy.f(y)^g(x)(y) Y ) Z )
             * Y : toTree(Lx.Ly.R(x,y))
             * Z : toTree(Ly.P(y))
             */
            else if (inner.getExprCat() == Expr.COUNT) {
                Count count = (Count) inner;
                Identifier y = count.getVariable();

                Expr innerCount = count.getInnerExpr().stripParens();
//                System.err.println(innerCount+"\t"+innerCount.getExprCat());

                if (innerCount.getExprCat() == Expr.AND) {

                    And and = (And) innerCount;
                    Expr Py = and.getLeft();
                    Expr Rxy = and.getRight();

                    if (involvesTwoVars(Rxy, x, y)) {

                        Var g = new Var(genVar(), new CompositeType(x.getType(), new CompositeType(y.getType(), Rxy.getType())), true);
                        Var f = new Var(genVar(), new CompositeType(y.getType(), Py.getType()), true);

                        Expr fy = new FunApp(f, y);
                        Expr gxy = new FunApp(new FunApp(g, x), y);
                        Lambda LgLfLxCyfygxy = new Lambda(g, new Lambda(f, new Lambda(x, new Count(y, new And(fy, gxy), true), true), true), true);

                        Lambda LxLyRxy = new Lambda(x, new Lambda(y, Rxy, true), true);
                        Lambda LyPy = new Lambda(y, Py, true);
                        Type Y = LxLyRxy.getType();
                        Type Z = LyPy.getType();
                        
                        ArrayList<LambdaProd> children1 = this.decompose(LxLyRxy);
                        ArrayList<LambdaProd> children2 = this.decompose(LyPy);
                        
                        ExprToken token_LgLfLxCyfygxy = this.toExprToken(LgLfLxCyfygxy);
                        TypeToken token_Y = this.toTypeToken(Y);
                        TypeToken token_Z = this.toTypeToken(Z);
                        LambdaProd treeRule = this.toLambdaProd(token_X, token_LgLfLxCyfygxy, token_Y, token_Z);
                        treeRule.setChildren(children1, children2);
                        
                        if(debug) checkRestore(LgLfLxCyfygxy, LxLyRxy, LyPy, exp);

                        treeRules.add(treeRule);
                    }
                    else {
                    	throw new NotImplementedException("10600\t" + exp);
                    }
                }
            }

            /** (10700)
             * X : toTree ( Lx.Ly.p(x,y))
             * <=>
             * X : Lx.Ly.p(x,y)
             */
            else if (inner.getExprCat() == Expr.LAMBDA) {
                Lambda innerLambda = (Lambda) inner;
                Identifier y = innerLambda.getVariable();

                Expr innerInnerLambda = innerLambda.getInnerExpr().stripParens();

                /** (10701)
                 * X : toTree ( Lx.Ly.p(x,y) )
                 * <=>
                 * X : Lx.Ly.p(x,y)
                 */
                if (isDoubleBinaryFunc(innerInnerLambda, x, y)) {
                	//do nothing, it's already there.
                }

                /** (10702)
                 * X : toTree ( Lx.Ly.EXP_y )
                 * <=>
                 * X : (Lf.Lx.f Y)
                 * Y : toTree(Ly.EXP_y)
                 */
                else if (!involvesOneVar(innerInnerLambda, x)) {
                	
                    Expr EXP_y = innerInnerLambda;
                    
                    Var f = new Var(genVar(), EXP_y.getType(), true);
                    Lambda LfLxf = new Lambda(f, new Lambda(x, f, true), true);
                    Lambda LyEXP_y = new Lambda(y, EXP_y, true);
                    Type Y = LyEXP_y.getType();
                    
                    ArrayList<LambdaProd> children = this.decompose(LyEXP_y);
                    
                    ExprToken token_LfLxf = this.toExprToken(LfLxf);
                    TypeToken token_Y = this.toTypeToken(Y);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLxf, token_Y);
                    treeRule.setChildren(children);
                    
                    if(debug) checkRestore(LfLxf, LyEXP_y, exp);
                    
                    treeRules.add(treeRule);
                }

                /** (10703)
                 * X : toTree ( Lx.Ly.P(y)^R(x,y))
                 * <=>
                 * X : ( ( Lg.Lf.Lx.Ly.f(y)^g(x)(y) Y ) Z )
                 * Y : toTree(Lx.Ly.R(x,y))
                 * Z : toTree(Ly.P(y))
                 */
                else if (innerInnerLambda.getExprCat() == Expr.AND) {
                	
                    And and = (And) innerInnerLambda;
                    Expr Py = and.getLeft();
                    Expr Rxy = and.getRight();

                    if (involvesTwoVars(Rxy, x, y)) {
                        Var g = new Var(genVar(), new CompositeType(x.getType(), new CompositeType(y.getType(), Rxy.getType())), true);
                        Var f = new Var(genVar(), new CompositeType(y.getType(), Py.getType()), true);

                        Expr fy = new FunApp(f, y);
                        Expr gxy = new FunApp(new FunApp(g, x), y);
                        Lambda LgLfLxCyfygxy = new Lambda(g, new Lambda(f, new Lambda(x, new Lambda(y, new And(fy, gxy), true), true), true), true);

                        Lambda LxLyRxy = new Lambda(x, new Lambda(y, Rxy, true), true);
                        Lambda LyPy = new Lambda(y, Py, true);
                        Type Y = LxLyRxy.getType();
                        Type Z = LyPy.getType();
                        
                        ArrayList<LambdaProd> children1 = this.decompose(LxLyRxy);
                        ArrayList<LambdaProd> children2 = this.decompose(LyPy);
                        
                        ExprToken token_LgLfLxCyfygxy = this.toExprToken(LgLfLxCyfygxy);
                        TypeToken token_Y = this.toTypeToken(Y);
                        TypeToken token_Z = this.toTypeToken(Z);
                        LambdaProd treeRule = this.toLambdaProd(token_X, token_LgLfLxCyfygxy, token_Y, token_Z);
                        treeRule.setChildren(children1, children2);

                        if(debug) checkRestore(LgLfLxCyfygxy, LxLyRxy, LyPy, exp);

                        treeRules.add(treeRule);
                    }
                    else {
                    	throw new NotImplementedException("10703\t" + exp);
                    }
                }
                else {
                	throw new NotImplementedException("10700\t" + exp);
                }
            }

            /** (10800)
             * X : Lx.ArgMax(EXP1(x), EXP2(x))
             * <=>
             * X : toTree ( ((Lf.Lg.Lx.ArgMax(f(x),g(x)) Y) Z) )
             * Y : toTree(Lx.EXP1(x))
             * Z : toTree(Lx.EXP2(x))
             */
            else if (inner.getExprCat() == Expr.ARGMAX) {

                ArgMax argmax = (ArgMax) inner;
                Identifier argmax_var = argmax.getVariable();
                Expr argmax_e = ((FunApp) argmax.getInnerExpr()).getFunc();
                ArgList innerArgmax = ((ArgList) ((FunApp) argmax.getInnerExpr()).getArg());
                Expr[] args = innerArgmax.getArgs();
                Expr EXP1x = args[0];
                Expr EXP2x = args[1];

                Var f = new Var(genVar(), new CompositeType(x.getType(), EXP1x.getType()), true);
                Var g = new Var(genVar(), new CompositeType(x.getType(), EXP2x.getType()), true);
                FunApp fx = new FunApp(f, x);
                FunApp gx = new FunApp(g, x);

                Lambda LfLgLxArgMaxfxg = new Lambda(f,
                        new Lambda(g, new Lambda(x, new ArgMax(argmax_var, new FunApp(argmax_e, new ArgList(new Expr[] {fx, gx})), true), true), true), true);
                Lambda LxEXP1x = new Lambda(x, EXP1x, true);
                Lambda LxEXP2x = new Lambda(x, EXP2x, true);

                Type Y = LxEXP1x.getType();
                Type Z = LxEXP2x.getType();
                
                ArrayList<LambdaProd> children1 = this.decompose(LxEXP1x);
                ArrayList<LambdaProd> children2 = this.decompose(LxEXP2x);

                ExprToken token_LfLgLxArgMaxfxg = this.toExprToken(LfLgLxArgMaxfxg);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLgLxArgMaxfxg, token_Y, token_Z);
                treeRule.setChildren(children1, children2);

                if(debug) checkRestore(LfLgLxArgMaxfxg, LxEXP1x, LxEXP2x, exp);

                treeRules.add(treeRule);
            }

            /** (10900)
             * X : Lx.ArgMin(EXP(x), EXP2)
             * <=>
             * X : ??
             */
            else if (inner.getExprCat() == Expr.ARGMIN) {
                throw new NotImplementedException("10900\t" + exp);
            }
            
            else {
            	throw new NotImplementedException("11000\t" + exp);
            }

        }
        
        if(treeRules.size() == size_before){
            LambdaProd expProd = this.toLambdaProd(token_X, token_self);
        	treeRules.add(expProd);
        }
        
//        ps.println("Not supported:"+exp.toString()+" with type:"+exp.getExprCat());
//        System.exit(1);
		
        return treeRules;
	}
	

	private ArrayList<LambdaProd> decompose_single(Expr exp) throws UnsupportedEncodingException, TypeEvaluationException{
		
		ArrayList<LambdaProd> treeRules = new ArrayList<LambdaProd>();
		
		exp = exp.stripAnyDoubleParens().stripOutermostParens();
		
        TypeToken token_X = this.toTypeToken(exp.getType());
    	ExprToken token_self = this.toExprToken(exp);
        
        /**
         * If it is a constant, add one transit prod.
         */
    	/**/
        if(exp instanceof Const){
            LambdaProd expProd = this.toLambdaProd(token_X, token_self);
            Var x = new Var(genVar(), exp.getType(), true);
            Expr Lxx = new Lambda(x, x, true);
        	ExprToken token_Lxx = this.toExprToken(Lxx);
        	LambdaProd transProd = this.toLambdaProd(token_X, token_Lxx, token_X);
        	ArrayList<LambdaProd> children = new ArrayList<LambdaProd>();
        	children.add(expProd);
        	transProd.setChildren(children);
        	treeRules.add(transProd);
        	if(debug) checkRestore(Lxx, exp, exp);
        	return treeRules;
        }
        /**/
        
        int size_before = treeRules.size();
		
		//ps.println(exp);

        /** (1000)
         * If the outer most is constant
         * X : toTree ( c )
         * <=>
         * X : c
         */
        if (exp.getExprCat() == Expr.CONST) {
        	//do nothing, already added..
        }
        /** (2000)
         * If the outer most is Count
         * X : toTree ( Cx.P(x) )
         * <=>
         * X : ( Lf.Cx.f(x) Y )
         * Y : toTree ( Lx.P(x) )
         */
        else if (exp.getExprCat() == Expr.COUNT) {
        	Count count = (Count) exp;
            Expr inner = count.getInnerExpr().stripParens();
            
            Identifier x = count.getVariable();
            
            Var f = new Var(genVar(), Type.ET, true);
            Lambda LfCxfx = new Lambda(f, new Count(x, new FunApp(f, x), true), true);
            Lambda LxPx = new Lambda(x, inner, true);
            
            Type Y = LxPx.getType();
            
            ExprToken token_exp = this.toExprToken(LfCxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_exp, token_Y);
            ArrayList<LambdaProd> treeRule_children = this.decompose_single(LxPx);
            
            treeRule.setChildren(treeRule_children);
            
            if(debug) checkRestore(LfCxfx, LxPx, exp);
            
            treeRules.add(treeRule);
        }
        /** (3000)
         * If the outer most is FunApp
         */
        else if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr p = funapp.getFunc().stripParens();
            Expr EXP = funapp.getArg().stripParens();
            
            /** (3001)
             * X : toTree ( p(EXP) )
             * <=>
             * X : ( ( Lf.Lx.f(x) Y ) Z )
             * Y : toTree ( Lx.p(x) )
             * Z : toTree ( EXP )
             */

            if (EXP.getExprCat() != Expr.ARGLIST && p.getExprCat() == Expr.CONST) {
                Var x = new Var(genVar(), EXP.getType(), true);
                Lambda Lxpx = new Lambda(x, new FunApp(p, x), true);
                Type Y = Lxpx.getType();
                Type Z = EXP.getType();
                
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                Var f = new Var(genVar(), Y, true);
                Expr LfLxfx = new Lambda(f, new Lambda(x, new FunApp(f,x), true), true);
                ExprToken token_LfLxfx = this.toExprToken(LfLxfx);
                
                LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLxfx, token_Y, token_Z);
                ArrayList<LambdaProd> dec_Lxpx = this.decompose_single(Lxpx);
                ArrayList<LambdaProd> dec_EXP = this.decompose_single(EXP);
                treeRule1.setChildren(dec_Lxpx, dec_EXP);
                
                if(debug) checkRestore(LfLxfx, Lxpx, EXP, exp);
                
                treeRules.add(treeRule1);
            }
            /** (3002)
             * X : toTree ( p(EXP1, EXP2) )
             * <=>
             * X : ( (Lx.Ly.p(x,y) Y) Z)
             * Y : toTree (EXP1)
             * Z : toTree (EXP2)
             */
            else if (EXP.getExprCat() == Expr.ARGLIST) {
                Expr[] args = ((ArgList) EXP).getArgs();
                Expr EXP1 = args[0];
                Expr EXP2 = args[1];

                Var x = new Var(genVar(), EXP1.getType(), true);
                Var y = new Var(genVar(), EXP2.getType(), true);

                Type Y = EXP1.getType();
                Type Z = EXP2.getType();

                Lambda LxLypxy = new Lambda(x, new Lambda(y, new FunApp(p, new ArgList(new Expr[] {x, y})), true), true);
                
                ExprToken token_LxLypxy = this.toExprToken(LxLypxy);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                LambdaProd treeRule = this.toLambdaProd(token_X, token_LxLypxy, token_Y, token_Z);
                
                ArrayList<LambdaProd> dec_EXP1 = this.decompose_single(EXP1);
                ArrayList<LambdaProd> dec_EXP2 = this.decompose_single(EXP2);
                
                treeRule.setChildren(dec_EXP1, dec_EXP2);

                if(debug) checkRestore(LxLypxy, EXP1, EXP2, exp);
                
                treeRules.add(treeRule);
            }
            else{
            	throw new NotImplementedException("3003 not implemented!");
            }
        }
        /** (4000)
         * If the outer most is Iota
         * X : toTree ( Ix.EXP )
         * <=>
         * X : ( Lf.Ix.f(x) Y )
         * Y : toTree ( Lx.EXP )
         */
        else if (exp.getExprCat() == Expr.IOTA) {

            Iota iota = (Iota) exp;
            Identifier x = iota.getVariable();
            Expr EXP = iota.getInnerExpr().stripParens();
            Lambda lambda = new Lambda(x, EXP, true);
            Type Y = lambda.getType();
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXP.getType()), true);
            Lambda LfIxfx = new Lambda(f, new Iota(x, new FunApp(f, x), true), true);
            Lambda LxEXP = new Lambda(x, EXP, true);

            ExprToken token_LfIxfx = this.toExprToken(LfIxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfIxfx, token_Y);
            treeRule.setChildren(this.decompose_single(LxEXP));
            
            if(debug) checkRestore(LfIxfx, LxEXP, exp);
            
            treeRules.add(treeRule);
        }
        /** (5000)
         * If the outer most is Not
         * X : toTree ( ~EXP )
         * <=>
         * X : ( Lv.~v Y )
         * Y : toTree ( EXP )
         */
        else if (exp.getExprCat() == Expr.NOT) {
            Not not = (Not) exp;
            Expr EXP = not.getInnerExpr().stripParens();
            Type X = exp.getType();
            Type Y = EXP.getType();
            
            Var v = new Var(genVar(), X, true);
            Lambda LvNOTv = new Lambda(v, new Not(v), true);

            ExprToken token_LvNOTv = this.toExprToken(LvNOTv);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LvNOTv, token_Y);
            treeRule.setChildren(this.decompose_single(EXP));
            
            if(debug) checkRestore(LvNOTv, EXP, exp);

            treeRules.add(treeRule);
        }
        /** (6000)
         * If the outer most is Exists
         * X : toTree ( Ex.EXP )
         * <=>
         * X : ( Lf.Ex.f(x) Y )
         * Y : toTree ( Lx.EXP )
         */
        else if (exp.getExprCat() == Expr.EXISTS) {
            Exists exists = (Exists) exp;
            Expr EXP = exists.getInnerExpr().stripParens();
            Identifier x = exists.getVariable();
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXP.getType()), true);
            Lambda LfExfx = new Lambda(f, new Exists(x, new FunApp(f, x), true), true);
            Lambda LxEXP = new Lambda(x, EXP, true);
            Type Y = LxEXP.getType();

            ExprToken token_LfExfx = this.toExprToken(LfExfx);
            TypeToken token_Y = this.toTypeToken(Y);
            
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfExfx, token_Y);
            treeRule.setChildren(this.decompose_single(LxEXP));
            
            if(debug) checkRestore(LfExfx, LxEXP, exp);

            treeRules.add(treeRule);
        }
        /** (7000)
         * If the outer most is Sum
         * X : toTree ( S(EXP1, EXP2) )
         * <=>
         * X : ( ((Lf.Lg.S(f,g) Y) Z) )
         * Y : toTree ( EXP1 )
         * Z : toTree ( EXP2 )
         */
        else if (exp.getExprCat() == Expr.SUM) {
            Sum sum = (Sum) exp;
            Identifier x = sum.getVariable();
            FunApp fa = (FunApp) sum.getInnerExpr().stripParens();
            Expr sum_e = fa.getFunc(); //sum
            Expr[] args = ((ArgList) fa.getArg()).getArgs();
            Expr EXP1 = args[0].stripParens();
            Expr EXP2 = args[1].stripParens();
            Type Y = EXP1.getType();
            Type Z = EXP2.getType();

            Var f = new Var(genVar(), Y, true);
            Var g = new Var(genVar(), Z, true);

            Lambda LfLgSumfg = new Lambda(f, new Lambda(g, new Sum(x, new FunApp(sum_e, new ArgList(new Expr[] {f, g})), true), true), true);
            
            ExprToken token_LfLgSumfg = this.toExprToken(LfLgSumfg);
            TypeToken token_Y = this.toTypeToken(Y);
            TypeToken token_Z = this.toTypeToken(Z);
            
            LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgSumfg, token_Y, token_Z);
            ArrayList<LambdaProd> children1 = this.decompose_single(EXP1);
            ArrayList<LambdaProd> children2 = this.decompose_single(EXP2);
            treeRule1.setChildren(children1, children2);
            
            if(debug) checkRestore(LfLgSumfg, EXP1, EXP2, exp);
            
            treeRules.add(treeRule1);
        }
        /** (8000)
         * If the outer most is ArgMax
         * X : toTree ( ArgMax(EXP1, EXP2) )
         * <=>
         * X : ( ((Lf.Lg.ArgMax(f,g) Y) Z) )
         * Y : toTree ( EXP1 )
         * Z : toTree ( EXP2 )
         */
        else if (exp.getExprCat() == Expr.ARGMAX) {
        	
            ArgMax argmax = (ArgMax) exp;
            Identifier x = argmax.getVariable();
            Expr inner = argmax.getInnerExpr().stripParens();
            FunApp funapp = (FunApp) inner;
            Expr argmax_e = funapp.getFunc(); //argmax
            Expr[] args = ((ArgList) funapp.getArg()).getArgs();
            Expr EXP1 = args[0].stripParens();
            Expr EXP2 = args[1].stripParens();
            
            
            Type Y = EXP1.getType();
            Type Z = EXP2.getType();
            
            Var f = new Var(genVar(), EXP1.getType(), true);
            Var g = new Var(genVar(), EXP2.getType(), true);
            Lambda LfLgArgMaxfg = new Lambda(f, new Lambda(g, new ArgMax(x, new FunApp(argmax_e, new ArgList(new Expr[] {f, g})), true), true), true);
            
            ExprToken token_LfLgArgMaxfg = this.toExprToken(LfLgArgMaxfg);
            TypeToken token_Y = this.toTypeToken(Y);
            TypeToken token_Z = this.toTypeToken(Z);
            
            LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgArgMaxfg, token_Y, token_Z);
            ArrayList<LambdaProd> children1 = this.decompose_single(EXP1);
            ArrayList<LambdaProd> children2 = this.decompose_single(EXP2);
            treeRule1.setChildren(children1, children2);

            treeRules.add(treeRule1);
            if(debug) checkRestore(LfLgArgMaxfg, EXP1, EXP2, exp);
            
        }

        /** (9000)
         * If the outer most is ArgMin
         * X : toTree ( ArgMin(EXP1, EXP2) )
         * <=>
         * X : ( ((Lf.Lg.ArgMin(f,g) Y) Z) )
         * Y : toTree ( EXP1 )
         * Z : toTree ( EXP2 )
         */
        else if (exp.getExprCat() == Expr.ARGMIN) {

        	ArgMin argmin = (ArgMin) exp;
            Identifier x = argmin.getVariable();
            Expr inner = argmin.getInnerExpr().stripParens();
            FunApp funapp = (FunApp) inner;
            Expr argmin_e = funapp.getFunc(); //argmin
            Expr[] args = ((ArgList) funapp.getArg()).getArgs();
            Expr EXP1 = args[0].stripParens();
            Expr EXP2 = args[1].stripParens();

            Type Y = EXP1.getType();
            Type Z = EXP2.getType();

            Var f = new Var(genVar(), EXP1.getType(), true);
            Var g = new Var(genVar(), EXP2.getType(), true);
            Lambda LfLgArgMinfg = new Lambda(f, new Lambda(g, new ArgMin(x, new FunApp(argmin_e, new ArgList(new Expr[] {f, g})), true), true), true);
            
            ExprToken token_LfLgArgMinfg = this.toExprToken(LfLgArgMinfg);
            TypeToken token_Y = this.toTypeToken(Y);
            TypeToken token_Z = this.toTypeToken(Z);
            
            LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgArgMinfg, token_Y, token_Z);
            ArrayList<LambdaProd> children1 = this.decompose_single(EXP1);
            ArrayList<LambdaProd> children2 = this.decompose_single(EXP2);
            treeRule1.setChildren(children1, children2);
            
            if(debug) checkRestore(LfLgArgMinfg, EXP1, EXP2, exp);

            treeRules.add(treeRule1);
        }
        /** (80000)
         * X : toTree( Max x. EXP(x))
         * <=>
         * X : ( Lf.Max x.f(x) Y )
         * Y : toTree(Lx.EXP(x))
         */
        else if (exp.getExprCat() == Expr.MAX) {
            Max max = (Max) exp;
            FunApp fa = (FunApp) max.getInnerExpr();
            Expr max_e = fa.getFunc();
            Expr EXPx = fa.getArg();
            Identifier x = max.getVariable();
            
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
            Lambda LfMaxxfx = new Lambda(f, new Max(x, new FunApp(max_e, new ArgList(new Expr[] {f, x})), true), true);
            Lambda LxEXPx = new Lambda(x, EXPx, true);
            
            Type Y = LxEXPx.getType();
            
            ExprToken token_LfMaxxfx = this.toExprToken(LfMaxxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfMaxxfx, token_Y);
            treeRule.setChildren(this.decompose_single(LxEXPx));
            
            if(debug) checkRestore(LfMaxxfx, LxEXPx, exp);

            treeRules.add(treeRule);
        }

        /** (90000)
         * X : toTree( Min x. EXP(x))
         * <=>
         * X : ( Lf.Min x.f(x) Y )
         * Y : toTree(Lx.EXP(x))
         */
        else if (exp.getExprCat() == Expr.MIN) {
        	Min min = (Min) exp;
            FunApp fa = (FunApp) min.getInnerExpr();
            Expr min_e = fa.getFunc();
            Expr EXPx = fa.getArg();
            Identifier x = min.getVariable();
            
            Var f = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
            Lambda LfMinxfx = new Lambda(f, new Min(x, new FunApp(min_e, new ArgList(new Expr[] {f, x})), true), true);
            Lambda LxEXPx = new Lambda(x, EXPx, true);
            
            Type Y = LxEXPx.getType();
            
            ExprToken token_LfMinxfx = this.toExprToken(LfMinxfx);
            TypeToken token_Y = this.toTypeToken(Y);
            LambdaProd treeRule = this.toLambdaProd(token_X, token_LfMinxfx, token_Y);
            treeRule.setChildren(this.decompose_single(LxEXPx));
            
            if(debug) checkRestore(LfMinxfx, LxEXPx, exp);

            treeRules.add(treeRule);
        }


        /** (10000)
         * If the outer most is Lambda
         */
        else if (exp.getExprCat() == Expr.LAMBDA) {
            Lambda lambda = (Lambda) exp;
            Expr inner = lambda.getInnerExpr().stripParens();
            Identifier x = lambda.getVariable();

            /** (10100)
             * X : toTree ( Lx.EXP1 ^ EXP2 )
             * <=>
             * X : ( ((Lf.Lg.Lx.f(x) ^ g(x) Y) Z) )
             * Y : toTree ( Lx.EXP1 )
             * Z : toTree ( Lx.EXP2 )
             */
            if (inner.getExprCat() == Expr.AND) {
                And and = (And) inner;
                Expr EXP1 = and.getLeft().stripParens();
                Expr EXP2 = and.getRight().stripParens();
                
                Lambda LxEXP1 = new Lambda(x, EXP1, true);
                Lambda LxEXP2 = new Lambda(x, EXP2, true);
                Type Y = LxEXP1.getType();
                Type Z = LxEXP2.getType();
                
                Var f = new Var(genVar(), Y, true);
                Var g = new Var(genVar(), Z, true);
                
                FunApp fx = new FunApp(f, x);
                FunApp gx = new FunApp(g, x);
                
                Lambda LfLgLxfxANDgx = new Lambda(f, new Lambda(g, new Lambda(x, new And(fx, gx), true), true), true);
                
                ExprToken token_LfLgLxfxANDgx = this.toExprToken(LfLgLxfxANDgx);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                ArrayList<LambdaProd> children1 = this.decompose_single(LxEXP1);
                ArrayList<LambdaProd> children2 = this.decompose_single(LxEXP2);
                
                LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxfxANDgx, token_Y, token_Z);
                treeRule1.setChildren(children1, children2);
                
                if(debug) checkRestore(LfLgLxfxANDgx, LxEXP1, LxEXP2, exp);
                
                treeRules.add(treeRule1);
            }

            /** (10200)
             * X : toTree ( Lx.EXP1 v EXP2 )
             * <=>
             * X : ( ((Lf.Lg.Lx.f(x) v g(x) Y) Z) )
             * Y : toTree ( Lx.EXP1 )
             * Z : toTree ( Lx.EXP2 )
             */
            else if (inner.getExprCat() == Expr.OR) {
                Or or = (Or) inner;
                Expr EXP1 = or.getLeft();
                Expr EXP2 = or.getRight();

                Lambda LxEXP1 = new Lambda(x, EXP1, true);
                Lambda LxEXP2 = new Lambda(x, EXP2, true);
                Type Y = LxEXP1.getType();
                Type Z = LxEXP2.getType();

                Var f = new Var(genVar(), Y, true);
                Var g = new Var(genVar(), Z, true);

                FunApp fx = new FunApp(f, x);
                FunApp gx = new FunApp(g, x);

                Lambda LfLgLxfxORgx = new Lambda(f, new Lambda(g, new Lambda(x, new Or(fx, gx), true), true), true);

                ExprToken token_LfLgLxfxORgx = this.toExprToken(LfLgLxfxORgx);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                
                ArrayList<LambdaProd> children1 = this.decompose_single(LxEXP1);
                ArrayList<LambdaProd> children2 = this.decompose_single(LxEXP2);
                
                LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxfxORgx, token_Y, token_Z);
                treeRule1.setChildren(children1, children2);
                
                if(debug) checkRestore(LfLgLxfxORgx, LxEXP1, LxEXP2, exp);
                
                treeRules.add(treeRule1);
            }

            /** (10300)
             * X : toTree ( Lx.Ey.P(y)^R(x,y))
             * <=>
             * X : ( ( Lg.Lf.Lx.Ey.f(y)^g(x)(y) Y ) Z )
             * Y : toTree(Lx.Ly.R(x,y))
             * Z : toTree(Ly.P(y))
             */
            else if (inner.getExprCat() == Expr.EXISTS) {
                Exists exists = (Exists) inner;
                Identifier y = exists.getVariable();

                Expr innerExists = exists.getInnerExpr().stripParens();

                if (innerExists.getExprCat() == Expr.AND) {
                	
                    And and = (And) innerExists;
                    Expr Py = and.getLeft();
                    Expr Rxy = and.getRight();

                    /** (10301)
                     * X : toTree ( Lx.Ey.P(y)^R(x,y) )
                     * <=>
                     * X : ( ( Lg.Lf.Lx.Ey.f(y)^g(x)(y) Y ) Z )
                     * Y : toTree( Lx.Ly.R(x,y) )
                     * Z : toTree( Ly.P(y) )
                     */
                    if (involvesTwoVars(Rxy, x, y)) {
                        Var g = new Var(genVar(), new CompositeType(x.getType(), new CompositeType(y.getType(), Rxy.getType())), true);
                        Var f = new Var(genVar(), new CompositeType(y.getType(), Py.getType()), true);

                        Expr fy = new FunApp(f, y);
                        Expr gxy = new FunApp(new FunApp(g, x), y);
                        Lambda LgLfLxEyfygxy = new Lambda(g, new Lambda(f, new Lambda(x, new Exists(y, new And(fy, gxy), true), true), true), true);

                        Lambda LxLyRxy = new Lambda(x, new Lambda(y, Rxy, true), true);
                        Lambda LyPy = new Lambda(y, Py, true);
                        Type Y = LxLyRxy.getType();
                        Type Z = LyPy.getType();
                        
                        ExprToken token_LgLfLxCyfygxy = this.toExprToken(LgLfLxEyfygxy);
                        TypeToken token_Y = this.toTypeToken(Y);
                        TypeToken token_Z = this.toTypeToken(Z);
                        
                        ArrayList<LambdaProd> children1 = this.decompose_single(LxLyRxy);
                        ArrayList<LambdaProd> children2 = this.decompose_single(LyPy);
                        
                        LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LgLfLxCyfygxy, token_Y, token_Z);
                        treeRule1.setChildren(children1, children2);

                        if(debug) checkRestore(LgLfLxEyfygxy, LxLyRxy, LyPy, exp);
                        
                        treeRules.add(treeRule1);
                    }

                    /** (10302)
                     * X : toTree ( Lx.Ey.P(y)^Q(x)>=<R(y) )
                     * <=>
                     * ??
                     */
                    else if (Rxy.getExprCat() == Expr.EQUALITY) {
                        System.err.println(Rxy.getFreeVars());
                        throw new NotImplementedException("10302\t" + exp);
                    }

                    else {
                    	throw new NotImplementedException(">>10300\t" + exp);
                    }
                }
            }

            /** (10350)
             * X : toTree ( Lx.P(x) >=< EXP )
             * <=>
             * X : ( ( Lf.Lg.Lx.f(x) >=< g Y) Z)
             * Y : toTree(Lx.P(x))
             * Z : toTree(EXP)
             */
            else if (inner.getExprCat() == Expr.EQUALITY) {
                Equality eq = (Equality) inner;
                int equality = eq.getEquality();
                Expr left = eq.getLeft();
                Expr right = eq.getRight();

                /** (10351)
                 * X : toTree ( Lx.P(x) >=< EXP )
                 * <=>
                 * X : ( ( Lf.Lg.Lx.f(x) >=< g Y) Z)
                 * Y : toTree(Lx.P(x))
                 * Z : toTree(EXP)
                 */
                if (involvesOneVar(left, x) && !involvesOneVar(right, x)) {

                    Expr Px = left;
                    Expr EXP = right;
                    Var f = new Var(genVar(), new CompositeType(x.getType(), Px.getType()), true);
                    Var g = new Var(genVar(), EXP.getType(), true);

                    FunApp fx = new FunApp(f, x);

                    Lambda LfLgLxfxEQg = new Lambda(f, new Lambda(g, new Lambda(x, new Equality(fx, g, equality), true), true), true);
                    Lambda LxPx = new Lambda(x, Px, true);

                    Type Y = LxPx.getType();
                    Type Z = EXP.getType();

                    ArrayList<LambdaProd> children1 = this.decompose_single(LxPx);
                    ArrayList<LambdaProd> children2 = this.decompose_single(EXP);
                    
                    ExprToken token_LfLgLxfxEQg = this.toExprToken(LfLgLxfxEQg);
                    TypeToken token_Y = this.toTypeToken(Y);
                    TypeToken token_Z = this.toTypeToken(Z);
                    
                    LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxfxEQg, token_Y, token_Z);
                    treeRule1.setChildren(children1, children2);

                    if(debug) checkRestore(LfLgLxfxEQg, LxPx, EXP, exp);

                    treeRules.add(treeRule1);
                }

                /** (10352)
                 * X : toTree ( Lx.EXP >=< P(x) )
                 * <=>
                 * X : ( ( Lg.Lf.Lx.g >=< f(x) Y) Z)
                 * Y : toTree(EXP)
                 * Z : toTree(Lx.P(x))
                 */
                else if (!involvesOneVar(left, x) && involvesOneVar(right, x)) {

                    Expr EXP = left;
                    Expr Px = right;
                    Var f = new Var(genVar(), new CompositeType(x.getType(), Px.getType()), true);
                    Var g = new Var(genVar(), EXP.getType(), true);

                    FunApp fx = new FunApp(f, x);

                    Lambda LfLgLxgEQfx = new Lambda(g, new Lambda(f, new Lambda(x, new Equality(g, fx, equality), true), true), true);
                    Lambda LxPx = new Lambda(x, Px, true);

                    Type Y = EXP.getType();
                    Type Z = LxPx.getType();

                    ArrayList<LambdaProd> children1 = this.decompose_single(EXP);
                    ArrayList<LambdaProd> children2 = this.decompose_single(LxPx);
                    
                    ExprToken token_LfLgLxgEQfx = this.toExprToken(LfLgLxgEQfx);
                    TypeToken token_Y = this.toTypeToken(Y);
                    TypeToken token_Z = this.toTypeToken(Z);
                    
                    LambdaProd treeRule1 = this.toLambdaProd(token_X, token_LfLgLxgEQfx, token_Y, token_Z);
                    treeRule1.setChildren(children1, children2);

                    if(debug) checkRestore(LfLgLxgEQfx, EXP, LxPx, exp);

                    treeRules.add(treeRule1);
                }
                else {
                	throw new NotImplementedException("10350\t" + exp);
                }
            }

            /** (10400)
             * X : toTree ( Lx.p(x.) )
             */
            else if (inner.getExprCat() == Expr.FUNAPP) {
                FunApp funapp = (FunApp) inner;
                Expr p = funapp.getFunc().stripParens();
                Expr arg = funapp.getArg().stripParens();

                /** (10401)
                 * X : toTree ( Lx.p(x) )
                 * <=>
                 * X : Lx.p(x)
                 */
                if (arg.equals(x)) {
                	//do nothing.. it's already added
                }

                /** (10402)
                 * X : toTree ( Lx.p(x,EXP) )
                 * <=>
                 * X : ( Ly.Lx.p(x,y) Y )
                 * Y : toTree(EXP)
                 */
                else if (isLeftBinaryFunc(inner, x)) {
                	
                    Expr[] args = ((ArgList) arg).getArgs();
                    Expr EXP = args[1].stripParens();
                    Var y = new Var(genVar(), EXP.getType(), true);
                    Lambda LyLxpxy = new Lambda(y, new Lambda(x, new FunApp(p, new ArgList(new Expr[] {x, y})), true), true);
                    Type Y = EXP.getType();
                    
                    ExprToken token_LyLxpxy = this.toExprToken(LyLxpxy);
                    TypeToken token_Y = this.toTypeToken(Y);
                    
                    ArrayList<LambdaProd> children = this.decompose_single(EXP);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LyLxpxy, token_Y);
                    treeRule.setChildren(children);

                    if(debug) checkRestore(LyLxpxy, EXP, exp);

                    treeRules.add(treeRule);
                }

                /** (10403)
                 * X : toTree ( Lx.p(EXP,x) )
                 * <=>
                 * X : ( Ly.Lx.p(y,x) Y )
                 * Y : toTree(EXP)
                 */
                else if (isRightBinaryFunc(inner, x)) {

                    Expr[] args = ((ArgList) arg).getArgs();
                    Expr EXP = args[0].stripParens();
                    Var y = new Var(genVar(), EXP.getType(), true);
                    Lambda LyLxpyx = new Lambda(y, new Lambda(x, new FunApp(p, new ArgList(new Expr[] {y, x})), true), true);
                    Type Y = EXP.getType();
                    
                    ExprToken token_LyLxpyx = this.toExprToken(LyLxpyx);
                    TypeToken token_Y = this.toTypeToken(Y);
                    
                    ArrayList<LambdaProd> children = this.decompose_single(EXP);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LyLxpyx, token_Y);
                    treeRule.setChildren(children);

                    if(debug) checkRestore(LyLxpyx, EXP, exp);

                    treeRules.add(treeRule);
                }

                /** (10409)
                 * X : toTree ( Lx.p(EXP(x)) )
                 * <=>
                 * X : ((Lf.Lg.Lx.f(g(x)) Y) Z)
                 * Y : toTree(Lx.p(x))
                 * Z : toTree(Lx.EXP(x))
                 */
                else if (!arg.equals(x) && arg.getExprCat() != Expr.ARGLIST && involvesOneVar(arg, x)) {

                    Expr EXPx = arg;

                    Var g = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
                    FunApp gx = new FunApp(g, x);
                    Var f = new Var(genVar(), new CompositeType(gx.getType(), funapp.getType()), true);

                    FunApp fgx = new FunApp(f, gx);
                    FunApp px = new FunApp(p, x);

                    Lambda LfLgLxfgx = new Lambda(f, new Lambda(g, new Lambda(x, fgx, true), true), true);
                    Lambda Lxpx = new Lambda(x, px, true);
                    Lambda LxEXPx = new Lambda(x, EXPx, true);

                    Type Y = Lxpx.getType();
                    Type Z = LxEXPx.getType();

                    ArrayList<LambdaProd> children1 = this.decompose_single(Lxpx);
                    ArrayList<LambdaProd> children2 = this.decompose_single(LxEXPx);
                    
                    ExprToken token_LfLgLxfgx = this.toExprToken(LfLgLxfgx);
                    TypeToken token_Y = this.toTypeToken(Y);
                    TypeToken token_Z = this.toTypeToken(Z);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLgLxfgx, token_Y, token_Z);
                    treeRule.setChildren(children1, children2);

                    checkRestore(LfLgLxfgx, Lxpx, LxEXPx, exp);

                    treeRules.add(treeRule);
                }
                else{
                	throw new NotImplementedException("10400\t" + exp);
                }
            }

            /** (10500)
             * X : toTree ( Lx.~EXP(x) )
             * <=>
             * 
             * Lf.~f Lx.EXP(x)
             * 
             * X : ( Lf.Lx.~f(x) Y )
             * Y : toTree(Lx.EXP(x))
             */
            else if (inner.getExprCat() == Expr.NOT) {
            	
                Not not = (Not) inner;
                Expr EXPx = not.getInnerExpr();
                Var f = new Var(genVar(), new CompositeType(x.getType(), EXPx.getType()), true);
                Lambda LfLxNOTfx = new Lambda(f, new Lambda(x, new Not(new FunApp(f, x)), true), true);
                Expr LxEXPx = new Lambda(x, EXPx, true);
                Type Y = LxEXPx.getType();
                
                ArrayList<LambdaProd> children = this.decompose_single(LxEXPx);
                
                ExprToken token_LfLxNOTfx = this.toExprToken(LfLxNOTfx);
                TypeToken token_Y = this.toTypeToken(Y);
                LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLxNOTfx, token_Y);
                treeRule.setChildren(children);
                
                if(debug) checkRestore(LfLxNOTfx, LxEXPx, exp);

                treeRules.add(treeRule);
            }

            /** (10600)
             * X : toTree ( Lx.Cy.P(y)^R(x,y))
             * <=>
             * 
             * X : ( ( Lg.Lf.Lx.Cy.f(y)^g(x)(y) Y ) Z )
             * Y : toTree(Lx.Ly.R(x,y))
             * Z : toTree(Ly.P(y))
             */
            else if (inner.getExprCat() == Expr.COUNT) {
                Count count = (Count) inner;
                Identifier y = count.getVariable();

                Expr innerCount = count.getInnerExpr().stripParens();
//                System.err.println(innerCount+"\t"+innerCount.getExprCat());

                if (innerCount.getExprCat() == Expr.AND) {

                    And and = (And) innerCount;
                    Expr Py = and.getLeft();
                    Expr Rxy = and.getRight();

                    if (involvesTwoVars(Rxy, x, y)) {

                        Var g = new Var(genVar(), new CompositeType(x.getType(), new CompositeType(y.getType(), Rxy.getType())), true);
                        Var f = new Var(genVar(), new CompositeType(y.getType(), Py.getType()), true);

                        Expr fy = new FunApp(f, y);
                        Expr gxy = new FunApp(new FunApp(g, x), y);
                        Lambda LgLfLxCyfygxy = new Lambda(g, new Lambda(f, new Lambda(x, new Count(y, new And(fy, gxy), true), true), true), true);

                        Lambda LxLyRxy = new Lambda(x, new Lambda(y, Rxy, true), true);
                        Lambda LyPy = new Lambda(y, Py, true);
                        Type Y = LxLyRxy.getType();
                        Type Z = LyPy.getType();
                        
                        ArrayList<LambdaProd> children1 = this.decompose_single(LxLyRxy);
                        ArrayList<LambdaProd> children2 = this.decompose_single(LyPy);
                        
                        ExprToken token_LgLfLxCyfygxy = this.toExprToken(LgLfLxCyfygxy);
                        TypeToken token_Y = this.toTypeToken(Y);
                        TypeToken token_Z = this.toTypeToken(Z);
                        LambdaProd treeRule = this.toLambdaProd(token_X, token_LgLfLxCyfygxy, token_Y, token_Z);
                        treeRule.setChildren(children1, children2);
                        
                        if(debug) checkRestore(LgLfLxCyfygxy, LxLyRxy, LyPy, exp);

                        treeRules.add(treeRule);
                    }
                    else {
                    	throw new NotImplementedException("10600\t" + exp);
                    }
                }
            }

            /** (10700)
             * X : toTree ( Lx.Ly.p(x,y))
             * <=>
             * X : Lx.Ly.p(x,y)
             */
            else if (inner.getExprCat() == Expr.LAMBDA) {
                Lambda innerLambda = (Lambda) inner;
                Identifier y = innerLambda.getVariable();

                Expr innerInnerLambda = innerLambda.getInnerExpr().stripParens();

                /** (10701)
                 * X : toTree ( Lx.Ly.p(x,y) )
                 * <=>
                 * X : Lx.Ly.p(x,y)
                 */
                if (isDoubleBinaryFunc(innerInnerLambda, x, y)) {
                	//do nothing, it's already there.
                }

                /** (10702)
                 * X : toTree ( Lx.Ly.EXP_y )
                 * <=>
                 * X : (Lf.Lx.f Y)
                 * Y : toTree(Ly.EXP_y)
                 */
                else if (!involvesOneVar(innerInnerLambda, x)) {
                	
                    Expr EXP_y = innerInnerLambda;
                    
                    Var f = new Var(genVar(), EXP_y.getType(), true);
                    Lambda LfLxf = new Lambda(f, new Lambda(x, f, true), true);
                    Lambda LyEXP_y = new Lambda(y, EXP_y, true);
                    Type Y = LyEXP_y.getType();
                    
                    ArrayList<LambdaProd> children = this.decompose_single(LyEXP_y);
                    
                    ExprToken token_LfLxf = this.toExprToken(LfLxf);
                    TypeToken token_Y = this.toTypeToken(Y);
                    LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLxf, token_Y);
                    treeRule.setChildren(children);
                    
                    if(debug) checkRestore(LfLxf, LyEXP_y, exp);
                    
                    treeRules.add(treeRule);
                }

                /** (10703)
                 * X : toTree ( Lx.Ly.P(y)^R(x,y))
                 * <=>
                 * X : ( ( Lg.Lf.Lx.Ly.f(y)^g(x)(y) Y ) Z )
                 * Y : toTree(Lx.Ly.R(x,y))
                 * Z : toTree(Ly.P(y))
                 */
                else if (innerInnerLambda.getExprCat() == Expr.AND) {
                	
                    And and = (And) innerInnerLambda;
                    Expr Py = and.getLeft();
                    Expr Rxy = and.getRight();

                    if (involvesTwoVars(Rxy, x, y)) {
                        Var g = new Var(genVar(), new CompositeType(x.getType(), new CompositeType(y.getType(), Rxy.getType())), true);
                        Var f = new Var(genVar(), new CompositeType(y.getType(), Py.getType()), true);

                        Expr fy = new FunApp(f, y);
                        Expr gxy = new FunApp(new FunApp(g, x), y);
                        Lambda LgLfLxCyfygxy = new Lambda(g, new Lambda(f, new Lambda(x, new Lambda(y, new And(fy, gxy), true), true), true), true);

                        Lambda LxLyRxy = new Lambda(x, new Lambda(y, Rxy, true), true);
                        Lambda LyPy = new Lambda(y, Py, true);
                        Type Y = LxLyRxy.getType();
                        Type Z = LyPy.getType();
                        
                        ArrayList<LambdaProd> children1 = this.decompose_single(LxLyRxy);
                        ArrayList<LambdaProd> children2 = this.decompose_single(LyPy);
                        
                        ExprToken token_LgLfLxCyfygxy = this.toExprToken(LgLfLxCyfygxy);
                        TypeToken token_Y = this.toTypeToken(Y);
                        TypeToken token_Z = this.toTypeToken(Z);
                        LambdaProd treeRule = this.toLambdaProd(token_X, token_LgLfLxCyfygxy, token_Y, token_Z);
                        treeRule.setChildren(children1, children2);

                        if(debug) checkRestore(LgLfLxCyfygxy, LxLyRxy, LyPy, exp);

                        treeRules.add(treeRule);
                    }
                    else {
                    	throw new NotImplementedException("10703\t" + exp);
                    }
                }
                else {
                	throw new NotImplementedException("10701\t" + exp);
                }
            }

            /** (10800)
             * X : Lx.ArgMax(EXP1(x), EXP2(x))
             * <=>
             * X : toTree ( ((Lf.Lg.Lx.ArgMax(f(x),g(x)) Y) Z) )
             * Y : toTree(Lx.EXP1(x))
             * Z : toTree(Lx.EXP2(x))
             */
            else if (inner.getExprCat() == Expr.ARGMAX) {

                ArgMax argmax = (ArgMax) inner;
                Identifier argmax_var = argmax.getVariable();
                Expr argmax_e = ((FunApp) argmax.getInnerExpr()).getFunc();
                ArgList innerArgmax = ((ArgList) ((FunApp) argmax.getInnerExpr()).getArg());
                Expr[] args = innerArgmax.getArgs();
                Expr EXP1x = args[0];
                Expr EXP2x = args[1];

                Var f = new Var(genVar(), new CompositeType(x.getType(), EXP1x.getType()), true);
                Var g = new Var(genVar(), new CompositeType(x.getType(), EXP2x.getType()), true);
                FunApp fx = new FunApp(f, x);
                FunApp gx = new FunApp(g, x);

                Lambda LfLgLxArgMaxfxg = new Lambda(f,
                        new Lambda(g, new Lambda(x, new ArgMax(argmax_var, new FunApp(argmax_e, new ArgList(new Expr[] {fx, gx})), true), true), true), true);
                Lambda LxEXP1x = new Lambda(x, EXP1x, true);
                Lambda LxEXP2x = new Lambda(x, EXP2x, true);

                Type Y = LxEXP1x.getType();
                Type Z = LxEXP2x.getType();
                
                ArrayList<LambdaProd> children1 = this.decompose_single(LxEXP1x);
                ArrayList<LambdaProd> children2 = this.decompose_single(LxEXP2x);

                ExprToken token_LfLgLxArgMaxfxg = this.toExprToken(LfLgLxArgMaxfxg);
                TypeToken token_Y = this.toTypeToken(Y);
                TypeToken token_Z = this.toTypeToken(Z);
                LambdaProd treeRule = this.toLambdaProd(token_X, token_LfLgLxArgMaxfxg, token_Y, token_Z);
                treeRule.setChildren(children1, children2);

                if(debug) checkRestore(LfLgLxArgMaxfxg, LxEXP1x, LxEXP2x, exp);

                treeRules.add(treeRule);
            }

            /** (10900)
             * X : Lx.ArgMin(EXP(x), EXP2)
             * <=>
             * X : ??
             */
            else if (inner.getExprCat() == Expr.ARGMIN) {
                throw new NotImplementedException("10900\t" + exp);
            }
            
            else {
            	throw new NotImplementedException("11000\t" + exp);
            }

        }
        
        
        if(treeRules.size() == size_before){
            LambdaProd expProd = this.toLambdaProd(token_X, token_self);
        	treeRules.add(expProd);
        }
        
        
//        ps.println("Not supported:"+exp.toString()+" with type:"+exp.getExprCat());
//        System.exit(1);
		
        return treeRules;
	}
	
	
	
	private static void checkRestore(Expr exp1, Expr exp2, Expr correct) throws TypeEvaluationException, UnsupportedEncodingException {
    	PrintStream ps = new PrintStream(System.out, true, "UTF-8");
        Expr predict = new FunApp(exp1, exp2).simplifyFully();
        if (!predict.alphaEquivalent(correct)) {
            ps.println(predict);
            ps.println(correct);
            throw new RuntimeException("Error!");
        }
//        System.err.println("Good!");
    }
    
    private static void checkRestore(Expr exp1, Expr exp2, Expr exp3, Expr correct) throws TypeEvaluationException, UnsupportedEncodingException{
    	PrintStream ps = new PrintStream(System.out, true, "UTF-8");
        Expr predict;
        try {
            predict = new FunApp(new FunApp(exp1, exp2), exp3).simplifyFully();
        } catch (Exception e) {
            ps.println(exp1);
            ps.println(exp2);
            ps.println(exp3);
            ps.println(correct);
            throw new RuntimeException("Critical Error!");
        }

        if (!predict.alphaEquivalent(correct)) {
            ps.println(exp1);
            ps.println(exp2);
            ps.println(exp3);
            ps.println(predict);
            ps.println(correct);
            throw new RuntimeException("Error!");
        }
//        System.err.println("Good!");
    }
    
    private static boolean involvesOneVar(Expr exp, Identifier var) {
        Set<?> fvs = exp.getFreeVars();
        return fvs.contains(var);
    }

    private static boolean involvesTwoVars(Expr exp, Identifier var1, Identifier var2) {
        Set<?> fvs = exp.getFreeVars();
        return fvs.contains(var1) && fvs.contains(var2);
    }

    private static boolean isDoubleBinaryFunc(Expr exp, Identifier var1, Identifier var2) {
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr arg = funapp.getArg();
            if (arg.getExprCat() == Expr.ARGLIST) {
                Expr[] args = ((ArgList) arg).getArgs();
                if (args[0].equals(var1) && args[1].equals(var2)) {
                    return true;
                } else if (args[1].equals(var1) && args[0].equals(var2)) {
                    return true;
                }
            }
        }
        return false;
    }

    private static boolean isLeftBinaryFunc(Expr exp, Identifier boundedVariable) {
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr arg = funapp.getArg();
            if (arg.getExprCat() == Expr.ARGLIST) {
                Expr[] args = ((ArgList) arg).getArgs();
                if (args.length == 2) {
                    if (args[0].equals(boundedVariable)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    private static boolean isRightBinaryFunc(Expr exp, Identifier boundedVariable) {
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr arg = funapp.getArg();
            if (arg.getExprCat() == Expr.ARGLIST) {
                Expr[] args = ((ArgList) arg).getArgs();
                if (args.length == 2) {
                    if (args[1].equals(boundedVariable)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public static Expr reGroup(Expr exp) {
        return reGroupHelper(exp, new Vector<Identifier>());
    }

    private static Expr reGroupHelper(Expr exp, Vector<Identifier> boundedVars) {
        if (exp instanceof Binder) {
            Binder binder = (Binder) exp;
            Identifier var = binder.getVariable();
            boundedVars.add(0, var);
            Expr inner = reGroupHelper(binder.getInnerExpr(), boundedVars);
            binder.setInnerExpr(inner);
            boundedVars.remove(0);
            return binder;
        } else if (exp instanceof Unary) {
            Unary unary = (Unary) exp;
            Expr inner = reGroupHelper(unary.getInnerExpr(), boundedVars);
            unary.setInnerExpr(inner);
            return unary;
        } else if (exp instanceof And) {
            And and = (And) exp;
            And newAnd = reGroupAnd(and, boundedVars);
            return newAnd;
        } else if (exp instanceof Binary) {
            Binary binary = (Binary) exp;
            Expr left = binary.getLeft();
            Expr right = binary.getRight();
            binary.setLeft(reGroupHelper(left, boundedVars));
            binary.setRight(reGroupHelper(right, boundedVars));
            return binary;
        } else if (exp instanceof ArgList) {
            ArgList arglist = (ArgList) exp;
            Expr[] args = arglist.getArgs();
            Expr[] newArgs = new Expr[args.length];
            for (int i = 0; i < args.length; i++) {
                Expr arg = args[i];
                newArgs[i] = reGroupHelper(arg, boundedVars);
            }
            arglist.setArgs(newArgs);
            return arglist;
        } else {
            return exp;
        }
    }
    
    private static And reGroupAnd(And oldAnd, Vector<Identifier> boundedVars) {
        Vector<Expr> elements = new Vector<Expr>();

        Vector<Expr> varUnary = new Vector<Expr>();
        Vector<Expr> unaryPredicates = new Vector<Expr>();
        Vector<Expr> doubleBinaryPredicates = new Vector<Expr>();
        Vector<Expr> others = new Vector<Expr>();

        andSplitter(oldAnd, boundedVars, elements);

        for (int i = 0; i < elements.size(); i++) {
            Expr element = elements.get(i);
            if (isVarUnary(element, boundedVars)) {
                varUnary.add(element);
            } else if (isUnaryPredicate(element, boundedVars)) {
                unaryPredicates.add(0, element);
            } else if (isDoubleBinaryPredicate(element, boundedVars)) {
                doubleBinaryPredicates.add(element);
            } else {
                others.add(element);
            }
        }

        Vector<Vector<Expr>> exps = new Vector<Vector<Expr>>();
        exps.add(varUnary);
        exps.add(unaryPredicates);
        exps.add(others);
        exps.add(doubleBinaryPredicates);

        Vector<Expr> allExps = merge(exps);

        Expr result = constructAnd(allExps);

        return (And) result;
    }

    private static Vector<Expr> merge(Vector<Vector<Expr>> vs) {
        Vector<Expr> allExps = new Vector<Expr>();
        for (int i = 0; i < vs.size(); i++) {
            Vector<Expr> exps = vs.elementAt(i);
            for (int j = 0; j < exps.size(); j++) {
                allExps.add(exps.elementAt(j));
            }
        }
        return allExps;
    }

    // P Q R => P ^ ( Q ^ R)
    private static Expr constructAnd(Vector<Expr> elements) {
        Expr exp = elements.remove(elements.size() - 1);
        if (elements.size() == 0) {
            return exp;
        }

        return new And(constructAnd(elements), exp);
    }
    
    private static void andSplitter(And oldAnd, Vector<Identifier> boundedVars, Vector<Expr> elements) {

        Expr left = oldAnd.getLeft().stripParens();
        Expr right = oldAnd.getRight().stripParens();

        if (left.getExprCat() == Expr.AND) {
            andSplitter((And) left, boundedVars, elements);
        } else {
            elements.add(reGroupHelper(left, boundedVars));
        }

        if (right.getExprCat() == Expr.AND) {
            andSplitter((And) right, boundedVars, elements);
        } else {
            elements.add(reGroupHelper(right, boundedVars));
        }

    }
    
    /**
     * check if it is f(x)
     */
    private static boolean isVarUnary(Expr exp, Vector<Identifier> vars) {
        Identifier var = vars.get(0);
        exp = exp.stripParens();
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr func = funapp.getFunc();
            Expr arg = funapp.getArg();
            if (!vars.contains(func)) {
                return false;
            }
            if (arg.equals(var)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * check if it is p(x)
     */
    private static boolean isUnaryPredicate(Expr exp, Vector<Identifier> vars) {
        Identifier var = vars.get(0);
        exp = exp.stripParens();
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr func = funapp.getFunc();
            Expr arg = funapp.getArg();
            if (vars.contains(func)) {
                return false;
            }
            if (arg.equals(var)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * check if it is p(x,y)
     */
    private static boolean isDoubleBinaryPredicate(Expr exp, Vector<Identifier> vars) {
        Identifier var = vars.get(0);
        exp = exp.stripParens();
        if (exp.getExprCat() == Expr.FUNAPP) {
            FunApp funapp = (FunApp) exp;
            Expr func = funapp.getFunc();
            if (vars.contains(func)) {
                return false;
            }
            Expr arg = funapp.getArg();
            if (arg.getExprCat() == Expr.ARGLIST) {
                Expr[] args = ((ArgList) arg).getArgs();
                if (args.length == 2) {
                    boolean flag1 = vars.contains(args[0]) && vars.contains(args[1]) && args[0] != args[1];
                    boolean flag2 = var.equals(args[0]) || var.equals(args[1]);
                    return flag1 && flag2;
                }
            }
        }
        return false;
    }
    
	
	
	/**
	 * Private methods
	 */
	/**
	 * Serialization methods...
	 */
	private void writeObject(ObjectOutputStream out) throws IOException{
		
		out.writeInt(this._allTypes.size());
		for(int i =0; i<this._allTypes.size(); i++){
			out.writeObject(this._allTypes.get(i));
		}
		
		out.writeInt(this._allExprs.size());
		for(int i=0; i<this._allExprs.size(); i++){
			out.writeObject(this._allExprs.get(i));
		}

		out.writeInt(this._allGMRs.size());
		for(int i=0; i<this._allGMRs.size(); i++){
			out.writeObject(this._allGMRs.get(i));
		}

		out.writeInt(this._typeTable.size());
		Iterator<Type> keys_type = this._typeTable.keySet().iterator();
		while(keys_type.hasNext()){
			Type type = keys_type.next();
			out.writeObject(type);
			out.writeObject(this._typeTable.get(type));
		}
		
		out.writeInt(this._exprTable.size());
		Iterator<Expr> keys_expr = this._exprTable.keySet().iterator();
		while(keys_expr.hasNext()){
			Expr expr = keys_expr.next();
			out.writeObject(expr);
			out.writeObject(this._exprTable.get(expr));
		}
		
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
		int size;

		size = in.readInt();
		this._allTypes = new ArrayList<Type>(size);
		for(int i = 0 ;i<size; i++){
			this._allTypes.add((Type)in.readObject());
		}
		
		size = in.readInt();
		this._allExprs = new ArrayList<Expr>(size);
		for(int i = 0; i<size; i++){
			this._allExprs.add((Expr)in.readObject());
		}
		
		size = in.readInt();
		this._allGMRs = new ArrayList<LambdaGProd>(size);
		for(int i = 0; i<size; i++){
			this._allGMRs.add((LambdaGProd)in.readObject());
		}
		
		size = in.readInt();
		this._typeTable = new HashMap<Type, TypeToken>();
		for(int i = 0; i<size; i++){
			this._typeTable.put((Type)in.readObject(), (TypeToken)in.readObject());
		}
		
		size = in.readInt();
		this._exprTable = new HashMap<Expr, ExprToken>();
		for(int i = 0; i<size; i++){
			this._exprTable.put((Expr)in.readObject(), (ExprToken)in.readObject());
		}
	}
	
}






