/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.data;

import java.io.*;
import java.util.*;

/**
 * @author Wei LU
 * @version 1.0
 */

public class Vocabulary implements Serializable{

	private static final long serialVersionUID = -3998138915076326046L;
	
	private ArrayList<Word> _words_byStr;
	private ArrayList<Word> _words_byId;
	
	public Vocabulary(){
		this._words_byStr = new ArrayList<Word>();
		this._words_byId = new ArrayList<Word>();
		this.createSpecialWords();
	}

	private void createSpecialWords(){
		Word._BEGIN = this.createWord("[BEG]");
		Word._X1 = this.createWord("[X,1]");
		Word._X2 = this.createWord("[X,2]");
		Word._END = this.createWord("[END]");
		Word._UNK = this.createWord("[UNK]");
	}
	
	private void mapSpecialWords(){
		Word._BEGIN = this.retrieveWord("[BEG]");
		Word._X1 = this.retrieveWord("[X,1]");
		Word._X2 = this.retrieveWord("[X,2]");
		Word._END = this.retrieveWord("[END]");
		Word._UNK = this.retrieveWord("[UNK]");
	}
	
	/* Find the Word with a given id */
	public Word getWordById(int id){
		return this._words_byId.get(id);
	}
	
	public Word retrieveWord(String w){
		Word word = new Word(w);
		int index = Collections.binarySearch(this._words_byStr, word);
		if(index >= 0){
			word = this._words_byStr.get(index);
		} else {
			//if not found, return the unknown word
			word = Word._UNK;
		}
		return word;
	}
	
	public Word createWord(String w){
		Word word = new Word(w);
		int index = Collections.binarySearch(this._words_byStr, word);
		if(index >= 0){
			word = this._words_byStr.get(index);
		} else {
			this._words_byStr.add(-1-index, word);
			word.setId(this._words_byId.size());
			this._words_byId.add(word);
		}
		return word;
	}
	
	@Override
	public String toString(){
		StringBuffer sb = new StringBuffer();

		sb.append("-Words-\n");
		for(int i =0; i<Math.min(10, this._words_byId.size()); i++){
			sb.append(this._words_byId.get(i));
			sb.append('\n');
		}
		sb.append("...\n");
		for(int i=Math.min(2, this._words_byId.size()); i>0; i--){
			sb.append(this._words_byId.get(this._words_byId.size()-i));
			sb.append('\n');
		}
		
		return sb.toString();
	}
	
	public ArrayList<Word> getAllWords(){
		return this._words_byId;
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		int num_srcWords = in.readInt();
		this._words_byId = new ArrayList<Word>(num_srcWords);
		this._words_byStr = new ArrayList<Word>(num_srcWords);
		for(int i=0; i<num_srcWords; i++){
			Word word = new Word(in.readUTF());
			word.setId(i);
			this._words_byId.add(word);
			this._words_byStr.add(word);
		}
		Collections.sort(this._words_byStr);
		
		this.mapSpecialWords();
	}
	
	private void writeObject(ObjectOutputStream out) throws IOException {
		out.writeInt(this._words_byId.size());
		for(int i=0; i<this._words_byId.size(); i++){
			Word word = this._words_byId.get(i);
			out.writeUTF(word.getWord());
		}
	}
	
}
