/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.gen;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Iterator;

/**
 * @author Wei LU
 * @version 1.0
 */

public class JavaLM {
	
	private HashMap<String, Integer> _word2Id;
	
	private HashMap<Long, Double> _probTable_gram1;
	private HashMap<Long, Double> _probTable_gram2;
	private HashMap<Long, Double> _probTable_gram3;
	
	private HashMap<Long, Double> _probTable_gram1_backoff;
	private HashMap<Long, Double> _probTable_gram2_backoff;
	
	public JavaLM(){
		this._word2Id = new HashMap<String, Integer>();

		this._probTable_gram1 = new HashMap<Long, Double>();
		this._probTable_gram2 = new HashMap<Long, Double>();
		this._probTable_gram3 = new HashMap<Long, Double>();
		
		this._probTable_gram1_backoff = new HashMap<Long, Double>();
		this._probTable_gram2_backoff = new HashMap<Long, Double>();
		
		this.initMapping();
	}
	
	public JavaLM(String lmfile){
		this();
		this.readTrainedLMFromFile(lmfile);
	}
	
	public int toId(String word){
		if(!this._word2Id.containsKey(word)){
			this._word2Id.put(word, this._word2Id.size()+1);
		}

		return this._word2Id.get(word);
	}
	
	public int[] mapFinalIds(String sent){
		String[] words = sent.split("\\s");
		int[] ids = new int[words.length+2];
		ids[0] = this._word2Id.get("<s>");
		for(int i = 0; i<words.length; i++){
			String word = words[i];
			if(!this._word2Id.containsKey(word)){
				ids[i+1] = this._word2Id.get("<unk>");
			} else {
				ids[i+1] = this._word2Id.get(word);
			}
		}
		ids[words.length+1] = this._word2Id.get("</s>");
		
		return ids;
	}

	
	public int[] mapSuffixIds(String sent, int n){
		String[] words = sent.split("\\s");
		int[] ids = new int[words.length];
		for(int i = 0; i<words.length; i++){
			String word = words[i];
			if(!this._word2Id.containsKey(word)){
				ids[i] = this._word2Id.get("<unk>");
			} else {
				ids[i] = this._word2Id.get(word);
			}
		}
		
		return ids;
	}
	
	public void viewVoc(){
		StringBuffer sb = new StringBuffer();
		
		Iterator<String> words = this._word2Id.keySet().iterator();
		while(words.hasNext()){
			String word = words.next();
			int id = this._word2Id.get(word);
			sb.append(word);
			sb.append('\t');
			sb.append(id);
			sb.append('\n');
		}
		
		System.err.println(sb.toString());
	}
	
	public void initMapping(){

		this.toId("<unk>");
		this.toId("<s>");
		this.toId("</s>");
		
	}
	
	private long toKey(String[] words){
		long val = this._word2Id.get(words[0]);
		for(int i =1; i<words.length; i++){
			val = val * 1000 + this._word2Id.get(words[i]);
		}
		return val;
	}
	
	private long toKey(int id2, int id1, int id){
		return id2 * 1000 * 1000 + id1 * 1000 + id;
	}
	
	private long toKey(int id1, int id){
		return id1 * 1000 + id;
	}
	
	private long toKey(int id){
		return id;
	}
	
	private double getUnigram(int id){
		long key = this.toKey(id);
		double val;
		if(this._probTable_gram1.containsKey(key)){
			val = this._probTable_gram1.get(key);
		}else {
			val = Double.NaN;
		}
		return val;
	}
	
	private double getUnigram_backoff(int id){
		long key = this.toKey(id);
		double val;
		if(this._probTable_gram1_backoff.containsKey(key)){
			val = this._probTable_gram1_backoff.get(key);
		} else {
			val = Double.NaN;
		}
		return val;
	}
	
	private double getBigram(int id1, int id){
		long key = this.toKey(id1, id);
		double val;
		if(this._probTable_gram2.containsKey(key)){
			val = this._probTable_gram2.get(key);
		} else {
			val = Double.NaN;
		}
		return val;
	}
	
	private double getBigram_backoff(int id1, int id){
		long key = this.toKey(id1, id);
		double val;
		if(this._probTable_gram2_backoff.containsKey(key)){
			val = this._probTable_gram2_backoff.get(key);
		} else {
			val = Double.NaN;
		}
		return val;
	}
	
	private double getTrigram(int id2, int id1, int id){
		long key = this.toKey(id2, id1, id);
		double val;
		if(this._probTable_gram3.containsKey(key)){
			val =  this._probTable_gram3.get(key);
		} else {
			val = Double.NaN;
		}
		return val;
	}
	
	private double getProb(int[] ids, int curr){
		
		if(curr==1){
			int id1 = ids[0];
			int id = ids[1];
			double val = this.getBigram(id1, id);
			if(Double.isNaN(val)){
				val = this.getUnigram(id) + this.getUnigram_backoff(id1);
			}
			return val;
		}
		
		int id2 = ids[curr-2];
		int id1 = ids[curr-1];
		int id = ids[curr];
		double val_trigram = this.getTrigram(id2, id1, id);
		if(Double.isNaN(val_trigram)){
			double val_bigram = this.getBigram(id1, id);
			if(Double.isNaN(val_bigram)){
				double val_unigram = this.getUnigram(id);
				double val_bigram_backoff = this.getBigram_backoff(id2, id1);
				if(Double.isNaN(val_bigram_backoff)){
					val_bigram_backoff = 0;
				}
				double val_unigram_backoff = this.getUnigram_backoff(id1);
				if(Double.isNaN(val_unigram_backoff)){
					val_unigram_backoff = 0;
				}
				double val = val_unigram + val_bigram_backoff + val_unigram_backoff;
				return val;
			} else {
				double val_bigram_backoff = this.getBigram_backoff(id2, id1);
				if(Double.isNaN(val_bigram_backoff)){
					val_bigram_backoff = 0;
				}
				double val = val_bigram + val_bigram_backoff;
				return val;
			}
		} else {
			return val_trigram;
		}
	}
	
	public void readTrainedLMFromFile(String file) {
		try{
			FileInputStream infile = new FileInputStream(file);
			BufferedReader stdin = new BufferedReader(new InputStreamReader(infile, "UTF-8"));
			
			String line;
			int gram = 0;
			while((line=stdin.readLine())!=null){
				if(line.startsWith("\\1-grams:")){
					gram = 1;
					continue;
				} else if(line.startsWith("\\2-grams:")){
					gram = 2;
					continue;
				} else if(line.startsWith("\\3-grams:")){
					gram = 3;
					continue;
				} else if(line.startsWith("\\end\\")){
					break;
				}
				this.processGram(line, gram);
			}
		}catch(IOException e){
			throw new RuntimeException(e);
		}
	}
	
	private void processGram(String line, int gram){
		if(line.trim().equals("") || gram == 0){
			return;
		}
		String[] tokens = line.split("\\s");
		double prob = Double.parseDouble(tokens[0]);
		
		String[] words = new String[gram];
		for(int i =0; i<gram; i++){
			words[i] = tokens[i+1];
			if(gram==1){
				this.toId(words[i]);
			}
		}
		long key = this.toKey(words);
		
		if(tokens.length > gram + 1){
			double prob_backoff = Double.parseDouble(tokens[tokens.length-1]);
			if(gram == 1){
				this._probTable_gram1_backoff.put(key, prob_backoff);
			} else if(gram == 2){
				this._probTable_gram2_backoff.put(key, prob_backoff);
			} else if(gram == 3){
				throw new RuntimeException("impossible...");
			}
		}
		
		if(gram == 1){
			this._probTable_gram1.put(key, prob);
		} else if(gram == 2){
			this._probTable_gram2.put(key, prob);
		} else if(gram == 3){
			this._probTable_gram3.put(key, prob);
		}
	}
	
	
	public static void main(String args[])throws IOException{
		JavaLM lm = new JavaLM("model/srilm.3.lm");
		
		FileInputStream infile = new FileInputStream("train.txt");
		BufferedReader stdin = new BufferedReader(new InputStreamReader(infile,"UTF-8"));

		String sent;
		while((sent=stdin.readLine())!=null){
			double logprob = lm.getFinalLogProb(sent);
			System.err.println("sum="+logprob);
		}
	}
	
	/**
	 * [0,1,2,..,n-2]
	 */
	public double getPrefixLogProb(String prev, String sent, int n){
		int[] ids = this.mapSuffixIds(sent, n);
		double logprob = 0;
		for(int i = 0 ; i<=n-2; i++){
			double val = this.getProb(ids, i);
			logprob += val;
		}
		return logprob;
	}
	
	/**
	 * 0,1,2,..,n-2,[n-1,n,..,W]
	 */
	public double getSuffixLogProb(String sent, int n){
		int[] ids = this.mapSuffixIds(sent, n);
		double logprob = 0;
		for(int i = n - 1 ; i<ids.length; i++){
			double val = this.getProb(ids, i);
			logprob += val;
		}
		return logprob;
	}
	
	/**
	 * <s> 0,1,2,..,W,</s>
	 */
	public double getFinalLogProb(String sent){
		int[] ids = this.mapFinalIds(sent);
		double logprob = 0;
		for(int i = 1 ; i<ids.length; i++){
			double val = this.getProb(ids, i);
			logprob += val;
		}
		return logprob;
	}
}
