/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.gen;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import lambdagen.data.LambdaProd;
import lambdagen.data.Word;
import lambdagen.grammar.LambdaGRule;
import lambdagen.grammar.LambdaRuleFactory;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaGenSpan implements Serializable, Comparable<LambdaGenSpan> {
	
	private static final long serialVersionUID = -7622814298637717734L;
	
	private LambdaGRule _grule;
	private LambdaProd _prod; // the prod.
	private Weights _weights;
	
	private double[] _fvscores; // aggregated score, but not weighted.
	private double _lmscore; // aggregated score, but not weighted.
	
	private LambdaGenSpan _child1; // child1 should correspond to X1.
	private LambdaGenSpan _child2; // child2 should correspond to X2.
	
	private String _sent = null;
	private int _n; // n-gram
	
	public String toString(LambdaRuleFactory rulefac){
		if(this == _empty){
			return "-empty-";
		}
		StringBuffer sb = new StringBuffer();
		sb.append(this._grule);
		sb.append('\n');
		sb.append(this._prod);
		return sb.toString();
	}
	
	public static final LambdaGenSpan _empty = new LambdaGenSpan(null, null, null, null, null);
	
	public String toTree(LambdaRuleFactory rulefac){
		if(this.equals(_empty)){
			return "--empty tree--";
		}
		StringBuffer sb = new StringBuffer();
		this.toTree_helper(sb, rulefac, 0);
		return sb.toString();
	}
	
	public void toTree_helper(StringBuffer sb, LambdaRuleFactory rulefac, int indent){
		
		for(int k=0; k<indent; k++){
			sb.append('\t');
		}
		sb.append(this._grule);
		sb.append('\n');
		ArrayList<Word> words = this._grule.getWords();
		for(int i = 0; i<words.size(); i++){
			Word word = words.get(i);
			if(word.equals(Word._X1)){
				this._child1.toTree_helper(sb, rulefac, indent+1);
			} else if(word.equals(Word._X2)){
				this._child2.toTree_helper(sb, rulefac, indent+1);
			} else {
				for(int k=0; k<indent+1; k++){
					sb.append('\t');
				}
				sb.append(word.getWord());
				sb.append('\n');
			}
		}
	}
	
	/**
	 * Constructor...
	 */
	public LambdaGenSpan(LambdaGRule grule, double[] fvscores, LambdaProd prod, LambdaGenSpan[] children, Weights weights){
		this._grule = grule;
		this._fvscores = fvscores;
		this._prod = prod;
		if(children == null || children.length == 0){
		} else if(children.length == 1){
			if(children[0] == _empty){
				throw new RuntimeException("x10");
			}
			this._child1 = children[0];
		} else if(children.length == 2){
			this._child1 = children[0];
			this._child2 = children[1];
			if(children[0] == _empty){
				throw new RuntimeException("x1");
			}
			if(children[1] == _empty){
				throw new RuntimeException("x2");
			}
		}
		this._weights = weights;
	}
	
	public Weights getWeights(){
		return this._weights;
	}
	
	public LambdaProd getProd(){
		return this._prod;
	}
	
	public String toNbestFormat(int lineno, LambdaRuleFactory rulefac){
		if(this.equals(_empty)){
			if(true){
//				throw new RuntimeException("Should not happen!");
			}
			
			StringBuffer sb = new StringBuffer();
			
			sb.append(lineno);
			sb.append(' ');
			sb.append('|');
			sb.append('|');
			sb.append('|');
			sb.append(' ');
//			sb.append("-EMPTY-");
			sb.append(' ');
			sb.append('|');
			sb.append('|');
			sb.append('|');
			sb.append(' ');
			sb.append(0.0);
			
			for(int i = 0; i<4; i++){
				sb.append(' ');
				sb.append(0.0);
			}
			
			sb.append(' ');
			sb.append('|');
			sb.append('|');
			sb.append('|');
			sb.append(' ');
			
			sb.append(0.0);
			
			return sb.toString();
		} else {
			StringBuffer sb = new StringBuffer();
			
			sb.append(lineno);
			sb.append(' ');
			sb.append('|');
			sb.append('|');
			sb.append('|');
			sb.append(' ');
			sb.append(this.toSentence(rulefac));
			sb.append(' ');
			sb.append('|');
			sb.append('|');
			sb.append('|');
			sb.append(' ');
			sb.append(this._lmscore);

			for(int i = 0; i<this._fvscores.length; i++){
				sb.append(' ');
				sb.append(this._fvscores[i]);
			}
			
			sb.append(' ');
			sb.append('|');
			sb.append('|');
			sb.append('|');
			sb.append(' ');
			
			sb.append(this.getWeightedTotalScore());
			
			return sb.toString();
		}
	}
	
	public LambdaGRule getRule(LambdaRuleFactory rulefac){
		return this._grule;
	}
	
	public double[] getFeatureScores(){
		return this._fvscores;
	}
	
	public double getLMScore(){
		return this._lmscore;
	}
	
	public void setLMScore(double lmscore){
		this._lmscore = lmscore;
	}
	
	//in fact an approximate..due to the lm score..
	public double getWeightedTotalScore(){
		if(this.equals(_empty)){
			return Double.NEGATIVE_INFINITY;
		}
		
		double total = this._lmscore * this._weights.getLMWeight();
		for(int i = 0; i<this._fvscores.length; i++){
			total += this._fvscores[i] * this._weights.getFeatureWeights()[i];
		}
		
		if(Double.isNaN(total)){
			System.err.println(this._lmscore+"x"+this._weights.getLMWeight());
			for(int i = 0; i<this._fvscores.length; i++){
				System.err.println(this._fvscores[i] +"\t"+ this._weights.getFeatureWeights()[i]);
			}
			System.exit(1);
		}
		
		return total;
	}
	
	public LambdaGenSpan getFirstChild(){
		return this._child1;
	}
	
	public LambdaGenSpan getSecondChild(){
		return this._child2;
	}
	
	@Override
	public int compareTo(LambdaGenSpan span) {
		double score = this.getWeightedTotalScore();
		double score_span = span.getWeightedTotalScore();
		
		if(score > score_span){
			return +1;
		} else if(score < score_span){
			return -1;
		} else {
			return 0;
		}
	}
	
	/**
	 * To string representation of the sentence constructed from the spans 
	 * rooted by this span.
	 */
	public String toSentence(LambdaRuleFactory rulefac){
		
		if(this._sent != null){
			return this._sent;
		}
		
		if(this._grule == null){
			this._sent = "-EMPTY-";
		} else {
			StringBuffer sb = new StringBuffer();
			ArrayList<Word> words = this._grule.getWords();
			
			for(int i =0; i< words.size(); i++){
				if(i!=0){
					sb.append(' ');
				}
				Word word = words.get(i);
				if(word.equals(Word._X1)){
					sb.append(this._child1.toSentence(rulefac));
				} else if(word.equals(Word._X2)){
					sb.append(this._child2.toSentence(rulefac));
				} else {
					sb.append(word.getWord());
				}
			}
			
			this._sent = sb.toString();
		}
		
		return this._sent;
	}
	
	/**
	 * For serialization...
	 */
	private void writeObject(ObjectOutputStream out)throws IOException{
		out.writeObject(this._grule);
		out.writeInt(this._n);
		out.writeInt(this._fvscores.length);
		for(int i = 0; i<this._fvscores.length; i++){
			out.writeDouble(this._fvscores[i]);
		}
		out.writeDouble(this._lmscore);
		if(this._child1== null){
			out.writeInt(0);
		} else if(this._child2 == null){
			out.writeInt(1);
			out.writeObject(this._child1);
		} else {
			out.writeInt(2);
			out.writeObject(this._child1);
			out.writeObject(this._child2);
		}
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
		this._grule = (LambdaGRule)in.readObject();
		this._n = in.readInt();
		this._fvscores = new double[in.readInt()];
		for(int i =0; i<this._fvscores.length; i++){
			this._fvscores[i] = in.readDouble();
		}
		this._lmscore = in.readDouble();
		int size = in.readInt();
		if(size == 2){
			this._child1 = (LambdaGenSpan)in.readObject();
			this._child2 = (LambdaGenSpan)in.readObject();
		} else if(size == 1){
			this._child1 = (LambdaGenSpan)in.readObject();
		}
		
	}
	
}
