/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.gen;

import lambdagen.grammar.LambdaGRule;
import lambdagen.grammar.LambdaRuleFactory;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaGenSpanTriple implements Comparable<LambdaGenSpanTriple>{
	
	private LambdaGRule _grule;
    private LambdaGenSpan _span1;
    private LambdaGenSpan _span2;
    private double _score;
    private int[] _position;
    
    private LambdaRuleFactory _rulefac;
    private LambdaLM _lm;
    private Weights _weights;
    
    public LambdaGenSpanTriple(LambdaGRule grule, LambdaGenSpan[] span1, LambdaGenSpan[] span2, int i, int j, LambdaRuleFactory rulefac, LambdaLM lm, Weights weights) {
    	this._grule = grule;
        this._span1 = span1[i];
        this._span2 = span2[j];
        this._position = new int[] {i, j};
        
        this._rulefac = rulefac;
        this._lm = lm;
        this._weights = weights;
        
        this.computeScores();
    }
    
    private void computeScores(){
    	
    	if(this._span1.equals(LambdaGenSpan._empty) || this._span2.equals(LambdaGenSpan._empty)){
    		this._score = Double.NEGATIVE_INFINITY;
    		return;
    	}

    	//first, compute the feature scores
    	double fvscore = this._weights.computeWeightedFeatureScore(this._grule) 
    		+ this._weights.computeWeightedFeatureScore(this._span1) 
    		+ this._weights.computeWeightedFeatureScore(this._span2);
    	
    	this._score = fvscore;
    	
    	String sent = this._grule.toSentence(new LambdaGenSpan[]{this._span1, this._span2}, this._rulefac);
		double lmscore = this._weights.getLMWeight() * this._lm.computeNgram(sent);
		
    	//third, add the lm score...
    	this._score += lmscore;
    	
    	//TODO 
    	/**
    	 * [NOTE] computation of the lm score should be done in a smarter way. this method has redundant computations.
    	 */

    }
    
    public LambdaGRule getGRule(){
    	return this._grule;
    }
    
    public boolean isPositive(){
    	return this._score != Double.NEGATIVE_INFINITY;
    }

    public int[] getPosition() {
        return this._position;
    }

    @Override
	public int compareTo(LambdaGenSpanTriple triple) {
        return this._score > triple._score ? -1 : this._score == triple._score ? 0 : +1;
    }

    public LambdaGenSpan getFirstSpan() {
        return this._span1;
    }

    public LambdaGenSpan getSecondSpan() {
        return this._span2;
    }

    public LambdaGenSpan[] toArray() {
        return new LambdaGenSpan[] {this._span1, this._span2};
    }

}
