/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.gen;

import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;

import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaProd;
import lambdagen.data.Vocabulary;
import lambdagen.grammar.LambdaGRule;
import lambdagen.grammar.LambdaRuleFactory;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaGenerator {
	
	private ArrayList<LambdaProd> _prods;//this specifies the tree already.
	private LambdaProd _rootProd;
	private ArrayList<LambdaGenCell> _cells;// each cell is associated with exactly one rule
	private ArrayList<LambdaGenCell> _rootCells;
	
	private int _K; // k-best list
	private int _n; // n-gram
	
	private Vocabulary _voc;
	private LambdaRuleFactory _rulefac;
	private Weights _weights;
	private LambdaLM _lm;
	
	private boolean debug = false;
	
	public LambdaGenerator(LambdaProd rootProd, Vocabulary voc, LambdaRuleFactory rulefac, Weights weights, LambdaLM lm, int K, int n){
		this._rootProd = rootProd;
		this._prods = rootProd.getAllProds();
		this._voc = voc;
		this._rulefac = rulefac;
		this._weights = weights;
		this._lm = lm;
		this._K = K;
		this._n = n;
		this.init();
	}
	
	private void init() {

		//de-duplicate..
		for(int i = 0 ; i<this._prods.size(); i++){
			LambdaProd prod1 = this._prods.get(i);
			for(int j = i+1 ; j<this._prods.size(); j++){
				LambdaProd prod2 = this._prods.get(j);
				if(prod1.equals(prod2)){
					this._prods.remove(j);
					j--;
				}
			}
		}
		
		this._rootCells = new ArrayList<LambdaGenCell>();
		
		ArrayList<LambdaGRule> allGRules = new ArrayList<LambdaGRule>();
		
		this._cells = new ArrayList<LambdaGenCell>();
		for(int i = 0 ; i < this._prods.size(); i++){
			LambdaProd prod = this._prods.get(i);
            prod.setSeq(i);//set the sequence number...
			LambdaGProd gp = prod.getGP();
			
			ArrayList<LambdaGRule> grules = this._rulefac.getGRules(gp);
			if(debug &&grules.size() == 0){
				System.err.println("The gp "+gp.toString()+" has no rule!");
//				System.exit(1);
			}
			
			if(grules.size()==0){
//				LambdaGRule grule = new LambdaGRule(gp, );
				
//				System.err.println("The gp "+gp.toString()+" has no rule!");
//				System.exit(1);
			}
			
			for(int j = 0; j<grules.size(); j++){
				LambdaGRule grule = grules.get(j);

				LambdaGenCell cell = new LambdaGenCell(grule, this._K);
				if(prod.equals(this._rootProd)){
					this._rootCells.add(cell);
				}
				
				//de-duplicate, but preserve the order.
				if(!allGRules.contains(grule)){
					this._cells.add(cell);
					allGRules.add(grule);
				} else {
					if(prod.equals(this._rootProd)){
						System.err.println("Error! Cell associated with rootProd gets removed?!");
					}
				}
			}
		}
		
//		debug = false;
		if(debug){
			try{
				PrintStream ps = new PrintStream(System.err, true, "UTF-8");
				
				ps.println(this._rootProd.toString());
				
				for(int i =0; i<this._cells.size(); i++){
					ps.println(i+"\t"+this._cells.get(i).getGRule());
				}
			} catch(UnsupportedEncodingException e){
				System.err.println(e);
				System.exit(1);
			}
		}
	}
	
	/**
	 * Generate the k-best list of sentences.
	 * @return
	 */
	public LambdaGenSpan[] computeGen(){
		
		for(int i =0; i<this._prods.size(); i++){
			
			LambdaProd prod = this._prods.get(i);
			
	        for (int m = 0; m < this._cells.size(); m++) {
	            LambdaGenCell cell = this._cells.get(m);
	            
	            LambdaGProd gp = cell.getGRule().getGP();
	            if(!gp.equals(prod.getGP())){
	            	continue;
	            }
	            
	            int arity = cell._grule.getArity();
	            
	            if(arity == 0){
	            	this.computeGen_terminal(cell, prod);
	            }
	            
	            else if(arity == 1){
	            	this.computeGen_unary(cell, prod);
	            }
	            
	            else if(arity == 2){
	            	this.computeGen_binary(cell, prod);
	            }
	        }
		}
		
		ArrayList<LambdaGenSpan> rs = new ArrayList<LambdaGenSpan>();
		for(int i = 0; i<this._rootCells.size(); i++){
			LambdaGenSpan[] result = this._rootCells.get(i).getSpansByProd(this._rootProd);
//			System.err.println("+++"+result[result.length-1].getRule(_rulefac));
			for(int k = result.length -1; k >=0; k--){
				rs.add(result[k]);
			}
		}
		Collections.sort(rs);
		LambdaGenSpan[] results = new LambdaGenSpan[this._K];
		for(int i = 0; i<results.length; i++){
			results[i] = rs.get(rs.size()-1-i);
		}
		
		/*
		LambdaGenSpan[] results = new LambdaGenSpan[this._K];
		Arrays.fill(results, LambdaGenSpan._empty);
		for(int i = 0; i<this._rootCells.size(); i++){
			LambdaGenSpan[] result = this._rootCells.get(i).getSpansByProd(this._rootProd);
			for(int k = result.length -1; k >=0; k--){
				if(result[k].equals(LambdaGenSpan._empty)){
					continue;
				}
				int index = Arrays.binarySearch(results, result[k]);
				if(index < 0){
					if(-1-index == 0){
						//do nothing... do not insert..
					} else {
						int insert_index = -1 - index -1;
						for(int j = 0; j<= insert_index - 1; j++){
							results[j] = results[j+1];
						}
						results[insert_index] = result[k];
					}
				} else {
					int insert_index = index -1;
					for(int j = 0; j<= insert_index - 1; j++){
						results[j] = results[j+1];
					}
					results[insert_index] = result[k];
				}
			}
		}
		*/
		
		return results;
        
	}
	
	private void computeGen_terminal(LambdaGenCell cell, LambdaProd prod){
		LambdaGRule grule = cell.getGRule();
		LambdaGenSpan span = new LambdaGenSpan(grule, grule.getFeatures(), prod, new LambdaGenSpan[0], this._weights);
		boolean f = cell.addSpan(span, prod);
	}
	
	private void computeGen_unary(LambdaGenCell cell, LambdaProd prod){
		
		LambdaGRule grule = cell.getGRule();
		
		ArrayList<LambdaProd> children_prods = prod.getFirstChildren();
		for(int pIndex1 = 0; pIndex1<children_prods.size(); pIndex1++){
			LambdaProd child_prod = children_prods.get(pIndex1);
			
			ArrayList<LambdaGRule> children_rules = this._rulefac.getGRules(child_prod.getGP());

			/*
			if(cell.getGRule().getId()==2947 && prod.getGP().getId()==157){
				System.err.println("ssssss1\n"+child_prod.getGP());
				for(int k =0; k< children_rules.size(); k++){
					System.err.println("child rule:\t"+children_rules.get(k).toPhraseTableEntry());
				}
				System.exit(1);
			}
			*/
			
			//consider all possible rules.
			for(int cIndex1 = 0; cIndex1<this._cells.size(); cIndex1++){
				LambdaGenCell child_cell = this._cells.get(cIndex1);
				LambdaGRule child_rule = child_cell.getGRule();
				
				//only to find those relevant child rules.
				if(children_rules.contains(child_rule)){
					LambdaGenSpan[] child_spans = child_cell.getSpansByProd(child_prod);
					
					/*
					if(cell.getGRule().getId()==2947 && prod.getGP().getId()==157){
						double v1 = child_spans[child_spans.length-1].getWeightedTotalScore();
						System.err.println(v1 + "\t" + child_rule.toPhraseTableEntry());
						System.exit(1);
					}
					*/
					
					if(debug){
						checkMonotonicity(child_spans);
					}

					for(int c1 = child_spans.length-1; c1>=0; c1--){
						LambdaGenSpan child_span  = child_spans[c1];
						
						if(child_span.equals(LambdaGenSpan._empty)){
							break;
						}
						
						double[] fvs = grule.getFeatures().clone();
						for(int k =0; k<fvs.length; k++){
							fvs[k] += child_span.getFeatureScores()[k];
						}
						
						LambdaGenSpan span = new LambdaGenSpan(grule, fvs, prod, new LambdaGenSpan[]{child_span}, this._weights);
						double lmscore = this._lm.computeNgram(span.toSentence(_rulefac));
						span.setLMScore(lmscore);
						boolean f = cell.addSpan(span, prod);
//						System.err.println(">>>"+span.getRule(_rulefac)+"\t"+f);						
					}
				}
			}
		}

		
		if(debug){
			LambdaGenSpan[] spans = cell.getSpansByProd(prod);
			this.checkMonotonicity(spans);
			if(spans[spans.length-1].equals(LambdaGenSpan._empty)){
				System.err.println(cell.getGRule()+"\tat\t"+prod.toString()+" has nothing valid...");
				System.exit(1);
			}
		}
	}
	
	private void checkMonotonicity(LambdaGenSpan[] spans){
		double oldVal = Double.NEGATIVE_INFINITY;
		for(int i = 0; i<spans.length; i++){
			double val = spans[i].getWeightedTotalScore();
			if(val < oldVal){
				throw new RuntimeException("old:"+oldVal+"\nnew:"+val);
			}
			oldVal = val;
		}
	}
	
	private void computeGen_binary(LambdaGenCell cell, LambdaProd prod){
		
		
		
		LambdaGRule grule = cell.getGRule();

		ArrayList<LambdaProd> children_prods1 = prod.getFirstChildren();
		ArrayList<LambdaProd> children_prods2 = prod.getSecondChildren();
		
		for(int pIndex1 = 0; pIndex1<children_prods1.size(); pIndex1++){
			LambdaProd child_prod1 = children_prods1.get(pIndex1);
			
			ArrayList<LambdaGRule> children_rules1 = this._rulefac.getGRules(child_prod1.getGP());
			
			//consider all possible rules.
			for(int cIndex1 = 0; cIndex1<this._cells.size(); cIndex1++){
				LambdaGenCell child_cell1 = this._cells.get(cIndex1);
				LambdaGRule child_rule1 = child_cell1.getGRule();
				
				//only to find those relevant child rules.
				if(Collections.binarySearch(children_rules1, child_rule1)>=0){
					LambdaGenSpan[] child_spans1 = child_cell1.getSpansByProd(child_prod1);

					if(debug){
						checkMonotonicity(child_spans1);
					}
					
					for(int pIndex2 = 0; pIndex2<children_prods2.size(); pIndex2++){
						LambdaProd child_prod2 = children_prods2.get(pIndex2);
						
						ArrayList<LambdaGRule> children_rules2 = this._rulefac.getGRules(child_prod2.getGP());
						
						//consider all possible rules.
						for(int cIndex2 = 0; cIndex2<this._cells.size(); cIndex2++){
							LambdaGenCell child_cell2 = this._cells.get(cIndex2);
							LambdaGRule child_rule2 = child_cell2.getGRule();
							
							//only to find those relevant child rules.
							if(Collections.binarySearch(children_rules2, child_rule2)>=0){
								LambdaGenSpan[] child_spans2 = child_cell2.getSpansByProd(child_prod2);

								if(debug){
									checkMonotonicity(child_spans2);
								}
								
								LambdaGenCube cube = new LambdaGenCube(grule, child_spans1, child_spans2, this._rulefac, this._lm, this._weights);
								LambdaGenSpan[] children;
								
								/*
								if(cell.getGRule().getId()==2941 && prod.getGP().getId()==155){
									double v1 = child_spans1[child_spans1.length-1].getWeightedTotalScore();
									double v2 = child_spans2[child_spans2.length-1].getWeightedTotalScore();
									System.err.println(v1 + "\t" + child_rule1.toPhraseTableEntry());
									System.err.println(v2 + "\t" + child_rule2.toPhraseTableEntry());
									System.exit(1);
								}
								*/
								
								while((children = cube.getNext())!= null){
									double[] fvs = grule.getFeatures().clone();
									for(int k =0; k<fvs.length; k++){
										fvs[k] += children[0].getFeatureScores()[k];
										fvs[k] += children[1].getFeatureScores()[k];
									}
									
									LambdaGenSpan span = new LambdaGenSpan(grule, fvs, prod, children, this._weights);
									double lmscore = this._lm.computeNgram(span.toSentence(_rulefac));
									span.setLMScore(lmscore);
									boolean f = cell.addSpan(span, prod);
									//System.err.println(">>>BBBB>>>>"+span.getRule(_rulefac)+"\t"+f+"\t"+span.getWeightedTotalScore());						
								}
								
							}
						}
						
					}
					
				}
			}
		}
		
		LambdaGenSpan[] spans = cell.getSpansByProd(prod);
		if(debug){
			this.checkMonotonicity(spans);
			if(spans[spans.length-1].equals(LambdaGenSpan._empty)){
				System.err.println(cell.getGRule()+"\tat\t"+prod.toString()+" has nothing valid...");
				System.exit(1);
			}
		}
		
	}
	
}
