/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.grammar;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaPattern;
import lambdagen.data.TypeToken;
import lambdagen.data.Word;
import lambdagen.gen.LambdaGenSpan;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaGRule implements Comparable<LambdaGRule>, Serializable{
	
	private static final long serialVersionUID = 4888435730331535376L;
	
	/**
	 * Static fields
	 */
	public static final int _TERM = 0; //terminal, i.e. no RHS : Expr
	public static final int _APP1 = 1; //one application, i.e. 1 RHS : (Expr Y)
	public static final int _APP2 = 2; //two applications, i.e. 2 RHS : ((Expr Y) Z)
	
	/**
	 * Private object fields
	 */
	private int _id = -1;
	private LambdaGProd _gm;
	private LambdaPattern _pattern;
	private ArrayList<Word> _words;
	private double[] _features;//the features..
	
	private static final double _phi = 0;
	private static final double _mu = 1.0;
//	private static final double _alpha_trans = 0.00001;//the Dirichlet prior for transition 
	
	private ArrayList<LambdaGRule> _originalRules;//if the rule is constructed from several atomic rules, i.e. type (2).
	
	public void setOriginalRules(ArrayList<LambdaGRule> originalRules){
		this._originalRules = originalRules;
	}
	
	public ArrayList<LambdaGRule> getOriginalRules(){
		return this._originalRules;
	}
	
	private int _type = 1; // default is type 1, can be 0 or 2
	
	public void setType(int type){
		this._type = type;
	}
	
	public int getType(){
		return this._type;
	}

	@Override
	public boolean equals(Object o){
		if(o instanceof LambdaGRule){
			LambdaGRule grule = (LambdaGRule)o;
			if(this._id == -1 || grule._id == -1){
				
				if(!this._gm.equals(grule._gm)){
					return false;
				}
				if(!this._pattern.equals(grule._pattern)){
					return false;
				}
				if(this._words.size() != grule._words.size()){
					return false;
				}
				for(int i = 0; i<this._words.size(); i++){
					if(!this._words.get(i).equals(grule._words.get(i))){
						return false;
					}
				}
				return true;
			} else {
				return this._id == grule._id;
			}
		}
		return false;
	}
	
	@Override
	public int compareTo(LambdaGRule rule){
		if(this._gm.getId() > rule._gm.getId()){
			return +1;
		} else if(this._gm.getId() < rule._gm.getId()){
			return -1;
		}
		
		if(this._pattern.getId() > rule._pattern.getId()){
			return +1;
		} else if(this._pattern.getId() < rule._pattern.getId()){
			return -1;
		}
		
		if(this._words.size() > rule._words.size()){
			return +1;
		} else if(this._words.size() < rule._words.size()){
			return -1;
		}
		
		for(int i = 0 ; i<this._words.size(); i++){
			int val = this._words.get(i).compareTo(rule._words.get(i));
			if(val > 0){
				return +1;
			} else if(val < 0){
				return -1;
			}
		}
		
		return 0;
	}
	
	/**
	 * Return a PhraseTable entry for the moment...
	 */
	@Override
	public String toString(){
		return this.toPhraseTableEntry();
	}
	
	public String toPhraseTableEntry(){
		StringBuffer sb = new StringBuffer();
		
		sb.append(this._id);
		sb.append('(');
		sb.append(this._type);
		sb.append(')');
		sb.append(' ');
		sb.append('|');
		sb.append('|');
		sb.append('|');
		sb.append(' ');
		
		sb.append(this._gm.getLHSTypeToken().toString());
		sb.append(' ');
		sb.append('|');
		sb.append('|');
		sb.append('|');
		if(this._gm.getExprToken()!=null){
			sb.append(' ');
			sb.append(this._gm.getExprToken().getExpr());
		}
		TypeToken[] rhs = this._gm.getRHSTypeTokens();
		for(int i =0; i<rhs.length; i++){
			sb.append(' ');
			sb.append(rhs[i].toString());
		}
		sb.append(' ');
		sb.append('|');
		sb.append('|');
		sb.append('|');
		for(int i =0; i<this._words.size(); i++){
			sb.append(' ');
			sb.append(this._words.get(i).getWord());
		}
		sb.append(' ');
		sb.append('|');
		sb.append('|');
		sb.append('|');
		if(this._features != null){
			for(int i =0; i<this._features.length; i++){
				sb.append(' ');
				sb.append(this._features[i]);
			}
		}
		
		return sb.toString();
	}
	
	/**
	 * The constructor
	 * @param constructType
	 * @param lhsT
	 * @param exprT
	 * @param rhsT
	 */
	public LambdaGRule(LambdaGProd gm, LambdaPattern pattern, ArrayList<Word> words){
		this._gm = gm;
		this._pattern = pattern;
		this._words = words;
	}
	
	public void setId(int id){
		this._id = id;
	}
	
	public int getId(){
		return this._id;
	}
	
	public LambdaPattern getPattern(){
		return this._pattern;
	}
	
	public ArrayList<Word> getWords(){
		return this._words;
	}
	
	public LambdaGProd getGP(){
		return this._gm;
	}
	
	public int getArity(){
		return this._gm.getArity();
	}
	
	public void setFeatureValue(int index, double feature){
		this._features[index] = feature;
	}
	
	public void setFeatures(double[] features){
		this._features = features;
	}
	
	public double[] getFeatures(){
		return this._features;
	}
	
	public boolean isRoot(){
		return this._gm.isRoot();
	}
	
	public String toSentence(LambdaGenSpan[] span, LambdaRuleFactory rulefac){
		
		if(this.getArity() != span.length){
			throw new RuntimeException("Error code 207!");
		}
		
		StringBuffer sb = new StringBuffer();

		for(int i =0; i< this._words.size(); i++){
			if(i!=0){
				sb.append(' ');
			}
			Word word = this._words.get(i);
			if(word.equals(Word._X1)){
				sb.append(span[0].toSentence(rulefac));
			} else if(word.equals(Word._X2)){
				sb.append(span[1].toSentence(rulefac));
			} else {
				sb.append(word.getWord());
			}
		}
		
		return sb.toString();
	}
	
	private HashMap<Integer, Double> _mrTable_a = null;
	private HashMap<Integer, Double> _mrTable_b = null;
	
	private HashMap<Integer, Double> _mrTable_a_new = new HashMap<Integer, Double>();
	private HashMap<Integer, Double> _mrTable_b_new = new HashMap<Integer, Double>();
	
	public double getMRTransProb_train(int pos, LambdaGRule m2) {
		
		//in case it's the first time to run...
		if(this._mrTable_a == null && this._mrTable_b == null){
			if(this.getGP().getRHSTypeTokens()[pos].equals(m2.getGP().getLHSTypeToken())){
				//a pre-defined small value...
				return _mu;
			} else {
				return 0;
			}
		}
		
		if(pos == 0){
			return this.getVal(this._mrTable_a, m2.getId());
		} else {
			return this.getVal(this._mrTable_b, m2.getId());
		}
	}

	public double getMRTransProb_test(int pos, LambdaGRule m2) {

		if(this._mrTable_a == null && this._mrTable_b == null){
			if(this.getGP().getRHSTypeTokens()[pos].equals(m2.getGP().getLHSTypeToken())){
				//a pre-defined small value...
				return _phi;
			} else {
				return 0;
			}
		}
		
		if(pos == 0){
			double prob = this.getVal(this._mrTable_a, m2.getId());
			if(prob > _phi){
				return prob;
			}
			if(this.getGP().getRHSTypeTokens()[pos].equals(m2.getGP().getLHSTypeToken())){
				//a pre-defined small value...
				return _phi;
			} else {
				return 0;
			}
		} else {
			double prob = this.getVal(this._mrTable_b, m2.getId());
			if(prob > _phi){
				return prob;
			}
			if(this.getGP().getRHSTypeTokens()[pos].equals(m2.getGP().getLHSTypeToken())){
				//a pre-defined small value...
				return _phi;
			} else {
				return 0;
			}
		}
	}
	
	private double getVal(HashMap<Integer, Double> map, int key){
		Double val = map.get(key);
		if(val == null){
			return 0;
		}
		return val;
	}
	
	/**
	 * normalization for the HashMap...
	 * @param map
	 */
	private boolean norm(HashMap<Integer, Double> map){
		if(map.size() == 0){
			return false;
		}
		Iterator<Integer> keys;
		keys = map.keySet().iterator();
		double sum = 0;
		while(keys.hasNext()){
			int key = keys.next();
			double val = map.get(key);
			sum += val;
		}
		keys = map.keySet().iterator();
		while(keys.hasNext()){
			int key = keys.next();
			double val = map.get(key);
			map.put(key, val / sum);
		}
		return true;
	}
	
	/**
	 * Update function.
	 */
	public void updates(){
		
		if(this.norm(this._mrTable_a_new)){
			this._mrTable_a = this._mrTable_a_new;
			this._mrTable_a_new = new HashMap<Integer, Double>();
		}

		if(this.norm(this._mrTable_b_new)){
			this._mrTable_b = this._mrTable_b_new;
			this._mrTable_b_new = new HashMap<Integer, Double>();
		}
		
	}
	
	/**
	 * Private methods for serialization...
	 */
	private void writeObject(ObjectOutputStream out) throws IOException{
		out.writeInt(this._id);
		out.writeInt(this._type);
		out.writeObject(this._gm);
		out.writeInt(this._words.size());
		for(int i =0; i<this._words.size(); i++){
			out.writeObject(this._words.get(i));
		}
		out.writeObject(this._pattern);
		
		if(this._features==null){
			out.writeInt(-1);
		} else {
			out.writeInt(this._features.length);
			for(int i=0; i<this._features.length; i++){
				out.writeDouble(this._features[i]);
			}
		}

		this.writeHashMap(this._mrTable_a, out);
		this.writeHashMap(this._mrTable_b, out);
		
		if(this._originalRules == null){
			out.writeInt(-1);
		} else {
			out.writeInt(this._originalRules.size());
			for(int i =0; i<this._originalRules.size(); i++){
				out.writeObject(this._originalRules.get(i));
			}
		}
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
		this._id = in.readInt();
		this._type = in.readInt();
		this._gm = (LambdaGProd)in.readObject();
		int size = in.readInt();
		this._words = new ArrayList<Word>();
		for(int i =0; i<size; i++){
			this._words.add((Word)in.readObject());
		}
		this._pattern = (LambdaPattern)in.readObject();
		
		size = in.readInt();
		if(size == -1){
			this._features = null;
		} else {
			this._features = new double[size];
			for(int i =0; i<this._features.length; i++){
				this._features[i] = in.readDouble();
			}
		}

		this._mrTable_a = this.readHashMap(in);
		this._mrTable_b = this.readHashMap(in);
		this._mrTable_a_new = new HashMap<Integer, Double>();
		this._mrTable_b_new = new HashMap<Integer, Double>();
		
		size = in.readInt();
		if(size >= 0){
			this._originalRules = new ArrayList<LambdaGRule>(size);
			for(int i = 0; i<size; i++){
				this._originalRules.add((LambdaGRule)in.readObject());
			}
		}
	}
	
	//aux methods..
	
	private void writeHashMap(HashMap<Integer, Double> map, ObjectOutputStream out) throws IOException{
		if (map == null) {
			out.writeInt(-1);
			return;
		}
		out.writeInt(map.size());
		Iterator<Integer> ids = map.keySet().iterator();
		while(ids.hasNext()){
			int id = ids.next();
			double val = map.get(id);
			out.writeInt(id);
			out.writeDouble(val);
		}
	}

	private HashMap<Integer, Double> readHashMap(ObjectInputStream in) throws IOException{
		int size = in.readInt();
		if(size == -1){
			return null;
		}
		HashMap<Integer, Double> map = new HashMap<Integer, Double>();
		for(int i = 0; i<size; i++){
			int key = in.readInt();
			double val = in.readDouble();
			map.put(key, val);
		}
		return map;
	}
}
