/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.grammar;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;

import lambdacalc.logic.Expr;
import lambdacalc.logic.Lambda;
import lambdacalc.logic.Type;
import lambdacalc.logic.Var;
import lambdagen.data.ExprToken;
import lambdagen.data.Instance;
import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaPattern;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.TypeToken;
import lambdagen.data.Word;
import lambdagen.viterbi.LambdaAlignSpan;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaRuleFactory implements Serializable{
	
	private static final long serialVersionUID = 4819926586943794990L;
	
	private LambdaProdFactory _prodfac;
	private ArrayList<LambdaGRule> _rules;
	private HashMap<LambdaGRule, Double> _countTable;
	private HashMap<Integer, ArrayList<LambdaGRule>> _gp2grTable;
	
	public LambdaRuleFactory(LambdaProdFactory prodfac){
		this._prodfac = prodfac;
		this._rules = new ArrayList<LambdaGRule>();
		this._countTable = new HashMap<LambdaGRule, Double>();
		this._gp2grTable = new HashMap<Integer, ArrayList<LambdaGRule>>();
	}
	
	public ArrayList<LambdaGRule> getGRules(LambdaGProd gp){

		if(this._gp2grTable.containsKey(gp.getId())){
			return this._gp2grTable.get(gp.getId());
		} else {
			ArrayList<LambdaGRule> grules = new ArrayList<LambdaGRule>();
			for(int i = 0; i < this._rules.size(); i++){
				LambdaGRule grule = this._rules.get(i);
				if(grule.getGP().equals(gp)){
					grules.add(grule);
				}
			}
			Collections.sort(grules);//sort the rules
			this._gp2grTable.put(gp.getId(), grules);
			return grules;
		}
		
	}
	
	public LambdaGRule getRuleById(int rid){
		return this._rules.get(rid);
	}
	
	public LambdaProdFactory getProdFac(){
		return this._prodfac;
	}
	
	public ArrayList<LambdaGRule> getAllRules(){
		return this._rules;
	}
	
	public HashMap<LambdaGRule, Double> getCountTable(){
		return this._countTable;
	}
	
	private void initFeatures(int num_features){
		//set the empty feature arrays...
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule rule = this._rules.get(i);
			rule.setFeatures(new double[num_features]);
		}
	}
	
	private double getHybridSequenceProb_Bigram(LambdaGRule rule){
		
		ArrayList<Word> words = rule.getWords();
		LambdaGProd gp = rule.getGP();
		
		double prob = 0;
		Word word1 = Word._BEGIN;
		Word word;
		for(int i =0; i<words.size(); i++){
			word = words.get(i);
			prob += gp.getBigramProb_test(word1.getId(), word.getId());
			word1 = word;
		}
		word = Word._END;
		prob += gp.getBigramProb_test(word1.getId(), word.getId());
		
		prob /= (words.size() + 1);
		
		double logprob = Math.log10(prob);
		
		return logprob;
		
	}
	
	private double getHybridSequenceProb_Unigram(LambdaGRule rule){
		
		ArrayList<Word> words = rule.getWords();
		LambdaGProd gp = rule.getGP();
		
		double prob = 0;
		Word word;
		for(int i =0; i<words.size(); i++){
			word = words.get(i);
			prob += gp.getUnigramProb_test(word.getId());
		}
		word = Word._END;
		prob += gp.getUnigramProb_test(word.getId());
		
		prob /= (words.size() + 1);
		
		double logprob = Math.log10(prob);
		
		return logprob;
	}
	
	private String wordsToStr(ArrayList<Word> words){
		StringBuffer sb = new StringBuffer();
		
		for(int i =0; i<words.size(); i++){
			if(i!=0){
				sb.append(' ');
			}
			sb.append(words.get(i).getWord());
		}
		
		return sb.toString();
	}
	
	private void setBigramProbs(int index){

		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			double logprob = this.getHybridSequenceProb_Bigram(grule);
			//if did not train bigram table...
			if(logprob == Double.NEGATIVE_INFINITY){
				logprob = 0;
			}
			grule.setFeatureValue(index, logprob);
		}
		
	}
	
	private void setUnigramProbs(int index){
		
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			double logprob = this.getHybridSequenceProb_Unigram(grule);
			//if the prod never appeared in the training set, this value will be -Inf.
			if(logprob == Double.NEGATIVE_INFINITY){
				logprob = 0;
			}
			grule.setFeatureValue(index, logprob);
		}
		
	}

	private void setExpGivenRule2(int index){
		
		HashMap<String, Double> gps = new HashMap<String, Double>();
		
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			ArrayList<Word> words = grule.getWords();
			String key = this.wordsToStr(words);
			double count = this._countTable.get(grule);
			if(!gps.containsKey(key)){
				gps.put(key, count);
			} else {
				double oldCount = gps.get(key);
				gps.put(key, oldCount + count);
			}
		}
		
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			ArrayList<Word> words = grule.getWords();
			//if it is UNK, assign a strong penalty..
			if(words.size()==1 && words.get(0).equals(Word._UNK)){
				grule.setFeatureValue(index, -1000);
			} else {
				String key = this.wordsToStr(words);
				double count = this._countTable.get(grule);
				double freq  = count / gps.get(key);
				freq = Math.log10(freq);
				grule.setFeatureValue(index, freq);
			}
		}
	}
	
	private void setExpGivenRule(int index){
		
		HashMap<String, Double> gps = new HashMap<String, Double>();
		
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			ArrayList<Word> words = grule.getWords();
			String nont = grule.isRoot() ? "R" : grule.getGP().getLHSTypeToken().getType().toString();
			String key = this.wordsToStr(words) + "\t"+ nont;
			double count = this._countTable.get(grule);
			if(!gps.containsKey(key)){
				gps.put(key, count);
			} else {
				double oldCount = gps.get(key);
				gps.put(key, oldCount + count);
			}
		}
		
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			ArrayList<Word> words = grule.getWords();
			//if it is UNK, assign a strong penalty..
			if(words.size()==1 && words.get(0).equals(Word._UNK)){
				grule.setFeatureValue(index, -1000);
			} else {
				String nont = grule.isRoot() ? "R" : grule.getGP().getLHSTypeToken().getType().toString();
				String key = this.wordsToStr(words) + "\t"+ nont;
				double count = this._countTable.get(grule);
				double freq  = count / gps.get(key);
				freq = Math.log10(freq);
				grule.setFeatureValue(index, freq);
			}
		}
	}
	
	/**
	 * The relative frequency of the rule given the same source expression.
	 * @return
	 */
	private void setRuleGivenExp(int index){
		
		HashMap<Integer, Double> gps = new HashMap<Integer, Double>();
		
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			double count = this._countTable.get(grule);
			int key = grule.getGP().getId();
			if(!gps.containsKey(key)){
				gps.put(key, count);
			} else {
				double oldCount = gps.get(key);
				gps.put(key, oldCount + count);
			}
		}
		
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			double count = this._countTable.get(grule);
			int key = grule.getGP().getId();
			double freq  = count / gps.get(key);
			freq = Math.log10(freq);
			grule.setFeatureValue(index, freq);
		}
	}
	
	private void setWordPenalty(int index){
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			double val = grule.getWords().size() - grule.getArity();
			//for non-word case, make it zero?!
			if(val == 0){
				val = 0;
			} else {
				val = - Math.log10(Math.E) * val;
			}
			grule.setFeatureValue(index, val);
		}
	}
	
	private void setPhrasePenalty(int index){
		
		double val = - Math.log10(Math.E);
		for(int i = 0; i<this._rules.size(); i++){
			LambdaGRule grule = this._rules.get(i);
			if(grule.getType()==2){
				grule.setFeatureValue(index, val);
			} else {
				grule.setFeatureValue(index, val);
			}
		}
		
	}
	
	private void computeFeaturesForAllRules(){

		//for now, consider the following features..
		this.initFeatures(3);
		
		this.setRuleGivenExp(0);
		this.setExpGivenRule(1);
		this.setWordPenalty(2);

//		this.setExpGivenRule2(2);
//		this.setPhrasePenalty(4);
		
//		this.setBigramProbs(5);
//		this.setUnigramProbs(5);
	}
	
	//make sure one GP should have at least one Rule
	public void amendRules(){
		
		double weight = 0.001;
		
		ArrayList<LambdaGProd> gps = this._prodfac.getAllGProds();
		for(int i = 0; i < gps.size(); i++){
			LambdaGProd gp = gps.get(i);
			ArrayList<LambdaGRule> rules = this.getGRules(gp);
			if(rules.size() == 0){
				int arity = gp.getArity();
				LambdaPattern p;
				if(arity == 0){
//					throw new RuntimeException("There should not be a terminal gp that maps to no grule:\n"+gp.toString());
//					System.err.println("[amended]\t"+gp.toString());
					p = LambdaPattern._w;
					ArrayList<Word> words = new ArrayList<Word>();
					words.add(Word._UNK);
					LambdaGRule grule = new LambdaGRule(gp, p, words);
					rules.add(grule);
					
					this._gp2grTable.put(gp.getId(), rules);
					
					this.pushRules(rules, weight);
				} else if(arity == 1){
//					System.err.println("[amended]\t"+gp.toString());
					p = LambdaPattern._X;
					ArrayList<Word> words = new ArrayList<Word>();
					words.add(Word._X1);
					LambdaGRule grule = new LambdaGRule(gp, p, words);
					rules.add(grule);
					
					this._gp2grTable.put(gp.getId(), rules);
					
					this.pushRules(rules, weight);
				} else{
//					System.err.println("[amended]\t"+gp.toString());
					ArrayList<Word> words1 = new ArrayList<Word>();
					p = LambdaPattern._XY;
					words1.add(Word._X1);
					words1.add(Word._X2);
					LambdaGRule grule1 = new LambdaGRule(gp, p, words1);
					rules.add(grule1);
					
					ArrayList<Word> words2 = new ArrayList<Word>();
					p = LambdaPattern._YX;
					words2.add(Word._X2);
					words2.add(Word._X1);
					LambdaGRule grule2 = new LambdaGRule(gp, p, words2);
					rules.add(grule2);
					
					this._gp2grTable.put(gp.getId(), rules);

					this.pushRules(rules, weight);
				}
			}
		}
		
	}
	
	@Override
	public void finalize(){
		this.amendRules();
		this.assignIds();
		//finally, compute the features...
		this.computeFeaturesForAllRules();
	}
	
	public void assignIds(){
		for(int i = 0; i<this._rules.size(); i++){
			this._rules.get(i).setId(i);
		}
	}
	
	public void extractRules(LambdaAlignSpan span, Instance inst){
		ArrayList<Word> words = inst.getWords();
		ArrayList<LambdaProd> prods = inst.getAllProds(this._prodfac);
		
		double weight = inst.getWeight();
		
		ArrayList<LambdaGRule> rules = new ArrayList<LambdaGRule>();
		
		this.extractRules_Level0(null, 0, span, prods, words, rules);
		this.extractRules_Level1(null, 0, span, prods, words, rules);
		this.extractRules_Level2(null, 0, span, prods, words, rules);
		
		this.pushRules(rules, weight);
	}
	
	private void pushRules(ArrayList<LambdaGRule> rules, double weight){
		
		for(int i=0; i<rules.size(); i++){
			LambdaGRule rule = rules.get(i);
			
			int index = Collections.binarySearch(this._rules, rule);
			double oldVal = 0;
			if(index < 0){
				this._rules.add(-1-index, rule);
			} else {
				rule = this._rules.get(index);
				oldVal = this._countTable.get(rule);
			}
			this._countTable.put(rule, weight + oldVal);
		}
	}
	
	public void expandRules(LambdaProd prod, Instance inst){
		
		this.expandRules_Level0(null, 0, prod, new ArrayList<LambdaProd>());
		this.expandRules_Level2(null, 0, prod, new ArrayList<LambdaProd>());
		
	}
	
	private void expandRules_Level2(LambdaProd prod_parent, int cIndex, LambdaProd prod, ArrayList<LambdaProd> visitedProds){

		//means this is a newly added prod..
		if(prod.getSeq()==-1){
			return;
		}

		int arity = prod.getArity();
		
		if(arity == 0){
			//do nothing...
		}
		
		else if(arity == 1){
			
			ArrayList<LambdaProd> childProds = prod.getFirstChildren();
			
			for(int i =0; i<childProds.size(); i++){
				LambdaProd childProd = childProds.get(i);
				
				if(childProd.getSeq()==-1){
					continue;
//					throw new RuntimeException("invalid child seq:"+childProd.toString());
				}
				
				ArrayList<LambdaProd> grandChildProds1 = childProd.getFirstChildren();
				
				if(grandChildProds1!=null && grandChildProds1.size()>0){
					for(int j =0; j<grandChildProds1.size(); j++){
						LambdaProd grandChildProd = grandChildProds1.get(j);
						if(grandChildProd.getSeq()==-1){
							continue;
//							throw new RuntimeException("Invalid seq:"+grandChildProd.toString());
						}
						
						ArrayList<Integer> seqIds = new ArrayList<Integer>();
						seqIds.add(childProd.getSeq());
						seqIds.add(grandChildProd.getSeq());
						
						ArrayList<Var> var = new ArrayList<Var>();
						Expr exp = prod.toExpr(grandChildProd.getSeq(), var, seqIds);
						
						Type type = null;
						try{
							type = exp.getType();
							exp = new Lambda(var.get(0), exp, true);
						}catch(Exception e){
							throw new RuntimeException(e);
						}
						ExprToken token_exp = this._prodfac.toExprToken(exp);
						TypeToken token_type = this._prodfac.toTypeToken(type);
						TypeToken token_gchild_type = grandChildProd.getGP().getLHSTypeToken();

						LambdaProd prod_new = this._prodfac.toLambdaProd_test(token_type, token_exp, token_gchild_type);
						if(prod_new!=null)
						{
							ArrayList<LambdaProd> prod_new_children = new ArrayList<LambdaProd>();
							prod_new_children.add(grandChildProd);
							prod_new.setChildren(prod_new_children);
						}
						if(prod_parent!=null && prod_new!=null){
							if(cIndex == 0){
								ArrayList<LambdaProd> children = prod_parent.getFirstChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							} else {
								ArrayList<LambdaProd> children = prod_parent.getSecondChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							}
						}
					}
				}//end grandChildProd1
				
				ArrayList<LambdaProd> grandChildProds2 = childProd.getSecondChildren();
				
				if(grandChildProds2!=null && grandChildProds2.size()>0){
					for(int j =0; j<grandChildProds2.size(); j++){
						LambdaProd grandChildProd = grandChildProds2.get(j);
						if(grandChildProd.getSeq()==-1){
							continue;
//							throw new RuntimeException("Invalid seq:"+grandChildProd.toString());
						}

						ArrayList<Integer> seqIds = new ArrayList<Integer>();
						seqIds.add(childProd.getSeq());
						seqIds.add(grandChildProd.getSeq());
						
						ArrayList<Var> var = new ArrayList<Var>();
						Expr exp = prod.toExpr(grandChildProd.getSeq(), var, seqIds);
						
						Type type = null;
						try{
							type = exp.getType();
							exp = new Lambda(var.get(0), exp, true);
						}catch(Exception e){
							throw new RuntimeException(e);
						}
						ExprToken token_exp = this._prodfac.toExprToken(exp);
						TypeToken token_type = this._prodfac.toTypeToken(type);
						TypeToken token_gchild_type = grandChildProd.getGP().getLHSTypeToken();

						LambdaProd prod_new = this._prodfac.toLambdaProd_test(token_type, token_exp, token_gchild_type);
						if(prod_new!=null)
						{
							ArrayList<LambdaProd> prod_new_children = new ArrayList<LambdaProd>();
							prod_new_children.add(grandChildProd);
							prod_new.setChildren(prod_new_children);
//							System.err.println("try:"+prod_new.getGP().toString());
						}
						if(prod_parent!=null && prod_new!=null){
							if(cIndex == 0){
								ArrayList<LambdaProd> children = prod_parent.getFirstChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							} else {
								ArrayList<LambdaProd> children = prod_parent.getSecondChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							}
						}
					}
				}//end grandChildProd2
				
				//child.
				this.expandRules_Level2(prod, 0, childProd, visitedProds);
				
			}//end childProd
			
		}
		
		else if(arity == 2){
			
			ArrayList<LambdaProd> childProds;
			
			childProds = prod.getFirstChildren();
			
			for(int i =0; i<childProds.size(); i++){
				LambdaProd childProd = childProds.get(i);
				
				if(childProd.getSeq()==-1){
					continue;
//					throw new RuntimeException("invalid child seq:"+childProd.toString());
				}

				ArrayList<LambdaProd> grandChildProds1 = childProd.getFirstChildren();
				
				if(grandChildProds1!=null && grandChildProds1.size()>0){
					for(int j =0; j<grandChildProds1.size(); j++){
						LambdaProd grandChildProd = grandChildProds1.get(j);
						if(grandChildProd.getSeq()==-1){
							continue;
//							throw new RuntimeException("Invalid seq:"+grandChildProd.toString());
						}

						ArrayList<Integer> seqIds = new ArrayList<Integer>();
						seqIds.add(childProd.getSeq());
						seqIds.add(grandChildProd.getSeq());
						
						ArrayList<Var> var = new ArrayList<Var>();
						Expr exp = prod.toExpr(grandChildProd.getSeq(), var, seqIds);
						
						Type type = null;
						try{
							type = exp.getType();
							exp = new Lambda(var.get(0), exp, true);
						}catch(Exception e){
							throw new RuntimeException(e);
						}
						ExprToken token_exp = this._prodfac.toExprToken(exp);
						TypeToken token_type = this._prodfac.toTypeToken(type);
						TypeToken token_gchild_type = grandChildProd.getGP().getLHSTypeToken();

						LambdaProd prod_new = this._prodfac.toLambdaProd_test(token_type, token_exp, token_gchild_type);
						if(prod_new!=null)
						{
							ArrayList<LambdaProd> prod_new_children = new ArrayList<LambdaProd>();
							prod_new_children.add(grandChildProd);
							prod_new.setChildren(prod_new_children);
//							System.err.println("try:"+prod_new.getGP().toString());
						}
						if(prod_parent!=null && prod_new!=null){
							if(cIndex == 0){
								ArrayList<LambdaProd> children = prod_parent.getFirstChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							} else {
								ArrayList<LambdaProd> children = prod_parent.getSecondChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							}
						}
					}
				}//end grandChildProd1

				ArrayList<LambdaProd> grandChildProds2 = childProd.getSecondChildren();
				
				if(grandChildProds2!=null && grandChildProds2.size()>0){
					for(int j =0; j<grandChildProds2.size(); j++){
						LambdaProd grandChildProd = grandChildProds2.get(j);
						if(grandChildProd.getSeq()==-1){
							continue;
//							throw new RuntimeException("Invalid seq:"+grandChildProd.toString());
						}

						ArrayList<Integer> seqIds = new ArrayList<Integer>();
						seqIds.add(childProd.getSeq());
						seqIds.add(grandChildProd.getSeq());
						
						ArrayList<Var> var = new ArrayList<Var>();
						Expr exp = prod.toExpr(grandChildProd.getSeq(), var, seqIds);
						
						Type type = null;
						try{
							type = exp.getType();
							exp = new Lambda(var.get(0), exp, true);
						}catch(Exception e){
							throw new RuntimeException(e);
						}
						ExprToken token_exp = this._prodfac.toExprToken(exp);
						TypeToken token_type = this._prodfac.toTypeToken(type);
						TypeToken token_gchild_type = grandChildProd.getGP().getLHSTypeToken();

						LambdaProd prod_new = this._prodfac.toLambdaProd_test(token_type, token_exp, token_gchild_type);
						if(prod_new!=null)
						{
							ArrayList<LambdaProd> prod_new_children = new ArrayList<LambdaProd>();
							prod_new_children.add(grandChildProd);
							prod_new.setChildren(prod_new_children);
//							System.err.println("try:"+prod_new.getGP().toString());
						}
						if(prod_parent!=null && prod_new!=null){
							if(cIndex == 0){
								ArrayList<LambdaProd> children = prod_parent.getFirstChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							} else {
								ArrayList<LambdaProd> children = prod_parent.getSecondChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							}
						}
					}
				}//end grandChildProd2
				
				//child.
				this.expandRules_Level2(prod, 0, childProd, visitedProds);
				
			}//end childProd1
			

			childProds = prod.getSecondChildren();
			
			for(int i =0; i<childProds.size(); i++){
				LambdaProd childProd = childProds.get(i);

				if(childProd.getSeq()==-1){
					continue;
//					throw new RuntimeException("invalid child seq:"+childProd.toString());
				}

				ArrayList<LambdaProd> grandChildProds1 = childProd.getFirstChildren();
				
				if(grandChildProds1!=null && grandChildProds1.size()>0){
					for(int j =0; j<grandChildProds1.size(); j++){
						LambdaProd grandChildProd = grandChildProds1.get(j);
						if(grandChildProd.getSeq()==-1){
							continue;
//							throw new RuntimeException("Invalid seq:"+grandChildProd.toString());
						}

						ArrayList<Integer> seqIds = new ArrayList<Integer>();
						seqIds.add(childProd.getSeq());
						seqIds.add(grandChildProd.getSeq());
						
						ArrayList<Var> var = new ArrayList<Var>();
						Expr exp = prod.toExpr(grandChildProd.getSeq(), var, seqIds);
						
						Type type = null;
						try{
							type = exp.getType();
							exp = new Lambda(var.get(0), exp, true);
						}catch(Exception e){
							throw new RuntimeException(e);
						}
						ExprToken token_exp = this._prodfac.toExprToken(exp);
						TypeToken token_type = this._prodfac.toTypeToken(type);
						TypeToken token_gchild_type = grandChildProd.getGP().getLHSTypeToken();

						LambdaProd prod_new = this._prodfac.toLambdaProd_test(token_type, token_exp, token_gchild_type);
						if(prod_new!=null)
						{
							ArrayList<LambdaProd> prod_new_children = new ArrayList<LambdaProd>();
							prod_new_children.add(grandChildProd);
							prod_new.setChildren(prod_new_children);
//							System.err.println("try:"+prod_new.getGP().toString());
						}
						if(prod_parent!=null && prod_new!=null){
							if(cIndex == 0){
								ArrayList<LambdaProd> children = prod_parent.getFirstChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							} else {
								ArrayList<LambdaProd> children = prod_parent.getSecondChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							}
						}
					}
				}//end grandChildProd1
				
				ArrayList<LambdaProd> grandChildProds2 = childProd.getSecondChildren();
				if(grandChildProds2!=null && grandChildProds2.size()>0){
					for(int j =0; j<grandChildProds2.size(); j++){
						LambdaProd grandChildProd = grandChildProds2.get(j);
						if(grandChildProd.getSeq()==-1){
							continue;
//							throw new RuntimeException("Invalid seq:"+grandChildProd.toString());
						}
						
						ArrayList<Integer> seqIds = new ArrayList<Integer>();
						seqIds.add(childProd.getSeq());
						seqIds.add(grandChildProd.getSeq());
						
						ArrayList<Var> var = new ArrayList<Var>();
						Expr exp = prod.toExpr(grandChildProd.getSeq(), var, seqIds);
						
						Type type = null;
						try{
							type = exp.getType();
							exp = new Lambda(var.get(0), exp, true);
						}catch(Exception e){
							throw new RuntimeException(e);
						}
						ExprToken token_exp = this._prodfac.toExprToken(exp);
						TypeToken token_type = this._prodfac.toTypeToken(type);
						TypeToken token_gchild_type = grandChildProd.getGP().getLHSTypeToken();

						LambdaProd prod_new = this._prodfac.toLambdaProd_test(token_type, token_exp, token_gchild_type);
						if(prod_new!=null)
						{
							ArrayList<LambdaProd> prod_new_children = new ArrayList<LambdaProd>();
							prod_new_children.add(grandChildProd);
							prod_new.setChildren(prod_new_children);
//							System.err.println("try:"+prod_new.getGP().toString());
						}
						if(prod_parent!=null && prod_new!=null){
							if(cIndex == 0){
								ArrayList<LambdaProd> children = prod_parent.getFirstChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							} else {
								ArrayList<LambdaProd> children = prod_parent.getSecondChildren();
								boolean added = false;
								for(int k = 0; k<children.size(); k++){
									if(children.get(k).getGP().equals(prod_new.getGP())){
										added = true;
										break;
									}
								}
								if(!added){
									children.add(prod_new);
								}
							}
						}
					}
				}//end grandChildProd2
				
				//child.
				this.expandRules_Level2(prod, 1, childProd, visitedProds);
			}//end childProd2
			
		}
	}
	
	private void expandRules_Level0(LambdaProd prod_parent, int cIndex, LambdaProd prod, ArrayList<LambdaProd> visitedProds){

		//check if the prod is already visited...
		if(visitedProds.contains(prod)){
			return;
		} else {
			visitedProds.add(prod);
		}
		
		int arity = prod.getArity();
		
		if(arity == 1){
			
			ArrayList<LambdaProd> children = prod.getFirstChildren();
			for(int i = 0; i<children.size(); i++){
				LambdaProd child = children.get(i);
				this.expandRules_Level0(prod, 0, child, visitedProds);
			}
			
		} else if(arity == 2){
			
			ArrayList<LambdaProd> children1 = prod.getFirstChildren();
			for(int i = 0; i<children1.size(); i++){
				LambdaProd child = children1.get(i);
				this.expandRules_Level0(prod, 0, child, visitedProds);
			}
			ArrayList<LambdaProd> children2 = prod.getSecondChildren();
			for(int i = 0; i<children2.size(); i++){
				LambdaProd child = children2.get(i);
				this.expandRules_Level0(prod, 1, child, visitedProds);
			}
		} else {
			return;
		}
		
		Expr exp = prod.toExpr();
		ExprToken token_exp = this._prodfac.toExprToken(exp);
		Type type = null;
		try{
			type = exp.getType();
		}catch(Exception e){
			throw new RuntimeException(e);
		}
		TypeToken token_type = this._prodfac.toTypeToken(type);
		LambdaProd prod_new = this._prodfac.toLambdaProd_test(token_type, token_exp);
		
		//if the parent is not null, insert the newly created LambdaProd into the existing tree.
		if(prod_parent!=null && prod_new!=null){
			if(cIndex == 0){
				ArrayList<LambdaProd> children = prod_parent.getFirstChildren();
				boolean added = false;
				for(int k = 0; k<children.size(); k++){
					if(children.get(k).getGP().equals(prod_new.getGP())){
						added = true;
						break;
					}
				}
				if(!added){
					children.add(prod_new);
				}
			} else {
				ArrayList<LambdaProd> children = prod_parent.getSecondChildren();
				boolean added = false;
				for(int k = 0; k<children.size(); k++){
					if(children.get(k).getGP().equals(prod_new.getGP())){
						added = true;
						break;
					}
				}
				if(!added){
					children.add(prod_new);
				}
			}
		}
	}
		
	private void extractRules_Level2(LambdaAlignSpan parent, int cIndex, LambdaAlignSpan span, ArrayList<LambdaProd> prods, ArrayList<Word> words, ArrayList<LambdaGRule> rules){
		
		LambdaProd prod = prods.get(span.getSeq());
		
		int arity = prod.getArity();
		
		if(arity == 0){
			//do nothing...
		}
		
		else if(arity == 1){
			int bIndex = span.getBIndex();
			int eIndex = span.getEIndex();
			
			LambdaAlignSpan childSpan = span.getChild();
			
			LambdaAlignSpan[] grandChildSpans = childSpan.getChildren();
			
			if(grandChildSpans!=null && grandChildSpans.length > 0){
				for(int i = 0; i<grandChildSpans.length; i++){
					LambdaAlignSpan grandChildSpan = grandChildSpans[i];
					LambdaProd grandChild_prod = prods.get(grandChildSpan.getSeq());
					int bIndex_gc = grandChildSpan.getBIndex();
					int eIndex_gc = grandChildSpan.getEIndex();
					
					ArrayList<Integer> seqIds = new ArrayList<Integer>();
					seqIds.add(childSpan.getSeq());
					seqIds.add(grandChild_prod.getSeq());
					
					ArrayList<Var> var = new ArrayList<Var>();
					Expr exp = prod.toExpr(grandChild_prod.getSeq(), var, seqIds);

					Type type = null;
					try{
						type = exp.getType();
						exp = new Lambda(var.get(0), exp, true);
					}catch(Exception e){
						throw new RuntimeException(e);
					}
					ExprToken token_exp = this._prodfac.toExprToken(exp);
					TypeToken token_type = this._prodfac.toTypeToken(type);
					TypeToken token_type_child = grandChild_prod.getGP().getLHSTypeToken();
					LambdaGProd gm = this._prodfac.toLambdaGProd(token_type, token_exp, token_type_child);
					
					ArrayList<Word> wds = new ArrayList<Word>();
					for(int j = bIndex; j<bIndex_gc; j++){
						wds.add(words.get(j));
					}
					wds.add(Word._X1);
					for(int j = eIndex_gc; j<eIndex; j++){
						wds.add(words.get(j));
					}
					LambdaPattern p;
					if(bIndex < bIndex_gc && eIndex_gc < eIndex){
						p = LambdaPattern._wXw;
					} else if(bIndex<bIndex_gc){
						p = LambdaPattern._wX;
					} else if(eIndex_gc < eIndex){
						p = LambdaPattern._Xw;
					} else {
						p = LambdaPattern._X;
					}
					LambdaGRule rule = new LambdaGRule(gm, p, wds);
					rule.setType(2);
					rules.add(rule);
				}
			}
			
			//child.
			this.extractRules_Level2(span, 0, childSpan, prods, words, rules);
		}
		
		else if(arity == 2){
			int bIndex = span.getBIndex();
			int eIndex = span.getEIndex();
			
			LambdaAlignSpan[] childrenSpan = span.getChildren();
			
			for(int pos = 0; pos<2; pos++){
				
				LambdaAlignSpan childSpan = childrenSpan[pos];
				
				LambdaAlignSpan[] grandChildSpans = childSpan.getChildren();
				
				if(grandChildSpans!=null && grandChildSpans.length > 0){
					for(int i = 0; i<grandChildSpans.length; i++){
						LambdaAlignSpan grandChildSpan = grandChildSpans[i];
						LambdaProd grandChild_prod = prods.get(grandChildSpan.getSeq());
						int bIndex_gc = grandChildSpan.getBIndex();
						int eIndex_gc = grandChildSpan.getEIndex();
						
						ArrayList<Integer> seqIds = new ArrayList<Integer>();
						seqIds.add(childSpan.getSeq());
						seqIds.add(grandChild_prod.getSeq());

						ArrayList<Var> var = new ArrayList<Var>();
						Expr exp = prod.toExpr(grandChild_prod.getSeq(), var, seqIds);

						Type type = null;
						try{
							/*
							PrintStream ps = new PrintStream(System.err, true, "UTF-8");
							*/
							type = exp.getType();
							/*
							ps.println(exp);
							ps.println(prod.toString());
							ps.println(var.get(0));
							*/
							exp = new Lambda(var.get(0), exp, true);
						}catch(Exception e){
							throw new RuntimeException(e);
						}
						ExprToken token_exp = this._prodfac.toExprToken(exp);
						TypeToken token_type = this._prodfac.toTypeToken(type);
						TypeToken token_type_child = grandChild_prod.getGP().getLHSTypeToken();
						LambdaGProd gm = this._prodfac.toLambdaGProd(token_type, token_exp, token_type_child);
						
						ArrayList<Word> wds = new ArrayList<Word>();
						for(int j = bIndex; j<bIndex_gc; j++){
							wds.add(words.get(j));
						}
						wds.add(Word._X1);
						for(int j = eIndex_gc; j<eIndex; j++){
							wds.add(words.get(j));
						}
						LambdaPattern p;
						if(bIndex < bIndex_gc && eIndex_gc < eIndex){
							p = LambdaPattern._wXw;
						} else if(bIndex<bIndex_gc){
							p = LambdaPattern._wX;
						} else if(eIndex_gc < eIndex){
							p = LambdaPattern._Xw;
						} else {
							p = LambdaPattern._X;
						}
						LambdaGRule rule = new LambdaGRule(gm, p, wds);
						rule.setType(2);
						rules.add(rule);
					}
				}
				
				//child.
				this.extractRules_Level2(span, pos, childSpan, prods, words, rules);
			}
		}
		
	}
	
	private void extractRules_Level1(LambdaAlignSpan parent, int cIndex, LambdaAlignSpan span, ArrayList<LambdaProd> prods, ArrayList<Word> words, ArrayList<LambdaGRule> rules){
		
		LambdaProd prod = prods.get(span.getSeq());
		
		int arity = prod.getArity();
		
		if(arity == 0){
			
			int bIndex = span.getBIndex();
			int eIndex = span.getEIndex();
			ArrayList<Word> wds = new ArrayList<Word>();
			for(int i = bIndex; i< eIndex; i++){
				wds.add(words.get(i));
			}
			LambdaGRule rule = new LambdaGRule(prod.getGP(), span.getPattern(), wds);
			rules.add(rule);
		}
		
		else if(arity == 1){
			
			int bIndex = span.getBIndex();
			int eIndex = span.getEIndex();
			LambdaAlignSpan child = span.getChild();
			int bIndex_c = child.getBIndex();
			int eIndex_c = child.getEIndex();
			ArrayList<Word> wds = new ArrayList<Word>();
			for(int i = bIndex; i < bIndex_c; i++){
				wds.add(words.get(i));
			}
			wds.add(Word._X1);
			for(int i = eIndex_c; i< eIndex; i++){
				wds.add(words.get(i));
			}
			LambdaGRule rule = new LambdaGRule(prod.getGP(), span.getPattern(), wds);
			rules.add(rule);
			
			this.extractRules_Level1(span, 0, child, prods, words, rules);
		}
		
		else if(arity == 2){
			int bIndex = span.getBIndex();
			int eIndex = span.getEIndex();
			LambdaAlignSpan[] children = span.getChildren();
			LambdaAlignSpan child1 = children[0];
			LambdaAlignSpan child2 = children[1];
			int bIndex_c1 = child1.getBIndex();
			int eIndex_c1 = child1.getEIndex();
			int bIndex_c2 = child2.getBIndex();
			int eIndex_c2 = child2.getEIndex();
			ArrayList<Word> wds = new ArrayList<Word>();
			for(int i = bIndex; i < bIndex_c1; i++){
				wds.add(words.get(i));
			}
			if(span.getPattern().isReversed()){
				wds.add(Word._X2);
			} else {
				wds.add(Word._X1);
			}
			for(int i = eIndex_c1; i< bIndex_c2; i++){
				wds.add(words.get(i));
			}
			if(span.getPattern().isReversed()){
				wds.add(Word._X1);
			} else {
				wds.add(Word._X2);
			}
			for(int i = eIndex_c2; i< eIndex; i++){
				wds.add(words.get(i));
			}
			LambdaGRule rule = new LambdaGRule(prod.getGP(), span.getPattern(), wds);
			rules.add(rule);
			
			this.extractRules_Level1(span, 0, child1, prods, words, rules);
			this.extractRules_Level1(span, 1, child2, prods, words, rules);
		}
	}
	
	private void extractRules_Level0(LambdaAlignSpan parent, int cIndex, LambdaAlignSpan span, ArrayList<LambdaProd> prods, ArrayList<Word> words, ArrayList<LambdaGRule> rules) {
		
		LambdaProd prod = prods.get(span.getSeq());
		
		int arity = prod.getArity();
		
		if(arity == 0){
			return;
		}
		
		int bIndex = span.getBIndex();
		int eIndex = span.getEIndex();

		ArrayList<Word> wds = new ArrayList<Word>();
		for(int i = bIndex; i<eIndex; i++){
			wds.add(words.get(i));
		}
		
		Expr exp = prod.toExpr();
		ExprToken token_exp = this._prodfac.toExprToken(exp);
		Type type = null;
		try{
			type = exp.getType();
		}catch(Exception e){
			throw new RuntimeException(e);
		}
		TypeToken token_type = this._prodfac.toTypeToken(type);
		LambdaProd prod_new = this._prodfac.toLambdaProd(token_type, token_exp);
		LambdaGProd gm_new = prod_new.getGP();
		
		LambdaGRule rule = new LambdaGRule(gm_new, LambdaPattern._w, wds);
		rule.setType(0);
		rules.add(rule);
		
		if(arity == 1){
			LambdaAlignSpan[] children = span.getChildren();
			LambdaAlignSpan child1 = children[0];
			
			this.extractRules_Level0(span, 0, child1, prods, words, rules);
		} else if(arity == 2){
			LambdaAlignSpan[] children = span.getChildren();
			LambdaAlignSpan child1 = children[0];
			LambdaAlignSpan child2 = children[1];
			
			this.extractRules_Level0(span, 0, child1, prods, words, rules);
			this.extractRules_Level0(span, 1, child2, prods, words, rules);
		}
		
	}
	
	/**
	 * methods for serialization...
	 */
	
	private void writeObject(ObjectOutputStream out) throws IOException{
		out.writeObject(this._prodfac);
		out.writeInt(this._rules.size());
		for(int i =0; i<this._rules.size(); i++){
			LambdaGRule rule = this._rules.get(i);
			double val = this._countTable.get(rule);
			
			out.writeObject(rule);
			out.writeDouble(val);
		}
		
		out.writeInt(this._gp2grTable.size());
		
//		System.err.println("size out = "+this._gp2grTable.size());
		
		Iterator<Integer> keys = this._gp2grTable.keySet().iterator();
		while(keys.hasNext()){
			int key = keys.next();
//			System.err.print(key+" ");
			ArrayList<LambdaGRule> rules = this._gp2grTable.get(key);
			out.writeInt(key);
			out.writeInt(rules.size());
			for(int i = 0; i<rules.size(); i++){
				out.writeObject(rules.get(i));
			}
		}
//		System.err.println();
	}
	
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
		this._prodfac = (LambdaProdFactory) in.readObject();
		int size = in.readInt();

		this._rules = new ArrayList<LambdaGRule>(size);
		this._countTable = new HashMap<LambdaGRule, Double>();
		for(int i = 0; i<size; i++){
			LambdaGRule rule = (LambdaGRule)in.readObject();
			double val = in.readDouble();
			
			this._rules.add(rule);
			this._countTable.put(rule, val);
		}
		
		this._gp2grTable = new HashMap<Integer, ArrayList<LambdaGRule>>();
		size = in.readInt();
//		System.err.println("size in  = "+size);
		for(int i = 0; i<size; i++){
			int key = in.readInt();
//			System.err.print(key+" ");
			int num_rules = in.readInt();
			ArrayList<LambdaGRule> rules = new ArrayList<LambdaGRule>(num_rules);
			for(int j = 0; j<num_rules; j++){
				rules.add((LambdaGRule)in.readObject());
			}
			this._gp2grTable.put(key, rules);
		}
//		System.err.println();
	}
	
}
