/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.learn;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import lambdacalc.logic.Expr;
import lambdacalc.logic.Identifier;
import lambdacalc.logic.Type;
import lambdagen.data.Instance;
import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.data.Word;
import lambdagen.util.LambdaUtil;
import lambdagen.util.Logger;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaInitLearner {
	
	private LambdaProdFactory _fac;
	private Vocabulary _voc;
	private ArrayList<Instance> _instances;
	private final int[] cats = new int[] {Expr.ARGMAX, Expr.ARGMIN, Expr.NOT, Expr.IOTA, Expr.EXISTS, Expr.COUNT, Expr.SUM, Expr.AND, Expr.EQUALITY, Expr.OR};

	private int _iterations = 100;//the number of iterations for ibm model 1
	
	public LambdaInitLearner(LambdaProdFactory fac, Vocabulary voc, ArrayList<Instance> instances){
		this._fac = fac;
		this._voc = voc;
		this._instances = instances;
	}
	
	/**
	 * Perform IBM model 1 style alignment algorithm
	 */
	public void ibm1(){

		System.err.println("IBM1");
		for(int k=0; k<this._iterations; k++){
			if(k%10==0){
				System.err.print('.');
			}
//			System.err.print("\r"+k+"/"+this._iterations);
			this.em_step();
		}
		System.err.println();
		
		//store it as the initial probabilities..
		this.assignProbsToGM();
		
		Logger.alert("IBM model1 trained, and initialized with those probabilities..");

		/*
		try{
			this.viewIbms();
		}catch(Exception e){
			throw new RuntimeException(e);
		}
		*/

//		System.exit(1);
	}
	
	/**
	 * Assign the learned probabilities to GMs...
	 */
	private void assignProbsToGM(){
		
		//only consider those relevant gms used in training...
		ArrayList<LambdaGProd> gms = new ArrayList<LambdaGProd>();
		try{
		for(int i=0; i<this._instances.size(); i++){
			Instance inst = this._instances.get(i);
			ArrayList<LambdaProd> prods = inst.getAllProds(_fac);
			for(LambdaProd prod : prods){
				if(!gms.contains(prod.getGP())){
					gms.add(prod.getGP());
				}
			}
		}
		}catch(Exception e){
			throw new RuntimeException(e);
		}
		
		for(int i=0; i<gms.size(); i++){
			LambdaGProd gm = gms.get(i);
			if(gm.getExprToken()!=null){
				Expr exp = gm.getExprToken().getExpr();
				ArrayList<Identifier> ids = this.getConsts(exp);
				for(int k = 0; k<ids.size(); k++){
					Identifier id = ids.get(k);
					HashMap<Word, Double> map = this._probs.get(id);
					
					//this means the gm is not in from the testing set...
					if(map==null){
						continue;
					}
					Iterator<Word> words = map.keySet().iterator();
					while(words.hasNext()){
						Word word = words.next();
						double val = map.get(word);
						gm.putUnigramCount(word.getId(), val);
					}
				}
				
				for(int cat : cats){
					if(LambdaUtil.contains(exp, cat)){
						HashMap<Word, Double> map = this._probs_cat.get(cat);
						//this means the gm is not in from the testing set...
						if(map==null){
							continue;
						}
						Iterator<Word> words = map.keySet().iterator();
						while(words.hasNext()){
							Word word = words.next();
							double val = map.get(word);
							gm.putUnigramCount(word.getId(), val);
						}
					}
				}
			}
			
			try{
				if(gm.isRoot()){
					int key;
					if(gm.getRHSTypeTokens()[0].getType().equals(Type.E)){
						key = -1000;
					} else if(gm.getRHSTypeTokens()[0].getType().equals(Type.ET)){
						key = -2000;
					} else {
						key = -3000;
					}

					HashMap<Word, Double> map = this._probs_cat.get(key);
					//this means the gm is not in from the testing set...
					if(map==null){
						continue;
					}
					Iterator<Word> words = map.keySet().iterator();
					while(words.hasNext()){
						Word word = words.next();
						double val = map.get(word);
						gm.putUnigramCount(word.getId(), val);
//						System.err.println(val);
					}
				}
			} catch(Exception e){
				System.err.println(e);
				System.exit(1);
			}
			
			if(gm.getArity()>=0){
				gm.putUnigramCount(Word._END.getId(), 1.0);
			}
			if(gm.getArity()>=1){
				gm.putUnigramCount(Word._X1.getId(), 1.0);
			}
			if(gm.getArity()==2){
				gm.putUnigramCount(Word._X2.getId(), 1.0);
			}
			
			ArrayList<Word> allwords = this._voc.getAllWords();
			for(int j=0; j<allwords.size(); j++){
				gm.putUnigramCount(allwords.get(j).getId(), 0.01);
			}
			
			gm.updates();
			
			/*
			try{
				gm.viewUnigramTable(this._voc);
			}catch(Exception e){
				throw new RuntimeException(e);
			}
			/**/
		}
		
//		System.exit(1);
		
	}
	
	private HashMap<Identifier, HashMap<Word, Double>> _probs = null;
	private HashMap<Identifier, HashMap<Word, Double>> _probs_new = new HashMap<Identifier, HashMap<Word, Double>>();
	private HashMap<Integer, HashMap<Word, Double>> _probs_cat = null;
	private HashMap<Integer, HashMap<Word, Double>> _probs_cat_new = new HashMap<Integer, HashMap<Word, Double>>();
	
	private void em_step(){
		
		for(int k=0; k<this._instances.size(); k++){
			Instance inst = this._instances.get(k);
			Expr exp = inst.getExpr();
			
			ArrayList<Identifier> ids = this.getConsts(exp);//identifiers, words are generated from them...
			ArrayList<Word> words = inst.getWords();
			
			ArrayList<Integer> ts = this.getRelevantTypes(cats, exp);

			try{
				if(exp.getType().equals(Type.E)){
					ts.add(-1000);
				} else if(exp.getType().equals(Type.ET)){
					ts.add(-2000);
				} else {
					ts.add(-3000);
				}
			}catch(Exception e){
				System.err.println(e);
				System.exit(1);
			}
			
			double weight = inst.getWeight();
			
			for(int j=0; j<words.size(); j++){
				Word word = words.get(j);
				double sum = 0;
				for(int i=0; i<ids.size(); i++){
					Identifier id = ids.get(i);
					double prob = this.getProb(id, word);
					sum += prob;
				}
				for(int i=0; i<ts.size(); i++){
					int t = ts.get(i);
					double prob = this.getProb2(t, word);
					sum += prob;
				}
				for(int i=0; i<ids.size(); i++){
					Identifier id = ids.get(i);
					double prob = this.getProb(id, word);
					this.putProb(id, word, prob/sum * weight);
				}
				for(int i=0; i<ts.size(); i++){
					int t = ts.get(i);
					double prob = this.getProb2(t, word);
					this.putProb2(t, word, prob/sum * weight);
				}
			}
		}
		this.norm_step();
	}
	
	private ArrayList<Integer> getRelevantTypes(int[] types, Expr exp){
		ArrayList<Integer> ids = new ArrayList<Integer>();
		for(int i = 0; i<types.length; i++){
			if(LambdaUtil.contains(exp, types[i])){
				ids.add(types[i]);
			}
		}
		return ids;
	}
	
	private void putProb2(int id, Word word, double value){
		if(!this._probs_cat_new.containsKey(id)){
			this._probs_cat_new.put(id, new HashMap<Word, Double>());
		}
		HashMap<Word, Double> map = this._probs_cat_new.get(id);
		double oldValue = 0;
		if(map.containsKey(word)){
			oldValue = map.get(word);
		}
		map.put(word, oldValue + value);
	}
	
	private void putProb(Identifier id, Word word, double value){
		if(!this._probs_new.containsKey(id)){
			this._probs_new.put(id, new HashMap<Word, Double>());
		}
		HashMap<Word, Double> map = this._probs_new.get(id);
		double oldValue = 0;
		if(map.containsKey(word)){
			oldValue = map.get(word);
		}
		map.put(word, oldValue + value);
	}
	
	private double getProb2(int expCat, Word word){
		if(this._probs_cat == null){
			return 0.1;
		}
		return this._probs_cat.get(expCat).get(word);
	}
	
	private double getProb(Identifier id, Word word){
		if(this._probs == null){
			return 0.1;
		}
		return this._probs.get(id).get(word);
	}
	
	private void norm_step(){
		
		Iterator<Identifier> ids = this._probs_new.keySet().iterator();
		while(ids.hasNext()){
			Identifier id = ids.next();
			HashMap<Word, Double> subMap = this._probs_new.get(id);
			Iterator<Word> words = subMap.keySet().iterator();
			double sum = 0;
			while(words.hasNext()){
				Word word = words.next();
				double val = subMap.get(word);
				sum += val;
			}
			words = subMap.keySet().iterator();
			while(words.hasNext()){
				Word word = words.next();
				double val = subMap.get(word);
				subMap.put(word, val/sum);
			}
		}
		
		Iterator<Integer> ts = this._probs_cat_new.keySet().iterator();
		while(ts.hasNext()){
			int t = ts.next();
			HashMap<Word, Double> subMap = this._probs_cat_new.get(t);
			Iterator<Word> words = subMap.keySet().iterator();
			double sum = 0;
			while(words.hasNext()){
				Word word = words.next();
				double val = subMap.get(word);
				sum += val;
			}
			words = subMap.keySet().iterator();
			while(words.hasNext()){
				Word word = words.next();
				double val = subMap.get(word);
				subMap.put(word, val/sum);
			}
		}
		
		this._probs = this._probs_new;
		this._probs_cat = this._probs_cat_new;
		this._probs_new = new HashMap<Identifier, HashMap<Word, Double>>();
		this._probs_cat_new = new HashMap<Integer, HashMap<Word, Double>>();
	}
	
	private ArrayList<Identifier> getConsts(Expr exp){
		
		ArrayList<Identifier> rs = new ArrayList<Identifier>();
		
		if(exp instanceof Identifier){
			rs.add((Identifier)exp);
			return rs;
		}
		
		HashSet<Identifier> results = new HashSet<Identifier>();
		exp.getConsts(results);
		Iterator<Identifier> ids = results.iterator();
		while(ids.hasNext()){
			rs.add(ids.next());
		}
		return rs;
	}
}




