/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.learn;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;

import lambdacalc.logic.TypeEvaluationException;
import lambdagen.data.Instance;
import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaPattern;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Word;
import lambdagen.util.Logger;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaLearner {
	
	public static boolean _learnEmission = true;
	public static boolean _learnPattern = true;
	public static boolean _learnTransition = true;
	
	public static final short _UNIGRAM = 10;
	public static final short _BIGRAM = 20;
	
	protected double _globalBeta = 0;
	
	protected Instance _inst;
	
	protected int[] _words;
	
	protected int _N;
	
	protected ArrayList<LambdaProd> _prod;
	
	protected LambdaProd _rootProd;
	
	protected double _weight;
	
	protected short _gram = -1; //default value is not valid, must specify it!
	
	private LambdaLearnCell[] _cells;
    
    private int BEGIN_WORD_ID = -1;
    private int END_WORD_ID = -1;
    private int CHILD_A_WORD_ID = -1;
    private int CHILD_B_WORD_ID = -1;
    
	private boolean _debug = false;
	
	public LambdaLearner(LambdaProdFactory fac, Instance inst, short gram) throws UnsupportedEncodingException, TypeEvaluationException{
		this._inst = inst;
		this._words = new int[inst.getWords().size()];
		for(int i = 0; i< this._words.length; i++){
			this._words[i] = inst.getWords().get(i).getId();
		}
		this._N = this._words.length;
		this._prod = inst.getAllProds(fac);
		this._weight = inst.getWeight();
		this._gram = gram;
		//check if the gram value is valid.
		if(this._gram!=_UNIGRAM && this._gram!=_BIGRAM){
			throw new RuntimeException("Invalid gram: "+gram);
		}
		this.init();

		if(_debug){
			this.checkValidOrder();
		}
		if(_debug){
			System.err.println(_learnEmission + "\t" + _learnPattern + "\t" + _learnTransition);
		}
//		System.err.println(inst.getMRProd(fac));

	}
	
	private void checkValidOrder(){
		ArrayList<LambdaProd> visited = new ArrayList<LambdaProd>();
		for(int i =0; i<this._cells.length; i++){
			LambdaLearnCell cell = this._cells[i];
			LambdaProd prod = cell.getProd();
			visited.add(prod);
			ArrayList<LambdaProd> cProds = prod.getAllProds();
			for(LambdaProd cProd : cProds){
				if(!visited.contains(cProd)){
					System.err.println(cProd.getGP().toString()+" is not visited yet!!!");
					System.exit(1);
				}
			}
		}
		System.err.println("Ordered correctly!");
	}
	
	private void init(){
		//de-duplicate..
		for(int i = 0 ; i<this._prod.size(); i++){
			LambdaProd prod1 = this._prod.get(i);
			for(int j = i+1 ; j<this._prod.size(); j++){
				LambdaProd prod2 = this._prod.get(j);
				if(prod1.equals(prod2)){
					this._prod.remove(j);
					j--;
				}
			}
		}
		
		//creating one cell for one prod
		this._cells = new LambdaLearnCell[this._prod.size()];
		for(int i = 0 ; i<this._prod.size(); i++){
			LambdaProd prod = this._prod.get(i);
			this._cells[i] = new LambdaLearnCell(prod, this._N);
			if(i == this._prod.size()-1){
				this._rootProd = prod;
			}
			//set the learn cell...
			prod.setLearnCell(this._cells[i]);
//			System.err.println(prod.getGM()+"\t|"+prod.hashCode()+"|"+this._cells[i].hashCode());
		}
		//index the ids for the words
		this.BEGIN_WORD_ID = Word._BEGIN.getId();
		this.END_WORD_ID = Word._END.getId();
		this.CHILD_A_WORD_ID = Word._X1.getId();
		this.CHILD_B_WORD_ID = Word._X2.getId();
	}
	
	/**
	 * Perform one EM step.
	 * @return the global inside score, possibly zero.
	 */
	public double EM_step(){
		
        long initTime = System.currentTimeMillis();
        this.computeBeta();
        long betaTime = System.currentTimeMillis();
        Logger.info("[BETA TIME] \t" + (betaTime - initTime) / 1000.0 + " seconds.");
        
        //perform outside computation only if the inside score is non-zero.
        if(this._globalBeta > 0){
            this.computeAlpha();
            long alphaTime = System.currentTimeMillis();
            Logger.info("[ALFA TIME]\t" + (alphaTime - betaTime) / 1000.0 + " seconds.");
        }
        
//        System.err.println("Beta = "+this._globalBeta);
//        this.checkUpdateCorrectness();
//        viewWordTable();
//        System.err.println(this._globalBeta);
//        System.exit(1);
        
        return this._globalBeta;
	}
	
	private void computeBeta_terminal(LambdaLearnCell cell, int bIndex, int eIndex){
		
		int len = eIndex - bIndex;

        /* terminal rule */
        /** compute the beta_w **/
        {
            double wProb;
            if (len == 1) {
                wProb = 1.0;
            } else {
                int word1 = this._words[eIndex - 2];
                int word = this._words[eIndex - 1];
                wProb = cell.getBetaProb_w(bIndex, eIndex - 1) *
                        this.computeOneWordProb(cell, word1, word);
            }
            cell.addBetaProb_w(bIndex, eIndex, wProb);
        }

        //case : w
        int bWord = this._words[bIndex];
        int eWord = this._words[eIndex - 1];
        double prob = this.computePatternProb(cell, LambdaPattern._w) *
                      this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                      cell.getBetaProb_w(bIndex, eIndex) * this.computeOneWordProb(cell, eWord, END_WORD_ID);

        cell.addBetaProb(bIndex, eIndex, prob);
	}
	
	private void computeBeta_unary(LambdaLearnCell cell, int bIndex, int eIndex){
		
		int len = eIndex - bIndex;

        double prob = 0;
        int pos = 0;
        
        int cWord = CHILD_A_WORD_ID;
        
        //case of TWO parts
        if (len >= 2) {
            double subProb;
            LambdaPattern p;

            // wX = w + ( X | wX )
            {
                p = LambdaPattern._wX;

                double prob_X = cell.getBetaProb_X(pos, bIndex + 1, eIndex);
                double prob_wX = cell.getBetaProb_wX(pos, bIndex + 1, eIndex);
                double prob_pattern = this.computePatternProb(cell, p);
                
                subProb = 0;
                
                if (prob_X > 0) {
                    subProb += this.computeOneWordProb(cell, this._words[bIndex], cWord) * prob_X;
                }
                
                if (prob_wX > 0 && len >= 3) {
                    subProb += this.computeOneWordProb(cell, this._words[bIndex], this._words[bIndex + 1]) * prob_wX;
                }
                
                if (subProb > 0) {
                    cell.addBetaProb_wX(pos, bIndex, eIndex, subProb);

                    if (prob_pattern > 0) {
                        prob += prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) * subProb *
                                this.computeOneWordProb(cell, cWord, END_WORD_ID);
                    }
                }
            }

            // Xw = ( X | Xw ) + w
            {
                p = LambdaPattern._Xw;

                double prob_X = cell.getBetaProb_X(pos, bIndex, eIndex - 1);
                double prob_Xw = cell.getBetaProb_Xw(pos, bIndex, eIndex - 1);
                double prob_pattern = this.computePatternProb(cell, p);

                subProb = 0;

                if (prob_X > 0) {
                    subProb += prob_X * this.computeOneWordProb(cell, cWord, this._words[eIndex - 1]);
                }

                if (prob_Xw > 0 && len >= 3) {
                    subProb += prob_Xw * this.computeOneWordProb(cell, this._words[eIndex - 2], this._words[eIndex - 1]);
                }

                if (subProb > 0) {
                    cell.addBetaProb_Xw(pos, bIndex, eIndex, subProb);

                    if (prob_pattern > 0) {
                        prob += prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, cWord) * subProb *
                                this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);
                    }
                }
            }
        }

        //case of THREE parts
        if (len >= 3) {
            double subProb;
            LambdaPattern p;

            // wXw = ( wX | wXw ) + w
            {
                p = LambdaPattern._wXw;
                
                double prob_wX = cell.getBetaProb_wX(pos, bIndex, eIndex - 1);
                double prob_wXw = cell.getBetaProb_wXw(pos, bIndex, eIndex - 1);
                double prob_pattern = this.computePatternProb(cell, p);

                subProb = 0;

                if (prob_wX > 0) {
                    subProb += this.computeOneWordProb(cell, cWord, this._words[eIndex - 1]) * prob_wX;
                }

                if (prob_wXw > 0 && len >= 4) {
                    subProb += this.computeOneWordProb(cell, this._words[eIndex - 2], this._words[eIndex - 1]) * prob_wXw;
                }

                if (subProb > 0) {
                    cell.addBetaProb_wXw(pos, bIndex, eIndex, subProb);

                    if (prob_pattern > 0) {
                        prob += prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) * subProb *
                                this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);
                    }
                }
            }
        }

        cell.addBetaProb(bIndex, eIndex, prob);
	}

	private void computeBeta_binary(LambdaLearnCell cell, int bIndex, int eIndex){
		
		int len = eIndex - bIndex;

        double prob = 0;
        double subProb;

        //case of TWO parts..
        if (len >= 2) {
            subProb = 0;

            //case : X | Y / Y | X
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._XY : LambdaPattern._YX;
                double prob_pattern = this.computePatternProb(cell, p);
                

                if (prob_pattern > 0) {
                    int bWord = this.getChildWord(pos);
                    int eWord = this.getChildWord(1 - pos);
                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, bWord, eWord) *
                            this.computeOneWordProb(cell, eWord, END_WORD_ID);
                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {
                            sProb += cell.getBetaProb_X(pos, bIndex, sIndex) * cell.getBetaProb_X(1 - pos, sIndex, eIndex);
                        }
                        subProb += sProb * tProb;
                    }
                }
            }

            prob += subProb;
        }

        //case of THREE parts..
        if (len >= 3) {
            subProb = 0;

            //case : X w | Y / Y w | X
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._XwY : LambdaPattern._YwX;
                double prob_pattern = this.computePatternProb(cell, p);

                if (prob_pattern > 0) {
                    int bWord = this.getChildWord(pos);
                    int eWord = this.getChildWord(1 - pos);
                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, eWord, END_WORD_ID);
                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                            int word = this._words[sIndex-1];
                            sProb += cell.getBetaProb_Xw(pos, bIndex, sIndex) * this.computeOneWordProb(cell, word, eWord) *
                                    cell.getBetaProb_X(1 - pos, sIndex, eIndex);
                        }
                        subProb += sProb * tProb;
                    }
                }
            }

            //case : X | Yw / Y | Xw
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._XYw : LambdaPattern._YXw;
                double prob_pattern = this.computePatternProb(cell, p);

                if (prob_pattern > 0) {
                    int bWord = this.getChildWord(pos);
                    int word = this.getChildWord(1 - pos);
                    int eWord = this._words[eIndex - 1];
                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, bWord, word) * this.computeOneWordProb(cell, eWord, END_WORD_ID);
                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                            sProb += cell.getBetaProb_X(pos, bIndex, sIndex) * cell.getBetaProb_Xw(1 - pos, sIndex, eIndex);
                        }
                        subProb += sProb * tProb;
                    }
                }
            }

            //case : wX | Y / wY | X
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._wXY : LambdaPattern._wYX;
                double prob_pattern = this.computePatternProb(cell, p);

                if (prob_pattern > 0) {
                    int bWord = this._words[bIndex];
                    int word = this.getChildWord(pos);
                    int eWord = this.getChildWord(1 - pos);
                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, word, eWord) *
                            this.computeOneWordProb(cell, eWord, END_WORD_ID);
                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                            sProb += cell.getBetaProb_wX(pos, bIndex, sIndex) *
                                    cell.getBetaProb_X(1 - pos, sIndex, eIndex);
                        }
                        subProb += tProb * sProb;
                    }
                }
            }

            prob += subProb;
        }

        //case of FOUR parts
        if (len >= 4) {
            //case : X w1 | Y w2 / Y w1 | X w2
            subProb = 0;
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._XwYw : LambdaPattern._YwXw;
                double prob_pattern = this.computePatternProb(cell, p);
                if (prob_pattern > 0) {
                    int bWord = this.getChildWord(pos);
                    int word2 = this.getChildWord(1 - pos);
                    int eWord = this._words[eIndex - 1];
                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, eWord, END_WORD_ID);
                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                            int word1 = this._words[sIndex - 1];
                            sProb += cell.getBetaProb_Xw(pos, bIndex, sIndex) * this.computeOneWordProb(cell, word1, word2) *
                                    cell.getBetaProb_Xw(1 - pos, sIndex, eIndex);
                        }
                        subProb += tProb * sProb;
                    }
                }
            }

            //case : w1 X | w2 Y / w1 Y | w2 X
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._wXwY : LambdaPattern._wYwX;
                double prob_pattern = this.computePatternProb(cell, p);
                if (prob_pattern > 0) {
                    int bWord = this._words[bIndex];
                    int word1 = this.getChildWord(pos);
                    int eWord = this.getChildWord(1 - pos);
                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, eWord, END_WORD_ID);
                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                            int word2 = this._words[sIndex];
                            sProb +=
                                    cell.getBetaProb_wX(pos, bIndex, sIndex) * this.computeOneWordProb(cell, word1, word2) *
                                    cell.getBetaProb_wX(1 - pos, sIndex, eIndex);
                        }
                        subProb += sProb * tProb;
                    }
                }
            }

            //case : w1 X | Y w2 / w1 Y | X w2
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._wXYw : LambdaPattern._wYXw;
                double prob_pattern = this.computePatternProb(cell, p);
                if (prob_pattern > 0) {
                    int bWord = this._words[bIndex];
                    int word1 = this.getChildWord(pos);
                    int word2 = this.getChildWord(1 - pos);
                    int eWord = this._words[eIndex - 1];
                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, word1, word2) * this.computeOneWordProb(cell, eWord, END_WORD_ID);
                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                            sProb += cell.getBetaProb_wX(pos, bIndex, sIndex) *
                                    cell.getBetaProb_Xw(1 - pos, sIndex, eIndex);
                        }
                        subProb += tProb * sProb;
                    }
                }
            }
            prob += subProb;
        }

        //case of FIVE parts
        if (len >= 5) {
            subProb = 0;
            //case : w1 X w2 | Y w3 / w1 Y w2 | X w3
            for (int pos = 0; pos <= 1; pos++) {
                LambdaPattern p = pos == 0 ? LambdaPattern._wXwYw : LambdaPattern._wYwXw;
                double prob_pattern = this.computePatternProb(cell, p);
                if (prob_pattern > 0) {
                    int bWord = this._words[bIndex];
                    int word1 = this.getChildWord(1-pos);
                    int eWord = this._words[eIndex - 1];

                    double tProb = prob_pattern * this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, eWord, END_WORD_ID);

                    if (tProb > 0) {
                        double sProb = 0;
                        for (int sIndex = bIndex + 3; sIndex <= eIndex - 2; sIndex++) {
                            sProb += cell.getBetaProb_wXw(pos, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this._words[sIndex-1], word1) *
                                    cell.getBetaProb_Xw(1 - pos, sIndex, eIndex);
                        }
                        subProb += sProb * tProb;
                    }
                }
            }
            prob += subProb;
        }

        cell.addBetaProb(bIndex, eIndex, prob);
	}

	private void computeBeta_postprocess(int bIndex, int eIndex){

		int len = eIndex - bIndex;
		
        //post-processing starts ...
        for (int m = 0; m < this._cells.length; m++) {
        	
            LambdaLearnCell cell = this._cells[m];
            LambdaProd prod = cell.getProd();
            
//        	System.err.println(">>>>>"+cell.getProd().getGM()+"\t"+m);
        	
            int arity = prod.getArity();
            
            //ignore if it is terminal rule
            if(arity == 0){
            	continue;
            }
            
            // computing the beta_X probabilities for this window size "len"
            if (arity == 1) {
            	
                int pos = 0;
                
                ArrayList<LambdaProd> children = prod.getFirstChildren();
                
                //case : X
                double subProb = 0;
                for (int n = 0; n < children.size(); n++) {
                	LambdaProd cProd = children.get(n);
                	LambdaLearnCell c_cell = cProd.getLearnCell();
                    subProb += this.computeTransProb(prod.getGP(), pos, cProd.getGP()) * c_cell.getBetaProb(bIndex, eIndex);
                }
                cell.addBetaProb_X(pos, bIndex, eIndex, subProb);

                /** special treatment for the root and Lx.x rule... **/
                if((len == this._N && prod==this._rootProd) || prod.getGP().isLxx()){
                    LambdaPattern p = LambdaPattern._X;
                    double prob = this.computePatternProb(cell, p) * cell.getBetaProb_X(pos, bIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, CHILD_A_WORD_ID) *
                                  this.computeOneWordProb(cell, CHILD_A_WORD_ID, END_WORD_ID);
                    cell.addBetaProb(bIndex, eIndex, prob);
                    /*
                    if(prod.getGM().isLxx()){
                    	System.err.println("xxxxxxxxxxxx>>>>");
                    	this.viewBeta(cell, bIndex, eIndex);
                    	System.err.println("xxxxxxxxxxxx<<<<"+cell.getProd().hashCode()+"|"+cell.hashCode()+"|"+cell.getProd().getLearnCell().hashCode());
                    }
                    */
                }
                
            }

            else if (arity == 2) {

                for (int pos = 0; pos < arity; pos++) {
                    int cWord = this.getChildWord(pos);
                    
                    ArrayList<LambdaProd> children = pos == 0 ? prod.getFirstChildren() : prod.getSecondChildren();
                    
                    //case of ONE part
                    {
                        //case : X
                        double subProb = 0;
                        for (int n = 0; n < children.size(); n++) {
                        	LambdaProd cProd = children.get(n);
                        	LambdaLearnCell c_cell = cProd.getLearnCell();
                            subProb += this.computeTransProb(prod.getGP(), pos, cProd.getGP()) * c_cell.getBetaProb(bIndex, eIndex);
                        }
                        
                        cell.addBetaProb_X(pos, bIndex, eIndex, subProb);
                        
                    }

                    //case of TWO parts
                    if (len >= 2) {
                        double subProb = 0;
                        
                        // wX = w + ( X | wX )
                        {
                            subProb = 0;
                            double prob_X = cell.getBetaProb_X(pos, bIndex + 1, eIndex);
                            double prob_wX = cell.getBetaProb_wX(pos, bIndex + 1, eIndex);
                            
                            if (prob_X > 0) {
                                subProb += this.computeOneWordProb(cell, this._words[bIndex], cWord) * prob_X;
                            }
                            if (prob_wX > 0 && len >= 3) {
                                subProb += this.computeOneWordProb(cell, this._words[bIndex], this._words[bIndex + 1]) * prob_wX;
                            }
                            
                            if (subProb > 0) {
                                cell.addBetaProb_wX(pos, bIndex, eIndex, subProb);
                            }
                        }
                        
                        // Xw = ( X | Xw ) + w
                        {
                            subProb = 0;
                            double prob_X = cell.getBetaProb_X(pos, bIndex, eIndex - 1);
                            double prob_Xw = cell.getBetaProb_Xw(pos, bIndex, eIndex - 1);
                            
                            if (prob_X > 0) {
                                subProb += prob_X * this.computeOneWordProb(cell, cWord, this._words[eIndex - 1]);
                            }

                            if (prob_Xw > 0 && len >= 3) {
                                subProb += prob_Xw * this.computeOneWordProb(cell, this._words[eIndex - 2], this._words[eIndex - 1]);
                            }

                            if (subProb > 0) {
                                cell.addBetaProb_Xw(pos, bIndex, eIndex, subProb);
                            }
                        }
                    }

                    //case of THREE parts
                    if (len >= 3) {
                        double subProb = 0;
                        double prob_wX = cell.getBetaProb_wX(pos, bIndex, eIndex - 1);
                        double prob_wXw = cell.getBetaProb_wXw(pos, bIndex, eIndex - 1);

                        //case : wXw = ( wX | wXw ) + w

                        if (prob_wX > 0) {
                            subProb += prob_wX * this.computeOneWordProb(cell, cWord, this._words[eIndex - 1]);
                        }

                        if (prob_wXw > 0 && len >= 4) {
                            subProb += prob_wXw * this.computeOneWordProb(cell, this._words[eIndex - 2], this._words[eIndex - 1]);
                        }

                        if (subProb > 0) {
                            cell.addBetaProb_wXw(pos, bIndex, eIndex, subProb);
                        }
                    }
                }

            }
        }
        //end of post-processing.
	}
	
	/**
	 * Computing the inside probabilities.
	 */
	public void computeBeta(){
		
        for (short len = 1; len <= this._N; len++) {
            Logger.info(len + "/" + this._N);
//            System.err.println("len="+len);
            
            for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                int eIndex = bIndex + len;
                
                //for each possible LambdaGMRProd
                for (int m = 0; m < this._cells.length; m++) {
                    LambdaLearnCell cell = this._cells[m];
                    LambdaGProd gm = cell.getProd().getGP();

                    int arity = gm.getArity();

                    /* terminal rule */
                    if (arity == 0) {
                    	this.computeBeta_terminal(cell, bIndex, eIndex);
                    }

                    /* unary rule */
                    else if (arity == 1) {
                    	this.computeBeta_unary(cell, bIndex, eIndex);
                    }
                    
                    /* binary rule */
                    else if (arity == 2) {
                    	this.computeBeta_binary(cell, bIndex, eIndex);
                    }

                    else {
                        throw new RuntimeException("Invalid number of RHS! Currently only supports up to 2 RHS for GMRTokens!");
                    }
                    
                }
                
                this.computeBeta_postprocess(bIndex, eIndex);

                /*
                for (int m = 0; m < this._cells.length; m++) {
                    LambdaLearnCell cell = this._cells[m];
                    this.viewBeta(cell, bIndex, eIndex);
                }
                */
            }
        }
        
        this._globalBeta = this._rootProd.getLearnCell().getBetaProb(0, this._N);
	}
	
	private void computeAlpha_root(){
		
        LambdaLearnCell rootCell = this._rootProd.getLearnCell();
        
        double prob = 1.0;

        //initialize the alpha scores for the rootProd.
        rootCell.addAlphaProb(0, this._N, prob);
        for (int len = this._N - 1; len >= 1; len--) {
            for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                int eIndex = bIndex + len;
                rootCell.addAlphaProb(bIndex, eIndex, 0.0);
            }
        }

        this.computeAlpha_unary(rootCell, 0, this._N);
        
        double subProb;
        LambdaPattern p = LambdaPattern._X;
        int cIndex = 0;
        int cWord = CHILD_A_WORD_ID;
        
        LambdaLearnCell cell = rootCell;

        double prob_beta = cell.getBetaProb(0, this._N);

        if (prob > 0 && prob_beta > 0) {
            subProb = prob * this.computePatternProb(cell, p) *
                      this.computeOneWordProb(cell, BEGIN_WORD_ID, cWord) *
                      this.computeOneWordProb(cell, cWord, END_WORD_ID);

            //update the rule and word counts.
            if (_learnPattern || _learnEmission) {
                double count = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, 0, this._N)) - Math.log(this._globalBeta);
                count = Math.exp(count);

                if (count > 0) {
                    if (_learnPattern) {
                        this.updatePatternProb(cell, p, count);
                    }
                    if (_learnEmission) {
                        this.updateWordProb(cell, BEGIN_WORD_ID, cWord, count);
                        this.updateWordProb(cell, cWord, END_WORD_ID, count);
                    }
                }
            }

            //propagate
            if (subProb > 0) {
            	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();
                for (int n = 0; n < children.size(); n++) {
                	LambdaProd cProd = children.get(n);
                	LambdaLearnCell child = cProd.getLearnCell();

                    double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                    child.addAlphaProb(0, this._N, subAlpha);

                    if (_learnTransition) {
                        //update the transition probability
                        double count = Math.log(subAlpha) +
                                Math.log(child.getBetaProb(0, this._N)) -
                                Math.log(this._globalBeta);
                        count = Math.exp(count);

                        this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                    }
                }
            }
        }
	}
	
	private void computeAlpha_terminal(LambdaLearnCell cell, int bIndex, int eIndex){
		
		int len = eIndex - bIndex;

        double prob = cell.getAlphaProb(bIndex, eIndex);

        double subProb;
        double count;
        int bWord = this._words[bIndex];
        int eWord = this._words[eIndex - 1];

        //do this iff prob is greater than zero
        if (prob > 0) {
            subProb = prob * this.computePatternProb(cell, LambdaPattern._w) *
                      this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord) *
                      this.computeOneWordProb(cell, eWord, END_WORD_ID);

            cell.addAlphaProb_w(bIndex, eIndex, subProb);
        }

        //update the rule probabilities..
        if (_learnPattern && prob > 0) {
            double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, LambdaPattern._w)) +
                               Math.log(this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord)) +
                               Math.log(cell.getBetaProb_w(bIndex, eIndex)) +
                               Math.log(this.computeOneWordProb(cell, eWord, END_WORD_ID)) -
                               Math.log(this._globalBeta);
            ruleCount = Math.exp(ruleCount);

            this.updatePatternProb(cell, LambdaPattern._w, ruleCount);
        }

        if (len == 1) {
            //update...
            if (_learnEmission) {
                count = Math.log(cell.getAlphaProb_w(bIndex, eIndex)) + Math.log(cell.getBetaProb_w(bIndex, eIndex)) -
                        Math.log(this._globalBeta);
                count = Math.exp(count);

                this.updateWordProb(cell, BEGIN_WORD_ID, eWord, count);
            }
        } else {
            int eWord1 = this._words[eIndex - 2];
            subProb = cell.getAlphaProb_w(bIndex, eIndex) * this.computeOneWordProb(cell, eWord1, eWord);

            cell.addAlphaProb_w(bIndex, eIndex - 1, subProb);
            
            //update...
            if (_learnEmission) {
                count = Math.log(cell.getAlphaProb_w(bIndex, eIndex)) +
                        Math.log(cell.getBetaProb_w(bIndex, eIndex)) -
                        Math.log(this._globalBeta);
                count = Math.exp(count);

                this.updateWordProb(cell, eWord1, eWord, count);
            }
        }

        //for the END
        if (_learnEmission) {
            count = Math.log(cell.getAlphaProb(bIndex, eIndex)) + 
            		Math.log(cell.getBetaProb(bIndex, eIndex)) -
                    Math.log(this._globalBeta);
            count = Math.exp(count);

            this.updateWordProb(cell, eWord, END_WORD_ID, count);
        }
	}
	
	private void computeAlpha_unary(LambdaLearnCell cell, int bIndex, int eIndex){

		int len = eIndex - bIndex;

        double prob = cell.getAlphaProb(bIndex, eIndex);

        int cIndex = 0;

        int cWord = CHILD_A_WORD_ID;
        
    	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();

        double subProb;
        double tCount;

        LambdaLearnCell child;

        //do these if and only if prob is greater than zero!
        if (prob > 0) {

            double prob_pattern;
            LambdaPattern p;

            p = LambdaPattern._wXw;
            prob_pattern = this.computePatternProb(cell, p);

            if (prob_pattern > 0) {
                subProb = prob * prob_pattern *
                          this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                          this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);
                cell.addAlphaProb_wXw(cIndex, bIndex, eIndex, subProb);

                //update...
                if (_learnEmission && subProb > 0) {
                    tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wXw(cIndex, bIndex, eIndex)) - Math.log(this._globalBeta);
                    tCount = Math.exp(tCount);
                    this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                    this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                }
            }

            p = LambdaPattern._wX;
            prob_pattern = this.computePatternProb(cell, p);

            if (prob_pattern > 0) {
                subProb = prob * prob_pattern *
                          this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                          this.computeOneWordProb(cell, cWord, END_WORD_ID);
                cell.addAlphaProb_wX(cIndex, bIndex, eIndex, subProb);

                //update...
                if (_learnEmission && subProb > 0) {
                    tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, eIndex)) - Math.log(this._globalBeta);
                    tCount = Math.exp(tCount);
                    /*
                    if(tCount > 1.0001){
                    	System.err.println("x:"+Math.log(subProb)+"/"+Math.log(cell.getBetaProb_wX(cIndex, bIndex, eIndex))+"/"+Math.log(this._globalBeta)+"/"+Math.log(1.2E-322));
                    	System.err.println(Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, eIndex)) - Math.log(this._globalBeta));
                    	System.err.println(Double.MIN_VALUE);
                    }
                    */
                    this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                    this.updateWordProb(cell, cWord, END_WORD_ID, tCount);
                }
            }

            p = LambdaPattern._Xw;
            prob_pattern = this.computePatternProb(cell, p);

            if (prob_pattern > 0) {
                subProb = prob * prob_pattern *
                          this.computeOneWordProb(cell, BEGIN_WORD_ID, cWord) *
                          this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);
                cell.addAlphaProb_Xw(cIndex, bIndex, eIndex, subProb);

                //update...
                if (_learnEmission && subProb > 0) {
                    tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, bIndex, eIndex)) - Math.log(this._globalBeta);
                    tCount = Math.exp(tCount);
                    this.updateWordProb(cell, BEGIN_WORD_ID, cWord, tCount);
                    this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                }
            }
        }

        //update rule LambdaPattern probabilities...
        if (_learnPattern && prob > 0) {

            int bWord = this._words[bIndex];
            int eWord = this._words[eIndex - 1];

            double ruleCount;
            LambdaPattern p;
            double prob_pattern;

            p = LambdaPattern._wXw;
            prob_pattern = this.computePatternProb(cell, p);
            if (prob_pattern > 0) {
                ruleCount = Math.log(prob) + Math.log(prob_pattern) +
                            Math.log(this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord)) +
                            Math.log(cell.getBetaProb_wXw(cIndex, bIndex, eIndex)) +
                            Math.log(this.computeOneWordProb(cell, eWord, END_WORD_ID)) - Math.log(this._globalBeta);
                ruleCount = Math.exp(ruleCount);

                this.updatePatternProb(cell, p, ruleCount);
            }

            p = LambdaPattern._wX;
            prob_pattern = this.computePatternProb(cell, p);
            if (prob_pattern > 0) {
                ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                            Math.log(this.computeOneWordProb(cell, BEGIN_WORD_ID, bWord)) +
                            Math.log(cell.getBetaProb_wX(cIndex, bIndex, eIndex)) +
                            Math.log(this.computeOneWordProb(cell, cWord, END_WORD_ID)) - Math.log(this._globalBeta);
                ruleCount = Math.exp(ruleCount);

                this.updatePatternProb(cell, p, ruleCount);
            }

            p = LambdaPattern._Xw;
            prob_pattern = this.computePatternProb(cell, p);
            if (prob_pattern > 0) {
                ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                            Math.log(this.computeOneWordProb(cell, BEGIN_WORD_ID, cWord)) +
                            Math.log(cell.getBetaProb_Xw(cIndex, bIndex, eIndex)) +
                            Math.log(this.computeOneWordProb(cell, eWord, END_WORD_ID)) - Math.log(this._globalBeta);
                ruleCount = Math.exp(ruleCount);

                this.updatePatternProb(cell, p, ruleCount);
            }
        }

        // wXw = ( wX | wXw ) + w
        /**...enclosed by w1Xw2...**/
        if (len >= 3) {
            int eWord = this._words[eIndex - 1];

            if (len >= 4) {
                int eWord1 = this._words[eIndex - 2];
                subProb = cell.getAlphaProb_wXw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, eWord1, eWord);
                cell.addAlphaProb_wXw(cIndex, bIndex, eIndex - 1, subProb);

                //update:
                if (_learnEmission) {
                    double count = Math.log(subProb) + Math.log(cell.getBetaProb_wXw(cIndex, bIndex, eIndex - 1)) -
                            Math.log(this._globalBeta);
                    count = Math.exp(count);

                    this.updateWordProb(cell, eWord1, eWord, count);
                }
            }

            subProb = cell.getAlphaProb_wXw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, cWord, eWord);
            cell.addAlphaProb_wX(cIndex, bIndex, eIndex - 1, subProb);

            //update:
            if (_learnEmission) {
                double count = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, eIndex - 1)) -
                               Math.log(this._globalBeta);
                count = Math.exp(count);

                this.updateWordProb(cell, cWord, eWord, count);
            }
        }

        /**enclosed by Xw**/
        if (len >= 2) {
            int eWord = this._words[eIndex - 1];

            if (len >= 3) {
                int eWord1 = this._words[eIndex - 2];
                subProb = cell.getAlphaProb_Xw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, eWord1, eWord);
                cell.addAlphaProb_Xw(cIndex, bIndex, eIndex - 1, subProb);

                //update:
                if (_learnEmission) {
                    double count = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, bIndex, eIndex - 1)) -
                            Math.log(this._globalBeta);
                    count = Math.exp(count);

                    this.updateWordProb(cell, eWord1, eWord, count);
                }
            }

            subProb = cell.getAlphaProb_Xw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, cWord, eWord);

            //propagate
            if (subProb > 0) {
            	
                for (int n = 0; n < children.size(); n++) {
                	LambdaProd cProd = children.get(n);
                	child = cProd.getLearnCell();

                    double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                    child.addAlphaProb(bIndex, eIndex - 1, subAlpha);

                    if (_learnTransition) {
                        //update the transition probability
                        double count = Math.log(subAlpha) +
                                Math.log(child.getBetaProb(bIndex, eIndex - 1)) -
                                Math.log(this._globalBeta);
                        count = Math.exp(count);

                        this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                    }
                }
            }

            //update:
            if (_learnEmission && subProb > 0) {
                double count = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex, eIndex - 1)) -
                               Math.log(this._globalBeta);
                count = Math.exp(count);

                this.updateWordProb(cell, cWord, eWord, count);
            }
        }

        /**enclosed by wX**/
        if (len >= 2) {
            int bWord = this._words[bIndex];

            if (len >= 3) {
                int bWord_1 = this._words[bIndex + 1];
                subProb = cell.getAlphaProb_wX(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, bWord, bWord_1);
                cell.addAlphaProb_wX(cIndex, bIndex + 1, eIndex, subProb);

                //update:
                if (_learnEmission && subProb > 0) {
                    double count = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex + 1, eIndex)) -
                            Math.log(this._globalBeta);
                    count = Math.exp(count);

                    this.updateWordProb(cell, bWord, bWord_1, count);
                }
            }
            
            subProb = cell.getAlphaProb_wX(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, bWord, cWord);

            //propagate
            if (subProb > 0) {
                for (int n = 0; n < children.size(); n++) {
                	LambdaProd cProd = children.get(n);
                	child = cProd.getLearnCell();
                    
                    double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                    child.addAlphaProb(bIndex + 1, eIndex, subAlpha);

                    if (_learnTransition) {
                        //update the transition probability
                        double count = Math.log(subAlpha) +
                                Math.log(child.getBetaProb(bIndex + 1, eIndex)) -
                                Math.log(this._globalBeta);
                        count = Math.exp(count);

                        this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                    }
                }
            }

            //update:
            if (_learnEmission && subProb > 0) {
                double count = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex + 1, eIndex)) -
                               Math.log(this._globalBeta);
                count = Math.exp(count);

                this.updateWordProb(cell, bWord, cWord, count);
            }
        }
        
	}
	
	private void computeAlpha_binary(LambdaLearnCell cell, int bIndex, int eIndex){

		int len = eIndex - bIndex;

        double prob = cell.getAlphaProb(bIndex, eIndex);
        double subProb;

        //do the following iff prob is greater than zero!
        if (prob > 0) {

            //update rule probabilities...
            if (_learnPattern) {

                //CASE OF FIVE PARTS...
                if (len >= 5) {
                    // w1 X w2 | Y w3
                    // w1 Y w2 | X w3
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        subProb = 0;
                        LambdaPattern p = cIndex == 0 ? LambdaPattern._wXwYw : LambdaPattern._wYwXw;
                        for (int sIndex = bIndex + 3; sIndex <= eIndex - 2; sIndex++) {
                            subProb += cell.getBetaProb_wXw(cIndex, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex)) *
                                    cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex);
                        }
                        subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                                this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                        double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                                Math.log(subProb) - Math.log(this._globalBeta);
                        ruleCount = Math.exp(ruleCount);

                        this.updatePatternProb(cell, p, ruleCount);
                    }
                }

                //CASE OF FOUR PARTS..
                if (len >= 4) {
                    // w1 X | w2 Y
                    // w1 Y | w2 X
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        subProb = 0;
                        LambdaPattern p = cIndex == 0 ? LambdaPattern._wXwY : LambdaPattern._wYwX;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                            subProb += cell.getBetaProb_wX(cIndex, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this.getChildWord(cIndex), this._words[sIndex]) *
                                    cell.getBetaProb_wX(1 - cIndex, sIndex, eIndex);
                        }
                        subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                                this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                        double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                                Math.log(subProb) - Math.log(this._globalBeta);
                        ruleCount = Math.exp(ruleCount);

                        this.updatePatternProb(cell, p, ruleCount);
                    }

                    // X w1 | Y w2
                    // Y w1 | X w2
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        subProb = 0;
                        LambdaPattern p = cIndex == 0 ? LambdaPattern._XwYw : LambdaPattern._YwXw;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                            subProb += cell.getBetaProb_Xw(cIndex, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex)) *
                                    cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex);
                        }
                        subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                        double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                                Math.log(subProb) - Math.log(this._globalBeta);
                        ruleCount = Math.exp(ruleCount);

                        this.updatePatternProb(cell, p, ruleCount);
                    }

                    // w1 X | Y w2
                    // w1 Y | X w2
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        subProb = 0;
                        LambdaPattern p = cIndex == 0 ? LambdaPattern._wXYw : LambdaPattern._wYXw;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                            subProb += cell.getBetaProb_wX(cIndex, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                    cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex);
                        }
                        subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                                this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                        double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                                Math.log(subProb) - Math.log(this._globalBeta);
                        ruleCount = Math.exp(ruleCount);

                        this.updatePatternProb(cell, p, ruleCount);
                    }
                }

                //CASE OF THREE PARTS.
                if (len >= 3) {
                    // w X | Y
                    // w Y | X
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        subProb = 0;
                        LambdaPattern p = cIndex == 0 ? LambdaPattern._wXY : LambdaPattern._wYX;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                            subProb += cell.getBetaProb_wX(cIndex, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                    cell.getBetaProb_X(1 - cIndex, sIndex, eIndex);
                        }
                        subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                                this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                        double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                                Math.log(subProb) - Math.log(this._globalBeta);
                        ruleCount = Math.exp(ruleCount);

                        this.updatePatternProb(cell, p, ruleCount);
                    }

                    // X w | Y
                    // Y w | X
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        subProb = 0;
                        LambdaPattern p = cIndex == 0 ? LambdaPattern._XwY : LambdaPattern._YwX;
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                            subProb += cell.getBetaProb_Xw(cIndex, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex)) *
                                    cell.getBetaProb_X(1 - cIndex, sIndex, eIndex);
                        }
                        subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                        double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                                Math.log(subProb) - Math.log(this._globalBeta);
                        ruleCount = Math.exp(ruleCount);

                        this.updatePatternProb(cell, p, ruleCount);
                    }

                    // X | Y w
                    // Y | X w
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        subProb = 0;
                        LambdaPattern p = cIndex == 0 ? LambdaPattern._XYw : LambdaPattern._YXw;
                        for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                            subProb += cell.getBetaProb_X(cIndex, bIndex, sIndex) *
                                    this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                    cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex);
                        }
                        subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                        double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) +
                                Math.log(subProb) + -Math.log(this._globalBeta);
                        ruleCount = Math.exp(ruleCount);

                        this.updatePatternProb(cell, p, ruleCount);
                    }
                }
            } //above not yet optimized.

            //update the word probabilities...
            //stage 1: decomposition phase
            //CASE OF FIVE PARTS.
            if (len >= 5) {
                double tCount;
                double tProb = prob * this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                               this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                LambdaPattern p;

                //case : w1 X w2 | Y w3 / w1 Y w2 | X w3
                if (tProb > 0) {
                    for (int cIndex = 0; cIndex <= 1; cIndex++) {
                        p = cIndex == 0 ? LambdaPattern._wXwYw : LambdaPattern._wYwXw;
                        double prob_pattern = this.computePatternProb(cell, p);
                        if (prob_pattern > 0) {
                            for (int sIndex = bIndex + 3; sIndex <= eIndex - 2; sIndex++) {
                                double sProb = tProb * prob_pattern *
                                        this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex));

                                if (sProb > 0) {
                                    // w1 Y w2 | * X w3 *
                                    subProb = sProb * cell.getBetaProb_wXw(cIndex, bIndex, sIndex);
                                    cell.addAlphaProb_Xw(1 - cIndex, sIndex, eIndex, subProb);

                                    // * w1 Y w2 * | X w3
                                    subProb = sProb * cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex);
                                    cell.addAlphaProb_wXw(cIndex, bIndex, sIndex, subProb);

                                    if (_learnEmission) {
                                        tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wXw(cIndex, bIndex, sIndex)) -
                                                Math.log(this._globalBeta);
                                        tCount = Math.exp(tCount);

                                        this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                                        this.updateWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex), tCount);
                                        this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                                    }
                                }
                            }
                        }
                    }
                }
            }

            //CASE OF FOUR PARTS.
            if (len >= 4) {
                double tCount;
                for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                    //case : w1 X | w2 Y / w1 Y | w2 X
                    {
                        double tProb = prob * this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]);

                        int cIndex;
                        double sProb;

                        cIndex = 0;
                        sProb = tProb * this.computePatternProb(cell, LambdaPattern._wXwY) *
                                this.computeOneWordProb(cell, this.getChildWord(cIndex), this._words[sIndex]) *
                                this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);
                        // * w1 X * | w2 Y
                        subProb = sProb * cell.getBetaProb_wX(1 - cIndex, sIndex, eIndex);
                        cell.addAlphaProb_wX(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), this._words[sIndex], tCount);
                        }

                        // w1 X | * w2 Y *
                        subProb = sProb * cell.getBetaProb_wX(cIndex, bIndex, sIndex);
                        cell.addAlphaProb_wX(1 - cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(1 - cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID, tCount);
                        }

                        cIndex = 1;
                        sProb = tProb * this.computePatternProb(cell, LambdaPattern._wYwX) *
                                this.computeOneWordProb(cell, this.getChildWord(cIndex), this._words[sIndex]) *
                                this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);
                        // w1 Y | * w2 X *
                        subProb = sProb * cell.getBetaProb_wX(cIndex, bIndex, sIndex);
                        cell.addAlphaProb_wX(1 - cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(1 - cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(cIndex), this._words[sIndex], tCount);
                            this.updateWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID, tCount);
                        }

                        // * w1 Y * | w2 X
                        subProb = sProb * cell.getBetaProb_wX(1 - cIndex, sIndex, eIndex);
                        cell.addAlphaProb_wX(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                        }
                    }

                    //case : X w1 | Y w2 / Y w1 | X w2
                    {
                        double tProb = prob * this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                        int cIndex;

                        cIndex = 0;

                        // * X w1 * | Y w2
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._XwYw) *
                                  cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex));
                        cell.addAlphaProb_Xw(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex), tCount);
                        }

                        // Y w1 | * X w2 *
                        subProb = tProb * cell.getBetaProb_Xw(1 - cIndex, bIndex, sIndex) *
                                  this.computePatternProb(cell, LambdaPattern._YwXw) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(cIndex));
                        cell.addAlphaProb_Xw(cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this._words[sIndex - 1], this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                        }

                        cIndex = 1;

                        // * Y w1 * | X w2
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._YwXw) *
                                  cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex));
                        cell.addAlphaProb_Xw(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                        }

                        // X w1 | * Y w2 *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._XwYw) *
                                  cell.getBetaProb_Xw(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(cIndex));
                        cell.addAlphaProb_Xw(cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                        }
                    }

                    //case : w1 X | Y w2 / w1 Y | X w2
                    {
                        double tProb = prob * this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]) *
                                this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                        int cIndex;

                        cIndex = 0;
                        // * w1 X * | Y w2
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wXYw) *
                                  cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));
                        cell.addAlphaProb_wX(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);
                        }

                        // w1 Y | * X w2 *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wYXw) *
                                  cell.getBetaProb_wX(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));
                        cell.addAlphaProb_Xw(cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                        }

                        cIndex = 1;
                        // * w1 Y * | X w2
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wYXw) *
                                  cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));
                        cell.addAlphaProb_wX(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                        }

                        // w1 X | * Y w2 *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wXYw) *
                                  cell.getBetaProb_wX(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));
                        cell.addAlphaProb_Xw(cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                        }
                    }
                }
            }

            //CASE OF THREE PARTS.
            if (len >= 3) {
                double tCount;
                for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                    //case : w X | Y / w Y | X
                    {
                        double tProb = prob * this.computeOneWordProb(cell, BEGIN_WORD_ID, this._words[bIndex]);

                        int cIndex;

                        cIndex = 0;
                        LambdaLearnCell child;

                        // * w X * | Y
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wXY) * cell.getBetaProb_X(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                        cell.addAlphaProb_wX(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);
                        }

                        // w Y | * X *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wYX) *
                                  cell.getBetaProb_wX(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	
                        	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();
                        	
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();
                                
                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(sIndex, eIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) +
                                            Math.log(child.getBetaProb(sIndex, eIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), END_WORD_ID, tCount);
                        }

                        cIndex = 1;

                        // * w Y * | X
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wYX) * cell.getBetaProb_X(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                        cell.addAlphaProb_wX(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this._words[bIndex], tCount);
                        }

                        // w X | * Y *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._wXY) *
                                  cell.getBetaProb_wX(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getSecondChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(sIndex, eIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) +
                                            Math.log(child.getBetaProb(sIndex, eIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }

                            if (_learnEmission) {
                                tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, sIndex, eIndex)) -
                                        Math.log(this._globalBeta);
                                tCount = Math.exp(tCount);

                                this.updateWordProb(cell, this.getChildWord(cIndex), END_WORD_ID, tCount);
                            }
                        }
                    }

                    //case : X w | Y / Y w | X
                    {
                        int cIndex;

                        cIndex = 0;
                        LambdaLearnCell child;

                        // * X w * | Y
                        subProb = prob * this.computePatternProb(cell, LambdaPattern._XwY) * cell.getBetaProb_X(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);
                        cell.addAlphaProb_Xw(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex), tCount);
                        }

                        // Y w | * X *
                        subProb = prob * this.computePatternProb(cell, LambdaPattern._YwX) * cell.getBetaProb_Xw(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(sIndex, eIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) +
                                            Math.log(child.getBetaProb(sIndex, eIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this._words[sIndex - 1], this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), END_WORD_ID, tCount);
                        }

                        cIndex = 1;

                        // * Y w * | X
                        subProb = prob * this.computePatternProb(cell, LambdaPattern._YwX) * cell.getBetaProb_X(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);
                        cell.addAlphaProb_Xw(cIndex, bIndex, sIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                        }

                        // X w | * Y *
                        subProb = prob * this.computePatternProb(cell, LambdaPattern._XwY) * cell.getBetaProb_Xw(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this._words[sIndex - 1], this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getSecondChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(sIndex, eIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) +
                                            Math.log(child.getBetaProb(sIndex, eIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(cIndex), END_WORD_ID, tCount);
                        }
                    }
                }

                for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                    //case : X | Y w / Y | X w
                    {
                        double tProb = prob * this.computeOneWordProb(cell, this._words[eIndex - 1], END_WORD_ID);

                        int cIndex;

                        cIndex = 0;
                        LambdaLearnCell child;

                        // * X * | Y w
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._XYw) *
                                  cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(bIndex, sIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) +
                                            Math.log(child.getBetaProb(bIndex, sIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);
                        }

                        // Y | * X w *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._YXw) * cell.getBetaProb_X(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));

                        cell.addAlphaProb_Xw(cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                        }

                        cIndex = 1;

                        // * Y * | X w
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._YXw) *
                                  cell.getBetaProb_Xw(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getSecondChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                            	child = cProd.getLearnCell();
                            	
                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(bIndex, sIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) +
                                            Math.log(child.getBetaProb(bIndex, sIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                        }

                        // X | * Y w *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._XYw) * cell.getBetaProb_X(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));

                        cell.addAlphaProb_Xw(cIndex, sIndex, eIndex, subProb);

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this._words[eIndex - 1], END_WORD_ID, tCount);
                        }
                    }
                }
            }
        }
        
        this.computeAlpha_postprocess(cell, bIndex, eIndex);
	}
	
	private void computeAlpha_postprocess(LambdaLearnCell cell, int bIndex, int eIndex){

		int len = eIndex - bIndex;
        double prob = cell.getAlphaProb(bIndex, eIndex);
		
		double subProb;

        /**POST PROCESSING BEGINS**/
        {
            double count;

            for (int cIndex = 0; cIndex <= 1; cIndex++) {
            	
            	ArrayList<LambdaProd> children;
            	if(cIndex == 0){
            		children = cell.getProd().getFirstChildren();
            	} else {
            		children = cell.getProd().getSecondChildren();
            	}

                LambdaLearnCell child;
                int cWord = this.getChildWord(cIndex);
                // case : w1 X w2
                if (len >= 3) {
                    int eWord = this._words[eIndex - 1];

                    // w1 X w2 | w
                    if (len >= 4) {
                        int eWord1 = this._words[eIndex - 2];
                        subProb = cell.getAlphaProb_wXw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, eWord1, eWord);
                        cell.addAlphaProb_wXw(cIndex, bIndex, eIndex - 1, subProb);

                        //update...
                        if (_learnEmission) {
                            count = Math.log(subProb) + Math.log(cell.getBetaProb_wXw(cIndex, bIndex, eIndex - 1)) -
                                    Math.log(this._globalBeta);
                            count = Math.exp(count);

                            this.updateWordProb(cell, eWord1, eWord, count);
                        }
                    }

                    // w1 X | w
                    subProb = cell.getAlphaProb_wXw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, cWord, eWord);
                    cell.addAlphaProb_wX(cIndex, bIndex, eIndex - 1, subProb);

                    //update...
                    if (_learnEmission) {
                        count = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex, eIndex - 1)) -
                                Math.log(this._globalBeta);
                        count = Math.exp(count);

                        this.updateWordProb(cell, cWord, eWord, count);
                    }
                }

                // case : X w
                if (len >= 2) {
                    int eWord = this._words[eIndex - 1];

                    // X w | w
                    if (len >= 3) {
                        int eWord1 = this._words[eIndex - 2];
                        subProb = cell.getAlphaProb_Xw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, eWord1, eWord);
                        cell.addAlphaProb_Xw(cIndex, bIndex, eIndex - 1, subProb);

                        //update...
                        if (_learnEmission) {
                            count = Math.log(subProb) + Math.log(cell.getBetaProb_Xw(cIndex, bIndex, eIndex - 1)) -
                                    Math.log(this._globalBeta);
                            count = Math.exp(count);

                            this.updateWordProb(cell, eWord1, eWord, count);
                        }
                    }

                    // X | w
                    subProb = cell.getAlphaProb_Xw(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, cWord, eWord);

                    //propagate
                    if (subProb > 0) {
                        for (int n = 0; n < children.size(); n++) {
                        	LambdaProd cProd = children.get(n);
                            child = cProd.getLearnCell();

                            double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                            child.addAlphaProb(bIndex, eIndex - 1, subAlpha);

                            if (_learnTransition) {
                                //update the transition probability
                                count = Math.log(subAlpha) +
                                        Math.log(child.getBetaProb(bIndex, eIndex - 1)) -
                                        Math.log(this._globalBeta);
                                count = Math.exp(count);

                                this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                            }
                        }
                    }

                    //update...
                    if (_learnEmission) {
                        count = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex, eIndex - 1)) -
                                Math.log(this._globalBeta);
                        count = Math.exp(count);

                        this.updateWordProb(cell, cWord, eWord, count);
                    }
                }

                // case : w X
                if (len >= 2) {
                    int bWord = this._words[bIndex];

                    // w | w X
                    if (len >= 3) {
                        int bWord_1 = this._words[bIndex + 1];
                        subProb = cell.getAlphaProb_wX(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, bWord, bWord_1);
                        cell.addAlphaProb_wX(cIndex, bIndex + 1, eIndex, subProb);

                        //update..
                        if (_learnEmission) {
                            count = Math.log(subProb) + Math.log(cell.getBetaProb_wX(cIndex, bIndex + 1, eIndex)) -
                                    Math.log(this._globalBeta);
                            count = Math.exp(count);

                            this.updateWordProb(cell, bWord, bWord_1, count);
                        }
                    }

                    // w | X
                    subProb = cell.getAlphaProb_wX(cIndex, bIndex, eIndex) * this.computeOneWordProb(cell, bWord, cWord);

                    //propagate
                    if (subProb > 0) {
                        for (int n = 0; n < children.size(); n++) {
                        	LambdaProd cProd = children.get(n);
                            child = cProd.getLearnCell();

                            double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                            child.addAlphaProb(bIndex + 1, eIndex, subAlpha);

                            if (_learnTransition) {
                                //update the transition probability
                                count = Math.log(subAlpha) +
                                        Math.log(child.getBetaProb(bIndex + 1, eIndex)) -
                                        Math.log(this._globalBeta);
                                count = Math.exp(count);

                                this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                            }
                        }
                    }

                    //update..
                    if (_learnEmission) {
                        count = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex + 1, eIndex)) -
                                Math.log(this._globalBeta);
                        count = Math.exp(count);

                        this.updateWordProb(cell, bWord, cWord, count);
                    }
                }
            }
        }

        /**POST PROCESSING ENDS**/


        /** special treatment for the direct token only **/
        {
            if (_learnPattern && len >= 2) {
                // X | Y
                // Y | X
                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                    LambdaPattern p = cIndex == 0 ? LambdaPattern._XY : LambdaPattern._YX;

                    subProb = 0;
                    for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {
                        subProb += cell.getBetaProb_X(cIndex, bIndex, sIndex) *
                                this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                cell.getBetaProb_X(1 - cIndex, sIndex, eIndex);
                    }
                    subProb *= this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                            this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                    double ruleCount = Math.log(prob) + Math.log(this.computePatternProb(cell, p)) + Math.log(subProb) -
                            Math.log(this._globalBeta);
                    ruleCount = Math.exp(ruleCount);

                    this.updatePatternProb(cell, p, ruleCount);
                }
            }

            if (len >= 2) {
                double tCount;

                for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {
                    //case : X | Y / Y | X
                    {
                        double tProb = prob;

                        int cIndex;

                        cIndex = 0;
                        LambdaLearnCell child;

                        // * X * | Y
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._XY) *
                                  cell.getBetaProb_X(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(bIndex, sIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) + Math.log(child.getBetaProb(bIndex, sIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);
                        }

                        // Y | * X *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._YX) *
                                  cell.getBetaProb_X(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                            	child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(sIndex, eIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) + Math.log(child.getBetaProb(sIndex, eIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                            this.updateWordProb(cell, this.getChildWord(cIndex), END_WORD_ID, tCount);
                        }

                        cIndex = 1;

                        // * Y * | X
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._YX) *
                                  cell.getBetaProb_X(1 - cIndex, sIndex, eIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getSecondChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(bIndex, sIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) + Math.log(child.getBetaProb(bIndex, sIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex, sIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, BEGIN_WORD_ID, this.getChildWord(cIndex), tCount);
                        }

                        // X | * Y *
                        subProb = tProb * this.computePatternProb(cell, LambdaPattern._XY) *
                                  cell.getBetaProb_X(1 - cIndex, bIndex, sIndex) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, this.getChildWord(1 - cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                  this.computeOneWordProb(cell, this.getChildWord(cIndex), END_WORD_ID);

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getSecondChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                                child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(sIndex, eIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) + Math.log(child.getBetaProb(sIndex, eIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }

                        if (_learnEmission) {
                            tCount = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, sIndex, eIndex)) -
                                    Math.log(this._globalBeta);
                            tCount = Math.exp(tCount);

                            this.updateWordProb(cell, this.getChildWord(cIndex), END_WORD_ID, tCount);
                        }
                    }
                }
            }

        }
	}
	
	public void computeAlpha(){

        if (this._globalBeta == 0) {
            Logger.warn("Instance ignored due to zero inside probability.");
            return;
        }
        
        this.computeAlpha_root();
        
        for (int len = this._N; len >= 1; len--) {
        	
            Logger.info(len + "/" + this._N);
            
            for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                int eIndex = bIndex + len;
                
                /** deal with the Lx.x only **/
                for (int m = this._cells.length-1; m >= 0 ; m--) {
                	
                    LambdaLearnCell cell = this._cells[m];
                    if(!cell.getProd().getGP().isLxx()){
                    	continue;
                    }
                    
                    double subProb;
                    LambdaPattern p = LambdaPattern._X;
                    int cIndex = 0;
                    int cWord = CHILD_A_WORD_ID;
                    double prob = cell.getAlphaProb(bIndex, eIndex);

                    double prob_beta = cell.getBetaProb(bIndex, eIndex);

                    if (prob > 0 && prob_beta > 0) {
                        subProb = prob * this.computePatternProb(cell, p) *
                                  this.computeOneWordProb(cell, BEGIN_WORD_ID, cWord) *
                                  this.computeOneWordProb(cell, cWord, END_WORD_ID);

                        //update the rule and word counts.
                        if (_learnPattern || _learnEmission) {
                            double count = Math.log(subProb) + Math.log(cell.getBetaProb_X(cIndex, bIndex, eIndex)) - Math.log(this._globalBeta);
                            count = Math.exp(count);

                            if (count > 0) {
                                if (_learnPattern) {
                                    this.updatePatternProb(cell, p, count);
                                }
                                if (_learnEmission) {
                                    this.updateWordProb(cell, BEGIN_WORD_ID, cWord, count);
                                    this.updateWordProb(cell, cWord, END_WORD_ID, count);
                                }
                            }
                        }

                        //propagate
                        if (subProb > 0) {
                        	ArrayList<LambdaProd> children = cell.getProd().getFirstChildren();
                            for (int n = 0; n < children.size(); n++) {
                            	LambdaProd cProd = children.get(n);
                            	LambdaLearnCell child = cProd.getLearnCell();

                                double subAlpha = subProb * this.computeTransProb(cell.getProd().getGP(), cIndex, cProd.getGP());
                                child.addAlphaProb(bIndex, eIndex, subAlpha);

                                if (_learnTransition) {
                                    //update the transition probability
                                    double count = Math.log(subAlpha) +
                                            Math.log(child.getBetaProb(bIndex, eIndex)) -
                                            Math.log(this._globalBeta);
                                    count = Math.exp(count);

                                    this.updateTransProb(cell.getProd().getGP(), cIndex, cProd.getGP(), count);
                                }
                            }
                        }
                    }
                }
                
                //reverse the list
                for (int m = this._cells.length-1; m >= 0 ; m--) {

                    LambdaLearnCell cell = this._cells[m];
                    LambdaProd prod = cell.getProd();

                    //this trick will set the outside probability of a cell to zero if its inside is zero!
                    /*
                    if (cell.getBetaProb(bIndex, eIndex) == 0) {
                        double prob = cell.getAlphaProb(bIndex, eIndex);
                        cell.addAlphaProb(bIndex, eIndex, -prob);
                    }
                    */

                    if(cell.equals(this._rootProd.getLearnCell()) && len == this._N){
                    	continue;
                    }

                    int arity = prod.getArity();

                    //terminal rules...
                    if (arity == 0) {
                    	this.computeAlpha_terminal(cell, bIndex, eIndex);
                    }
                    
                    //unary rules...
                    else if (arity == 1) {
                    	this.computeAlpha_unary(cell, bIndex, eIndex);
                    }
                    
                    //binary rules...
                    else if (arity == 2) {
                    	this.computeAlpha_binary(cell, bIndex, eIndex);
                    }
                    
                    else {
                        System.out.println("Invalid number of children:" + cell.getProd().getGP().toString());
                        System.err.println("Currently only support at most two children.");
                        System.exit(1);
                    }
                    
//                    this.viewAlpha(cell, bIndex, eIndex);
                }
            }
        }

		
	}
	
	private int getChildWord(int pos) {
		return pos == 0 ? this.CHILD_A_WORD_ID : this.CHILD_B_WORD_ID;
	}
	
    /**
     * Methods for computing probabilities
     */
    private double computeTransProb(LambdaGProd m1, int pos, LambdaGProd m2) {
    	
    	/*
    	//if it is a prior instance..
    	if(this._inst.isPrior() && m1.isRoot()){
    		return 1.0;
    	}
    	*/
    	
    	double prob = m1.getMRTransProb_train(pos, m2);
    	return prob;
    }
    
    private double computePatternProb(LambdaLearnCell cell, LambdaPattern p) {
       	double prob = cell.getProd().getGP().getPatternProb_train(p);
       	return prob;
    }
    
    private double computeOneWordProb(LambdaLearnCell cell, int word1, int word) {

    	double prob;
    	if(this._gram == _UNIGRAM){
    		prob = cell.getProd().getGP().getUnigramProb_train(word);
    	} else {
    		prob = cell.getProd().getGP().getBigramProb_train(word1, word);
    	}
        return prob;
    }
    
    /**
     * Methods for updating probabilities
     */
    private void updatePatternProb(LambdaLearnCell cell, LambdaPattern p, double count) {
    	
        if (count > 0) {
        	cell.getProd().getGP().putPatternCount(p, count * this._weight);
        }
    }
    
    private void updateWordProb(LambdaLearnCell cell, int word1, int word2, double count) {
    	
        if (count > 0) {
        	if(this._gram == _UNIGRAM){
    			cell.getProd().getGP().putUnigramCount(word2, count*this._weight);
        	} else {
    			cell.getProd().getGP().putBigramCount(word1, word2, count*this._weight);
        	}
        }
    }
	

    private void updateTransProb(LambdaGProd m1, int pos, LambdaGProd m2, double count) {

        if (count > 0) {
        	m1.putMRTransCount(pos, m2.getId(), count * this._weight);
        }
    }
	
}
