/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.main;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;

import lambdacalc.logic.TypeEvaluationException;
import lambdagen.data.Instance;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.grammar.LambdaGRule;
import lambdagen.grammar.LambdaRuleFactory;
import lambdagen.learn.LambdaLearner;
import lambdagen.util.Config;
import lambdagen.viterbi.LambdaAlignSpan;
import lambdagen.viterbi.LambdaAligner;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaExtractor {

	private String _gram;
	private String _modeldir;
	private int _iterations;
	
	private boolean _debug;
	
	public static void main(String args[]) throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{

		if(args.length == 0){
			System.err.println("Usage: java LambdaExtractor <cfg_file>");
			System.exit(1);
		}
		
		Config cfg = new Config(args[0]);
		LambdaExtractor extractor = new LambdaExtractor(cfg);
		System.err.println(cfg);
		
		extractor.extract();
		
	}

	public LambdaExtractor(Config cfg){
		this._modeldir = cfg.readArgStr("modeldir", "");
		this._iterations = cfg.readArgInt("iterations", 100);
		this._gram = cfg.readArgStr("gram", "bigram");
		this._debug = cfg.readArgBool("debug", false);
	}

	public void extract() throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{
		
		long bTime = System.currentTimeMillis();
		
		String gmfacFile = _modeldir+"/lambda.gm."+_iterations+".fac";
		String vocFile = _modeldir+"/lambda."+_iterations+".voc";
		
		int K = 1;
		
		PrintStream ps = new PrintStream(System.out, true, "UTF-8");
		
		LambdaProdFactory gmfac = (LambdaProdFactory)(new ObjectInputStream(new FileInputStream(gmfacFile))).readObject();
		Vocabulary voc = (Vocabulary)(new ObjectInputStream(new FileInputStream(vocFile))).readObject();

		String instance_train_file = this._modeldir+"/train.instances";
		ArrayList<Instance> instances_train = LambdaReader.readInstances(instance_train_file);
		
		String instance_init_file = this._modeldir+"/init.instances";
		ArrayList<Instance> instances_init = LambdaReader.readInstances(instance_init_file);
		
		ArrayList<Instance> instances_train_init = new ArrayList<Instance>();
		for(int k=0; k<instances_train.size(); k++){
			instances_train_init.add(instances_train.get(k));
		}
		
		for(int k=0; k<instances_init.size(); k++){
			instances_train_init.add(instances_init.get(k));
		}
		
		ArrayList<Instance> instances = new ArrayList<Instance>();
		for(int i = 0; i<instances_train.size(); i++){
			instances.add(instances_train.get(i));
		}
		for(int i = 0; i<instances_init.size(); i++){
			instances.add(instances_init.get(i));
		}
		
		double failed = 0;
		double parsed = 0;
		
		short gram;
		
		if(_gram.equals("bigram")){
			gram = LambdaLearner._BIGRAM;
		} else if(_gram.equals("unigram")){
			gram = LambdaLearner._UNIGRAM;
		} else {
			throw new RuntimeException("Invalid gram:"+_gram);
		}
		
		LambdaRuleFactory rulefac = new LambdaRuleFactory(gmfac);
		
		for(int i = 0 ; i<instances.size(); i++){
			ps.print("\rExtracting :"+(i+1)+"/"+instances.size());
			
			try{
				Instance inst = instances.get(i);

				if(_debug){
					ps.println("[INST] "+i);
					ps.println("[NL] "+inst.getSent());
					LambdaProd prod = inst.getMRProd(gmfac);
					ps.println("[PRODUCTION]");
					ps.println(prod);
					ps.println("[EXPR]");
					ps.println(inst.getExpr());
				}
				
				LambdaAligner aligner = new LambdaAligner(gmfac, inst, gram, K, voc);
				LambdaAlignSpan[] spans = aligner.computeAlign();
				
				if(spans.length == 0 || spans[0].equals(LambdaAlignSpan._empty)){
				} else {
					parsed++;
				}
				for(int k = spans.length-1; k>=0; k--){
					if(_debug){
						ps.println(spans[k].toTree(voc, gmfac, inst));
					}
					if(spans[k].equals(LambdaAlignSpan._empty)){
						break;
					}
					rulefac.extractRules(spans[k], inst);
				}
			} catch(OutOfMemoryError error){
				failed ++;
			}
		}
		
		ps.println();
		
		//important...
		rulefac.finalize();
		
		if(_debug){
			ArrayList<LambdaGRule> rules = rulefac.getAllRules();
			HashMap<LambdaGRule, Double> ruleTable = rulefac.getCountTable();
			for(int i=0; i<rules.size(); i++){
				LambdaGRule rule = rules.get(i);
				double count = ruleTable.get(rule);
				ps.println(rule+"\t#times:"+count);
			}
		}
		
		long eTime = System.currentTimeMillis();
		
		String rulefacFile = _modeldir+"/lambda.rule."+_iterations+".fac";
		
		ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(new File(rulefacFile)));
		
		oos.writeObject(rulefac);
		oos.flush();
		oos.close();
		
		if(_debug)
		{
			ObjectInputStream ois = new ObjectInputStream(new FileInputStream(new File(rulefacFile)));
			rulefac = (LambdaRuleFactory)ois.readObject();
			ois.close();
		}
		
		
		String rulefacFile_txt = _modeldir+"/lambda.rule."+_iterations+".txt";
		FileOutputStream outfile = new FileOutputStream(rulefacFile_txt);
		BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(outfile , "UTF-8"));
		ArrayList<LambdaGRule> grules = rulefac.getAllRules();
		
		for(int i =0; i<grules.size(); i++){
			LambdaGRule grule = grules.get(i);
			writer.write(grule.toPhraseTableEntry());
			writer.write('\n');
			writer.flush();
		}
		writer.close();

		ps.println("Extraction done and the extracted rules are saved to: \nbinary\t:"+rulefacFile + "\ntext:\t"+rulefacFile_txt);

		ps.println("Total time: "+ (eTime-bTime)/1000.0 +" secs.");
		
	}
	
}
