/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.main;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Scanner;

import lambdacalc.logic.TypeEvaluationException;
import lambdagen.data.Instance;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.gen.LambdaGenSpan;
import lambdagen.gen.LambdaGenerator;
import lambdagen.gen.LambdaLM;
import lambdagen.gen.Weights;
import lambdagen.grammar.LambdaRuleFactory;
import lambdagen.util.Config;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaGen {

	private String _modeldir;
	private int _iterations;
	private short _n;
	private String _lm_file;
	
	private boolean _debug;
	
	public static void main(String args[]) throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{
		
		if(args.length < 4){
			System.err.println("Usage: java LambdaGen <cfg_file> <set:tune/test> <gen_cfg_file> <nbest_file>");
			System.exit(1);
		}
		
		short set;
		Config cfg = new Config(args[0]);
		if(args[1].equals("test")){
			set = LambdaReader._TEST;
		} else if(args[1].equals("tune")){
			set = LambdaReader._TUNE;
		} else if(args[1].equals("train")){
			set = LambdaReader._TRAIN;
		} else {
			throw new RuntimeException("Invalid option for set: "+args[4]);
		}
		String gen_cfg_file = args[2];
		String nbest_file = args[3];
		
		LambdaGen gener = new LambdaGen(cfg);
		gener.gen(nbest_file, gen_cfg_file, set);
		
	}
	
	public LambdaGen(Config cfg){
		this._modeldir = cfg.readArgStr("modeldir", "");
		this._n = (short)cfg.readArgInt("lmorder", 3);
		this._lm_file = cfg.readArgStr("lmfile", "");
		this._iterations = cfg.readArgInt("iterations", 100);
	}
	
	public void gen(String nbest_file, String cfg_file, short set) throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{
		
		PrintStream ps = new PrintStream(System.err, true, "UTF-8");
		
		String vocFile = _modeldir + "/lambda.voc";
		String rulefacFile = _modeldir + "/lambda.rule." +_iterations +".fac";
		
		Vocabulary voc = (Vocabulary)(new ObjectInputStream(new FileInputStream(vocFile))).readObject();
		LambdaRuleFactory rulefac = (LambdaRuleFactory)(new ObjectInputStream(new FileInputStream(rulefacFile))).readObject();
		LambdaProdFactory gmfac = rulefac.getProdFac();
		
		int K = 100; // default..
		
		double[] pms = new double[3];
		double lm_score = 1.0;
		
		Scanner cfg = new Scanner(new File(cfg_file));
		while(cfg.hasNextLine()){
			String line = cfg.nextLine();
			if(line.startsWith("top_n=")){
				K = Integer.parseInt(line.substring(6).trim());
			} else if(line.startsWith("RgivenE")){
				pms[0] = Double.parseDouble(line.split("\\s")[1]);
			} else if(line.startsWith("EgivenR")){
				pms[1] = Double.parseDouble(line.split("\\s")[1]);
			} else if(line.startsWith("WordP")){
				pms[2] = Double.parseDouble(line.split("\\s")[1]);
			} else if(line.startsWith("LM")){
				lm_score = Double.parseDouble(line.split("\\s")[1]);
			}
		}
		
		Weights weights = new Weights(pms, lm_score);
		LambdaLM lm = new LambdaLM(_lm_file, _n);
		
		String instance_file;
		if(set == LambdaReader._TUNE){
			instance_file = this._modeldir+"/tune.instances";
		} else if(set == LambdaReader._TEST){
			instance_file = this._modeldir+"/test.instances";
		} else if(set == LambdaReader._TRAIN){
			instance_file = this._modeldir+"/train.instances";
			System.err.println("[WARNING]\tYou are not supposed to decode the training set!");
		} else if(set == LambdaReader._INIT){
			instance_file = this._modeldir+"/init.instances";
			System.err.println("[WARNING]\tYou are not supposed to decode the init set!");
		} else {
			throw new RuntimeException("Invalid set! :" +set);
		}
		
		ArrayList<Instance> instances = LambdaReader.readInstances(instance_file);
		
		FileOutputStream outfile_nbest = new FileOutputStream(nbest_file);
		BufferedWriter writer_nbest = new BufferedWriter(new OutputStreamWriter(outfile_nbest, "UTF-8"));
		
		for(int i = 0; i<instances.size(); i++){
			ps.println("instance "+i);
			Instance inst = instances.get(i);
			ps.println("[exp]\t"+inst.getExpr());
			ps.println("[ref]\t"+inst.getSent());
			LambdaProd rootProd = inst.getMRProd(gmfac);
			
			if(rootProd != null){

				rootProd.labelSeqs();//label the seq numbers...
				
				boolean expand = true;
				if(expand){
					rulefac.expandRules(rootProd, inst);
					rootProd.labelSeqs();//label the seq numbers...
				}
				
				LambdaGenerator generator = new LambdaGenerator(rootProd, voc, rulefac, weights, lm, K, _n);
				LambdaGenSpan[] results = generator.computeGen();
				for(int k = 0; k< results.length; k++){
					LambdaGenSpan span = results[k];
					if(k==0){
						ps.println("[tst]\t"+span.toSentence(rulefac));
						if(_debug){
							ps.println(span.toTree(rulefac));
						}
					}
					if(!span.equals(LambdaGenSpan._empty)){
						writer_nbest.write(span.toNbestFormat(i, rulefac));
						writer_nbest.write('\n');
						writer_nbest.flush();
					} else if(k==0){
						writer_nbest.write(span.toNbestFormat(i, rulefac));
						writer_nbest.write('\n');
						writer_nbest.flush();
						
					}
				}
			} else {
				ps.println("[tst]\t"+LambdaGenSpan._empty.toSentence(rulefac));
				writer_nbest.write(LambdaGenSpan._empty.toNbestFormat(i, rulefac));
				writer_nbest.write('\n');
				writer_nbest.flush();
			}
		}
		writer_nbest.close();
	}
	
}
