/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.main;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.StringTokenizer;

import lambdacalc.logic.Expr;
import lambdacalc.logic.ExpressionParser;
import lambdacalc.logic.SyntaxException;
import lambdacalc.logic.TypeEvaluationException;
import lambdacalc.logic.ExpressionParser.ParseOptions;
import lambdagen.data.Instance;
import lambdagen.data.LambdaCalculusProcessor;
import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.data.Word;
import lambdagen.util.Config;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaLinearizer {
	
	private String _modeldir;
	private String _datadir;
	
	private static final short _TRAIN = 100;
	private static final short _TEST = 200;
	private static final short _INIT = 300;
	private static final short _TUNE = 400;
	
	private static boolean verbose = false;
	
	public static void main(String args[])throws IOException, TypeEvaluationException, ClassNotFoundException{
		
		if(args.length < 2){
			System.err.println("Usage: java LambdaLinearizer <config_file> <order>");
			System.exit(1);
		}
		
		Config cfg = new Config(args[0]);
		String order = args[1];
		if(order.equals("preorder") || order.equals("inorder") || order.equals("postorder") || order.equals("txt")){
		} else {
			throw new RuntimeException("invalid order : "+ order);
		}
		LambdaLinearizer linearizer = new LambdaLinearizer(cfg);
		linearizer.linearizd();
		
		if(order.equals("txt")){
			linearizer.writeLinearizedData_txt();
		} else {
			linearizer.writeLinearizedData(order);
		}
	}
	
	public LambdaLinearizer(Config cfg){
		this._modeldir = cfg.readArgStr("modeldir", "");
		this._datadir = cfg.readArgStr("datadir", "");
	}
	
	private ArrayList<String> toMRLTokens(String mrl){

		ArrayList<String> results = new ArrayList<String>();
		StringTokenizer st = new StringTokenizer(mrl,"() ");
		while(st.hasMoreTokens()){
			String token = st.nextToken();
			if(token.startsWith("$")){
				continue;
			} else if(token.equals("e")){
				continue;
			} else {
				results.add(token);
			}
		}
		
		return results;
	}
	
	private void writeLinearizedData_txt() throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{

		String order = "txt";
		
		String gmfacFile = _modeldir+"/lambda.gm.linearized.fac";
		String vocFile = _modeldir+"lambda.linearized.voc";
		
		LambdaProdFactory gmfac = (LambdaProdFactory)(new ObjectInputStream(new FileInputStream(gmfacFile))).readObject();
		Vocabulary voc = (Vocabulary)(new ObjectInputStream(new FileInputStream(vocFile))).readObject();
		
		ArrayList<Instance> instances_train = getGeoLambda_linearized(gmfac, voc, _TRAIN);
		ArrayList<Instance> instances_tune = getGeoLambda_linearized(gmfac, voc, _TUNE);
		ArrayList<Instance> instances_test = getGeoLambda_linearized(gmfac, voc, _TEST);
		ArrayList<Instance> instances_init = getGeoLambda_linearized(gmfac, voc, _INIT);
		
		BufferedWriter writer_src, writer_tgt;
		
		String train_src = _datadir+"/train.linearized."+order+".src";
		String train_tgt = _datadir+"/train.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(train_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(train_tgt), "UTF-8"));
		
		for(int i = 0; i<instances_train.size(); i++){
			System.err.println(">>> instance "+i);
			Instance inst = instances_train.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();
			
			String mrl = inst.getMRL();
			ArrayList<String> mrl_tokens = toMRLTokens(mrl);
			for(int k = 0; k<mrl_tokens.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				String token = mrl_tokens.get(k);
				writer_src.write(token);
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();

		String tune_src = _datadir+"/tune.linearized."+order+".src";
		String tune_tgt = _datadir+"/tune.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(tune_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(tune_tgt), "UTF-8"));

		for(int i = 0; i<instances_tune.size(); i++){
			System.err.println(">>> instance "+i);
			Instance inst = instances_tune.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();

			String mrl = inst.getMRL();
			ArrayList<String> mrl_tokens = toMRLTokens(mrl);
			for(int k = 0; k<mrl_tokens.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				String token = mrl_tokens.get(k);
				writer_src.write(token);
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();

		String test_src = _datadir+"/test.linearized."+order+".src";
		String test_tgt = _datadir+"/test.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(test_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(test_tgt), "UTF-8"));

		for(int i = 0; i<instances_test.size(); i++){
			System.err.println(">>> instance "+i);
			Instance inst = instances_test.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();

			String mrl = inst.getMRL();
			ArrayList<String> mrl_tokens = toMRLTokens(mrl);
			for(int k = 0; k<mrl_tokens.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				String token = mrl_tokens.get(k);
				writer_src.write(token);
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();

		String init_src = _datadir+"/init.linearized."+order+".src";
		String init_tgt = _datadir+"/init.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(init_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(init_tgt), "UTF-8"));

		for(int i = 0; i<instances_init.size(); i++){
			System.err.println(">>> instance "+i);
			Instance inst = instances_init.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();

			String mrl = inst.getMRL();
			ArrayList<String> mrl_tokens = toMRLTokens(mrl);
			for(int k = 0; k<mrl_tokens.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				String token = mrl_tokens.get(k);
				writer_src.write(token);
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();
		
		
	}
	
	private void writeLinearizedData(String order) throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{

		String gmfacFile = _modeldir+"/lambda.gm.linearized.fac";
		String vocFile = _modeldir+"lambda.linearized.voc";
		
		LambdaProdFactory gmfac = (LambdaProdFactory)(new ObjectInputStream(new FileInputStream(gmfacFile))).readObject();
		Vocabulary voc = (Vocabulary)(new ObjectInputStream(new FileInputStream(vocFile))).readObject();
		
		ArrayList<Instance> instances_train = getGeoLambda_linearized(gmfac, voc, _TRAIN);
		ArrayList<Instance> instances_tune = getGeoLambda_linearized(gmfac, voc, _TUNE);
		ArrayList<Instance> instances_test = getGeoLambda_linearized(gmfac, voc, _TEST);
		ArrayList<Instance> instances_init = getGeoLambda_linearized(gmfac, voc, _INIT);
		
		BufferedWriter writer_src, writer_tgt;
		
		String train_src = _datadir+"/train.linearized."+order+".src";
		String train_tgt = _datadir+"/train.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(train_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(train_tgt), "UTF-8"));
		
		for(int i = 0; i<instances_train.size(); i++){
			System.err.println(">>> instance "+i);
			Instance inst = instances_train.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();
			LambdaProd prod = inst.getMRProd_single(gmfac);
			
			ArrayList<LambdaProd> list = new ArrayList<LambdaProd>();
			if(order.equals("preorder")){
				prod.toPreOrder(list);
			} else if(order.equals("inorder")){
				prod.toInOrder(list);
			}  else if(order.equals("postorder")){
				prod.toPostOrder(list);
			}
			
			//check again...
			for(int k = 0; k<list.size(); k++){
				LambdaProd prod1 = list.get(k);
				for(int j = k +1 ; j<list.size(); j++){
					LambdaProd prod2 = list.get(j);
					if(prod1.equals(prod2)){
						System.err.println(prod.toString());
						throw new RuntimeException("Redandent at train: "+i);
					}
				}
			}
			for(int k = 0; k<list.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				LambdaProd cProd = list.get(k);
				writer_src.write("PROD"+cProd.getGP().getId());
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();

		String tune_src = _datadir+"/tune.linearized."+order+".src";
		String tune_tgt = _datadir+"/tune.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(tune_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(tune_tgt), "UTF-8"));

		for(int i = 0; i<instances_tune.size(); i++){
			Instance inst = instances_tune.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();
			LambdaProd prod = inst.getMRProd_single(gmfac);
			
			ArrayList<LambdaProd> list = new ArrayList<LambdaProd>();
			if(order.equals("preorder")){
				prod.toPreOrder(list);
			} else if(order.equals("inorder")){
				prod.toInOrder(list);
			}  else if(order.equals("postorder")){
				prod.toPostOrder(list);
			}
			
			//check again...
			for(int k = 0; k<list.size(); k++){
				LambdaProd prod1 = list.get(k);
				for(int j = k +1 ; j<list.size(); j++){
					LambdaProd prod2 = list.get(j);
					if(prod1.equals(prod2)){
						System.err.println(prod.toString());
						throw new RuntimeException("Redandent at tune: "+i);
					}
				}
			}
			for(int k = 0; k<list.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				LambdaProd cProd = list.get(k);
				writer_src.write("PROD"+cProd.getGP().getId());
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();

		String test_src = _datadir+"/test.linearized."+order+".src";
		String test_tgt = _datadir+"/test.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(test_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(test_tgt), "UTF-8"));

		for(int i = 0; i<instances_test.size(); i++){
			Instance inst = instances_test.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();
			LambdaProd prod = inst.getMRProd_single(gmfac);
			
			ArrayList<LambdaProd> list = new ArrayList<LambdaProd>();
			if(order.equals("preorder")){
				prod.toPreOrder(list);
			} else if(order.equals("inorder")){
				prod.toInOrder(list);
			}  else if(order.equals("postorder")){
				prod.toPostOrder(list);
			}
			
			//check again...
			for(int k = 0; k<list.size(); k++){
				LambdaProd prod1 = list.get(k);
				for(int j = k +1 ; j<list.size(); j++){
					LambdaProd prod2 = list.get(j);
					if(prod1.equals(prod2)){
						System.err.println(prod.toString());
						throw new RuntimeException("Redandent at test: "+i);
					}
				}
			}
			for(int k = 0; k<list.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				LambdaProd cProd = list.get(k);
				writer_src.write("PROD"+cProd.getGP().getId());
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();

		String init_src = _datadir+"/init.linearized."+order+".src";
		String init_tgt = _datadir+"/init.linearized."+order+".tgt";
		writer_src = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(init_src), "UTF-8"));
		writer_tgt = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(init_tgt), "UTF-8"));

		for(int i = 0; i<instances_init.size(); i++){
			Instance inst = instances_init.get(i);
			writer_tgt.write(inst.getSent());
			writer_tgt.write('\n');
			writer_tgt.flush();
			LambdaProd prod = inst.getMRProd_single(gmfac);
			
			ArrayList<LambdaProd> list = new ArrayList<LambdaProd>();
			if(order.equals("preorder")){
				prod.toPreOrder(list);
			} else if(order.equals("inorder")){
				prod.toInOrder(list);
			}  else if(order.equals("postorder")){
				prod.toPostOrder(list);
			}
			
			//check again...
			for(int k = 0; k<list.size(); k++){
				LambdaProd prod1 = list.get(k);
				for(int j = k +1 ; j<list.size(); j++){
					LambdaProd prod2 = list.get(j);
					if(prod1.equals(prod2)){
						System.err.println(prod.toString());
						throw new RuntimeException("Redandent at init: "+i);
					}
				}
			}
			for(int k = 0; k<list.size(); k++){
				if(k!= 0){
					writer_src.write(' ');
				}
				LambdaProd cProd = list.get(k);
				writer_src.write("PROD"+cProd.getGP().getId());
				writer_src.flush();
			}
			writer_src.write('\n');
			writer_src.flush();
		}
		writer_src.close();
		writer_tgt.close();
		
		
	}

	private void linearizd() throws IOException, TypeEvaluationException{
		
		String gmfacFile = _modeldir+"lambda.gm.linearized.fac";
		String vocFile = _modeldir+"lambda.linearized.voc";
		
		LambdaProdFactory fac = new LambdaProdFactory();
		
		processGeoLambda_linearized(fac, new Vocabulary(), 
				new ObjectOutputStream(new FileOutputStream(gmfacFile)), 
				new ObjectOutputStream(new FileOutputStream(vocFile)));
		
		ArrayList<LambdaGProd> gms = fac.getAllGProds();
		for(int i = 0; i<gms.size(); i++){
			System.err.println(gms.get(i));
		}
		
	}

	private void processGeoLambda_linearized(LambdaProdFactory fac, Vocabulary voc, ObjectOutputStream out_gmfac, ObjectOutputStream out_voc)throws IOException, TypeEvaluationException{
		
		PrintStream ps = new PrintStream(System.err, true, "UTF-8");
		
    	Scanner scan;
    	
    	int numInstances = 0;
    	
    	//training set...
    	{
        	scan = new Scanner(new File(_datadir+"/train.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line.trim();
        		String mrl = scan.nextLine();
        		numInstances ++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		ps.println();
        		ps.println("NL:\t"+nl);
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
                ps.println("[ORIGINAL]\t"+expr);
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR = fac.decomposeAsRoot_single(expr);
        		ps.println(rootMR);
        		System.err.println(rootMR.toString());
        	}
    	}
    	
    	//tune set...
    	{
        	scan = new Scanner(new File(_datadir+"/tune.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line;
        		String mrl = scan.nextLine();
        		numInstances++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		ps.println();
        		ps.println("NL:\t"+nl);
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
                ps.println("[ORIGINAL]\t"+expr);
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR = fac.decomposeAsRoot_single(expr);
        		ps.println(rootMR);            		
        		System.err.println(rootMR.toString());
        	}
    	}
    	
    	//test set...
    	{
        	scan = new Scanner(new File(_datadir+"/test.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line;
        		String mrl = scan.nextLine();
        		numInstances++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		ps.println();
        		ps.println("NL:\t"+nl);
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
                ps.println("[ORIGINAL]\t"+expr);
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR = fac.decomposeAsRoot_single(expr);
        		ps.println(rootMR);            		
        		System.err.println(rootMR.toString());
        	}
    	}
    	
    	//init set...
    	{
        	scan = new Scanner(new File(_datadir+"/init.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line;
        		String mrl = scan.nextLine();
        		numInstances++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		ps.println();
        		ps.println("NL:\t"+nl);
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
                ps.println("[ORIGINAL]\t"+expr);
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR = fac.decomposeAsRoot_prior(expr);
        		ps.println(rootMR);
        	}
    	}
    	
    	out_gmfac.writeObject(fac);
    	out_voc.writeObject(voc);
    	
    	ps.println("# instances = "+numInstances);
    	
    	ps.println("Data Processed OK.");
	}
	
	private ArrayList<Instance> getGeoLambda_linearized(LambdaProdFactory fac, Vocabulary voc, short set) throws UnsupportedEncodingException, FileNotFoundException, TypeEvaluationException{

		ArrayList<Instance> instances = new ArrayList<Instance>();
		
		PrintStream ps = new PrintStream(System.err, true, "UTF-8");
		
    	Scanner scan;
    	
    	int numInstances = 0;
    	
    	double weight = 1.0;
		boolean isPrior = false;
    	
    	//training set...
    	{
    		if(set == _TRAIN){
            	scan = new Scanner(new File(_datadir+"/train.corpus"), "UTF-8");
    		} else if(set == _TUNE){
            	scan = new Scanner(new File(_datadir+"/tune.corpus"), "UTF-8");
    		} else if(set == _TEST){
            	scan = new Scanner(new File(_datadir+"/test.corpus"), "UTF-8");
    		} else if(set == _INIT){
            	scan = new Scanner(new File(_datadir+"/init.corpus"), "UTF-8");
            	weight = 100.0;
            	isPrior = true;
    		} else {
    			throw new RuntimeException("Invalid set name : " + set);
    		}
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line.trim();
        		String mrl = scan.nextLine();
        		numInstances ++;
        		
        		String[] words = nl.split("\\s");
        		ArrayList<Word> wds = new ArrayList<Word>();
        		for(String word : words){
        			Word wd = voc.createWord(word);
        			wds.add(wd);
        		}
        		
        		if(verbose){
            		ps.println();
            		ps.println("NL:\t"+nl);
        		}
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
                if(verbose)
                	ps.println("[ORIGINAL]\t"+expr);
                
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR;

        		if(set == _INIT){
            		rootMR = fac.decomposeAsRoot_prior(expr);
        		} else {
            		rootMR = fac.decomposeAsRoot_single(expr);
            		System.err.println(rootMR);
        		}
        		
        		if(verbose)
        			ps.println(rootMR);
        		
        		Instance inst = new Instance(numInstances, weight, expr, wds, mrl, isPrior);
        		
        		instances.add(inst);
        	}
    	}
    	
    	if(verbose)
    		ps.println("# train instances: "+numInstances);
    	
    	return instances;
	}
}
