/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.main;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Scanner;

import lambdacalc.logic.Expr;
import lambdacalc.logic.ExpressionParser;
import lambdacalc.logic.SyntaxException;
import lambdacalc.logic.TypeEvaluationException;
import lambdacalc.logic.ExpressionParser.ParseOptions;
import lambdagen.data.Instance;
import lambdagen.data.LambdaCalculusProcessor;
import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.data.Word;
import lambdagen.util.Config;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaReader {
	
	public static final short _TRAIN = 1;
	public static final short _INIT = 2;
	public static final short _TUNE = 3;
	public static final short _TEST = 4;
	
	private String _modeldir;
	private String _datadir;
	
	private boolean _debug;
	
	public static void main(String args[]) throws FileNotFoundException, IOException, TypeEvaluationException{
		if(args.length==0){
			System.err.println("Usage: java LambdaReader <cfg_file>");
			System.exit(1);
		}
		
		Config cfg = new Config(args[0]);
		LambdaReader reader = new LambdaReader(cfg);
		System.err.println(cfg);
		
		reader.read();
	}
	
	public LambdaReader(Config cfg){
		this._modeldir = cfg.readArgStr("modeldir", "");
		this._datadir = cfg.readArgStr("datadir", "");
		this._debug = cfg.readArgBool("debug", false);
	}
	
	public void read() throws FileNotFoundException, IOException, TypeEvaluationException{
		
		long bTime = System.currentTimeMillis();
		
		File f = new File(this._modeldir);
		if(!f.exists()){
			f.mkdirs();
		}
		
		String gmfacFile = this._modeldir+"/lambda.gm.fac";
		String vocFile = this._modeldir+"/lambda.voc";
		
		LambdaProdFactory fac = new LambdaProdFactory();
		Vocabulary voc = new Vocabulary();
		
		this.processGeoLambda(fac, voc, 
				new ObjectOutputStream(new FileOutputStream(gmfacFile)), 
				new ObjectOutputStream(new FileOutputStream(vocFile)));
		
		//and save the instances into disk...
		String instance_train = this._modeldir+"/train.instances";
		ArrayList<Instance> inst_train = this.getGeoLambda(fac, voc, _TRAIN);
		writeInstances(inst_train, instance_train);
		
		String instance_tune = this._modeldir+"/tune.instances";
		ArrayList<Instance> inst_tune = this.getGeoLambda(fac, voc, _TUNE);
		writeInstances(inst_tune, instance_tune);
		
		String instance_test = this._modeldir+"/test.instances";
		ArrayList<Instance> inst_test = this.getGeoLambda(fac, voc, _TEST);
		writeInstances(inst_test, instance_test);
		
		String instance_init = this._modeldir+"/init.instances";
		ArrayList<Instance> inst_init = this.getGeoLambda(fac, voc, _INIT);
		writeInstances(inst_init, instance_init);
		
		long eTime = System.currentTimeMillis();
		
		System.err.println("Total time on data preparation: "+(eTime-bTime)/1000.0+" seconds.");
	}
	
	public static ArrayList<Instance> readInstances(String filename) throws FileNotFoundException, IOException, ClassNotFoundException{
		ArrayList<Instance> instances = new ArrayList<Instance>();
		ObjectInputStream in = new ObjectInputStream(new FileInputStream(filename));
		int size = in.readInt();
		for(int i = 0; i < size; i++){
			instances.add((Instance)in.readObject());
		}
		return instances;
	}
	
	public static void writeInstances(ArrayList<Instance> instances, String filename) throws FileNotFoundException, IOException{
		ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(filename));
		out.writeInt(instances.size());
		for(int i = 0; i<instances.size(); i++){
			out.writeObject(instances.get(i));
		}
		out.flush();
		out.close();
	}
	
	private void processGeoLambda(LambdaProdFactory fac, Vocabulary voc, ObjectOutputStream out_gmfac, ObjectOutputStream out_voc)throws IOException, TypeEvaluationException{
		
		PrintStream ps = new PrintStream(System.out, true, "UTF-8");
		
    	Scanner scan;
    	
    	int numInstances = 0;
    	
    	//training set...
    	{
        	scan = new Scanner(new File(this._datadir+"/train.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line.trim();
        		String mrl = scan.nextLine();
        		numInstances ++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		if(_debug){
            		ps.println();
            		ps.println("NL:\t"+nl);
        		}
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
        		if(_debug){
                    ps.println("[ORIGINAL]\t"+expr);
        		}
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR = fac.decomposeAsRoot(expr);
        		if(_debug){
        			ps.println(rootMR);
        			if(rootMR == null){
        			}
        		}
        	}
    	}
    	
    	//tune set...
    	{
    		scan = new Scanner(new File(this._datadir+"/tune.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line;
        		String mrl = scan.nextLine();
        		numInstances++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		if(_debug){
            		ps.println();
            		ps.println("NL:\t"+nl);
        		}
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
        		if(_debug){
        			ps.println("[ORIGINAL]\t"+expr);
        		}
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR = fac.decomposeAsRoot(expr);
        		if(_debug){
        			ps.println(rootMR);
        		}
        	}
    	}
    	
    	//test set...
    	{
    		scan = new Scanner(new File(this._datadir+"/test.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line;
        		String mrl = scan.nextLine();
        		numInstances++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		if(_debug){
        			ps.println();
        			ps.println("NL:\t"+nl);
        		}
        		
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
        		if(_debug){
        			ps.println("[ORIGINAL]\t"+expr);
        		}
        		
                expr = LambdaProdFactory.reGroup(expr);
        		LambdaProd rootMR = fac.decomposeAsRoot(expr);
        		
        		if(_debug){
        			ps.println(rootMR);
        		}
        	}
    	}
    	
    	//init set...
    	{
    		scan = new Scanner(new File(this._datadir+"/init.corpus"), "UTF-8");
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line;
        		String mrl = scan.nextLine();
        		numInstances++;
        		
        		String[] words = nl.split("\\s");
        		for(String word : words){
        			voc.createWord(word);
        		}
        		
        		if(_debug){
        			ps.println();
        			ps.println("NL:\t"+nl);
        		}
        		
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
        		if(_debug){
        			ps.println("[ORIGINAL]\t"+expr);
        		}
        		
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR = fac.decomposeAsRoot_prior(expr);
        		if(_debug){
        			ps.println(rootMR);
        		}
        	}
    	}
    	
    	out_gmfac.writeObject(fac);
    	out_voc.writeObject(voc);
    	
    	ps.println("# instances = "+numInstances);
    	
    	ps.println("Data Processed OK.");
    	
		if(_debug){
			ArrayList<LambdaGProd> gms = fac.getAllGProds();
			for(int i = 0; i<gms.size(); i++){
				ps.println(gms.get(i));
			}
		}

	}

	public ArrayList<Instance> getGeoLambda(LambdaProdFactory fac, Vocabulary voc, short set) throws UnsupportedEncodingException, FileNotFoundException, TypeEvaluationException{

		ArrayList<Instance> instances = new ArrayList<Instance>();
		
		PrintStream ps = new PrintStream(System.out, true, "UTF-8");
		
    	Scanner scan;
    	
    	int numInstances = 0;
    	
    	double weight = 1.0;
		boolean isPrior = false;
    	
    	{
    		if(set == _TRAIN){
            	scan = new Scanner(new File(this._datadir+"/train.corpus"), "UTF-8");
    		} else if(set == _TUNE){
            	scan = new Scanner(new File(this._datadir+"/tune.corpus"), "UTF-8");
    		} else if(set == _TEST){
            	scan = new Scanner(new File(this._datadir+"/test.corpus"), "UTF-8");
    		} else if(set == _INIT){
            	scan = new Scanner(new File(this._datadir+"/init.corpus"), "UTF-8");
            	weight = 100.0;
            	isPrior = true;
    		} else {
    			throw new RuntimeException("Invalid set name : " + set);
    		}
        	
        	while(scan.hasNextLine()){
        		String line = scan.nextLine();
        		if(line.equals("")){
        			continue;
        		}
        		if(line.startsWith("#") || line.startsWith("//")){
        			continue;
        		}
        		String nl = line.trim();
        		String mrl = scan.nextLine();
        		numInstances ++;
        		
        		String[] words = nl.split("\\s");
        		ArrayList<Word> wds = new ArrayList<Word>();
        		for(String word : words){
        			Word wd = voc.createWord(word);
        			wds.add(wd);
        		}
        		
        		if(_debug){
            		ps.println();
            		ps.println("NL:\t"+nl);
        		}
        		
        		String mrl_form = LambdaCalculusProcessor.preprocess(mrl);

                ParseOptions opts = new ParseOptions();
                opts.ASCII = true;
                
                Expr expr = null;
                try{
                	expr = ExpressionParser.parse(mrl_form, opts);
                	expr = expr.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
                } catch (SyntaxException se) {
                	System.err.println(se.getClass()+":"+se.getMessage());
                	System.exit(1);
                } catch (Exception e) {
                	System.err.println(e.getClass()+":"+e.getMessage());
                	System.exit(1);
                }
                
        		if(_debug){
                	ps.println("[ORIGINAL]\t"+expr);
        		}
                
                expr = LambdaProdFactory.reGroup(expr);
        		
        		LambdaProd rootMR;

        		if(set == _INIT){
            		rootMR = fac.decomposeAsRoot_prior(expr);
        		} else {
            		rootMR = fac.decomposeAsRoot(expr);
        		}
        		
        		if(_debug){
        			ps.println(rootMR);
        		}
        		
        		if(rootMR!=null || set!=_TRAIN){
            		Instance inst = new Instance(numInstances, weight, expr, wds, mrl, isPrior);
            		instances.add(inst);
        		}
        	}
    	}
    	
		if(_debug){
    		ps.println("# train instances: "+numInstances);
		}
    	
    	return instances;
    }
}
