/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.main;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import lambdacalc.logic.TypeEvaluationException;
import lambdagen.data.Instance;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.learn.LambdaInitLearner;
import lambdagen.learn.LambdaLearner;
import lambdagen.util.Config;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaTrainer {
	
	private String _gram;
	private String _modeldir;
	private int _iterations;
	private boolean _bayesian;
	
	private boolean _debug;
	
	public static void main(String args[]) throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{
		
		if(args.length == 0){
			System.err.println("Usage: java LambdaTrainer <config_file>");
			System.exit(1);
		}

		Config cfg = new Config(args[0]);
		LambdaTrainer trainer = new LambdaTrainer(cfg);
		System.err.println(cfg);
		
		trainer.train();
	}

	public LambdaTrainer(Config cfg){
		this._modeldir = cfg.readArgStr("modeldir", "");
		this._iterations = cfg.readArgInt("iterations", 100);
		this._gram = cfg.readArgStr("gram", "bigram");
		this._bayesian = cfg.readArgBool("bayesian", false);
		this._debug = cfg.readArgBool("debug", false);
	}
	
	public void train() throws FileNotFoundException, IOException, ClassNotFoundException, TypeEvaluationException{
		
		String gmfacFile = this._modeldir+"/lambda.gm.fac";
		String vocFile = this._modeldir+"/lambda.voc";
		
		PrintStream ps = new PrintStream(System.out, true, "UTF-8");
		
		LambdaProdFactory gmfac = (LambdaProdFactory)(new ObjectInputStream(new FileInputStream(gmfacFile))).readObject();
		Vocabulary voc = (Vocabulary)(new ObjectInputStream(new FileInputStream(vocFile))).readObject();
		
		String instance_train_file = this._modeldir+"/train.instances";
		ArrayList<Instance> instances_train = LambdaReader.readInstances(instance_train_file);
		
		String instance_init_file = this._modeldir+"/init.instances";
		ArrayList<Instance> instances_init = LambdaReader.readInstances(instance_init_file);
		
		ArrayList<Instance> instances_train_init = new ArrayList<Instance>();
		for(int k=0; k<instances_train.size(); k++){
			instances_train_init.add(instances_train.get(k));
		}
		
		for(int k=0; k<instances_init.size(); k++){
			instances_train_init.add(instances_init.get(k));
		}
		
		if(_debug){
			System.err.println("size of train : "+instances_train.size());
			System.err.println("size of init  : "+instances_init.size());
			System.err.println("size of total : "+instances_train_init.size());
		}
		
		//start with ibm model1
		
		LambdaInitLearner initLearner = new LambdaInitLearner(gmfac, voc, instances_train_init);
		initLearner.ibm1();
		
		short gram = LambdaLearner._UNIGRAM;
		
		int mid_k = 10;//initialize with 10 iterations of unigram EM first..
		
		if(_gram.equals("bigram") || _gram.equals("unigram")){
			//ok.
		} else {
			throw new RuntimeException("Invalid gram:"+_gram);
		}
		
		double logBeta_old = Double.NEGATIVE_INFINITY;
		
		System.err.println("Train with unigram assumption.");
		
		for(int k = 0; k< mid_k; k++){
			System.err.println("iteration "+k);
			
			double logBeta = 0;
			for(int i = 0; i<instances_train_init.size(); i++){
				Instance inst = instances_train_init.get(i);
				
				LambdaLearner learner = new LambdaLearner(gmfac, inst, gram);
				double beta = learner.EM_step();
				logBeta += Math.log(beta) * inst.getWeight();
			}
			
			double increment = (logBeta - logBeta_old);
			if(k > 1){
				ps.println("Total Log-likelihood = "+logBeta+"\t increment=\t"+ increment);
			}
			logBeta_old = logBeta;
			if(increment < 0 && k > 1){
				ps.println("Warning: loglikelihood decreased! (ignore this warning if bayesian option is on)");
			}
			
			if(_bayesian){
				System.err.println("Sorry this option is removed in this release.");
				System.exit(1);
			} else {
				gmfac.updateGMRs();
			}

		}
		
		gram = LambdaLearner._BIGRAM;
		
		System.err.println("Train with bigram assumption.");
		
		for(int k = 0; k< _iterations; k++){
			System.err.println("Iteration "+k);
			
			double logBeta = 0;
			for(int i = 0; i<instances_train_init.size(); i++){
				Instance inst = instances_train_init.get(i);
				
				LambdaLearner learner = new LambdaLearner(gmfac, inst, gram);
				double beta = learner.EM_step();
				logBeta += Math.log(beta) * inst.getWeight();
			}
			
			double increment = (logBeta - logBeta_old);
			if(k > 0){
				ps.println("Total Log-likelihood = "+logBeta+"\t increment=\t"+ increment);
			}
			logBeta_old = logBeta;
			if(increment < 0 && k > 1){
				ps.println("Warning: loglikelihood decreased! (ignore this warning if bayesian option is on)");
			}
			
			if(_bayesian){
				System.err.println("Sorry, this option is removed in this release.");
				System.exit(1);
			} else {
				gmfac.updateGMRs();
			}

		}
		
		if(_debug){
			//DEBUG purpose
			if(gram == LambdaLearner._UNIGRAM){
				gmfac.displayGMRs_unigram(voc);
			}
			if(gram == LambdaLearner._BIGRAM){
				gmfac.displayGMRs_bigram(voc);
			}
		}
		
		ObjectOutputStream out_gmfac = new ObjectOutputStream(new FileOutputStream(new File(_modeldir+"/lambda.gm."+_iterations+".fac")));
    	out_gmfac.writeObject(gmfac);
		out_gmfac.flush();
		out_gmfac.close();
		
		ObjectOutputStream out_voc = new ObjectOutputStream(new FileOutputStream(new File(_modeldir+"/lambda."+_iterations+".voc")));
		out_voc.writeObject(voc);
		out_voc.flush();
		out_voc.close();
		
	}
	
}
