/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.parse;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;

import lambdacalc.logic.Expr;
import lambdacalc.logic.FunApp;
import lambdacalc.logic.TypeEvaluationException;
import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaPattern;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.data.Word;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaParseSpan implements Comparable<LambdaParseSpan>, Serializable {
	
    private static final long serialVersionUID = 6058277521149744753L;
    
    private int _gid; // id of the LambdaSpanRule
    private int _key; // 1000* bIndex + eIndex;
    private double _prob;
    private LambdaParseSpan[] _children;
    private LambdaPattern _pattern;
    
    public LambdaPattern getPattern(){
    	return this._pattern;
    }
    
    public Expr toExpr(LambdaProdFactory fac) throws TypeEvaluationException{
    	if(this.equals(_empty)){
    		return null;
    	}
    	return this.toExpr_helper(fac);
    }
    
    private Expr toExpr_helper(LambdaProdFactory fac) throws TypeEvaluationException{
    	
    	LambdaGProd prod = fac.getAllGProds().get(this._gid);
    	
    	int arity = prod.getArity();
    	if(arity == 0){
        	Expr exp = prod.getExprToken().getExpr();
    		return exp;
    	} else if(arity == 1){
    		LambdaParseSpan child = this._children[0];
    		if(prod.isRoot()){
    			return child.toExpr_helper(fac);
    		}
    		
        	Expr exp = prod.getExprToken().getExpr();
    		Expr result = new FunApp(exp, child.toExpr_helper(fac)).simplifyFully().stripAnyDoubleParens().stripOutermostParens();
    		return result;
    	} else if(arity == 2){
    		LambdaParseSpan child1 = this._children[0];
    		LambdaParseSpan child2 = this._children[1];
    		if(this._pattern.isReversed()){
    			LambdaParseSpan child = child2;
    			child2 = child1;
    			child1 = child;
    		}
    		if(prod.getExprToken()==null){
        		Expr result = new FunApp(child1.toExpr_helper(fac), child2.toExpr_helper(fac))
        			.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
        		return result;
    		} else {
    			try{
            	Expr exp = prod.getExprToken().getExpr();
        		Expr result = new FunApp(new FunApp(exp, child1.toExpr_helper(fac)), child2.toExpr_helper(fac))
        			.simplifyFully().stripAnyDoubleParens().stripOutermostParens();
        		return result;
    			}catch(Exception e){
    				System.err.println(prod+"\t"+this.getBIndex());
    				System.err.println(fac.getAllGProds().get(child1._gid)+"\t"+child1.getBIndex());
    				System.err.println(fac.getAllGProds().get(child2._gid)+"\t"+child2.getBIndex());
    				System.exit(1);
    				return null;
    			}
    		}
    	} else {
    		return null;
    	}
    }
    
    public String toTree(Vocabulary voc, LambdaProdFactory fac, ArrayList<Word> words){
    	StringBuffer sb = new StringBuffer();
    	this.toTree_helper(voc, fac, words, sb, 0);
    	return sb.toString();
    }
    
    private void toTree_helper(Vocabulary voc, LambdaProdFactory fac, ArrayList<Word> words, StringBuffer sb, int indent){
    	StringBuffer prefix = new StringBuffer();
    	for(int i = 0; i<indent; i++){
			prefix.append('\t');
    	}
    	String prefix_str = prefix.toString();
    	
    	LambdaGProd prod = fac.getAllGProds().get(this._gid);//get the GM...
		sb.append(prefix_str);
    	sb.append(prod.toString());
    	sb.append('\t');
    	sb.append(this._pattern);
    	sb.append('\t');
    	sb.append(this._prob);
    	sb.append('\t');
    	sb.append('[');
    	sb.append(this.getBIndex());
    	sb.append(',');
    	sb.append(this.getEIndex());
    	sb.append(']');
    	sb.append('\n');
    	
    	int arity;
    	if(this._children == null){
    		for(int i = this.getBIndex(); i<this.getEIndex(); i++){
    			sb.append(prefix_str);
    			sb.append('\t');
    			sb.append(words.get(i).getWord());
    			sb.append('\n');
    		}
    		return;
    	} else {
    		arity = this._children.length;
    	}
    	
		LambdaParseSpan child1, child2;
    	
    	if(arity == 1){
    		child1 = this._children[0];
    		int index1 = this.getBIndex();
    		int index2 = child1.getBIndex();
    		int index3 = child1.getEIndex();
    		int index4 = this.getEIndex();
    		
    		for(int i = index1; i<index2; i++){
    			sb.append(prefix_str);
    			sb.append('\t');
    			sb.append(words.get(i).getWord());
    			sb.append('\n');
    		}
    		child1.toTree_helper(voc, fac, words, sb, indent+1);
    		for(int i = index3; i<index4; i++){
    			sb.append(prefix_str);
    			sb.append('\t');
    			sb.append(words.get(i).getWord());
    			sb.append('\n');
    		}
    		
    	} else if(arity == 2){
    		child1 = this._children[0];
    		child2 = this._children[1];
    		/*
    		if(this._pattern.isReversed()){
    			LambdaParseSpan child = child1;
    			child1 = child2;
    			child2 = child;
    		}
    		*/
    		
    		int index1 = this.getBIndex();
    		int index2 = child1.getBIndex();
    		int index3 = child1.getEIndex();
    		int index4 = child2.getBIndex();
    		int index5 = child2.getEIndex();
    		int index6 = this.getEIndex();

    		for(int i = index1; i<index2; i++){
    			sb.append(prefix_str);
    			sb.append('\t');
    			sb.append(words.get(i).getWord());
    			sb.append('\n');
    		}
    		child1.toTree_helper(voc, fac, words, sb, indent+1);
    		for(int i = index3; i<index4; i++){
    			sb.append(prefix_str);
    			sb.append('\t');
    			sb.append(words.get(i).getWord());
    			sb.append('\n');
    		}
    		child2.toTree_helper(voc, fac, words, sb, indent+1);
    		for(int i = index5; i<index6; i++){
    			sb.append(prefix_str);
    			sb.append('\t');
    			sb.append(words.get(i).getWord());
    			sb.append('\n');
    		}
    		
    	}
    	
    }
    
    /**
     * Private methods...
     * @throws IOException 
     */
    private void writeObject(ObjectOutputStream out) throws IOException{
    	out.writeInt(this._gid);
    	out.writeInt(this._key);
    	out.writeDouble(this._prob);
    	out.writeObject(this._pattern);
    }
    
    private void readObject(ObjectInputStream in)throws IOException, ClassNotFoundException{
    	this._gid = in.readInt();
    	this._key = in.readInt();
    	this._prob = in.readDouble();
    	this._pattern = (LambdaPattern)in.readObject();
    }
    
    public static final LambdaParseSpan _empty = new LambdaParseSpan(Integer.MIN_VALUE, Integer.MIN_VALUE, Integer.MIN_VALUE, null, 0);
    
    public LambdaParseSpan(int gid, int bIndex, int eIndex, LambdaPattern pattern, double prob) {
    	this._gid = gid;
    	this._key = 1000 * bIndex + eIndex;
    	this._pattern = pattern;
        this._prob = prob;
    }
    
    @Override
	public int compareTo(LambdaParseSpan entry) {
    	
        if (this._prob > entry._prob) {
            return +1;
        } else if (this._prob < entry._prob) {
            return -1;
        } else {
        	if(this.getBIndex() > entry.getBIndex()){
        		return +1;
        	} else if(this.getBIndex() < entry.getBIndex()){
        		return -1;
        	} else {
        		if(this._gid > entry._gid){
        			return +1;
        		} else if(this._gid < entry._gid){
        			return -1;
        		} else {
//        			throw new RuntimeException("Identical!!"+this._key+";"+this._prob+";"+this._gid);
        			return 0;
        		}
        	}
        }
    }
    
    public int getBIndex(){
    	return this._key / 1000;
    }
    
    public int getEIndex(){
    	return this._key % 1000;
    }
    
    public double getProb() {
        return this._prob;
    }
    
    public LambdaParseSpan getChild() {
        if (this._children != null) {
            return this._children[0];
        }
        return _empty;
    }
    
    public void setChildren(LambdaParseSpan[] children) {
        this._children = children;
    }
    
    public LambdaParseSpan[] getChildren() {
        return this._children;
    }
    
}