/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.parse;

import java.util.ArrayList;
import java.util.Collections;
import lambdagen.data.Instance;
import lambdagen.data.LambdaGProd;
import lambdagen.data.LambdaPattern;
import lambdagen.data.LambdaProd;
import lambdagen.data.LambdaProdFactory;
import lambdagen.data.Vocabulary;
import lambdagen.data.Word;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaParser{
	
	public static final short _UNIGRAM = 10;
	public static final short _BIGRAM = 20;
	
	private int _gram = _UNIGRAM;
	
	private int[] _words;
	
	private LambdaGProd[] _gms;
	
	private int _N;
	
	private int _K;
	
	private Vocabulary _vocabulary;
    
    private LambdaParseCell[] _cells; // note that these cells must be sorted
    
    private LambdaParseCell[] _rootCells; // the root cell
    
    private int _BEGIN_WORD_ID;
    private int _END_WORD_ID;
    private int _CHILD_A_WORD_ID;
    private int _CHILD_B_WORD_ID;
    private int _UNK_WORD_ID;
    
    private boolean _debug;
    
    private boolean checkRelevance(LambdaGProd gm){
    	
    	if(gm.isRoot() || gm.getExprToken()==null || gm.isLxx()){
    		return true;
    	}
    	
    	int word_1 = Word._BEGIN.getId();
    	int word;
    	
    	for(int i = 0; i<this._words.length; i++){
    		word = this._words[i];
    		
    		boolean flag = false;
    		if(this._gram == _BIGRAM){
        		if(gm.getBigramProb_train(word_1, word)>0){
        			return true;
        		}
        		if(gm.getBigramProb_train(Word._BEGIN.getId(), word)>0){
        			return true;
        		}
        		if(gm.getBigramProb_train(Word._X1.getId(), word)>0){
        			return true;
        		}
        		if(gm.getBigramProb_train(Word._X2.getId(), word)>0){
        			return true;
        		}
    		} else {
        		if(gm.getUnigramProb_train(word)>0){
        	    	//gm.viewSrcUnigram(_vocabulary);
        	    	//System.exit(1);
        			return true;
        		}
    		}
    		word_1 = word;
    	}
    	
    	return false;
    }

    //FOR DEBUGGING PURPOSE...
    public LambdaParser(ArrayList<LambdaProd> gms, Instance inst, int gram, int K, Vocabulary vocabulary) {
		this._words = new int[inst.getWords().size()];
		for(int i = 0; i< this._words.length; i++){
			this._words[i] = inst.getWords().get(i).getId();
		}
        this._N = _words.length;
        this._K = K;
        this._gram = gram;
        this._vocabulary = vocabulary;
        
        ArrayList<LambdaGProd> allgms = new ArrayList<LambdaGProd>();
        for(int i = 0; i<gms.size(); i++){
        	LambdaGProd gm = gms.get(i).getGP();
        	if(!allgms.contains(gm)){
            	allgms.add(gm);
        	}
        }
        
        ArrayList<LambdaGProd> relevantgms = new ArrayList<LambdaGProd>();
        for(LambdaGProd gm : allgms){
        	//if(this.checkRelevance(gm)){
        		relevantgms.add(gm);
        	//}
        }
        this._gms = new LambdaGProd[relevantgms.size()];
        int i =0;
        for(LambdaGProd gm : relevantgms){
        	this._gms[i++] = gm;
        }
        
//        Logger.alert("Selected "+this._gms.length+" from "+allgms.size()+" gms.");
        this.init();
    }
    
    public LambdaParser(LambdaProdFactory fac, Instance inst, int gram, int K, Vocabulary vocabulary) {
		this._words = new int[inst.getWords().size()];
		for(int i = 0; i< this._words.length; i++){
			this._words[i] = inst.getWords().get(i).getId();
		}
        this._N = _words.length;
        this._K = K;
        this._gram = gram;
        this._vocabulary = vocabulary;
        
        ArrayList<LambdaGProd> allgms = fac.getAllGProds();
        ArrayList<LambdaGProd> relevantgms = allgms;
        
        /*
        ArrayList<LambdaGMRProd> relevantgms = new ArrayList<LambdaGMRProd>();
        for(LambdaGMRProd gm : allgms){
        	//if(this.checkRelevance(gm)){
        		relevantgms.add(gm);
        	//}
        }
        */
        this._gms = new LambdaGProd[relevantgms.size()];
        int i =0;
        for(LambdaGProd gm : relevantgms){
        	this._gms[i++] = gm;
        }
        
        this.init();
    }

    private void init() {
        ArrayList<LambdaParseCell> rootCells = new ArrayList<LambdaParseCell>();
        
        this._cells = new LambdaParseCell[this._gms.length];
        for (int i = 0; i < this._gms.length; i++) {
            LambdaGProd gm = this._gms[i];
            this._cells[i] = new LambdaParseCell(gm, this._N, this._K);
        	if(gm.isRoot()){
        		rootCells.add(this._cells[i]);
        	}
        }
        this._rootCells = new LambdaParseCell[rootCells.size()];
        for(int i = 0; i<this._rootCells.length; i++){
        	this._rootCells[i] = rootCells.get(i);
        }
        
        this._BEGIN_WORD_ID = Word._BEGIN.getId();
        this._END_WORD_ID = Word._END.getId();
        this._CHILD_A_WORD_ID = Word._X1.getId();
        this._CHILD_B_WORD_ID = Word._X2.getId();
        this._UNK_WORD_ID = Word._UNK.getId();
    }

    private void computeParse_terminal(LambdaParseCell cell, int bIndex, int eIndex){
    	
    	int len = eIndex - bIndex;
    	
        /** compute the beta_w **/
        {
            double wProb;
            if (len == 1) {
                wProb = 1.0;
            } else {
                int word1 = this._words[eIndex - 2];
                int word = this._words[eIndex - 1];
                wProb = cell.getMaxProb_p(LambdaPattern._w, bIndex, eIndex - 1)[0].getProb() *
                        this.computeOneWordProb(cell, word1, word);
            }
            if(wProb > 0){
            	cell.setMaxProb_p(LambdaPattern._w, bIndex, eIndex, wProb, null);
            }
        }
        
        //case : w
        int bWord = this._words[bIndex];
        int eWord = this._words[eIndex - 1];
        double prob = this.computePatternProb(cell, LambdaPattern._w) *
                      this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                      cell.getMaxProb_p(LambdaPattern._w, bIndex, eIndex)[0].getProb() * this.computeOneWordProb(cell, eWord, _END_WORD_ID);
        
        if(prob >0){
        	cell.setMaxProb(bIndex, eIndex, prob, null, LambdaPattern._w);
        }
    }
    
    private void computeParse_unary(LambdaParseCell cell, int bIndex, int eIndex){
    	
        double prob = 0;
        
        int len = eIndex - bIndex;
        
        int cWord = _CHILD_A_WORD_ID;
        
        //case of TWO parts
        if (len >= 2) {
            double subProb;
            
            // wX = w + ( X | wX )
            {
                int bWord = this._words[bIndex];
                LambdaParseSpan[] c_entries;
                
                // w | X
                {
                    c_entries = cell.getMaxProb_p(LambdaPattern._X, bIndex + 1, eIndex);
                    int word_1 = cWord;

                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        subProb = this.computeOneWordProb(cell, bWord, word_1) * c_entry.getProb();
                        if (!cell.setMaxProb_p(LambdaPattern._wX, bIndex, eIndex, subProb, c_entry.getChild())) {
                            break;
                        }
                    }
                }

                // w | wX
                {
                    c_entries = cell.getMaxProb_p(LambdaPattern._wX, bIndex + 1, eIndex);
                    int word_1 = this._words[bIndex + 1];

                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        subProb = this.computeOneWordProb(cell, bWord, word_1) * c_entry.getProb();
                        if (!cell.setMaxProb_p(LambdaPattern._wX, bIndex, eIndex, subProb, c_entry.getChild())) {
                            break;
                        }
                    }
                }
            }

            // Xw = ( X | Xw ) + w
            {
            	int eWord = this._words[eIndex - 1];

                LambdaParseSpan[] c_entries;

                // X | w
                {
                    c_entries = cell.getMaxProb_p(LambdaPattern._X, bIndex, eIndex - 1);
                    int word1 = cWord;

                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        subProb = c_entry.getProb() * this.computeOneWordProb(cell, word1, eWord);
                        if (!cell.setMaxProb_p(LambdaPattern._Xw, bIndex, eIndex, subProb, c_entry.getChild())) {
                            break;
                        }
                    }
                }

                // Xw | w
                {
                    c_entries = cell.getMaxProb_p(LambdaPattern._Xw, bIndex, eIndex - 1);
                    int word1 = this._words[eIndex - 2];

                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        subProb = c_entry.getProb() * this.computeOneWordProb(cell, word1, eWord);
                        if (!cell.setMaxProb_p(LambdaPattern._Xw, bIndex, eIndex, subProb, c_entry.getChild())) {
                            break;
                        }
                    }
                }
            }
        }

        //case of THREE parts
        if (len >= 3) {
            double subProb;

            // wXw = ( wX | wXw ) + w
            {
            	int eWord = this._words[eIndex - 1];
                LambdaParseSpan[] c_entries;

                // wX | w
                {
                    c_entries = cell.getMaxProb_p(LambdaPattern._wX, bIndex, eIndex - 1);
                    int word1 = cWord;

                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        subProb = this.computeOneWordProb(cell, word1, eWord) * c_entry.getProb();
                        if (!cell.setMaxProb_p(LambdaPattern._wXw, bIndex, eIndex, subProb, c_entry.getChild())) {
                            break;
                        }
                    }
                }

                // wXw | w
                {
                    c_entries = cell.getMaxProb_p(LambdaPattern._wXw, bIndex, eIndex - 1);
                    int word1 = this._words[eIndex - 2];

                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        subProb = this.computeOneWordProb(cell, word1, eWord) * c_entry.getProb();
                        if (!cell.setMaxProb_p(LambdaPattern._wXw, bIndex, eIndex, subProb, c_entry.getChild())) {
                            break;
                        }
                    }
                }
            }
        }

        //add global max values
        LambdaPattern p;
        double wProb;
        LambdaParseSpan[] c_entries;

        int bWord = this._words[bIndex];
        int eWord = this._words[eIndex - 1];

        //wX
        if (len >= 2) {

        	p = LambdaPattern._wX;
            c_entries = cell.getMaxProb_p(LambdaPattern._wX, bIndex, eIndex);
            
            wProb = this.computePatternProb(cell, p) * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                    this.computeOneWordProb(cell, cWord, _END_WORD_ID);


            for (int k = c_entries.length - 1; k >= 0; k--) {
                LambdaParseSpan c_entry = c_entries[k];
                if (c_entry.getProb() <= 0) {
                    break;
                }
                prob = wProb * c_entry.getProb();
                if (!cell.setMaxProb(bIndex, eIndex, prob, new LambdaParseSpan[] {c_entry.getChild()}, p)) {
                    break;
                }
            }
        }

        //Xw
        if (len >= 2) {

            p = LambdaPattern._Xw;
            c_entries = cell.getMaxProb_p(LambdaPattern._Xw, bIndex, eIndex);

            wProb = this.computePatternProb(cell, p) * this.computeOneWordProb(cell, _BEGIN_WORD_ID, cWord) *
                    this.computeOneWordProb(cell, eWord, _END_WORD_ID);

            for (int k = c_entries.length - 1; k >= 0; k--) {
                LambdaParseSpan c_entry = c_entries[k];
                if (c_entry.getProb() <= 0) {
                    break;
                }
                prob = wProb * c_entry.getProb();
                if (!cell.setMaxProb(bIndex, eIndex, prob, new LambdaParseSpan[] {c_entry.getChild()}, p)) {
                    break;
                }
            }
        }

        //wXw
        if (len >= 3) {
            p = LambdaPattern._wXw;
            c_entries = cell.getMaxProb_p(LambdaPattern._wXw, bIndex, eIndex);

            wProb = this.computePatternProb(cell, p) * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                    this.computeOneWordProb(cell, eWord, _END_WORD_ID);

            for (int k = c_entries.length - 1; k >= 0; k--) {
                LambdaParseSpan c_entry = c_entries[k];
                if (c_entry.getProb() <= 0) {
                    break;
                }
                prob = wProb * c_entry.getProb();
                if (!cell.setMaxProb(bIndex, eIndex, prob, new LambdaParseSpan[] {c_entry.getChild()}, p)) {
                    break;
                }
            }
        }
        
    }
    
    private void computeParse_binary(LambdaParseCell cell, int bIndex, int eIndex){

    	int len = eIndex - bIndex;
    	
        double subProb;

        if (len >= 2) {
            subProb = 0;
            LambdaPattern p;

            //case : X | Y / Y | X
            for (int pos = 0; pos <= 1; pos++) {
                p = pos == 0 ? LambdaPattern._XY : LambdaPattern._YX;
                double prob_pattern = this.computePatternProb(cell, p);
                
                if (prob_pattern > 0) {
                    LambdaPattern[] ps = LambdaPattern.split2(p);
                    
                    int bWord = this.getChildWord(pos);
                    int eWord = this.getChildWord(1 - pos);
                    
                    double wProb = prob_pattern *
                            this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                            this.computeOneWordProb(cell, bWord, eWord) *
                            this.computeOneWordProb(cell, eWord, _END_WORD_ID);
                    for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {

                        LambdaParseSpan[] c_entries1;
                        LambdaParseSpan[] c_entries2;

                    	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                        c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);

                        LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);

                        LambdaParseSpan[] c_entries;
                        while ((c_entries = cube.getNext()) != null) {
                            subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;
                            if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                    c_entries[1].getChild()}, p)) {
                                break;
                            }
                        }
                    }
                }
            }
        }

        {
            //case of THREE parts..
            if (len >= 3) {
                subProb = 0;
                LambdaPattern p;

                for (int pos = 0; pos <= 1; pos++) {
                    p = pos == 0 ? LambdaPattern._XwY : LambdaPattern._YwX;
                    double prob_pattern = this.computePatternProb(cell, p);
                    
                    if (prob_pattern > 0) {
                        LambdaPattern[] ps = LambdaPattern.split2(p);
                        
                        int bWord = this.getChildWord(pos);
                        int eWord = this.getChildWord(1 - pos);

                        double tProb = prob_pattern *
                                this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                                this.computeOneWordProb(cell, eWord, _END_WORD_ID);
                        
                        //case : Xw | Y / Yw | X
                        for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                            int word = this._words[sIndex];
                            
                            LambdaParseSpan[] c_entries1;
                            LambdaParseSpan[] c_entries2;
                            
                        	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                        	c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);
                        	
                            double wProb = this.computeOneWordProb(cell, bWord, word) * tProb;

                            /*
                            if(cell._gm.getId()==2){
                        		System.err.println("777777:"+wProb+":"+bIndex+","+eIndex+","+sIndex);
                        		for(int t = 0; t<c_entries1.length; t++){
                        			System.err.println(c_entries1[t].getProb()+"\t"+ps[0]);
                        		}
                        	}
                        	*/
                            
                            LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);
                            
                            LambdaParseSpan[] c_entries;
                            while ((c_entries = cube.getNext()) != null) {
//                            	System.err.println("888888");
                                subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;

                                if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                        c_entries[1].getChild()}, p)) {
                                    break;
                                }
                            }
                        }
                    }
                }

                //case : X | Yw / Y | Xw
                for (int pos = 0; pos <= 1; pos++) {
                    p = pos == 0 ? LambdaPattern._XYw : LambdaPattern._YXw;
                    double prob_pattern = this.computePatternProb(cell, p);

                    if (prob_pattern > 0) {
                    	
                        LambdaPattern[] ps = LambdaPattern.split2(p);
                        
                    	int bWord = this.getChildWord(pos);
                    	int word = this.getChildWord(1 - pos);
                        int eWord = this._words[eIndex - 1];

                        double wProb = prob_pattern * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                                this.computeOneWordProb(cell, bWord, word) * this.computeOneWordProb(cell, eWord, _END_WORD_ID);

                        if (wProb > 0) {
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {

                                LambdaParseSpan[] c_entries1;
                                LambdaParseSpan[] c_entries2;
                                
                            	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                            	c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);

                                LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);

                                LambdaParseSpan[] c_entries;
                                while ((c_entries = cube.getNext()) != null) {
                                    subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;
                                    if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                            c_entries[1].getChild()}, p)) {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }

                //case : wX | Y / wY | X
                for (int pos = 0; pos <= 1; pos++) {
                    p = pos == 0 ? LambdaPattern._wXY : LambdaPattern._wYX;
                    
                    double prob_pattern = this.computePatternProb(cell, p);
                    
                    if (prob_pattern > 0) {
                        LambdaPattern[] ps = LambdaPattern.split2(p);
                        
                        int bWord = this._words[bIndex];
                        int word = this.getChildWord(pos);
                        int eWord = this.getChildWord(1 - pos);
                        
                        double wProb = prob_pattern * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                                this.computeOneWordProb(cell, word, eWord) * this.computeOneWordProb(cell, eWord, _END_WORD_ID);
                        
                        if (wProb > 0) {
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {

                                LambdaParseSpan[] c_entries1;
                                LambdaParseSpan[] c_entries2;
                                
                            	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                            	c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);

                                LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);

                                LambdaParseSpan[] c_entries;

                                while ((c_entries = cube.getNext()) != null) {
                                    subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;
                                    if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                            c_entries[1].getChild()}, p)) {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }

            //case of FOUR parts
            if (len >= 4) {
                LambdaPattern p;
                //case : X w1 | Y w2 / Y w1 | X w2
                subProb = 0;

                for (int pos = 0; pos <= 1; pos++) {
                    int bWord = this.getChildWord(pos); 
                    int word2 = this.getChildWord(1-pos);
                    int eWord = this._words[eIndex - 1];
                    p = pos == 0 ? LambdaPattern._XwYw : LambdaPattern._YwXw;

                    double prob_pattern = this.computePatternProb(cell, p);

                    if (prob_pattern > 0) {
                        LambdaPattern[] ps = LambdaPattern.split2(p);
                        
                        double tProb = prob_pattern * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                                this.computeOneWordProb(cell, eWord, _END_WORD_ID);

                        if (tProb > 0) {
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                            	int word1 = this._words[sIndex - 1];

                                LambdaParseSpan[] c_entries1;
                                LambdaParseSpan[] c_entries2;
                                
                            	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                            	c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);

                                double wProb = tProb * this.computeOneWordProb(cell, word1, word2);

                                LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);

                                LambdaParseSpan[] c_entries;
                                while ((c_entries = cube.getNext()) != null) {
                                    subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;
                                    if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                            c_entries[1].getChild()}, p)) {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }

                //case : w1 X | w2 Y / w1 Y | w2 X
                for (int pos = 0; pos <= 1; pos++) {
                    p = pos == 0 ? LambdaPattern._wXwY : LambdaPattern._wYwX;

                    double prob_pattern = this.computePatternProb(cell, p);

                    if (prob_pattern > 0) {
                        LambdaPattern[] ps = LambdaPattern.split2(p);
                        
                        int bWord = this._words[bIndex];
                        int word1 = this.getChildWord(pos);
                        int eWord = this.getChildWord(1 - pos);
                        double tProb = prob_pattern * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                                this.computeOneWordProb(cell, eWord, _END_WORD_ID);

                        if (tProb > 0) {
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {

                            	int word2 = this._words[sIndex];

                                LambdaParseSpan[] c_entries1;
                                LambdaParseSpan[] c_entries2;
                                
                            	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                            	c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);

                                double wProb = tProb * this.computeOneWordProb(cell, word1, word2);

                                LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);

                                LambdaParseSpan[] c_entries;
                                while ((c_entries = cube.getNext()) != null) {
                                    subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;
                                    if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                            c_entries[1].getChild()}, p)) {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }

                //case : w1 X | Y w2 / w1 Y | X w2
                for (int pos = 0; pos <= 1; pos++) {
                    p = pos == 0 ? LambdaPattern._wXYw : LambdaPattern._wYXw;

                    double prob_pattern = this.computePatternProb(cell, p);

                    if (prob_pattern > 0) {
                        LambdaPattern[] ps = LambdaPattern.split2(p);
                        
                    	int word1 = this.getChildWord(pos);
                    	int word2 = this.getChildWord(1 - pos);
                        int bWord = this._words[bIndex];
                        int eWord = this._words[eIndex - 1];
                        double wProb = prob_pattern * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                                this.computeOneWordProb(cell, word1, word2) * this.computeOneWordProb(cell, eWord, _END_WORD_ID);

                        if (wProb > 0) {
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {

                                LambdaParseSpan[] c_entries1;
                                LambdaParseSpan[] c_entries2;
                                
                            	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                            	c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);

                                LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);

                                LambdaParseSpan[] c_entries;
                                while ((c_entries = cube.getNext()) != null) {
                                    subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;
                                    if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                            c_entries[1].getChild()}, p)) {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }

            //case of FIVE parts
            if (len >= 5) {
                subProb = 0;
                LambdaPattern p;

                //case : w1 X w2 | Y w3 / w1 Y w2 | X w3
                for (int pos = 0; pos <= 1; pos++) {
                	int word1 = this.getChildWord(pos);
                    int bWord = this._words[bIndex];
                    int eWord = this._words[eIndex - 1];
                    p = pos == 0 ? LambdaPattern._wXwYw : LambdaPattern._wYwXw;

                    double prob_pattern = this.computePatternProb(cell, p);

                    if (prob_pattern > 0) {
                        LambdaPattern[] ps = LambdaPattern.split2(p);
                        
                        double tProb = prob_pattern * this.computeOneWordProb(cell, _BEGIN_WORD_ID, bWord) *
                                this.computeOneWordProb(cell, eWord, _END_WORD_ID);

                        if (tProb > 0) {
                            for (int sIndex = bIndex + 3; sIndex <= eIndex - 2; sIndex++) {

                            	int word2 = this._words[sIndex];

                                LambdaParseSpan[] c_entries1;
                                LambdaParseSpan[] c_entries2;
                                
                            	c_entries1 = cell.getMaxProb_p(ps[0], bIndex, sIndex);
                            	c_entries2 = cell.getMaxProb_p(ps[1], sIndex, eIndex);

                                double wProb = tProb * this.computeOneWordProb(cell, word1, word2);

                                LambdaParseSquare cube = new LambdaParseSquare(c_entries1, c_entries2);

                                LambdaParseSpan[] c_entries;
                                while ((c_entries = cube.getNext()) != null) {
                                    subProb = c_entries[0].getProb() * c_entries[1].getProb() * wProb;
                                    if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaParseSpan[] {c_entries[0].getChild(),
                                            c_entries[1].getChild()}, p)) {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    private void computeParse_postprocess(int bIndex, int eIndex){
    	
    	int len = eIndex - bIndex;

        for (int m = 0; m < this._cells.length; m++) {
            LambdaParseCell cell = this._cells[m];
            LambdaGProd gm = cell._gm;
            int arity = gm.getArity();

            if (arity == 0){
            	continue;
            }

            // computing the beta_X probabilities for this window size "len"
            else if (arity == 1) {

                //case : X
                double subProb = 0;
                for (int n = 0; n < this._cells.length; n++) {
                    LambdaParseCell c_cell = this._cells[n];
                    if (c_cell._gm.isRoot()) {
                        continue;
                    }

                    double tProb = this.computeTransProb(cell._gm, 0, c_cell._gm);

                    if (tProb <= 0) {
                        continue;
                    }
                    
                    LambdaParseSpan[] c_entries = c_cell.getMaxSpans(bIndex, eIndex);
                    
                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        subProb = tProb * c_entry.getProb();


                        if (!cell.setMaxProb_p(LambdaPattern._X, bIndex, eIndex, subProb, c_entry)) {
                            break;
                        }
                        
                        /** special treatment for the rule Lx.x **/
                        if (cell._gm.isLxx()) {
                            LambdaPattern p = LambdaPattern._X;

                            double prob = this.computePatternProb(cell, p) * subProb *
                                    this.computeOneWordProb(cell, _BEGIN_WORD_ID, _CHILD_A_WORD_ID) *
                                    this.computeOneWordProb(cell, _CHILD_A_WORD_ID, _END_WORD_ID);

                            if (!cell.setMaxProb(bIndex, eIndex, prob, new LambdaParseSpan[] {c_entry}, p)) {
                                break;
                            }
                        }
                    }

                }



            }

            else if (arity == 2) {

                for (int pos = 0; pos < arity; pos++) {

                	int cWord = this.getChildWord(pos);

                    //case of ONE part
                    {
                        //case : X
                        double subProb = 0;
                        for (int n = 0; n < this._cells.length; n++) {
                            LambdaParseCell c_cell = this._cells[n];

                            double tProb = this.computeTransProb(cell._gm, pos, c_cell._gm);
                            if (tProb <= 0) {
                                continue;
                            }

                            LambdaParseSpan[] c_entries = c_cell.getMaxSpans(bIndex, eIndex);

                            for (int k = c_entries.length - 1; k >= 0; k--) {
                                LambdaParseSpan c_entry = c_entries[k];
                                if (c_entry.getProb() <= 0) {
                                    break;
                                }
                                subProb = tProb * c_entry.getProb();

                                if(pos == 0){
                                	if (!cell.setMaxProb_p(LambdaPattern._X, bIndex, eIndex, subProb, c_entry)) {
                                		break;
                                	}
                                }else{
                                	if (!cell.setMaxProb_p(LambdaPattern._Y, bIndex, eIndex, subProb, c_entry)) {
                                		break;
                                	}
                                }
                            }
                        }
                        
                        /*
                        if(cell._gm.getId()==2){
                    		System.err.println("66666:"+subProb+":"+bIndex+","+eIndex);
                    	}
                    	*/
                    }

                    //case of TWO parts
                    if (len >= 2) {
                        double subProb = 0;

                        // wX = w + ( X | wX )
                        {
                            int bWord = this._words[bIndex];
                            LambdaParseSpan[] c_entries;

                            // w | X
                            {
                            	int word_1 = cWord;
                            	if(pos == 0){
                            		c_entries = cell.getMaxProb_p(LambdaPattern._X, bIndex + 1, eIndex);
                            	} else {
                            		c_entries = cell.getMaxProb_p(LambdaPattern._Y, bIndex + 1, eIndex);
                            	}
                                double tProb = this.computeOneWordProb(cell, bWord, word_1);

                                if (tProb > 0) {
                                    for (int k = c_entries.length - 1; k >= 0; k--) {
                                        LambdaParseSpan c_entry = c_entries[k];
                                        if (c_entry.getProb() <= 0) {
                                            break;
                                        }
                                        subProb = tProb * c_entry.getProb();

                                        if(pos == 0){
                                            if (!cell.setMaxProb_p(LambdaPattern._wX, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        }else{
                                            if (!cell.setMaxProb_p(LambdaPattern._wY, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        }
                                    }
                                }
                            }

                            // w | wX
                            {
                            	int word_1 = this._words[bIndex + 1];
                            	if(pos == 0){
                            		c_entries = cell.getMaxProb_p(LambdaPattern._wX, bIndex + 1, eIndex);
                            	} else {
                            		c_entries = cell.getMaxProb_p(LambdaPattern._wY, bIndex + 1, eIndex);
                            	}
                                double tProb = this.computeOneWordProb(cell, bWord, word_1);

                                if (tProb > 0) {
                                    for (int k = c_entries.length - 1; k >= 0; k--) {
                                        LambdaParseSpan c_entry = c_entries[k];
                                        if (c_entry.getProb() <= 0) {
                                            break;
                                        }
                                        subProb = tProb * c_entry.getProb();

                                        if(pos == 0){
                                            if (!cell.setMaxProb_p(LambdaPattern._wX, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        } else {
                                            if (!cell.setMaxProb_p(LambdaPattern._wY, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        // Xw = ( X | Xw ) + w
                        {
                            int eWord = this._words[eIndex - 1];
                            LambdaParseSpan[] c_entries;

                            // X | w
                            {
                            	int word1 = cWord;
                            	if(pos == 0){
                            		c_entries = cell.getMaxProb_p(LambdaPattern._X, bIndex, eIndex - 1);
                            	} else {
                            		c_entries = cell.getMaxProb_p(LambdaPattern._Y, bIndex, eIndex - 1);
                            	}
                                double tProb = this.computeOneWordProb(cell, word1, eWord);

                                if (tProb > 0) {
                                    for (int k = c_entries.length - 1; k >= 0; k--) {
                                        LambdaParseSpan c_entry = c_entries[k];
                                        if (c_entry.getProb() <= 0) {
                                            break;
                                        }
                                        subProb = c_entry.getProb() * tProb;

                                        if(pos == 0){
                                            if (!cell.setMaxProb_p(LambdaPattern._Xw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        } else {
//                                        	System.err.println("sssss"+bIndex+","+eIndex);
                                            if (!cell.setMaxProb_p(LambdaPattern._Yw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        }
                                    }
                                }
                            }

                            // Xw | w
                            {
                            	int word1 = this._words[eIndex - 2];
                            	if(pos == 0){
                            		c_entries = cell.getMaxProb_p(LambdaPattern._Xw, bIndex, eIndex - 1);
                            	} else {
                            		c_entries = cell.getMaxProb_p(LambdaPattern._Yw, bIndex, eIndex - 1);
                            	}
                                double tProb = this.computeOneWordProb(cell, word1, eWord);

                                if (tProb > 0) {
                                    for (int k = c_entries.length - 1; k >= 0; k--) {
                                        LambdaParseSpan c_entry = c_entries[k];
                                        if (c_entry.getProb() <= 0) {
                                            break;
                                        }
                                        subProb = c_entry.getProb() * tProb;

                                        if(pos == 0){
                                            if (!cell.setMaxProb_p(LambdaPattern._Xw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        } else {
                                            if (!cell.setMaxProb_p(LambdaPattern._Yw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                                break;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    //case of THREE parts
                    if (len >= 3) {
                        double subProb = 0;

                        //case : wXw = ( wX | wXw ) + w
                        int eWord = this._words[eIndex - 1];
                        LambdaParseSpan[] c_entries;

                        // wX | w
                        {
                        	int word1 = cWord;
                        	if(pos == 0){
                        		c_entries = cell.getMaxProb_p(LambdaPattern._wX, bIndex, eIndex - 1);
                        	} else {
                        		c_entries = cell.getMaxProb_p(LambdaPattern._wY, bIndex, eIndex - 1);
                        	}
                            double tProb = this.computeOneWordProb(cell, word1, eWord);

                            if (tProb > 0) {
                                for (int k = c_entries.length - 1; k >= 0; k--) {
                                    LambdaParseSpan c_entry = c_entries[k];
                                    if (c_entry.getProb() <= 0) {
                                        break;
                                    }
                                    subProb = c_entry.getProb() * tProb;

                                    if(pos == 0){
                                        if (!cell.setMaxProb_p(LambdaPattern._wXw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                            break;
                                        }
                                    } else {
                                        if (!cell.setMaxProb_p(LambdaPattern._wYw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                            break;
                                        }                                                	
                                    }
                                }
                            }
                        }

                        // wXw | w
                        {
                            int word1 = this._words[eIndex - 2];
                            if(pos == 0){
                            	c_entries = cell.getMaxProb_p(LambdaPattern._wXw, bIndex, eIndex - 1);
                            } else {
                            	c_entries = cell.getMaxProb_p(LambdaPattern._wYw, bIndex, eIndex - 1);
                            }
                            double tProb = this.computeOneWordProb(cell, word1, eWord);

                            if (tProb > 0) {
                                for (int k = c_entries.length - 1; k >= 0; k--) {
                                    LambdaParseSpan c_entry = c_entries[k];
                                    if (c_entry.getProb() <= 0) {
                                        break;
                                    }

                                    subProb = c_entry.getProb() * tProb;

                                    if(pos == 0){
                                        if (!cell.setMaxProb_p(LambdaPattern._wXw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                            break;
                                        }
                                    } else {
                                        if (!cell.setMaxProb_p(LambdaPattern._wYw, bIndex, eIndex, subProb, c_entry.getChild())) {
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    private LambdaParseSpan[] computeParse_root(){
    	
        ArrayList<LambdaParseSpan> rootSpans = new ArrayList<LambdaParseSpan>();
        
        for(int i = 0; i<this._rootCells.length; i++){
            LambdaParseCell rootCell = this._rootCells[i];
            
            //this.computeParse_unary(rootCell, 0, this._N);
            
            //working on the final root LambdaGMRProd
            int pos = 0;
            LambdaPattern p;

            for (int m = 0; m < this._cells.length; m++) {
                LambdaParseCell cell = this._cells[m];

                {
                    p = LambdaPattern._X;
                    
                    double prob_trans = this.computeTransProb(rootCell._gm, pos, cell._gm);
                    
                    if (prob_trans <= 0) {
                        continue;
                    }
                    
                    LambdaParseSpan[] c_entries = cell.getMaxSpans(0, this._N);
                    
                    double subProb = prob_trans * this.computePatternProb(rootCell, p) *
                                     this.computeOneWordProb(rootCell, _BEGIN_WORD_ID, _CHILD_A_WORD_ID) *
                                     this.computeOneWordProb(rootCell, _CHILD_A_WORD_ID, _END_WORD_ID);
                    
                    for (int k = c_entries.length - 1; k >= 0; k--) {
                        LambdaParseSpan c_entry = c_entries[k];
                        if (c_entry.getProb() <= 0) {
                            break;
                        }
                        double prob = subProb * c_entry.getProb();
                        
                        if (!rootCell.setMaxProb(0, this._N, prob, new LambdaParseSpan[] {c_entry}, p)) {
                            break;
                        }
                    }
                }
                
                
            }
            
            LambdaParseSpan[] subRootSpans = rootCell.getMaxSpans(0, this._N);
            for (int k = subRootSpans.length - 1; k >= 0; k--) {
            	LambdaParseSpan rootSpan = subRootSpans[k];
            	rootSpans.add(rootSpan);
            }
        }
        
        Collections.sort(rootSpans);
        
        LambdaParseSpan[] results = new LambdaParseSpan[this._K];
        for(int i = 0; i<this._K; i++){
        	if(i >= rootSpans.size()){
        		results[i] = LambdaParseSpan._empty;
        	} else {
        		results[i] = rootSpans.get(rootSpans.size()-1-i);
        	}
        }
        
        return results;
    }
    
    public LambdaParseSpan[] computeParse() {
    	
        for (int len = 1; len <= this._N; len++) {
        	
            for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                int eIndex = bIndex + len;
                
                /* for each possible LambdaGMRProd */
                for (int m = 0; m < this._cells.length; m++) {
                	
                    LambdaParseCell cell = this._cells[m];
                    int arity = cell._gm.getArity();
                    
                    /* terminal rule */
                    if (arity == 0) {
                    	this.computeParse_terminal(cell, bIndex, eIndex);
                    }
                    
                    /* unary rule */
                    else if (arity == 1) {
                    	this.computeParse_unary(cell, bIndex, eIndex);
                    }
                    
                    /* binary rule */
                    else if (arity == 2) {
                    	this.computeParse_binary(cell, bIndex, eIndex);
                    }
                    
                    else {
                        throw new RuntimeException("Invalid number of RHS! Currently only supports up to 2 RHS for GMRTokens!");
                    }
                    
                }
                
                /* post-processing starts */
                this.computeParse_postprocess(bIndex, eIndex);
                
            }
        }
        
        return this.computeParse_root();
        
    }
    
    private double computeOneWordProb(LambdaParseCell cell, int word1_id, int word_id) {
    	double prob ;
    	if(this._gram == _UNIGRAM){
    		prob = cell._gm.getUnigramProb_test(word_id);
    	} else {
//    		prob = cell._gm.getBigramProb_test(word1_id, word_id);
    		prob = ( cell._gm.getUnigramProb_test(word_id) + 100* cell._gm.getBigramProb_test(word1_id, word_id) ) / 101.0;
    	}
    	return Math.pow(prob, _weight_emission);
//    	return prob;
    }

    private double computeTransProb(LambdaGProd m1, int pos, LambdaGProd m2) {
    	double prob = m1.getMRTransProb_test(pos, m2);
    	return Math.pow(prob, _weight_trans);
//    	return prob;
    }
    
    private double computePatternProb(LambdaParseCell cell, LambdaPattern p) {
    	double prob = cell._gm.getPatternProb_test(p);
    	return Math.pow(prob, _weight_pattern);
//    	return prob;
    }
    
    private int getChildWord(int pos) {
        return pos == 0 ? _CHILD_A_WORD_ID : _CHILD_B_WORD_ID;
    }
    
    public static double _weight_trans = 1.0;
    public static double _weight_pattern = 1.0;
    public static double _weight_emission = 1.0;
    
}
