/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.phrase;


import java.util.*;
import lambdagen.data.LambdaPattern;
import lambdagen.data.Vocabulary;
import lambdagen.data.Word;
import lambdagen.grammar.LambdaGRule;
import lambdagen.grammar.LambdaRuleFactory;
import lambdagen.phrase.LambdaPhraseSpan;

/**
 * @author Wei LU
 * @version 1.0
 */

public class LambdaPhraser{

	private ArrayList<Word> _words;
	
	private ArrayList<LambdaPhraseCell> _cells;
	private ArrayList<LambdaPhraseCell> _rootCells;
	private SpanTable _spanTable;
	
	private int _N;//the number of words.
	private int _K;//top K
	private ArrayList<LambdaGRule> _rules;//the list of rules
	private Vocabulary _vocabulary;
	
	private static boolean debug = false;
	
	public LambdaPhraser(ArrayList<Word> words, LambdaRuleFactory fac, SpanTable spanTable, Vocabulary vocabulary, int K){
		
		this._words = words;
		
		this._N = words.size();
		this._K = K;
		this._rules = fac.getAllRules();
		this._spanTable = spanTable;
		
		this._cells = new ArrayList<LambdaPhraseCell>(this._rules.size());
		for(int i =0; i<this._rules.size(); i++){
			LambdaGRule rule = this._rules.get(i);
			LambdaPhraseCell cell = new LambdaPhraseCell(rule, this._N, this._K);
			
			this._cells.add(cell);
			if(rule.isRoot()){
				this._rootCells.add(cell);
			}
		}
		this._vocabulary = vocabulary;
	}
	
	private void handleMaxPartial(LambdaPhraseCell cell, LambdaPattern p, int bIndex, int eIndex, int wPos){
		LambdaPattern ps[] = LambdaPattern.split2(p);
		LambdaGRule rule = cell._rule;
		
		for(int cIndex = bIndex+1; cIndex<=eIndex-1; cIndex++){
			LambdaPhraseSpan[] spans1, spans2;

			int k = -1;
			if(ps[0] == LambdaPattern._w){
				spans1 = this.getWordSpans(rule, bIndex, cIndex, wPos);
				if(spans1 != null){
					spans2 = cell.getMaxProb_p(ps[1], cIndex, eIndex);
				} else {
					continue;
				}
				k = 1;
			} else if(ps[1] == LambdaPattern._w){
				spans2 = this.getWordSpans(rule, cIndex, eIndex, wPos);
				if(spans2 != null){
					spans1 = cell.getMaxProb_p(ps[0], bIndex, cIndex);
				} else {
					continue;
				}
				k = 0;
			} else{
				throw new RuntimeException("This should not happen!"+p);
			}
			
			LambdaPhraseSquare square = new LambdaPhraseSquare(spans1, spans2);
			
			LambdaPhraseSpan[] c_spans;
            while ((c_spans = square.getNext()) != null) {
                double subProb = c_spans[0].getProb() * c_spans[1].getProb();
                if (!cell.setMaxProb_p(p, bIndex, eIndex, subProb, c_spans[k].getChild())) {
                    break;
                }
            }
		}
	}
	
	private void handleMaxComplete1(LambdaPhraseCell cell, LambdaPattern p, int bIndex, int eIndex, int wPos){
		LambdaPattern ps[] = LambdaPattern.split2(p);
		LambdaGRule rule = cell._rule;
		
		for(int cIndex = bIndex+1; cIndex<=eIndex-1; cIndex++){
			LambdaPhraseSpan[] spans1;
			LambdaPhraseSpan[] spans2;

			int k = -1;
			if(ps[0] == LambdaPattern._w){
				spans1 = this.getWordSpans(rule, bIndex, cIndex, wPos);
				if(spans1 != null){
					spans2 = cell.getMaxProb_p(ps[1], cIndex, eIndex);
				} else {
					continue;
				}
				k = 1;
			} else if(ps[1] == LambdaPattern._w){
				spans2 = this.getWordSpans(rule, cIndex, eIndex, wPos);
				if(spans2 != null){
					spans1 = cell.getMaxProb_p(ps[0], bIndex, cIndex);
				} else {
					continue;
				}
				k = 0;
			} else{
				throw new RuntimeException("This should not happen!"+p);
			}
			
			LambdaPhraseSquare square = new LambdaPhraseSquare(spans1, spans2);
			
			LambdaPhraseSpan[] c_spans;
            while ((c_spans = square.getNext()) != null) {
                double subProb = c_spans[0].getProb() * c_spans[1].getProb();
                if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaPhraseSpan[] {c_spans[k].getChild()}, p)) {
                    break;
                }
            }
		}
	}
	
	private void handleMaxComplete2(LambdaPhraseCell cell, LambdaPattern p, int bIndex, int eIndex, int wPos){
		LambdaPattern ps[] = LambdaPattern.split2(p);
		
		for(int cIndex = bIndex+1; cIndex<=eIndex-1; cIndex++){
			LambdaPhraseSpan[] spans1 = cell.getMaxProb_p(ps[0], bIndex, cIndex);
			LambdaPhraseSpan[] spans2 = cell.getMaxProb_p(ps[1], cIndex, eIndex);
			
			LambdaPhraseSquare square = new LambdaPhraseSquare(spans1, spans2);
			
			LambdaPhraseSpan[] c_spans;
            while ((c_spans = square.getNext()) != null) {
                double subProb = c_spans[0].getProb() * c_spans[1].getProb();
                if (!cell.setMaxProb(bIndex, eIndex, subProb, new LambdaPhraseSpan[] {c_spans[0].getChild(), c_spans[1].getChild()}, p)) {
                    break;
                }
            }
		}
	}
	
	private double getSpanProb(LambdaGRule rule, int bIndex, int eIndex, int wPos){
		return this._spanTable.getProb(rule, bIndex, eIndex, wPos);
	}
	
	private LambdaPhraseSpan[] getWordSpans(LambdaGRule rule, int bIndex, int eIndex, int wPos){
		double prob = this.getSpanProb(rule, bIndex, eIndex, wPos);
		if(prob <= 0){
			return null;
		}
		LambdaPhraseSpan span = new LambdaPhraseSpan(rule.getId(), bIndex, eIndex, prob);
		return new LambdaPhraseSpan[] {span};
	}
	
	public void computeParse(){
		
		for(int len = 1; len <= this._N; len++ ){
			for(int bIndex = 0; bIndex<= this._N -len; bIndex++){
				
				int eIndex = bIndex + len;
				//the span is [bIndex, eIndex)
				for(int pIndex = 0; pIndex < this._cells.size(); pIndex++){
					LambdaPhraseCell cell = this._cells.get(pIndex);
					LambdaGRule rule = cell._rule;
					
					if(rule.isRoot()){
						continue;
					}
					
					int arity = rule.getArity();
					LambdaPattern p = rule.getPattern();
					
					if(arity == 0){
						double prob = this.getSpanProb(rule, bIndex, eIndex, 0);
						if(prob > 0){
							cell.setMaxProb(bIndex, eIndex, prob, null, p);
						}
					}
					
					else if(arity == 1){
						
						if(p == LambdaPattern._wX){
							// wX = w | X
							this.handleMaxComplete1(cell, p, bIndex, eIndex, 0);
						} else if(p == LambdaPattern._Xw){
							// Xw = X | w
							this.handleMaxComplete1(cell, p, bIndex, eIndex, 1);
						} else if(p == LambdaPattern._wXw){
							// wXw = w | Xw
							LambdaPattern[] ps = LambdaPattern.split2(p);
							this.handleMaxComplete1(cell, p, bIndex, eIndex, 0);
							
							// Xw = X | w
							this.handleMaxPartial(cell, ps[1], bIndex, eIndex, 1);
						}
						
						//TODO seems there is something missing here...
						
					}
					
					else if(arity == 2){
						
						LambdaPattern[] ps;
						
						// wXY
						// wYX
						if(p == LambdaPattern._wXY || p == LambdaPattern._wYX ){
							// wXY = wX | Y
							// wYX = wY | X
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
							// wX = w | X
							// wY = w | Y
							this.handleMaxPartial(cell, ps[0], bIndex, eIndex, 0);
						}
						
						// XwY
						// YwX
						else if(p == LambdaPattern._XwY || p == LambdaPattern._YwX){
							// XwY = Xw | Y
							// YwX = Yw | X
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
							// Xw = X | w
							// Yw = Y | w
							this.handleMaxPartial(cell, ps[0], bIndex, eIndex, 1);
						}
						
						// XYw
						// YXw
						else if(p == LambdaPattern._XYw || p == LambdaPattern._YXw){
							// XYw = X | Yw
							// YXw = Y | Xw
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
							// Yw = Y | w
							// Xw = X | w
							this.handleMaxPartial(cell, ps[1], bIndex, eIndex, 2);
						}
						
						// wXwY
						// wYwX
						else if(p == LambdaPattern._wXwY || p == LambdaPattern._wYwX){
							// wXwY = wX | wY
							// wYwX = wY | wX
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
							// wX = w | X
							// wY = w | Y
							this.handleMaxPartial(cell, ps[0], bIndex, eIndex, 0);
							// wY = w | Y
							// wX = w | X
							this.handleMaxPartial(cell, ps[1], bIndex, eIndex, 1);
						}
						
						// wXYw
						// wYXw
						else if(p == LambdaPattern._wXYw || p == LambdaPattern._wYXw){
							// wXYw = wX | Yw
							// wYXw = wY | Xw
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
							// wX = w | X
							// wY = w | Y
							this.handleMaxPartial(cell, ps[0], bIndex, eIndex, 0);
							// Yw = Y | w
							// Xw = X | w
							this.handleMaxPartial(cell, ps[1], bIndex, eIndex, 2);
						}
						
						// XwYw
						// YwXw
						else if(p == LambdaPattern._XwYw || p == LambdaPattern._YwXw){
							// XwYw = wX | Yw
							// YwXw = wY | Xw
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
							// Xw = X | w
							// Yw = Y | w
							this.handleMaxPartial(cell, ps[0], bIndex, eIndex, 1);
							// Yw = Y | w
							// Xw = X | w
							this.handleMaxPartial(cell, ps[1], bIndex, eIndex, 2);
						}
						
						else if(p == LambdaPattern._wXwYw || p == LambdaPattern._wYwXw){
							// wXwYw = wXw | Yw
							// wYwXw = wYw | Xw
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
							// wXw = w | Xw
							// wYw = w | Yw
							this.handleMaxPartial(cell, ps[0], bIndex, eIndex, 0);
							// Yw = Y | w
							// Xw = X | w
							this.handleMaxPartial(cell, ps[1], bIndex, eIndex, 2);
							// Yw = Y | w
							// Xw = X | w
							this.handleMaxPartial(cell, LambdaPattern.split2(ps[0])[1], bIndex, eIndex, 1);
						}
						
						else if(p == LambdaPattern._XY || p == LambdaPattern._YX){
							// XY = X | Y
							// YX = Y | X
							ps = LambdaPattern.split2(p);
							this.handleMaxComplete2(cell, p, bIndex, eIndex, -1);
						}
						
						else{
							//TODO .. seems there is something missing here.
//							throw new RuntimeException("LambdaGrammarRule "+gm.toForm(this._vocabulary)+" should have a fixed LambdaPattern!");
						}
					}
				}
				
				//processing the X/Y spans
				for(int pIndex = 0; pIndex< this._cells.size(); pIndex++){
					LambdaPhraseCell cell = this._cells.get(pIndex);
					LambdaGRule gm = cell._rule;
					if(gm.isRoot()){
						continue;
					}
					
					int arity = gm.getArity();
					
					if(arity >= 1){
						// X
						{
							for(LambdaPhraseCell cell_child : this._cells){
								LambdaPhraseSpan[] c_spans = cell_child.getMaxSpans(bIndex, eIndex);
								
								double tProb = gm.getMRTransProb_test(0, cell_child._rule);
								if(tProb > 0){
									for(int k = c_spans.length -1; k >= 0; k--){
										double subProb = tProb * c_spans[k].getProb();
										if(subProb > 0){
											if(!cell.setMaxProb_p(LambdaPattern._X ,bIndex, eIndex, subProb, c_spans[k])){
												break;
											}
										} else {
											break;
										}
									}
								}
							}
						}
						
						if(arity == 2){
							// Y
							{
								for(LambdaPhraseCell cell_child : this._cells){
									LambdaPhraseSpan[] c_spans = cell_child.getMaxSpans(bIndex, eIndex);
									
									double tProb = gm.getMRTransProb_test(1, cell_child._rule);
									if(tProb > 0){
										for(int k = c_spans.length -1; k >= 0; k--){
											double subProb = tProb * c_spans[k].getProb();
											if(subProb > 0){
												if(!cell.setMaxProb_p(LambdaPattern._Y ,bIndex, eIndex, subProb, c_spans[k])){
													break;
												}
											} else {
												break;
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		
		for(int n = 0; n<this._rootCells.size(); n++){
        	LambdaPhraseCell rootCell = this._rootCells.get(n);
        	LambdaGRule rootRule = rootCell._rule;
        	
        	LambdaPattern p = rootRule.getPattern();
        	
    		for(int pIndex = 0; pIndex < this._cells.size(); pIndex++){
    			LambdaPhraseCell cell = this._cells.get(pIndex);
            	LambdaGRule gm = cell._rule;
            	double transProb = rootRule.getMRTransProb_test(0, gm);
            	
            	if(transProb > 0){
    				LambdaPhraseSpan[] c_spans = cell.getMaxSpans(0, this._N);
    				for(int k = c_spans.length - 1; k >= 0; k--){
    					double subProb = transProb * c_spans[k].getProb();
    					if(subProb > 0){
    						if(!rootCell.setMaxProb(0, this._N, subProb, new LambdaPhraseSpan[]{c_spans[k]}, p)){
    							break;
    						}
    					}
    				}
            	}
            }
		}
		
		
	}
    
}





