/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.phrase;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Vector;

import lambdagen.data.Instance;
import lambdagen.data.Word;
import lambdagen.grammar.LambdaGRule;

/**
 * @author Wei LU
 * @version 1.0
 */

/**
 * Assumption required: the sentence length is less than 1000
 */
public class SpanTable {
	
	private ArrayList<Word> _words;
	private Hashtable<Integer, int[][]> _spansTable; // each gm is mapped to its possible spans
	
	private SpanTable(ArrayList<Word> words){
		this._words = words;
		this._spansTable = new Hashtable<Integer, int[][]>();
	}
	
	public static SpanTable create(Instance inst, ArrayList<LambdaGRule> rules){
		ArrayList<Word> words = inst.getWords();
		SpanTable spanTable = new SpanTable(words);
		for(LambdaGRule rule: rules){
			spanTable.findSpans(rule, words);
		}
		return spanTable;
	}
	
	/**
	 * Check if the interval [bIndex, eIndex) can be matched at a certain position by the LambdaGMRProd
	 */
	public double getProb(LambdaGRule rule, int bIndex, int eIndex, int pos){
		int[][] spans = this._spansTable.get(rule.getId());
		if(spans == null){
			throw new RuntimeException("The LambdaGMRProd "+rule+" should not be chosen!");
		}
		//found any match, then return 1.0
		for(int i =0; i<spans[pos].length; i++){
			if(spans[pos][i]==span(bIndex, eIndex)){
				return 1.0;
			}
		}
		//otherwise, return 0.0
		return 0;
	}
	
	public void findSpans(LambdaGRule rule, ArrayList<Word> words){
		int[][] results;
		if(rule.getArity()==0){
			results = new int[1][];
			results[0] = findSpans(words, 0, words.size());
			if(results[0] == null){
				return;
			}
		} else if(rule.getArity()==1){
			int child_a_index = -1;
			for(int i = 0; i<words.size(); i++){
				if(words.get(i).equals(Word._X1)){
					child_a_index = i;
					break;
				}
			}
			results = new int[2][];
			results[0] = findSpans(words, 0, child_a_index);
			if(results[0] == null){
				return;
			}
			results[1] = findSpans(words, child_a_index+1, words.size());
			if(results[1] == null){
				return;
			}
		} else {
			int child_a_index = -1;
			int child_b_index = -1;
			for(int i = 0; i<words.size(); i++){
				if(words.get(i).equals(Word._X1)){
					child_a_index = i;
				} else if(words.get(i).equals(Word._X2)){
					child_b_index = i;
				}
			}
			if(child_a_index > child_b_index){
				int tmp = child_a_index;
				child_a_index = child_b_index;
				child_b_index = tmp;
			}
			results = new int[3][];
			results[0] = findSpans(words, 0, child_a_index);
			if(results[0] == null){
				return;
			}
			results[1] = findSpans(words, child_a_index+1, child_b_index);
			if(results[1] == null){
				return;
			}
			results[2] = findSpans(words, child_b_index+1, words.size());
			if(results[2] == null){
				return;
			}
		}
		
		this._spansTable.put(rule.getId(), results);
	}
	
	/**
	 * Find all the matched spans for a given interval in an array
	 */
	private int[] findSpans(ArrayList<Word> words, int bIndex, int eIndex){
		if(eIndex - bIndex == 0){
			return new int[0];
		}
		
		Vector<Integer> spans = new Vector<Integer>();
		
		int n = eIndex - bIndex;
		int N = this._words.size();
		for(int i = 0; i <= N - n; i++){
			int j = i + n;
			boolean matched = true;
			for(int k = i ; k< j; k++){
				if(this._words.get(k) != words.get(k-i+bIndex)){
					matched = false;
				}
			}
			if(matched){
				spans.add(span(i, j));
			}
		}
		
		if(spans.size() == 0){
			return null;
		}
		
		int[] results = new int[spans.size()];
		for(int i =0; i<results.length; i++){
			results[i] = spans.get(i);
		}
		
		return results;
	}

	private int span(int bIndex, int eIndex){
		return 1000 * bIndex + eIndex;
	}
}
