/**
 *  A Probabilistic Forest-to-String Model for Language Generation from Typed Lambda Calculus Expressions
 *
 *  Copyright (C) 2011  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package lambdagen.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Random;
import java.util.Scanner;

/**
 * @author Wei LU
 * @version 1.0
 */

public class Resampler {
	
	private ArrayList<String> _reference;
	private ArrayList<String> _output1;
	private ArrayList<String> _output2;
	private ArrayList<String> _output0;
	private Random _rand;
	private int _N;
	
	public Resampler(String ref_file, String output1_file, String output2_file, String output0_file) throws FileNotFoundException{
		
		this._reference = new ArrayList<String>();
		this._output1 = new ArrayList<String>();
		this._output2 = new ArrayList<String>();
		this._output0 = new ArrayList<String>();
		
		this.readFile(ref_file, this._reference);
		this.readFile(output1_file, this._output1);
		this.readFile(output2_file, this._output2);
		this.readFile(output0_file, this._output0);
		
		if(this._reference.size()!= this._output1.size() || this._reference.size()!=this._output2.size()|| this._reference.size()!=this._output0.size()){
			System.err.println("size mismatch!");
			System.err.println("[ref0]"+this._reference.size());
			System.err.println("[out1]"+this._output1.size());
			System.err.println("[out2]"+this._output2.size());
			System.err.println("[out0]"+this._output0.size());
			System.exit(1);
		}
		this._N = this._reference.size();
		
		this._rand = new Random(12345678);
		
	}
	
	private int[] sample(){
		int[] ids = new int[this._N];
		
		for(int i = 0 ; i<ids.length; i++){
			ids[i] = this._rand.nextInt(this._N);
		}
		
		return ids;
	}
	
	private String[] toCands(int i) throws IOException{
		File cand1_file = File.createTempFile("cand1", "i-"+i);
		File cand2_file = File.createTempFile("cand2", "i-"+i);
		File cand0_file = File.createTempFile("cand0", "i-"+i);
		File ref_file = File.createTempFile("ref", "i-"+i);
		
		cand1_file.deleteOnExit();
		cand2_file.deleteOnExit();
		cand0_file.deleteOnExit();
		ref_file.deleteOnExit();
		
		int[] sampled_ids = this.sample();
		
		this.writeSample(ref_file, sampled_ids, this._reference);
		this.writeSample(cand1_file, sampled_ids, this._output1);
		this.writeSample(cand2_file, sampled_ids, this._output2);
		this.writeSample(cand0_file, sampled_ids, this._output0);
		
		return new String[]{cand1_file.getAbsolutePath(), cand2_file.getAbsolutePath(), cand0_file.getAbsolutePath(), ref_file.getAbsolutePath()};
	}
	
	private void writeSample(File file, int[] ids, ArrayList<String> arr) throws FileNotFoundException, UnsupportedEncodingException{
		PrintWriter p = new PrintWriter(file, "UTF-8");
		for(int i = 0; i<ids.length; i++){
			p.println(arr.get(ids[i]));
			p.flush();
		}
		p.close();
	}
	
	public double[] batchScore(int num, String scorer) throws IOException{
		double pos[] = {0,0};
		for(int i = 0; i<num; i++){
			String[] files = toCands(i);
			int score[] = score(i, files[0], files[1], files[2], files[3], scorer);
			if(score[0] > 0){
				pos[0] ++;
			}
			if(score[1] > 0){
				pos[1] ++;
			}
		}
		return new double[]{1.0 - pos[0] / num, 1.0 - pos[1] / num};
	}
	
	private int[] score(int index, String cand1_file, String cand2_file, String cand0_file, String ref_file, String scorer) throws IOException{
		String scorer_cmd = "-m BLEU 4 closest -txtNrm 0";
		if(scorer.equals("TER")){
			scorer_cmd = "-m TER nocase punc 20 50 lib/tercom.7.25.jar 1";
		}

		String cmd1 = "java -Djava.library.path=/Users/luwei/tools/joshua-1.3/lib " +
				"-cp /Users/luwei/tools/joshua-1.3/bin/ " +
				"-Xmx1000m -Xms1000m -Djava.util.logging.config.file=logging.properties " +
				"joshua.util.JoshuaEval " +
				"-cand "+cand1_file+" " +
				"-ref "+ref_file+" " +
				"-rps 1 " + scorer_cmd;

		String cmd2 = "java -Djava.library.path=/Users/luwei/tools/joshua-1.3/lib " +
				"-cp /Users/luwei/tools/joshua-1.3/bin/ " +
				"-Xmx1000m -Xms1000m -Djava.util.logging.config.file=logging.properties " +
				"joshua.util.JoshuaEval " +
				"-cand "+cand2_file+" " +
				"-ref "+ref_file+" " +
				"-rps 1 " + scorer_cmd;

		String cmd0 = "java -Djava.library.path=/Users/luwei/tools/joshua-1.3/lib " +
				"-cp /Users/luwei/tools/joshua-1.3/bin/ " +
				"-Xmx1000m -Xms1000m -Djava.util.logging.config.file=logging.properties " +
				"joshua.util.JoshuaEval " +
				"-cand "+cand0_file+" " +
				"-ref "+ref_file+" " +
				"-rps 1 " + scorer_cmd;
		
		double score1 = getScore(cmd1, scorer);
		double score2 = getScore(cmd2, scorer);
		double score0 = getScore(cmd0, scorer);
		
		System.err.println(index+"\t"+score1+"\t"+score2+"\t"+score0);
		
		int[] res = {0, 0};
		res[0] = score0 < score1 ? -1 : score0 > score1 ? +1 : 0;
		res[1] = score0 < score2 ? -1 : score0 > score2 ? +1 : 0;
		
		return res;
	}
	
	private double getScore(String cmd, String scorer) throws IOException{
		String line;
	    InputStream stdout = null;

	      Process process = Runtime.getRuntime ().exec (cmd);
	      stdout = process.getInputStream ();
	      
	      BufferedReader brCleanUp = new BufferedReader (new InputStreamReader (stdout));
	      while ((line = brCleanUp.readLine ()) != null) {
	        line = line.trim();
	        if(line.startsWith("=> BLEU = ")){
	        	int index = line.lastIndexOf("=");
	        	return Double.parseDouble(line.substring(index+1).trim());	        	
	        } else if(line.startsWith("TER = ")){
	        	int index = line.lastIndexOf("=");
	        	return 1-Double.parseDouble(line.substring(index+1).trim());	        	
	        }
	      }
	      brCleanUp.close();

	      return -1;
	}
	

	private static double getScore2(String cmd, String scorer) throws IOException{
		String line;
	    InputStream stdout = null;

	      Process process = Runtime.getRuntime ().exec (cmd);
	      stdout = process.getInputStream ();
	      
	      // clean up if any output in stdout
	      BufferedReader brCleanUp = new BufferedReader (new InputStreamReader (stdout));
	      while ((line = brCleanUp.readLine ()) != null) {
	        line = line.trim();
	        if(line.startsWith("=> BLEU = ") || line.startsWith("TER = ")){
	        	int index = line.lastIndexOf("=");
	        	return Double.parseDouble(line.substring(index+1).trim());	        	
	        }
	      }
	      brCleanUp.close();

	      return -1;
	}
	
	private void readFile(String filename, ArrayList<String> arr) throws FileNotFoundException{
		Scanner scan = new Scanner(new File(filename),"UTF-8");
		while(scan.hasNextLine()){
			String line = scan.nextLine().trim();
			arr.add(line);
		}
	}
	
	public static void main(String args[])throws IOException{
		
		/*
		String scorer = args[2];
		String scorer_cmd = "-m BLEU 4 closest -txtNrm 0";
		if(scorer.equals("TER")){
			scorer_cmd = "-m TER nocase punc 20 50 lib/tercom.7.25.jar 1";
		}

		String cmd1 = "java -Djava.library.path=/Users/luwei/tools/joshua-1.3/lib " +
				"-cp /Users/luwei/tools/joshua-1.3/bin/ " +
				"-Xmx1000m -Xms1000m -Djava.util.logging.config.file=logging.properties " +
				"joshua.util.JoshuaEval " +
				"-cand "+args[0]+" " +
				"-ref "+args[1]+" " +
				"-rps 1 " + scorer_cmd;
		
		double val = getScore2(cmd1, scorer);
		
		System.err.println(scorer+" = "+val);
		
		System.exit(1);
		*/
		
		if(args.length < 5){
			System.err.println("Usage: java Resampler <ref_file> <moses_output_file> <joshua_output_file> <mysys_output_file> <scorer>");
			System.exit(1);
		}
		
		Resampler resampler = new Resampler(args[0],args[1],args[2], args[3]);
		double p_value[] = resampler.batchScore(1000, args[4]);
		
		System.err.println("p_value: vs  moses = "+p_value[0]);
		System.err.println("p_value: vs joshua = "+p_value[1]);
		
	}
	
}
