package zmert;

/* This file is part of the Z-MERT Training Tool for MT systems.
 * 
 * Z-MERT is an open-source tool, licensed under the terms of the
 * GNU Lesser General Public License (LGPL). Therefore, it is free
 * for personal and scientific use by individuals and/or research
 * groups. It may not be modified or redistributed, publicly or
 * privately, unless the licensing terms are observed. If in doubt,
 * contact the author for clarification and/or an explicit
 * permission.
 *
 * If you use Z-MERT in your work, please cite the following paper:
 *
 *       Omar F. Zaidan. 2009. Z-MERT: A Fully Configurable Open
 *       Source Tool for Minimum Error Rate Training of Machine
 *       Translation Systems. The Prague Bulletin of Mathematical
 *       Linguistics, No. 91:79-88.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

import java.io.*;

public class TER extends EvaluationMetric
{
  private boolean caseSensitive;
  private boolean withPunctuation;
  private int beamWidth;
  private int maxShiftDist;

  public TER(String[] Metric_options)
  {
    // M_o[0]: case sensitivity, case/nocase
    // M_o[1]: with-punctuation, punc/nopunc
    // M_o[2]: beam width, positive integer
    // M_o[3]: maximum shift distance, positive integer

    // default in tercom-0.7.25: nocase, punc, 20, 50

    if (Metric_options[0].equals("case")) {
      caseSensitive = true;
    } else if (Metric_options[0].equals("nocase")) {
      caseSensitive = false;
    } else {
      System.out.println("Unknown case sensitivity string " + Metric_options[0] + ".");
      System.out.println("Should be one of case or nocase.");
      System.exit(1);
    }

    if (Metric_options[1].equals("punc")) {
      withPunctuation = true;
    } else if (Metric_options[1].equals("nopunc")) {
      withPunctuation = false;
    } else {
      System.out.println("Unknown with-punctuation string " + Metric_options[1] + ".");
      System.out.println("Should be one of punc or nopunc.");
      System.exit(1);
    }

    beamWidth = Integer.parseInt(Metric_options[2]);
    if (beamWidth < 1) {
      System.out.println("Beam width must be positive");
      System.exit(2);
    }

    maxShiftDist = Integer.parseInt(Metric_options[3]);
    if (maxShiftDist < 1) {
      System.out.println("Maximum shift distance must be positive");
      System.exit(2);
    }

    initialize(); // set the data members of the metric
  }

  @Override
protected void initialize()
  {
    metricName = "TER";
    toBeMinimized = true;
    suffStatsCount = 2;
  }

  @Override
public double bestPossibleScore() { return 0.0; }
  @Override
public double worstPossibleScore() { return (+1.0 / 0.0); }

  @Override
public int[] suffStats(String cand_str, int i)
  {
    // this method should never be used when the metric is TER,
    // because TER.java overrides suffStats(String[],int[]) below,
    // which is the only method that calls suffStats(Sting,int).
    return null;
  }

  @Override
public int[][] suffStats(String[] cand_strings, int[] cand_indices)
  {
    // calculate sufficient statistics for each sentence in an arbitrary set of candidates

    int candCount = cand_strings.length;
    if (cand_indices.length != candCount) {
      System.out.println("Array lengths mismatch in suffStats(String[],int[]); returning null.");
      return null;
    }

    int[][] stats = new int[candCount][suffStatsCount];

    try {

      // 1) Create input files for tercom.7.25.jar

      // 1a) Create hypothesis file
      FileOutputStream outStream = new FileOutputStream("hyp.txt.TER", false); // false: don't append
      OutputStreamWriter outStreamWriter = new OutputStreamWriter(outStream, "utf8");
      BufferedWriter outFile = new BufferedWriter(outStreamWriter);

      for (int d = 0; d < candCount; ++d) {
        writeLine(cand_strings[d] + " (ID" + d + ")",outFile);
      }

      outFile.close();

      // 1b) Create reference file
      outStream = new FileOutputStream("ref.txt.TER", false); // false: don't append
      outStreamWriter = new OutputStreamWriter(outStream, "utf8");
      outFile = new BufferedWriter(outStreamWriter);

      for (int d = 0; d < candCount; ++d) {
        for (int r = 0; r < refsPerSen; ++r) {
          writeLine(refSentences[cand_indices[d]][r] + " (ID" + d + ")",outFile);
        }
      }

      outFile.close();

      // 2) Launch tercom.7.25.jar as an external process

      String cmd_str = "java -Dfile.encoding=utf8 -jar tercom.7.25.jar -r ref.txt.TER -h hyp.txt.TER -o ter -n TER_out";
      cmd_str += " -b " + beamWidth;
      cmd_str += " -d " + maxShiftDist;
      if (caseSensitive) { cmd_str += " -S"; }
      if (withPunctuation) { cmd_str += " -P"; }

      Runtime rt = Runtime.getRuntime();
      Process p = rt.exec(cmd_str);

      StreamGobbler errorGobbler = new StreamGobbler(p.getErrorStream(), 0);
      StreamGobbler outputGobbler = new StreamGobbler(p.getInputStream(), 0);

      errorGobbler.start();
      outputGobbler.start();

      @SuppressWarnings("unused")
      int exitValue = p.waitFor();


      // 3) Read SS from output file produced by tercom.7.25.jar

      BufferedReader inFile = new BufferedReader(new FileReader("TER_out.ter"));
      String line = "";

      line = inFile.readLine(); // skip hyp line
      line = inFile.readLine(); // skip ref line

      for (int d = 0; d < candCount; ++d) {
        line = inFile.readLine(); // read info
        String[] strA = line.split("\\s+");

        stats[d][0] = (int)Double.parseDouble(strA[1]);
        stats[d][1] = (int)Double.parseDouble(strA[2]);
      }


      // 4) Delete TER files

      File fd;
      fd = new File("hyp.txt.TER"); if (fd.exists()) fd.delete();
      fd = new File("ref.txt.TER"); if (fd.exists()) fd.delete();
      fd = new File("TER_out.ter"); if (fd.exists()) fd.delete();

    } catch (IOException e) {
      System.err.println("IOException in TER.suffStats(String[],int[]): " + e.getMessage());
      System.exit(99902);
    } catch (InterruptedException e) {
      System.err.println("InterruptedException in TER.suffStats(String[],int[]): " + e.getMessage());
      System.exit(99903);
    }

    return stats;
  }

  @Override
public double score(int[] stats)
  {
    if (stats.length != suffStatsCount) {
      System.out.println("Mismatch between stats.length and suffStatsCount (" + stats.length + " vs. " + suffStatsCount + ") in TER.score(int[])");
      System.exit(1);
    }

    double sc = 0.0;

    sc = stats[0]/(double)stats[1];

    return sc;
  }

  @Override
public void printDetailedScore_fromStats(int[] stats, boolean oneLiner)
  {
    if (oneLiner) {
      System.out.println("TER = " + stats[0] + " / " + stats[1] + " = " + score(stats));
    } else {
      System.out.println("# edits = " + stats[0]);
      System.out.println("Reference length = " + stats[1]);
      System.out.println("TER = " + stats[0] + " / " + stats[1] + " = " + score(stats));
    }
  }

  private void writeLine(String line, BufferedWriter writer) throws IOException
  {
    writer.write(line, 0, line.length());
    writer.newLine();
    writer.flush();
  }

}

